#pragma once

#pragma comment(lib, "d3dx9.lib")
#pragma comment(lib, "d3d9.lib")

#define WIN32_MEAN_AND_LEAN
#define _USE_MATH_DEFINES // to get constants in cmath
#include <windows.h>
#include <d3dx9.h>
#include <cg/cg.h>

#include "../GraphicsEnvironment.h"

class D3D9GraphicsEnvironment : public GraphicsEnvironment {
	HWND hWnd;
	bool active;
	LPDIRECT3D9       m_pD3D;
	LPDIRECT3DDEVICE9 m_pDevice;
	bool m_bDeviceLost;
	D3DPRESENT_PARAMETERS d3dpp; 

	CGcontext cgContext;
	CGprofile cgVertexProfile;
	CGprofile cgPixelProfile;

	void initCG();
	void closeScreen();

	static LRESULT CALLBACK WndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );
	static void cgErrorCallback(void);

	bool handlePresentRetVal(HRESULT hResult);
	bool resetDevice();
	// Called when the device is lost or just reset
	void onDeviceLost();
	void onDeviceReset();

	virtual bool drawFrameStart();
public:
	//D3D9GraphicsEnvironment(int width,int height,int multisample,bool fullscreen,bool hdr);
	D3D9GraphicsEnvironment(VisionPrefs visionPrefs);
	virtual ~D3D9GraphicsEnvironment();

	virtual bool hasContext() const {
		return false;
	}

	// interface
#ifdef _WIN32
	virtual HWND getHWND() const {
		return this->hWnd;
	}
	virtual bool isTrueFullscreen() const {
		return isFullScreen();
	}
#endif
	virtual void setVisible(bool visible);
	virtual void setWindowTitle(const char *title);
	virtual void showCursor(bool show) {
		ShowCursor(show);
	}
	virtual void setCursor(unsigned char *data, unsigned char *mask, int w, int h, int hot_x, int hot_y) {
		Vision::setError(new VisionException(this, VisionException::V_NOT_YET_SUPPORTED, "setCursor not supported for this GraphicsEnvironment"));
	}
	virtual void swapBuffers();
	virtual void processEvents();
	virtual void waitForEvents();
	virtual void sleep(int time);
	virtual bool saveScreen(const char *filename,const char *type);
	virtual void checkError(const char *label) {
	}
	virtual bool hasShaders() const {
		return cgContext != NULL;
	}
	virtual bool isActive() const {
		return active;
	}
	virtual Shader *createShader(const char *file, const char *func, bool vertex_shader);
};
