//---------------------------------------------------------------------------
//#define DIRECTSOUND_VERSION 6

#define WIN32_MEAN_AND_LEAN
#include <windows.h>

#include "DXSound.h"
#include "wavread.h"
#include "../Resource.h"

#pragma comment(lib, "ole32.lib")
#pragma comment(lib, "strmiids.lib")

#define WM_GRAPHNOTIFY WM_APP + 1

//---------------------------------------------------------------------------

void printDSError(HRESULT hresult) {
	switch(hresult) {
		case DSERR_ALLOCATED:
			LOG("  DSERR_ALLOCATED\n");
			break;
		case DSERR_NOAGGREGATION:
			LOG("  DSERR_NOAGGREGATION\n");
			break;
		case DSERR_NODRIVER:
			LOG("  DSERR_NODRIVER\n");
			break;
		case DSERR_INVALIDPARAM:
			LOG("  DSERR_INVALIDPARAM\n");
			break;
		case DSERR_OUTOFMEMORY:
			LOG("  DSERR_OUTOFMEMORY\n");
			break;
		default :
			LOG("  Unknown DirectSound Error : %d\n",hresult);
			break;
	}
}


void DXSound::DS_Init() {
	HRESULT hresult;
	// setup DirectSound
	if( ( hresult = DirectSoundCreate(NULL,&lpds,NULL) ) != DS_OK ) {
		LOG("Failed to initialise DirectSound: %d\n", hresult);
		printDSError(hresult);
		//Vision::setError(new VisionException(this,VisionException::V_DIRECTSOUND_ERROR,"DirectSoundCreate() failed"));
	}
	else {
		this->initialised=true;
	}
}

//DXSound::DXSound(HWND hwnd) : Sound() {
DXSound::DXSound() : Sound() {
	for(int i=0;i<max_sounds_c;i++)
		lpdsbStatic[i] = NULL;
	this->lpds=NULL;
	pMediaGraph = NULL;
	pMediaControl = NULL;
	pMediaEvent = NULL;
	pMediaPosition = NULL;
	has_media = false;

	// initialize direct sound
	DS_Init();

	/*if( this->initialised ) {
		this->setCooperativeLevel(hwnd);
	}*/

	// DirectShow
	HRESULT hresult = CoInitialize(NULL);
	if( FAILED(hresult) ) {
		LOG("CoInitialize failed!: %d\n", hresult);
	}
	else if( FAILED( hresult = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, IID_IGraphBuilder, (void **)&pMediaGraph) ) ) {
		LOG("Failed to initialise DirectShow!: %d\n", hresult);
	}
	else if( FAILED( hresult = pMediaGraph->QueryInterface(IID_IMediaControl, (void **)&pMediaControl) ) ) {
		LOG("Failed to obtain control interface from DirectShow!: %d\n", hresult);
	}
	//else if( FAILED( hresult = pMediaGraph->QueryInterface(IID_IMediaEvent, (void **)&pMediaEvent) ) ) {
	else if( FAILED( hresult = pMediaGraph->QueryInterface(IID_IMediaEventEx, (void **)&pMediaEvent) ) ) {
		LOG("Failed to obtain event interface from DirectShow!: %d\n", hresult);
	}
	else if( FAILED( hresult = pMediaGraph->QueryInterface(IID_IMediaPosition, (void**)&pMediaPosition) ) ) {
		LOG("Failed to obtain media position interface from DirectShow!: %d\n", hresult);
	}
	else {
		has_media = true;

		/*IMediaEventEx *pMediaEventEx = NULL;
		pMediaGraph->QueryInterface(IID_IMediaEventEx, (void **)&MediapEventEx);
		pMediaEventEx->SetNotifyWindow((OAHWND)hwnd, WM_GRAPHNOTIFY, 0);*/
		//pMediaEvent->SetNotifyWindow((OAHWND)hwnd, WM_GRAPHNOTIFY, 0);
	}

	if( !has_media ) {
		// failed to start DirectShow
		if( pMediaControl )
			pMediaControl->Release();
		if( pMediaEvent )
			pMediaEvent->Release();
		if( pMediaPosition )
			pMediaPosition->Release();
		if( pMediaGraph )
			pMediaGraph->Release();
		pMediaControl = NULL;
		pMediaEvent = NULL;
		pMediaPosition = NULL;
		pMediaGraph = NULL;
	}
}

DXSound::~DXSound() {
	// free buffers
	for(int i=0;i<max_sounds_c;i++) {
		if(lpdsbStatic[i]!=NULL)
			lpdsbStatic[i]->Unlock(lpvAudio[i], dwBytes[i], NULL, 0);
	}

	if( lpds )
		lpds->Release();

	// DirectShow
	if( pMediaEvent ) {
		// Disable event notification before releasing the graph.
		pMediaEvent->SetNotifyWindow(NULL, 0, 0);
		pMediaEvent->Release();
		pMediaEvent = NULL;
	}
	if( pMediaControl ) {
		pMediaControl->Release();
		pMediaControl = NULL;
	}
	if( pMediaPosition ) {
		pMediaPosition->Release();
		pMediaPosition = NULL;
	}
	if( pMediaGraph ) {
		pMediaGraph->Release();
		pMediaGraph = NULL;
	}

	CoUninitialize();
}

//bool DXSound::setCooperativeLevel(HWND hwnd) {
bool DXSound::initHWND(HWND hwnd) {
	HRESULT hresult;
	if ( (hresult=lpds->SetCooperativeLevel(hwnd,DSSCL_PRIORITY))!=DS_OK) {
		printDSError(hresult);
		return false;
	}

	if( this->has_media ) {
		pMediaEvent->SetNotifyWindow((OAHWND)hwnd, WM_GRAPHNOTIFY, 0);
	}
	return true;
}

void DXSound::FreeSound(int s) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if(lpdsbStatic[s]!=NULL) {
		lpdsbStatic[s]->Unlock(lpvAudio[s], dwBytes[s], NULL, 0);
		lpdsbStatic[s]=NULL;
	}
}

bool DXSound::LoadSound(int s,const char *file) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if( !initialised ) {
		LOG("Sound system not initialised\n");
		return false;
	}
	CWaveSoundRead *cwave = new CWaveSoundRead();
	if( FAILED( cwave->Open(file) ) ) {
		LOG("failed to open sound file: %s\n", file);
		delete cwave;
		return false;
	}

	memset(&dsbdesc, 0, sizeof(DSBUFFERDESC));
	dsbdesc.dwSize = sizeof(DSBUFFERDESC);
	dsbdesc.dwFlags = DSBCAPS_STATIC;
	dsbdesc.dwBufferBytes = cwave->m_ckIn.cksize;
	dsbdesc.lpwfxFormat = cwave->m_pwfx;

	if FAILED(lpds->CreateSoundBuffer(&dsbdesc, &lpdsbStatic[s], NULL)) {
		LOG("failed to create sound buffer for: %s\n", file);
		delete cwave;
		return false;
	}
	// Remember how big the buffer is
	//bufferBytes[0] = dsbdesc.dwBufferBytes;

	if FAILED(lpdsbStatic[s]->Lock(
		0,              // Offset of lock start
		0,              // Size of lock; ignored in this case
		&lpvAudio[s],     // Address of lock start
		&dwBytes[s],      // Number of bytes locked
		NULL,           // Wrap around start; not used
		NULL,           // Wrap around size; not used
		DSBLOCK_ENTIREBUFFER))  // Flag
	{
		LOG("failed to lock sound buffer for file: %s\n", file);
		delete cwave;
		return false;
	}
	if(cwave->Read(dwBytes[s],(BYTE *) lpvAudio[s],&cbBytesRead)) {
		LOG("failed to read bytes for sound file: %s\n", file);
		delete cwave;
		return false;
	}
	//delete cwave; // don't need to delete???
	return true;
}

bool DXSound::PlaySound(int s,bool loop,int delay) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if(!initialised)
		return false;
	if(lpdsbStatic[s]==NULL)
		return false;
	if(delay!=-1) {
		if(IsSoundPlaying(s)) {
			DWORD pos;
			lpdsbStatic[s]->GetCurrentPosition(&pos,NULL);
			//printf(">>>%d %d\n",pos,delay);
			if((int)pos<delay)
				return false;
			StopSound(s,true);
		}
	}
	HRESULT hresult;
	DWORD dwLooped = loop ? DSBPLAY_LOOPING : 0L;
	if( FAILED( hresult = lpdsbStatic[s]->Play( 0, 0, dwLooped ) ) ) {
		return false;
	}
	return true;
}

bool DXSound::IsSoundPlaying(int s) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if(!initialised)
		return false;
	if(lpdsbStatic[s]==NULL)
		return false;
	DWORD dwStatus = 0;
	if( NULL == lpdsbStatic[s] )
		return false;
	lpdsbStatic[s]->GetStatus( &dwStatus );
	if( dwStatus & DSBSTATUS_PLAYING )
		return true;
	else
		return false;
}

void DXSound::StopSound(int s,bool bResetPosition ) {
	ASSERT( s >= 0 && s < max_sounds_c );
	if(!initialised)
		return;
	if(lpdsbStatic[s]==NULL)
		return;
	lpdsbStatic[s]->Stop();
	if( bResetPosition )
		lpdsbStatic[s]->SetCurrentPosition( 0L );
}

void DXSound::playMedia(const char *filename) {
	if( !has_media ) {
		return;
	}
	// faff for wide strings!
	int len = strlen(filename);
	int w_len = len+1;
	wchar_t *w_filename = new wchar_t[w_len];
	size_t value;
	mbstowcs_s(&value, w_filename, w_len, filename, len);
	HRESULT hresult = pMediaGraph->RenderFile(w_filename, NULL);
	if( FAILED( hresult ) ) {
		LOG("Failed to load music file\n");
	}
	else {
		hresult = pMediaControl->Run();
		if( FAILED( hresult ) ) {
			LOG("Failed to play music file\n");
		}
	}
	delete w_filename;
}

bool DXSound::isMediaPlaying() {
	if( !has_media ) {
		return false;
	}
	/*FILTER_STATE fs;
	if( FAILED( pControl->GetState(1000, (OAFilterState*)&fs) ) ) {
		LOG("Failed to get media state\n");
		return false;
	}
	return fs == State_Running;*/
    long evCode;
    LONG_PTR param1, param2;
    while( SUCCEEDED( pMediaEvent->GetEvent(&evCode, &param1, &param2, 0) ) ) {
        pMediaEvent->FreeEventParams(evCode, param1, param2);
        switch (evCode)
        {
        case EC_COMPLETE:  // Fall through.
        case EC_USERABORT: // Fall through.
        case EC_ERRORABORT:
			return true;
        }
    }
	return false;
}

void DXSound::processEvents(UINT message, WPARAM wParam, LPARAM lParam) {
	if( pMediaEvent == NULL ) {
		return;
	}
	if( message == WM_GRAPHNOTIFY ) {
		// Get all the events
		long evCode;
		LONG_PTR param1, param2;
		while( SUCCEEDED( pMediaEvent->GetEvent(&evCode, &param1, &param2, 0) ) ) {
			pMediaEvent->FreeEventParams(evCode, param1, param2);
			switch (evCode)
			{
			case EC_COMPLETE:  // Fall through.
			case EC_USERABORT: // Fall through.
			case EC_ERRORABORT:
				LOG("Music complete\n");
				// loop
				//pControl->
				//HRESULT hresult = pControl->Run();
				//HRESULT hresult = pMediaControl->Stop();
				HRESULT hresult = pMediaPosition->put_CurrentPosition(0);
				if( FAILED( hresult ) ) {
					LOG("Failed to play music file\n");
				}
				break;
			}
		} 
	}
}
