//---------------------------------------------------------------------------
#include <ctime>
#define _USE_MATH_DEFINES // needed to get maths constants in math.h when using Visual Studio
#include <cmath>
#include <cassert>

#include <algorithm>
using std::swap;

#include "GraphicsEngine.h"
#include "Renderer.h"
#include "GraphicsEnvironment.h"
#include "Entity.h"
#include "World.h"
#include "Terrain.h"
#include "RenderData.h"

//---------------------------------------------------------------------------

//const float splat_detail_scale_c = 100.0f; // scale for terrain splatting
//const float splat_detail_scale_c = 8.0f; // scale for terrain splatting; for SECTION_SIZE 64
//const float splat_detail_scale_c = 4.0f; // scale for terrain splatting; for SECTION_SIZE 32 (original)
//const float splat_detail_scale_c = 2.0f; // scale for terrain splatting; for SECTION_SIZE 16

Graphics3DShaders::Graphics3DShaders() {
	clear();
}

void Graphics3DShaders::clear() {
	memset(this, 0, sizeof(Graphics3DShaders));
}

Graphics3D::Graphics3D(GraphicsEnvironment *genv, bool want_hdr) {
	this->genv=genv;
	/*this->z_near = 0.1f;
	this->z_far = 1000.0f;*/
	//this->persp = true;
	/*this->width = genv->getWidth();
	this->height = genv->getHeight();*/
	this->world = NULL;
	//this->use_stencil = false;
	//this->use_stencil = true;
	this->use_shadow_volumes = false;
	this->want_hdr = want_hdr;
	this->hdr_enabled = false;
	this->bloom_enabled = false;
	this->hdr_fbo = NULL;
	this->temp_fbo = NULL;
	//this->shader_clipping = false;
	//this->textureID = 0;
	this->frames = 0;
	this->frame_time = 0;
	this->current_time = 0;
	this->fps = 0.0;
	//this->cgProgram_vertex_shader = NULL;
	//this->shader_enabled = false;
	this->worldToEyeMatrix = NULL; // created lazily

	//LOG("a\n");
	this->shaders = new vector<Shader *>();
	//LOG("b\n");
	freeShaders(); // to set shader pointers to NULL
	//LOG("c\n");

	/*this->c_vertex_shader = NULL;
	this->c_pixel_shader = NULL;*/

	this->c_queue = NULL;

}

Graphics3D::~Graphics3D() {
	freeShaders();
	delete shaders;
	delete hdr_fbo;
	delete temp_fbo;
	delete worldToEyeMatrix;
}

void Graphics3D::freeShaders() {
	LOG("Graphics3D::freeShaders()\n");
	for(unsigned int i=0;i<shaders->size();i++) {
		Shader *shader = shaders->at(i);
		delete shader;
	}
	shaders->clear();
	std_shaders.clear();

	/*this->vertex_shader_ambient = NULL;
	this->pixel_shader_ambient = NULL;
	this->vertex_shader_directional = NULL;
	this->pixel_shader_directional = NULL;
	this->vertex_shader_point = NULL;
	this->pixel_shader_point = NULL;
	this->bump_vertex_shader_directional = NULL;
	this->bump_vertex_shader_point = NULL;
	this->bump_pixel_shader_directional = NULL;
	this->bump_pixel_shader_point = NULL;
	//this->vertex_shader_shadow_point = NULL;
	this->splat_pixel_shader_ambient = NULL;
	this->splat_pixel_shader_directional = NULL;
	this->splat_pixel_shader_point = NULL;
	water_simple_pixel_shader = NULL;
	water_vertex_shader_ambient = NULL;
	water_pixel_shader_ambient = NULL;
	water_vertex_shader_directional = NULL;
	water_pixel_shader_directional = NULL;
	water_vertex_shader_point = NULL;
	water_pixel_shader_point = NULL;
	hdr_pixel_shader = NULL;
	blurw_pixel_shader = NULL;
	blurh_hdr_pixel_shader = NULL;
	vertex_shader_skyfromatmos = NULL;
	pixel_shader_skyfromatmos = NULL;*/
}

void Graphics3D::addShader(Shader *shader) {
	this->shaders->push_back(shader);
}

Renderer *Graphics3D::getRenderer() {
	return this->genv->getRenderer();
}

void Graphics3D::createHDRBuffers() {
	hdr_enabled = false;
	bloom_enabled = false;
	if( RendererInfo::framebufferObject && this->std_shaders.hdr_vertex_shader != NULL && this->std_shaders.hdr_pixel_shader != NULL ) {
		//if( false ) {
		LOG("creating HDR framebuffer...\n");

		/*int tex_width = genv->getAA()*width;
		int tex_height = genv->getAA()*height;*/
		/*int tex_width = width;
		int tex_height = height;*/
		int tex_width = genv->getWidth();
		int tex_height = genv->getHeight();

		//bool hdr_ok = genv->getRenderer()->generateHdrFbo(tex_width, tex_height);
		this->hdr_fbo = genv->getRenderer()->createFramebufferObject();
		bool hdr_ok = this->hdr_fbo->generate(tex_width, tex_height, true);

		if( !hdr_ok ) {
			LOG("Failed to create Framebuffer object!\n");
			//Vision::framebufferObject = false;
		}
		else {
			LOG("    done\n");
			hdr_enabled = true;

			if(
				//false &&
				this->std_shaders.blurw_pixel_shader != NULL && this->std_shaders.blurh_hdr_pixel_shader != NULL ) {

					this->temp_fbo = genv->getRenderer()->createFramebufferObject();
					bool temp_ok = this->temp_fbo->generate(tex_width, tex_height, false);
					if( !temp_ok ) {
						LOG("Failed to create temp Framebubber object!\n");
					}
					else {
						LOG("created temp Framebuffer okay\n");
						bloom_enabled = true;
					}
			}
		}
	}
	else {
		LOG("HDR not available\n");
	}
}

void Graphics3D::setZ(float z_near, float z_far, float fovy) {
	/*this->z_near = z_near;
	this->z_far = z_far;*/
	//ReSizeGLScene(width,height);
	ReSizeGLScene(z_near, z_far, genv->getWidth(), genv->getHeight(), genv->getRenderer()->getPersp(), fovy);
}

void Graphics3D::setPerspective(bool persp, float fovy) {
	//this->persp = persp;
	//this->ReSizeGLScene(this->width, this->height);
	ReSizeGLScene(genv->getRenderer()->getZNear(), genv->getRenderer()->getZFar(), genv->getWidth(), genv->getHeight(), persp, fovy);
}

/*void Graphics3D::InitGLSize(int width,int height,bool hdr) {
//this->hdr_enabled = hdr; // attempt to enable HDR?
this->want_hdr = hdr;
this->ReSizeGLScene(width, height);
}*/

void Graphics3D::setViewport() {
	/*if( this->hdr_enabled ) {
	// viewpoint size must take into account multisampling
	glViewport(0, 0, genv->getAA()*width, genv->getAA()*height);
	}
	else*/ {
		//glViewport(0, 0, width, height);
		//genv->getRenderer()->setViewport(width, height);
		genv->getRenderer()->setViewport(genv->getWidth(), genv->getHeight());
}
}

/* Called when the scene is resized.
*/
void Graphics3D::ReSizeGLScene(float z_near, float z_far, int width, int height, bool persp, int fovy) {
	LOG("Graphics3D::ReSizeGLScene(%f,%f,%d,%d,%d,%d)\n", z_near, z_far, width, height, persp, fovy);
	if( GraphicsEnvironment::existsOrHasContext() ) {
		// check needed, as this function can be called during initialisation of GraphicsEnvironment
		this->genv->checkError("GraphicsEnvironment::ReSizeGLScene BEGIN");
	}
	/*this->width = width;
	this->height = height;*/

	if( this->want_hdr ) {
		//genv->getRenderer()->freeHDRBuffers();
		createHDRBuffers();
	}

	// [note - even if using hdr, we don't want to take into account multisampling at this stage,
	// so we don't use setViewport(). - OUTDATED now that we don't do supersampling]

	//genv->getRenderer()->resizeScene(frustum.frustum, width, height, persp, z_near, z_far, fovy);
	//genv->getRenderer()->resizeScene(frustum.frustum, genv->getWidth(), genv->getHeight(), persp, z_near, z_far, fovy);
	genv->getRenderer()->resizeScene(genv->getFrustum()->frustum, genv->getWidth(), genv->getHeight(), persp, z_near, z_far, fovy);

	if( GraphicsEnvironment::existsOrHasContext() ) {
		this->genv->checkError("GraphicsEnvironment::ReSizeGLScene END");
	}
}


/*void trace_extractFrustum() {

float frustum[6][4];
double proj[16];
double modl[16];
double clip[16];
float t = 0.0f;

glGetDoublev( GL_PROJECTION_MATRIX, proj );

glMatrixMode(GL_MODELVIEW);
glPushMatrix();
glLoadIdentity();
glGetDoublev( GL_MODELVIEW_MATRIX, modl );
glPopMatrix();

// Combine the two matrices (multiply projection by modelview)
clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];

clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];

// Extract the numbers for the RIGHT plane
frustum[0][0] = (float)(clip[ 3] - clip[ 0]);
frustum[0][1] = (float)(clip[ 7] - clip[ 4]);
frustum[0][2] = (float)(clip[11] - clip[ 8]);
frustum[0][3] = (float)(clip[15] - clip[12]);
// Normalize the result
t = sqrt( frustum[0][0] * frustum[0][0] + frustum[0][1] * frustum[0][1] + frustum[0][2] * frustum[0][2] );
frustum[0][0] /= t;
frustum[0][1] /= t;
frustum[0][2] /= t;
frustum[0][3] /= t;
LOG("FRUSTUM RIGHT %f %f %f %f\n", frustum[0][0],  frustum[0][1], frustum[0][2], frustum[0][3]);

// Extract the numbers for the LEFT plane
frustum[1][0] = (float)(clip[ 3] + clip[ 0]);
frustum[1][1] = (float)(clip[ 7] + clip[ 4]);
frustum[1][2] = (float)(clip[11] + clip[ 8]);
frustum[1][3] = (float)(clip[15] + clip[12]);
// Normalize the result
t = sqrt( frustum[1][0] * frustum[1][0] + frustum[1][1] * frustum[1][1] + frustum[1][2] * frustum[1][2] );
frustum[1][0] /= t;
frustum[1][1] /= t;
frustum[1][2] /= t;
frustum[1][3] /= t;
LOG("FRUSTUM LEFT %f %f %f %f\n", frustum[1][0],  frustum[1][1], frustum[1][2], frustum[1][3]);

// Extract the BOTTOM plane
frustum[2][0] = (float)(clip[ 3] + clip[ 1]);
frustum[2][1] = (float)(clip[ 7] + clip[ 5]);
frustum[2][2] = (float)(clip[11] + clip[ 9]);
frustum[2][3] = (float)(clip[15] + clip[13]);
// Normalize the result
t = sqrt( frustum[2][0] * frustum[2][0] + frustum[2][1] * frustum[2][1] + frustum[2][2] * frustum[2][2] );
frustum[2][0] /= t;
frustum[2][1] /= t;
frustum[2][2] /= t;
frustum[2][3] /= t;
LOG("FRUSTUM BOTTOM %f %f %f %f\n", frustum[2][0],  frustum[2][1], frustum[2][2], frustum[2][3]);

// Extract the TOP plane
frustum[3][0] = (float)(clip[ 3] - clip[ 1]);
frustum[3][1] = (float)(clip[ 7] - clip[ 5]);
frustum[3][2] = (float)(clip[11] - clip[ 9]);
frustum[3][3] = (float)(clip[15] - clip[13]);
// Normalize the result
t = sqrt( frustum[3][0] * frustum[3][0] + frustum[3][1] * frustum[3][1] + frustum[3][2] * frustum[3][2] );
frustum[3][0] /= t;
frustum[3][1] /= t;
frustum[3][2] /= t;
frustum[3][3] /= t;
LOG("FRUSTUM TOP %f %f %f %f\n", frustum[3][0],  frustum[3][1], frustum[3][2], frustum[3][3]);

// Extract the FAR plane
frustum[4][0] = (float)(clip[ 3] - clip[ 2]);
frustum[4][1] = (float)(clip[ 7] - clip[ 6]);
frustum[4][2] = (float)(clip[11] - clip[10]);
frustum[4][3] = (float)(clip[15] - clip[14]);
// Normalize the result
t = sqrt( frustum[4][0] * frustum[4][0] + frustum[4][1] * frustum[4][1] + frustum[4][2] * frustum[4][2] );
frustum[4][0] /= t;
frustum[4][1] /= t;
frustum[4][2] /= t;
frustum[4][3] /= t;
LOG("FRUSTUM FAR %f %f %f %f\n", frustum[4][0],  frustum[4][1], frustum[4][2], frustum[4][3]);

// Extract the NEAR plane
frustum[5][0] = (float)(clip[ 3] + clip[ 2]);
frustum[5][1] = (float)(clip[ 7] + clip[ 6]);
frustum[5][2] = (float)(clip[11] + clip[10]);
frustum[5][3] = (float)(clip[15] + clip[14]);
// Normalize the result
t = sqrt( frustum[5][0] * frustum[5][0] + frustum[5][1] * frustum[5][1] + frustum[5][2] * frustum[5][2] );
frustum[5][0] /= t;
frustum[5][1] /= t;
frustum[5][2] /= t;
frustum[5][3] /= t;
LOG("FRUSTUM NEAR %f %f %f %f\n", frustum[5][0],  frustum[5][1], frustum[5][2], frustum[5][3]);
}*/

/*Vector3D Graphics3D::getScreenDirection(int x,int y) {
float alpha_x = ((float)x) / ((float)this->width - 1.0);
float alpha_y = ((float)y) / ((float)this->height - 1.0);
Vector3D dir = (1.0 - alpha_x) * dir_left + alpha_x * dir_right
+ (1.0 - alpha_y) * dir_up + alpha_y * dir_down;
return dir;
}*/

/*void unProject(Entity *entity,int x,int y) {
GLint viewport[4];
float proj[16];
float modl[16];

glGetIntegerv(GL_VIEWPORT, viewport);
glGetFloatv( GL_PROJECTION_MATRIX, proj );
glGetFloatv( GL_MODELVIEW_MATRIX, modl );

float winX = (float)x;
float winY = (float)viewport[3] - (float)y;
float winZ = 1.0;

double x, y, z;

gluUnProject(winX, winY, winZ, modl, proj, viewport, &x, &y, &z);
gluUnProject();
}*/

/* Must be called to set the 'world' that this Graphics3D should render.
*/
void Graphics3D::setWorld(World *world) {
	this->world = world;
	genv->getRenderer()->switchToModelview();
	genv->getRenderer()->loadIdentity();
}

/*void Graphics3D::setTexture(int new_textureID) {
if( new_textureID == 0 ) {
glDisable(GL_TEXTURE_2D);
}
else {
glEnable(GL_TEXTURE_2D);
}

if( this->textureID == new_textureID ) {
return;
}
if( new_textureID != 0 ) {
glBindTexture(GL_TEXTURE_2D, new_textureID);
}
this->textureID = new_textureID;
}*/

Renderable::Renderable(Entity *entity) {
	this->entity = entity;
	this->renderData = NULL;
}

RenderQueue::RenderQueue() {
	/*this->renderQueue = new Vector();
	this->renderQueue_lights = new Vector();
	this->renderQueue_infinite = new Vector();
	this->renderQueue_shadowcasters = new Vector();
	this->renderQueue_nonopaque = new Vector();*/
	//this->renderQueue_shadowplanes = new Vector();
	//this->lights = new Entity *[GL_LIGHTMAX - GL_LIGHT0];
	//this->renderQueue_shadowcasters = this->renderQueue;
	//this->renderQueue = new vector<Entity *>();
	//this->renderQueue = new vector<Renderable>();
	this->renderQueue = new RQRenderableCollection();
	/*this->renderQueue_lights = new vector<Entity *>();
	this->renderQueue_infinite = new vector<Entity *>();
	this->renderQueue_shadowcasters = new vector<Entity *>();*/
	this->renderQueue_lights = new RQEntityCollection();
	this->renderQueue_infinite = new RQEntityCollection();
	this->renderQueue_shadowcasters = new RQEntityCollection();
	//this->renderQueue_nonopaque = new vector<Entity *>();
	//this->renderQueue_nonopaque = new vector<Renderable>();
	this->renderQueue_nonopaque = new RQRenderableCollection();
	//this->renderQueue_shadowplanes = new vector<Entity *>();
	this->renderQueue_shadowplanes = new RQShadowplanesCollection();
}

RenderQueue::~RenderQueue() {
	delete this->renderQueue;
	delete this->renderQueue_lights;
	delete this->renderQueue_infinite;
	delete this->renderQueue_shadowcasters;
	delete this->renderQueue_nonopaque;
	delete this->renderQueue_shadowplanes;
	//delete [] this->lights;
}

void RenderQueue::flushRenderQueue() const {
	/*this->renderQueue->removeAll();
	this->renderQueue_lights->removeAll();
	this->renderQueue_infinite->removeAll();
	this->renderQueue_shadowcasters->removeAll();
	this->renderQueue_nonopaque->removeAll();*/
	//this->renderQueue_shadowplanes->removeAll();
	this->renderQueue->clear();
	this->renderQueue_lights->clear();
	this->renderQueue_infinite->clear();
	this->renderQueue_shadowcasters->clear();
	this->renderQueue_nonopaque->clear();
	this->renderQueue_shadowplanes->clear();
}

void RenderQueue::addOpaque(Entity *entity) const {
	Renderable renderable(entity);
	addOpaque(&renderable);
}

void RenderQueue::addOpaque(Renderable *renderable) const {
	this->renderQueue->push_back(*renderable);
}

void RenderQueue::addNonOpaque(Entity *entity) const {
	Renderable renderable(entity);
	addNonOpaque(&renderable);
}

void RenderQueue::addNonOpaque(Renderable *renderable) const {
	this->renderQueue_nonopaque->push_back(*renderable);
}

int RenderQueue::nOpaque() const {
	return this->renderQueue->size();
}

Renderable *RenderQueue::getOpaque(int i) const {
	return &this->renderQueue->at(i);
}

void Graphics3D::buildRenderQueue(SceneGraphNode *aobj) {
	genv->getRenderer()->pushMatrix();
	aobj->multGLMatrix();
	/*if( strcmp(aobj->getName(), "SPELL") == 0 ) {
		VI_log("");
	}*/

	if( aobj->renderRequired(genv) ) {
		aobj->renderQueue(this->c_queue, false);
	}
	else {
		// TODO: have a smaller radius, where we get lighting only if outside of this smaller radius
		// for now we use this to get the shadow casters
		aobj->renderQueue(this->c_queue, true);
		//aobj->renderQueue(this, false);
	}
	//children
	for(int i=0;i<aobj->getNChildren();i++) {
		SceneGraphNode *child = dynamic_cast<SceneGraphNode *>(aobj->getChild(i));
		buildRenderQueue(child);
	}
	genv->getRenderer()->popMatrix();
}

void Graphics3D::buildRenderQueue(Sector *sector) {
	V__ENTER;
	if( sector->spatialindex != NULL ) {
		sector->spatialindex->buildRenderQueue(this->genv);
	}
	else {
		buildRenderQueue(sector->rootObject);
	}
	if( world->getTerrain() != NULL ) {
		TerrainEngine *d_terrain = static_cast<TerrainEngine *>(world->getTerrain()); // cast to the internal representation
		d_terrain->renderQueue(this->c_queue, genv);
	}
	V__EXIT;
}

/* Render 'obj', and all its children.
*/
/*void Graphics3D::doObject(SceneGraphNode *aobj,Graphics3D::RenderPhase phase,bool lighting_pass) {
glPushMatrix();
aobj->multGLMatrix();

if( aobj->renderRequired(this) // needs to be first, as this sets the mvmatrix!
//|| phase == RENDER_LIGHTING
) {
aobj->render(this,phase,lighting_pass,NULL);
}
//children
for(int i=0;i<aobj->getNChildren();i++) {
doObject(aobj->getChildAt(i),phase,lighting_pass);
}
glPopMatrix();
}*/

void Graphics3D::disableLights() {
	/*for(unsigned int i=GL_LIGHT0;i<GL_LIGHTMAX;i++)
	glDisable((GLenum)i);*/
	genv->getRenderer()->disableLights();
}

void Graphics3D::disableLight(int index) {
	genv->getRenderer()->disableLight(index);
}

void Graphics3D::enableOneLight(int n_light,Entity *light_ent) {
	if( getRenderer()->getActiveVertexShader() != NULL || getRenderer()->getActivePixelShader() ) {
		// the engine should be coded so that we only change lights when shaders are not set - so we know that we
		// can set the shader variables for the lights when enabling the shader (as opposed to worrying about
		// updating the variables when we change the light)
		LOG("shouldn't be setting light when shaders enabled!\n");
		ASSERT(false);
	}
	Light *light = dynamic_cast<Light *>(light_ent->getObject());

	// position
	//Vector3D pos = light_ent->getCurrentWorldPos();
	Vector3D pos = light_ent->getLastRenderedEyePos();
	bool infinite = light_ent->getPoint()->isInfinite();
	float p[3];
	p[0] = pos.x;
	p[1] = pos.y;
	p[2] = pos.z;

	//genv->getRenderer()->enableLight(n_light, light, p, infinite);
	float lightAmbient[4], lightDiffuse[4], lightSpecular[4];
	light->getLightAmbient().floatArray(lightAmbient);
	light->getLightDiffuse().floatArray(lightDiffuse);
	light->getLightSpecular().floatArray(lightSpecular);
	const float *lightAttenuation = light->getLightAttenuation();
	genv->getRenderer()->enableLight(n_light, lightAmbient, lightDiffuse, lightSpecular, lightAttenuation, p, infinite);
}

void Graphics3D::enableLight(int index) {
	/*glPushMatrix();
	glLoadIdentity();*/
	genv->getRenderer()->pushMatrix();
	genv->getRenderer()->loadIdentity();
	Entity *light_ent = this->c_queue->renderQueue_lights->at(index);
	enableOneLight(0, light_ent);
	//glPopMatrix();
	genv->getRenderer()->popMatrix();
}

void Graphics3D::enableLights() {
	genv->getRenderer()->pushMatrix();
	genv->getRenderer()->loadIdentity();
	//for(int l_index=0;l_index<this->c_queue->lID-GL_LIGHT0;l_index++) {
	for(unsigned int l_index=0;l_index<this->c_queue->renderQueue_lights->size();l_index++) {
		Entity *light_ent = this->c_queue->renderQueue_lights->at(l_index);
		enableOneLight(l_index, light_ent);
	}
	genv->getRenderer()->popMatrix();
}

void Graphics3D::drawSky(RenderQueue *queue,Entity *portal) {
	V__ENTER;
	//if( world->getAtmosphere() != NULL || world->getSky() != NULL || world->getClouds() != NULL || world->isHorizonEnabled() ) {
	if( world->getAtmosphere() != NULL || world->getSky() != NULL || world->getClouds() != NULL ) {
		Clouds *d_clouds = static_cast<Clouds *>(world->getClouds()); // cast to the internal representation
		SkyBox *d_sky = static_cast<SkyBox *>(world->getSky()); // cast to the internal representation
		Atmosphere *d_atmos = static_cast<Atmosphere *>(world->getAtmosphere()); // cast to the internal representation
		genv->getRenderer()->pushMatrix();

		// so that the viewpoint is at the centre of the viewpoint
		genv->getRenderer()->loadIdentity();

		Rotation3D rot = world->getViewpoint()->getRotation();
		//rot.multGLMatrixTranspose();
		this->getRenderer()->transformInverse(rot.getQuaternion());
		// recurse up the hierarchy, if necessary
		VI_SceneGraphNode *parent = world->getViewpoint()->getParent();
		while(parent != NULL) {
			rot = parent->getRotation();
			//rot.multGLMatrixTranspose();
			this->getRenderer()->transformInverse(rot.getQuaternion());
			parent = parent->getParent();
		}
		if( portal != NULL ) {
			genv->getRenderer()->scale(1.0, -1.0, 1.0); // TODO: hardcoded
		}

		/* Drawing atmosphere before skybox causes bug on NVIDIA when not using shaders!
		* Alpha from atmosphere pixel shader always appears to be set to 1, meaning skybox isn't drawn with alpha blending
		*/
		bool blend = false;
		if( world->getSky() != NULL ) {
			d_sky->render(this->getRenderer());
			blend = true;
		}
		if( world->getClouds() != NULL ) {
			if( genv->hasShaders() ) {
				// need shaders

				blend = true;
				// the clouds need to be blended with the:
				// * skybox if there is one
				// * or the background colour, if there's no atmosphere
				if( world->getSky() != NULL || world->getAtmosphere() == NULL ) {
					//genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_ONE_BY_SRCALPHA);
					genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_BOTH);
				}
				d_clouds->render(this);
				if( d_sky != NULL || d_atmos == NULL ) {
					genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
				}
			}
		}

		if( world->getAtmosphere() != NULL ) {
			if( genv->hasShaders() ) {
				// need shaders
				if( blend ) {
					genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_ONE_BY_SRCALPHA);
				}
				d_atmos->render(this, portal);
				if( blend ) {
					genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
				}
			}
		}

		/*if( world->isHorizonEnabled() ) {
			genv->getRenderer()->enableLighting(false);
			genv->getRenderer()->enableTexturing(false);
			genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_NONE);
			//genv->getRenderer()->enableCullFace(false);
			//genv->getRenderer()->setColor3(world->groundColor.getR(),world->groundColor.getG(),world->groundColor.getB());
			const float h = 1.0;
			//const float d = 1000.0;
			float d = genv->getRenderer()->getZFar();
			float vertex_data[] = {
				-d, -h, -d,
				-d, -h, d,
				d, -h, d,
				d, -h, -d
			};
			const Color *groundColor = world->getGroundColor();
			unsigned char color_data[3] = { groundColor->getR(), groundColor->getG(), groundColor->getB() };
			this->getRenderer()->render(VertexArray::DRAWINGMODE_QUADS, vertex_data, NULL, 4, color_data, false, 4);
			genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READWRITE);
			genv->getRenderer()->enableLighting(true);
			genv->getRenderer()->enableTexturing(true);
			//genv->getRenderer()->enableCullFace(true);
		}*/
		genv->getRenderer()->popMatrix();
	}

	// infinite distance objects
	genv->getRenderer()->pushMatrix();
	for(RQEntityIter iter = queue->renderQueue_infinite->begin();iter != queue->renderQueue_infinite->end();++iter) {
		Entity *node = *iter;
		node->getMatrix()->set(15, 1);
		node->getMatrix()->load();
		node->render(this->genv, this->current_time, ObjectData::RENDER_OPAQUE, false, false, NULL);
	}
	genv->getRenderer()->popMatrix();
	V__EXIT;
}

void Graphics3D::setShaders(Entity *light_ent,ObjectData *obj) {
	V__ENTER;
	/*LOG("light_ent %d\n", light_ent);
	LOG("obj %d\n", obj);
	LOG("vertex_shader_ambient %d\n", vertex_shader_ambient);
	LOG("pixel_shader_ambient %d\n", pixel_shader_ambient);*/
	const Shader *std_vertex_shader = light_ent == NULL ? std_shaders.vertex_shader_ambient : light_ent->getPoint()->isInfinite() ? std_shaders.vertex_shader_directional : std_shaders.vertex_shader_point;
	const Shader *std_pixel_shader = light_ent == NULL ? std_shaders.pixel_shader_ambient : light_ent->getPoint()->isInfinite() ? std_shaders.pixel_shader_directional : std_shaders.pixel_shader_point;
	const Shader *this_vertex_shader = (obj != NULL) ? ( light_ent == NULL ? obj->getVertexShaderAmbient() : light_ent->getPoint()->isInfinite() ? obj->getVertexShaderDirectional() : obj->getVertexShaderPoint() ) : NULL;
	const Shader *this_pixel_shader = (obj != NULL) ? ( light_ent == NULL ? obj->getPixelShaderAmbient() : light_ent->getPoint()->isInfinite() ? obj->getPixelShaderDirectional() : obj->getPixelShaderPoint() ) : NULL;

	if( this_vertex_shader == NULL )
		this_vertex_shader = std_vertex_shader;

	if( this_pixel_shader == NULL )
		this_pixel_shader = std_pixel_shader;

	//this_pixel_shader = NULL;
	//this_vertex_shader = NULL;

	const Shader *c_vertex_shader = getRenderer()->getActiveVertexShader();
	const Shader *c_pixel_shader = getRenderer()->getActivePixelShader();
	/*if( c_pixel_shader == this->std_shaders.water_simple_pixel_shader ) {
		LOG("!!!\n");
	}*/
	if( this_vertex_shader == c_vertex_shader && this_pixel_shader == c_pixel_shader ) {
		V__EXIT;
		return; // quick exit
	}

	if( this_vertex_shader != NULL && this_vertex_shader != c_vertex_shader ) {
		this_vertex_shader->SetUniformParameter1f("time", (float)current_time);
	}
	if( this_pixel_shader != NULL && this_pixel_shader != c_pixel_shader ) {
		this_pixel_shader->SetUniformParameter1f("time", (float)current_time);
		//this_pixel_shader->SetUniformParameter1f("splat_detail_scale", splat_detail_scale_c);
	}

	if( light_ent == NULL ) {
		if( this_vertex_shader != NULL && this_vertex_shader != c_vertex_shader ) {
			//this_vertex_shader->setLightAmbient(world->lightAmbient.getRf(), world->lightAmbient.getGf(),  world->lightAmbient.getBf());
			float col[3];
			world->getAmbience(&col[0], &col[1], &col[2]);
			this_vertex_shader->setLightAmbient(col[0], col[1], col[2]);
			this_vertex_shader->enable();
		}

		if( this_pixel_shader != NULL && this_pixel_shader != c_pixel_shader ) {
			//this_pixel_shader->SetUniformParameter1f("time", (float)current_time);
			this_pixel_shader->enable();
		}
	}
	else
	{
		Light *light = dynamic_cast<Light *>(light_ent->getObject());

		Color ambient = light->getLightAmbient();
		Color diffuse = light->getLightDiffuse();

		Color specular = light->getLightSpecular();

		Vector3D pos = light_ent->getLastRenderedEyePos();
		if( this_vertex_shader != NULL && this_vertex_shader != c_vertex_shader ) {
			if( c_vertex_shader != NULL )
				c_vertex_shader->disable();

			this_vertex_shader->setLightAmbient(ambient.getRf(), ambient.getGf(), ambient.getBf());
			this_vertex_shader->setLightDiffuse(diffuse.getRf(), diffuse.getGf(), diffuse.getBf());
			this_vertex_shader->setLightSpecular(specular.getRf(), specular.getGf(), specular.getBf());
			this_vertex_shader->setLightAtten(light->getLightAttenuation(0), light->getLightAttenuation(1), light->getLightAttenuation(2));
			this_vertex_shader->setLightPos(pos.x, pos.y, pos.z, -pos.x, -pos.y, -pos.z);
			this_vertex_shader->enable();
		}

		if( this_pixel_shader != NULL && this_pixel_shader != c_pixel_shader ) {
			if( c_pixel_shader != NULL )
				c_pixel_shader->disable();

			this_pixel_shader->setLightAmbient(ambient.getRf(), ambient.getGf(), ambient.getBf());
			this_pixel_shader->setLightDiffuse(diffuse.getRf(), diffuse.getGf(), diffuse.getBf());
			this_pixel_shader->setLightSpecular(specular.getRf(), specular.getGf(), specular.getBf());
			this_pixel_shader->setLightAtten(light->getLightAttenuation(0), light->getLightAttenuation(1), light->getLightAttenuation(2));
			this_pixel_shader->setLightPos(pos.x, pos.y, pos.z, -pos.x, -pos.y, -pos.z);
			this_pixel_shader->enable();
		}
	}

	if( this_vertex_shader == NULL && c_vertex_shader != NULL )
		c_vertex_shader->disable();
	if( this_pixel_shader == NULL && c_pixel_shader != NULL )
		c_pixel_shader->disable();
	V__EXIT;
}

/* Render shadow of 'obj', and all its children.
*/
void Graphics3D::doObjectShadow(SceneGraphNode *aobj,ShadowPlane *shadowplane,int light_index) {
	V__ENTER;
	Entity *light = this->c_queue->renderQueue_lights->at(light_index);

	aobj->multGLMatrix();

	if( aobj->getClass() == V_CLASS_ENTITY // yuck
		&& light->collision(aobj) ) {
			Entity *entity = static_cast<Entity *>(aobj);
			ObjectData *d_obj = dynamic_cast<ObjectData *>(entity->getObject()); // cast to the internal representation

			if( d_obj != NULL && d_obj->getShadowPlane() != shadowplane && d_obj->isShadowCaster() ) {
				aobj->render(this->genv, this->current_time, ObjectData::RENDER_STENCIL, false, true, NULL);
			}
	}

	//children
	for(int i=0;i<aobj->getNChildren();i++) {
		genv->getRenderer()->pushMatrix();
		SceneGraphNode *child = dynamic_cast<SceneGraphNode *>(aobj->getChild(i));
		doObjectShadow(child,shadowplane,light_index);
		genv->getRenderer()->popMatrix();
	}
	V__EXIT;
}

void Graphics3D::doProjectiveShadows(Sector *sector) {
	V__ENTER;
	SceneGraphNode *root = sector->rootObject;
	for(RQShadowplanesIter iter = c_queue->renderQueue_shadowplanes->begin();iter != c_queue->renderQueue_shadowplanes->end();++iter) {
		Entity *mirror = *iter;
		ObjectData *d_obj = dynamic_cast<ObjectData *>(mirror->getObject()); // cast to the internal representation
		ShadowPlane *shadow = d_obj == NULL ? NULL : d_obj->getShadowPlane();
		// shadow shouldn't be NULL!
		if( !shadow->isShadow() ) {
			continue; // no shadows - draw later on
		}
		for(unsigned int j=0;j<this->c_queue->renderQueue_lights->size();j++) {
			// do opaque objects' shadows
			genv->getRenderer()->enableLighting(false);
			genv->getRenderer()->enableTexturing(false);
			genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READONLY);

			genv->getRenderer()->enableColorMask(false);
			// mark out regions where shadows are
			genv->getRenderer()->clear(false, false, true);
			genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_REPLACE);
			Entity *light = this->c_queue->renderQueue_lights->at(j);
			float lightPos[4];
			Vector3D lightWorldPos = light->getCurrentWorldPos();
			lightPos[0] = lightWorldPos.x;
			lightPos[1] = lightWorldPos.y;
			lightPos[2] = lightWorldPos.z;
			lightPos[3] = 1.0;
			const Plane *plane = shadow->getPlane();
			//TransformationMatrix *shadowMat = this->genv->getRenderer()->createTransformationMatrix();
			TransformationMatrix *shadowMat = this->genv->getRenderer()->getStaticTransformationMatrix();

			float dot = plane->normal.x * lightPos[0] + plane->normal.y * lightPos[1] + plane->normal.z * lightPos[2] - plane->D * lightPos[3];

			(*shadowMat)[0] = dot - lightPos[0] * plane->normal.x;
			(*shadowMat)[4] = - lightPos[0] * plane->normal.y;
			(*shadowMat)[8] = - lightPos[0] * plane->normal.z;
			(*shadowMat)[12] = lightPos[0] * plane->D;

			(*shadowMat)[1] = - lightPos[1] * plane->normal.x;
			(*shadowMat)[5] = dot - lightPos[1] * plane->normal.y;
			(*shadowMat)[9] = - lightPos[1] * plane->normal.z;
			(*shadowMat)[13] = lightPos[1] * plane->D;

			(*shadowMat)[2] = - lightPos[2] * plane->normal.x;
			(*shadowMat)[6] = - lightPos[2] * plane->normal.y;
			(*shadowMat)[10] = dot - lightPos[2] * plane->normal.z;
			(*shadowMat)[14] = lightPos[2] * plane->D;

			(*shadowMat)[3] = - lightPos[3] * plane->normal.x;
			(*shadowMat)[7] = - lightPos[3] * plane->normal.y;
			(*shadowMat)[11] = - lightPos[3] * plane->normal.z;
			(*shadowMat)[15] = dot + lightPos[3] * plane->D;

			genv->getRenderer()->pushMatrix();
			shadowMat->mult();
			doObjectShadow(root,shadow,j);
			genv->getRenderer()->popMatrix();

			genv->getRenderer()->enableLighting(true);
			genv->getRenderer()->enableTexturing(true);
			genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READWRITE);

			// enable new light
			enableLight(j);
			float col[4] = {0.0, 0.0, 0.0, 0.0};
			genv->getRenderer()->setAmbient(col);

			genv->getRenderer()->enableColorMask(true);
			// only draw where stencil isn't; disallow overdraw
			genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_DRAWATNOSTENCIL);

			genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_ADDITIVE);

			// now draw in the mirror with lighting
			genv->getRenderer()->setDepthBufferFunc(Renderer::RENDERSTATE_DEPTHFUNC_EQUAL);
			genv->getRenderer()->pushMatrix();
			mirror->getMatrix()->load();
			ObjectData *d_obj = dynamic_cast<ObjectData *>(mirror->getObject()); // cast to the internal representation
			setShaders(light, d_obj);
			mirror->render(this->genv, this->current_time, ObjectData::RENDER_STENCIL, false, false, NULL);
			genv->getRenderer()->popMatrix();

			genv->getRenderer()->setDepthBufferFunc(Renderer::RENDERSTATE_DEPTHFUNC_LESS);
			genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);

			//world->lightAmbient.floatArray(col);
			world->getAmbience(&col[0], &col[1], &col[2]);
			col[3] = 1.0f;
			genv->getRenderer()->setAmbient(col);
			disableLight(j);
			//delete shadowMat;

			// disable shaders
			const Shader *c_vertex_shader = getRenderer()->getActiveVertexShader();
			const Shader *c_pixel_shader = getRenderer()->getActivePixelShader();
			if( c_vertex_shader != NULL )
				c_vertex_shader->disable();
			if( c_pixel_shader != NULL )
				c_pixel_shader->disable();
		}
		genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_NONE);
	}
	V__EXIT;
}

void Graphics3D::drawMirrors() {
	V__ENTER;
	// draw mirrors for stencil method (done separately as we need to enable blending)
	for(RQShadowplanesIter iter = c_queue->renderQueue_shadowplanes->begin();iter != c_queue->renderQueue_shadowplanes->end();++iter) {
		Entity *mirror = *iter;
		ObjectData *d_obj = dynamic_cast<ObjectData *>(mirror->getObject()); // cast to the internal representation
		ShadowPlane *shadow = d_obj == NULL ? NULL : d_obj->getShadowPlane();
		// shadow shouldn't be NULL!
		//shadow->shadow = false;
		if( shadow->isReflection() && !shadow->isRenderToTexture() ) {
			// need to blend in with reflection
			genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
		}
		//if( use_shadow_volumes || !shadow->shadow )
		{
			/*// if using projected shadows, we won't be using shaders
			// NB - bug is that projected shadows won't work with render_to_texture mode!*/
			ObjectData *d_obj = dynamic_cast<ObjectData *>(mirror->getObject()); // cast to the internal representation
			setShaders(NULL, d_obj);
		}
		genv->getRenderer()->pushMatrix();
		mirror->getMatrix()->load();
		mirror->render(this->genv, this->current_time, ObjectData::RENDER_STENCIL, false, false, NULL);
		genv->getRenderer()->popMatrix();
		if( shadow->isReflection() && !shadow->isRenderToTexture() ) {
			genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
		}
	}
	const Shader *c_vertex_shader = getRenderer()->getActiveVertexShader();
	const Shader *c_pixel_shader = getRenderer()->getActivePixelShader();
	if( c_vertex_shader != NULL )
		c_vertex_shader->disable();
	if( c_pixel_shader != NULL )
		c_pixel_shader->disable();
	V__EXIT;
}


int sortRenderQueue(const void *ptr1, const void *ptr2) {
	/*Entity *e1 = *(Entity **)ptr1;
	Entity *e2 = *(Entity **)ptr2;*/
	Renderable *p1 = (Renderable *)ptr1;
	Renderable *p2 = (Renderable *)ptr2;
	Entity *e1 = p1->entity;
	Entity *e2 = p2->entity;

	/*
	// reduce texture swaps
	ObjectData *obj1 = e1->getObject();
	ObjectData *obj2 = e2->getObject();
	if( obj1 < obj2 )
	return -1;
	else if( obj1 > obj2 )
	return 1;*/

	// reduce shader swaps
	ObjectData *obj1 = dynamic_cast<ObjectData *>(e1->getObject()); // cast to the internal representation
	ObjectData *obj2 = dynamic_cast<ObjectData *>(e2->getObject()); // cast to the internal representation
	if( obj1->getVertexShaderAmbient() < obj2->getVertexShaderAmbient() )
		return -1;
	else if( obj1->getVertexShaderAmbient() > obj2->getVertexShaderAmbient() )
		return 1;
	else if( obj1->getVertexShaderDirectional() < obj2->getVertexShaderDirectional() )
		return -1;
	else if( obj1->getVertexShaderDirectional() > obj2->getVertexShaderDirectional() )
		return 1;
	else if( obj1->getVertexShaderPoint() < obj2->getVertexShaderPoint() )
		return -1;
	else if( obj1->getVertexShaderPoint() > obj2->getVertexShaderPoint() )
		return 1;
	else if( obj1->getPixelShaderAmbient() < obj2->getPixelShaderAmbient() )
		return -1;
	else if( obj1->getPixelShaderAmbient() > obj2->getPixelShaderAmbient() )
		return 1;
	else if( obj1->getPixelShaderDirectional() < obj2->getPixelShaderDirectional() )
		return -1;
	else if( obj1->getPixelShaderDirectional() > obj2->getPixelShaderDirectional() )
		return 1;
	else if( obj1->getPixelShaderPoint() < obj2->getPixelShaderPoint() )
		return -1;
	else if( obj1->getPixelShaderPoint() > obj2->getPixelShaderPoint() )
		return 1;

	// reduce texture swaps
	if( p1->renderData != NULL && p2->renderData != NULL ) {

		if( p1->renderData->texture < p2->renderData->texture ) {
			return -1;
		}
		else if( p1->renderData->texture > p2->renderData->texture ) {
			return 1;
		}
	}
	else if( p1->renderData == NULL && p2->renderData != NULL ) {
		return -1;
	}
	else if( p1->renderData != NULL && p2->renderData == NULL ) {
		return 1;
	}

	// sort by Z distance
	/*float dist1 = -e1->mvmatrix[14];
	float dist2 = -e2->mvmatrix[14];*/
	float dist1 = - e1->getMatrix()->get(14);
	float dist2 = - e2->getMatrix()->get(14);
	/*dist1 = - dist1;
	dist2 = - dist2;*/
	if( dist1 < dist2 )
		return -1;
	else if( dist1 > dist2 )
		return 1;

	/*if( ptr1 < ptr2 )
	return -1;
	else if( ptr1 > ptr2 )
	return 1;*/
	return 0;
}

bool stl_sortRenderQueue(const Renderable& p1, const Renderable& p2) {
	return sortRenderQueue(&p1, &p2) < 0;
}

void Graphics3D::drawFog() {
	const Fog *fog = this->world->getFog();
	if( fog->isEnabled() ) {
		if( fog->fogtype == Fog::FOG_LINEAR ) {
			GraphicsEnvironment::getSingleton()->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_LINEAR, fog->start, fog->end, fog->density);
		}
		else if( fog->fogtype == Fog::FOG_EXP ) {
			GraphicsEnvironment::getSingleton()->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_EXP, fog->start, fog->end, fog->density);
		}
		else if( fog->fogtype == Fog::FOG_EXP2 ) {
			GraphicsEnvironment::getSingleton()->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_EXP2, fog->start, fog->end, fog->density);
		}
		float fog_rgba[4] = {0.0f, 0.0f, 0.0f, 0.0f};
		fog->floatArray(fog_rgba);
		GraphicsEnvironment::getSingleton()->getRenderer()->setFogColor(fog_rgba);
	}
	else {
		GraphicsEnvironment::getSingleton()->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_NONE, 0.0f, 0.0f, 0.0f);
	}
}

void Graphics3D::drawOpaque(int recursion) {
	//LOG("drawOpaque enter: %d\n", recursion);
	V__ENTER;
	_DEBUG_CHECK_ERROR_;
	//genv->checkGLError("Graphics3D::drawOpaque BEGIN");
	//unsigned int n_shadows = c_queue->renderQueue_shadowplanes->size();
	//qsort(queue->renderQueue->getData(), queue->renderQueue->size(), sizeof(void *), sortRenderQueue);
	//qsort(c_queue->renderQueue->begin(), c_queue->renderQueue->size(), sizeof(void *), sortRenderQueue);
	//qsort(&*c_queue->renderQueue->begin(), c_queue->renderQueue->size(), sizeof(Renderable), sortRenderQueue);
	//LOG(">>> %d\n", c_queue->renderQueue->size());
	sort(c_queue->renderQueue->begin(), c_queue->renderQueue->end(), stl_sortRenderQueue);
	{
		// enable fog for ambient pass
		//world->getFog()->draw();
		drawFog();

		// enable vertex shader
		// no pixel shader for ambient
		// should be drawing with ambient light only already (N.B. - mirrors are already drawn in ambient)
		genv->getRenderer()->pushMatrix();
		for(int i=0;i<c_queue->nOpaque();i++) {
			Renderable *renderable = &c_queue->renderQueue->at(i);
			Entity *node = renderable->entity;
			//ObjectData *obj = node->getObject();
			ObjectData *d_obj = dynamic_cast<ObjectData *>(node->getObject()); // cast to the internal representation
			setShaders(NULL, d_obj);
			node->getMatrix()->load();
			_DEBUG_CHECK_ERROR_;
			node->render(this->genv, this->current_time, ObjectData::RENDER_OPAQUE, false, false, renderable->renderData);
			_DEBUG_CHECK_ERROR_;
		}
		genv->getRenderer()->popMatrix();
		{
			const Shader *c_vertex_shader = getRenderer()->getActiveVertexShader();
			const Shader *c_pixel_shader = getRenderer()->getActivePixelShader();
			if( c_vertex_shader != NULL )
				c_vertex_shader->disable();
			if( c_pixel_shader != NULL )
				c_pixel_shader->disable();
		}
		if( world->getFog()->isEnabled() )
			genv->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_NONE, 0.0f, 0.0f, 0.0f);

		//return; // uncomment for ambient only
		//LOG(">>> %d\n", c_queue->renderQueue_lights->size());

		_DEBUG_CHECK_ERROR_;
		for(unsigned int j=0;j<this->c_queue->renderQueue_lights->size();j++) {
			//continue;

			Entity *light_ent = this->c_queue->renderQueue_lights->at(j);
			/*if( !light_ent->getPoint()->isInfinite() )
			continue;*/

			bool infinite = light_ent->getPoint()->isInfinite();
			Vector3D lightEyePos = light_ent->getLastRenderedEyePos(); // position in eye space
			//LOG("%f , %f, %f\n", lightEyePos.x, lightEyePos.y, lightEyePos.z);
			float light_radius = light_ent->getBoundingRadius();
			/*if( !infinite && ( -lightEyePos.z + light_radius < this->z_near ||
			-lightEyePos.z - light_radius > this->z_far ) ) {
			// should have been clipped away?
			continue;
			}*/
			bool scissor_test = false;
			int scissor_x0, scissor_y0, scissor_x1, scissor_y1;
			//if( false ) {
			if( !infinite && abs(lightEyePos.z) > V_TOL_LINEAR ) {
				scissor_test = true;
				Vector3D scissor_left = lightEyePos;
				Vector3D scissor_right = lightEyePos;
				Vector3D scissor_bottom = lightEyePos;
				Vector3D scissor_top = lightEyePos;
				scissor_left.x -= light_radius;
				scissor_right.x += light_radius;
				scissor_bottom.y -= light_radius;
				scissor_top.y += light_radius;
				/*MouseInputHandler *mouse = this->genv->getInputSystem()->getMouse();
				Vector3D mouse_eye = this->getDirectionOfPoint(mouse->mouseX, mouse->mouseY);
				scissor_x0 = 0;
				scissor_y0 = 0;
				this->eyeToScreen(&scissor_x1, &scissor_y1, &mouse_eye);*/
				int dummy;
				genv->eyeToScreen(&scissor_x0, &dummy, &scissor_left);
				genv->eyeToScreen(&scissor_x1, &dummy, &scissor_right);
				genv->eyeToScreen(&dummy, &scissor_y1, &scissor_bottom);
				genv->eyeToScreen(&dummy, &scissor_y0, &scissor_top);
				if( scissor_x0 > scissor_x1 )
					swap(scissor_x0, scissor_x1);
				else if( scissor_x0 == scissor_x1 )
					scissor_test = false;
				if( scissor_y0 > scissor_y1 )
					swap(scissor_y0, scissor_y1);
				else if( scissor_y0 == scissor_y1 )
					scissor_test = false;
				/*assert(scissor_x1 > scissor_x0);
				assert(scissor_y1 > scissor_y0);*/
				/*scissor_y0 = this->height - scissor_y0;
				scissor_y1 = this->height - scissor_y1;*/
			}
			scissor_test = false;
			if( scissor_test ) {
				//genv->getRenderer()->enableScissor(0, 0, this->width/2.0, this->height/2.0);
				//LOG("%d , %d , %d", scissor_y0, scissor_y1, scissor_y1-scissor_y0);
				// remember to convert to OpenGL coordinates
				//genv->getRenderer()->enableScissor(scissor_x0, this->height-scissor_y1, scissor_x1-scissor_x0, scissor_y1-scissor_y0);
				genv->getRenderer()->enableScissor(scissor_x0, genv->getHeight()-scissor_y1, scissor_x1-scissor_x0, scissor_y1-scissor_y0);
			}

			if( use_shadow_volumes && c_queue->renderQueue_shadowcasters->size() > 0 )
			{
				genv->getRenderer()->clear(false, false, true);
				// draw shadow volumes
				genv->getRenderer()->enableTexturing(false);
				genv->getRenderer()->enableLighting(false);
				genv->getRenderer()->enableColorMask(false);
				genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READONLY);

				//bool infinite = light_ent->getPoint()->isInfinite();
				//Vector3D lightEyePos = light_ent->getCurrentEyePos(); // position in eye space
				if( infinite ) {
					lightEyePos.normalise();
				}

				/*if( vertex_shader_shadow_point != NULL ) {
				vertex_shader_shadow_point->enable();
				}*/

				genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_SHADOWPASS_ONE);

				// 1st pass
				genv->getRenderer()->pushMatrix();
				for(unsigned int i=0;i<c_queue->renderQueue_shadowcasters->size();i++) {
					Entity *node = c_queue->renderQueue_shadowcasters->at(i);
					if( !light_ent->collision(node) ) {
						continue;
					}
					// TODO: clipping for (point) shadow volumes
					if( infinite ) {
						Sphere boundingSphere = node->getObject()->getBoundingSphere();
						Vector3D centre = node->localToEyePos(&boundingSphere.centre);
						//if( !frustum.infiniteShadowVolumeInFrustum(centre.x, centre.y, centre.z, boundingSphere.radius, -lightEyePos.x, -lightEyePos.y, -lightEyePos.z) ) {
						if( !genv->getFrustum()->infiniteShadowVolumeInFrustum(centre.x, centre.y, centre.z, boundingSphere.radius, -lightEyePos.x, -lightEyePos.y, -lightEyePos.z) ) {
							//printf("");
							continue;
						}
					}
					// need to convert the lightpos into the node's local space
					Vector3D lightPos = lightEyePos;
					//LOG("***\n");
					//LOG("%f , %f, %f\n", lightPos.x, lightPos.y, lightPos.z);

					// TODO: improve performance
					Matrix m;
					float values[16];
					for(int indx=0;indx<16;indx++) {
						values[indx] = node->getMatrix()->get(indx);
						//LOG(">>> matrix[%d] = %f\n", indx, values[indx]);
					}
					m.setOpenGLTranspose(values);
					//m.setOpenGLTranspose(node->mvmatrix);

					if( !infinite ) {
						Vector3D nodeEyePos = node->getLastRenderedEyePos();
						//LOG("eye %f , %f, %f\n", nodeEyePos.x, nodeEyePos.y, nodeEyePos.z);
						lightPos -= nodeEyePos;
						//LOG("%f , %f, %f\n", lightPos.x, lightPos.y, lightPos.z);
					}
					m.transformVector(&lightPos);
					node->getMatrix()->load();
					//LOG("%f , %f, %f\n", lightPos.x, lightPos.y, lightPos.z);
					node->castShadow(this->getRenderer(), this->genv, &lightPos, infinite);
				}
				genv->getRenderer()->popMatrix();

				if( genv->getRenderer()->nShadowPasses() == 1 ) {
					genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_SHADOWPASS_OFF);
				}
				//else if( false ) {
				else {
					// 2nd pass
					genv->getRenderer()->setFrontFace( recursion % 2 != 0 );
					genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_SHADOWPASS_TWO);
					genv->getRenderer()->pushMatrix();
					for(unsigned int i=0;i<c_queue->renderQueue_shadowcasters->size();i++) {
						Entity *node = c_queue->renderQueue_shadowcasters->at(i);
						if( !light_ent->collision(node) ) {
							continue;
						}
						// TODO: clipping for shadow volumes
						// need to convert the lightpos into the node's local space
						Vector3D lightPos = lightEyePos;

						// TODO: improve performance
						Matrix m;
						float values[16];
						for(int indx=0;indx<16;indx++)
							values[indx] = node->getMatrix()->get(indx);
						m.setOpenGLTranspose(values);
						//m.setOpenGLTranspose(node->mvmatrix);

						if( !infinite ) {
							Vector3D nodeEyePos = node->getLastRenderedEyePos();
							lightPos -= nodeEyePos;
						}
						m.transformVector(&lightPos);
						node->getMatrix()->load();
						node->castShadow(this->getRenderer(), this->genv, &lightPos, infinite);
					}
					genv->getRenderer()->popMatrix();
					genv->getRenderer()->setFrontFace( recursion % 2 == 0 );
				}

				genv->getRenderer()->enableTexturing(true);
				genv->getRenderer()->enableLighting(true);
				genv->getRenderer()->enableColorMask(true);
				genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READWRITE);

				{
					const Shader *c_vertex_shader = getRenderer()->getActiveVertexShader();
					const Shader *c_pixel_shader = getRenderer()->getActivePixelShader();
					if( c_vertex_shader != NULL )
						c_vertex_shader->disable();
					if( c_pixel_shader != NULL )
						c_pixel_shader->disable();
				}
			}

			float col[4] = {0.0, 0.0, 0.0, 0.0};
			genv->getRenderer()->setAmbient(col);
			enableLight(j);
			//genv->getRenderer()->enableColorMask(true);
			if( use_shadow_volumes && c_queue->renderQueue_shadowcasters->size() > 0 )
			{
				// only draw where stencil isn't; disallow overdraw
				genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_DRAWATNOSTENCIL);
			}
			genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_ADDITIVE);
			genv->getRenderer()->setDepthBufferFunc(Renderer::RENDERSTATE_DEPTHFUNC_EQUAL);
			genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READONLY);
			/*if( scissor_test ) {
			//genv->getRenderer()->enableScissor(0, 0, this->width/2.0, this->height/2.0);
			genv->getRenderer()->enableScissor(scissor_x0, scissor_y0, scissor_x1-scissor_x0, scissor_y1-scissor_y0);
			}*/

			// draw in mirrors with lighting
			for(RQShadowplanesIter iter = c_queue->renderQueue_shadowplanes->begin();iter != c_queue->renderQueue_shadowplanes->end();++iter) {
				Entity *mirror = *iter;
				if( !light_ent->collision(mirror) ) {
					continue;
				}
				ObjectData *d_obj = dynamic_cast<ObjectData *>(mirror->getObject()); // cast to the internal representation
				ShadowPlane *shadow = d_obj == NULL ? NULL : d_obj->getShadowPlane();
				// projected-shadow-receivers have already been drawn
				if( !shadow->isReflection() || use_shadow_volumes ) {
					ObjectData *d_obj = dynamic_cast<ObjectData *>(mirror->getObject()); // cast to the internal representation
					setShaders(light_ent, d_obj);
					genv->getRenderer()->pushMatrix();
					mirror->getMatrix()->load();
					mirror->render(this->genv, this->current_time, ObjectData::RENDER_STENCIL, true, false, NULL);
					genv->getRenderer()->popMatrix();
				}
			}
			{
				const Shader *c_vertex_shader = getRenderer()->getActiveVertexShader();
				const Shader *c_pixel_shader = getRenderer()->getActivePixelShader();
				if( c_vertex_shader != NULL )
					c_vertex_shader->disable();
				if( c_pixel_shader != NULL )
					c_pixel_shader->disable();
			}

			// draw in opaque objects with lighting
			genv->getRenderer()->pushMatrix();
			for(RQRenderableIter iter = c_queue->renderQueue->begin();iter != c_queue->renderQueue->end();++iter) {
				Renderable *renderable = &(*iter);
				Entity *node = renderable->entity;
				if( !light_ent->collision(node) ) {
					continue;
				}
				node->getMatrix()->load();
				//ObjectData *obj = node->getObject();
				ObjectData *d_obj = dynamic_cast<ObjectData *>(node->getObject()); // cast to the internal representation
				/*if( obj->vertex_shader_ambient != NULL || obj->pixel_shader_ambient != NULL ) {
				printf("");
				continue; // custom shaders don't support lighting atm
				}*/
				setShaders(light_ent, d_obj);
				node->render(this->genv, this->current_time, ObjectData::RENDER_OPAQUE, true, false, renderable->renderData);
			}
			genv->getRenderer()->popMatrix();

			if( scissor_test )
			{
				genv->getRenderer()->disableScissor();
			}
			genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
			genv->getRenderer()->setDepthBufferFunc(Renderer::RENDERSTATE_DEPTHFUNC_LESS);
			genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READWRITE);
			//world->lightAmbient.floatArray(col);
			world->getAmbience(&col[0], &col[1], &col[2]);
			col[3] = 1.0f;
			genv->getRenderer()->setAmbient(col);
			disableLight(j);
			{
				const Shader *c_vertex_shader = getRenderer()->getActiveVertexShader();
				const Shader *c_pixel_shader = getRenderer()->getActivePixelShader();
				if( c_vertex_shader != NULL )
					c_vertex_shader->disable();
				if( c_pixel_shader != NULL )
					c_pixel_shader->disable();
			}
		}
		if( use_shadow_volumes && c_queue->renderQueue_shadowcasters->size() > 0 )
		{
			genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_NONE);
		}

	}
	_DEBUG_CHECK_ERROR_;
	V__EXIT;
	//LOG("drawOpaque exit: %d\n", recursion);
}

void Graphics3D::drawTransparent() {
	//return;
	// [
	/*if( vertex_shader_ambient != NULL ) {
	cgGLSetParameter4f( vertex_shader_ambient->cgparam_AmbLight, world->lightAmbient.getRf(), world->lightAmbient.getGf(),  world->lightAmbient.getBf(), 1.0f );
	vertex_shader_ambient->enable();
	}*/
	// no pixel shader for ambient
	// ]
	if( c_queue->renderQueue_nonopaque->size() == 0 ) {
		return; // quick exit
	}

	V__ENTER;
	// currently done using fixed pipeline, unless custom shader specified
	enableLights();
	genv->getRenderer()->pushMatrix();
	genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READONLY);
	for(RQRenderableIter iter = c_queue->renderQueue_nonopaque->begin();iter != c_queue->renderQueue_nonopaque->end();++iter) {
		Entity *node = (*iter).entity;
		node->getMatrix()->load();

		ObjectData *obj = dynamic_cast<ObjectData *>(node->getObject()); // cast to the internal representation
		{
			// if no custom shaders, then use the fixed pipeline
			const Shader *c_vertex_shader = getRenderer()->getActiveVertexShader();
			const Shader *c_pixel_shader = getRenderer()->getActivePixelShader();
			const Shader *this_vertex_shader = (obj != NULL) ? obj->getVertexShaderAmbient() : NULL;
			if( this_vertex_shader != c_vertex_shader ) {
				if( c_vertex_shader != NULL )
					c_vertex_shader->disable();

				if( this_vertex_shader != NULL ) {
					//this_vertex_shader->setLightAmbient(world->lightAmbient.getRf(), world->lightAmbient.getGf(),  world->lightAmbient.getBf());
					float col[3];
					world->getAmbience(&col[0], &col[1], &col[2]);
					this_vertex_shader->setLightAmbient(col[0], col[1], col[2]);
					this_vertex_shader->SetUniformParameter1f("time", (float)current_time);
					this_vertex_shader->enable();
				}
			}
			const Shader *this_pixel_shader = (obj != NULL) ? obj->getPixelShaderAmbient() : NULL;
			if( this_pixel_shader != c_pixel_shader ) {
				if( c_pixel_shader != NULL )
					c_pixel_shader->disable();

				if( this_pixel_shader != NULL ) {
					//this_pixel_shader->setLightAmbient(world->lightAmbient.getRf(), world->lightAmbient.getGf(),  world->lightAmbient.getBf());
					float col[3];
					world->getAmbience(&col[0], &col[1], &col[2]);
					this_pixel_shader->setLightAmbient(col[0], col[1], col[2]);
					this_pixel_shader->SetUniformParameter1f("time", (float)current_time);
					//this_pixel_shader->SetUniformParameter1f("splat_detail_scale", splat_detail_scale_c);
					this_pixel_shader->enable();
				}
			}
		}

		node->render(this->genv, this->current_time, ObjectData::RENDER_NONOPAQUE, false, false, NULL);

		if( obj->getShadowPlane() != NULL && obj->getShadowPlane()->isRenderToTexture() && genv->hasShaders() ) {
			// do lighting pass
			for(RQEntityIter iter = this->c_queue->renderQueue_lights->begin();iter != this->c_queue->renderQueue_lights->end();++iter) {
				Entity *light_ent = *iter;
				if( !light_ent->collision(node) ) {
					continue;
				}
				setShaders(light_ent, obj);
				node->render(this->genv, this->current_time, ObjectData::RENDER_NONOPAQUE, true, false, NULL);
			}
		}
	}
	genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
	genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READWRITE);
	genv->getRenderer()->popMatrix();
	const Shader *c_vertex_shader = getRenderer()->getActiveVertexShader();
	const Shader *c_pixel_shader = getRenderer()->getActivePixelShader();
	if( c_vertex_shader != NULL )
		c_vertex_shader->disable();
	if( c_pixel_shader != NULL )
		c_pixel_shader->disable();
	V__EXIT;
}

inline float sgn(float a) {
	if (a > 0.0F) return (1.0F);
	if (a < 0.0F) return (-1.0F);
	return (0.0F);
}

void Graphics3D::setClipPlane(const double pl[4]) {
	genv->getRenderer()->enableClipPlane(pl);

	/*if( !genv->hasShaders() ) {
	// no shaders
	return;
	}

	// user clip planes seem to be supported for fixed pipeline, but not always
	// when using shaders (e.g., NVIDIA 8600GT, Intel GMA950).
	// we need to set with _both_ methods, because a scene can mix shaders and non-shaders!

	TransformationMatrix *model_matrix = this->genv->getRenderer()->createTransformationMatrix();

	genv->getRenderer()->pushMatrix();
	genv->getRenderer()->translate((float)(pl[3]*pl[0]), (float)(pl[3]*pl[1]), (float)(pl[3]*pl[2]));
	model_matrix->save();
	genv->getRenderer()->popMatrix();
	Vector3D new_p = model_matrix->getPos();

	genv->getRenderer()->pushMatrix();
	genv->getRenderer()->translate((float)pl[0], (float)pl[1], (float)pl[2]);
	model_matrix->save();
	genv->getRenderer()->popMatrix();
	Vector3D new_n = model_matrix->getPos();
	model_matrix->save();
	new_n.x -= (*model_matrix)[12];
	new_n.y -= (*model_matrix)[13];
	new_n.z -= (*model_matrix)[14];
	assert( new_n.magnitude() > 0 );
	new_n.normalise();
	delete model_matrix;

	{
	// shader clipping
	this->shader_clipping = true;
	this->shader_clipping_plane[0] = new_n.x;
	this->shader_clipping_plane[1] = new_n.y;
	this->shader_clipping_plane[2] = new_n.z;
	this->shader_clipping_plane[3] = new_n % new_p;
	return;
	}*/

	// oblique code

	/*float matrix[16];
	// Grab the current projection matrix from OpenGL
	glGetFloatv(GL_PROJECTION_MATRIX, matrix);

	pl[0] = new_n.x;
	pl[1] = new_n.y;
	pl[2] = new_n.z;
	pl[3] = new_n % new_p;
	//pl[3] = 2;

	//pl[3] = 10;
	//assert( pl[3] <= 0 );
	//glGetFloatv(GL_MODELVIEW_INVERSE_MATRIX, model_matrix);

	// Calculate the clip-space corner point opposite the clipping plane
	// as (sgn(clipPlane.x), sgn(clipPlane.y), 1, 1) and
	// transform it into camera space by multiplying it
	// by the inverse of the projection matrix
	float q_x = (sgn((float)pl[0]) + matrix[8]) / matrix[0];
	float q_y = (sgn((float)pl[1]) + matrix[9]) / matrix[5];
	float q_z = -1.0f;
	float q_w = (1.0f + matrix[10]) / matrix[14];

	// Calculate the scaled plane vector
	//Vector4D c = clipPlane * (2.0F / Dot(clipPlane, q));
	float dot = q_x * (float)pl[0] + q_y * (float)pl[1] + q_z * (float)pl[2] + q_w * (float)pl[3];
	float scale = 2.0f / dot;

	//scale *= -5.0;
	//scale = 100;
	// Replace the third row of the projection matrix
	matrix[2] = (float)pl[0] * scale;
	matrix[6] = (float)pl[1] * scale;
	matrix[10] = (float)pl[2] * scale + 1.0f;
	//matrix[10] = pl[2] * scale - 1.0F;
	//matrix[10] = 0;
	//matrix[10] = pl[2] * scale;
	matrix[14] = (float)pl[3] * scale;

	//matrix[2] = matrix[10] = 0;
	//matrix[14] = - matrix[6];

	// Load it back into OpenGL
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadMatrixf(matrix);
	glMatrixMode(GL_MODELVIEW);
	//trace_extractFrustum();
	*/
}

void Graphics3D::unsetClipPlane() {
	genv->getRenderer()->disableClipPlane();

	//this->shader_clipping = false;
	return;
}

void Graphics3D::resetClipPlane() {
	genv->getRenderer()->reenableClipPlane();
	/*if( genv->hasShaders() ) {
	this->shader_clipping = true;
	}*/
}

void Graphics3D::drawSector(Sector *sector,Entity *portal,int recursion) {
	V__ENTER;
	_DEBUG_CHECK_ERROR_;
	SceneGraphNode *root = sector->rootObject;
	const int MAX_RECURSION = 1;
	RenderQueue *queue = new RenderQueue(); // TODO: preallocate
	this->c_queue = queue;

	//if( recursion < MAX_RECURSION && world->getShadowNodes()->size() > 0 ) {
	if( recursion < MAX_RECURSION ) {
		//root->getShadows(this,shadows,&n_shadows,MAX_SHADOWS,portal);
		root->getShadows(this->c_queue, genv, portal);
	}

	TerrainEngine *d_terrain = static_cast<TerrainEngine *>(world->getTerrain()); // cast to the internal representation
	if( recursion < MAX_RECURSION && d_terrain != NULL && d_terrain->getSeaReflection() != NULL ) {
		// TODO: improve
		//shadows[n_shadows++] = world->terrain->entity_sea_dummy;
		//this->c_queue->renderQueue_shadowplanes->addIfAbsent(world->terrain->entity_sea_dummy);
		/*if( find(this->c_queue->renderQueue_shadowplanes->begin(), this->c_queue->renderQueue_shadowplanes->end(), world->terrain->entity_sea_dummy) == this->c_queue->renderQueue_shadowplanes->end() ) {
		this->c_queue->renderQueue_shadowplanes->push_back(world->terrain->entity_sea_dummy);
		}*/
		//addIfAbsent_vec(this->c_queue->renderQueue_shadowplanes, world->terrain->entity_sea_dummy);
		this->c_queue->renderQueue_shadowplanes->insert(d_terrain->getSeaReflection());
		if( d_terrain->getSeaReflection()->getMatrix() == NULL ) {
			d_terrain->getSeaReflection()->setMatrix( this->getRenderer()->createTransformationMatrix() );
		}
		//world->terrain->entity_sea_dummy->getShadows(this,shadows,&n_shadows,MAX_SHADOWS,portal);
	}

	//int n_shadows = c_queue->renderQueue_shadowplanes->size();
	if( recursion >= MAX_RECURSION ) {
		c_queue->renderQueue_shadowplanes->clear();
		//n_shadows = 0;
	}

	// mirrors/reflections - render to texture method (water)
	if( c_queue->renderQueue_shadowplanes->size() > 0 && genv->hasShaders() && this->hasWaterShaders() ) {
		bool set_res = false;
		//for(unsigned int i=0;i<c_queue->renderQueue_shadowplanes->size();i++) {
		for(RQShadowplanesIter iter = c_queue->renderQueue_shadowplanes->begin();iter != c_queue->renderQueue_shadowplanes->end();++iter) {
			// determine visibility of shadowplane
			Entity *mirror = *iter;
			ObjectData *d_obj = dynamic_cast<ObjectData *>(mirror->getObject()); // cast to the internal representation
			ShadowPlane *shadow = d_obj == NULL ? NULL : d_obj->getShadowPlane();
			// shadow shouldn't be NULL!
			if( !shadow->isReflection() || !shadow->isRenderToTexture() ) {
				continue;
			}
			if( !shadow->needUpdate() ) {
				continue;
			}
			//continue;

			//if( false )
			{
				//if( !Vision::framebufferObject ) {
				if( !set_res ) {
					// first time - buffers already cleared
					int texture_size = this->getRenderToTextureSize();
					genv->getRenderer()->setViewport(texture_size, texture_size);
					set_res = true;
				}
				else {
					// need to reclear buffers
					this->clearFrame();
				}
			}

			const Plane *plane = shadow->getPlane();
			double pl[4] = { -plane->normal.x, -plane->normal.y, -plane->normal.z, -plane->D };
			setClipPlane(pl);

			genv->getRenderer()->pushMatrix();
			Vector3D point = plane->normal * plane->D;
			genv->getRenderer()->translate(point.x, point.y, point.z);
			genv->getRenderer()->scale(1.0, -1.0, 1.0); // TODO: transformation is hardcoded atm
			genv->getRenderer()->translate(-point.x, -point.y, -point.z);

			genv->getRenderer()->setFrontFace(false); // TODO: fix for multiple recursions
			drawSector(sector, mirror, recursion+1);
			this->c_queue = queue; // need to reset!
			genv->getRenderer()->setFrontFace(true); // TODO: fix for multiple recursions
			genv->getRenderer()->popMatrix();

			unsetClipPlane();

			//if( false )
			{
				int texture_size = this->getRenderToTextureSize();
				shadow->getRenderTexture()->copyTo(texture_size, texture_size);
				// need to clear frame again
			}
		}
		if( set_res )
		{
			this->setViewport();
			// need to reclear frame
			this->clearFrame();
		}
	}
	/*if( recursion == 0 )
	return;*/

	if( recursion > 0 ) {
		unsetClipPlane();
	}
	//genv->getRenderer()->disableFog(); // don't render fog on the skybox etc
	if( world->getFog()->isEnabled() )
		genv->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_NONE, 0.0f, 0.0f, 0.0f); // don't render fog on the skybox etc
	drawSky(queue, portal);
	if( recursion > 0 ) {
		resetClipPlane();
	}

	// mirrors/reflections - stencil method
	/*if( n_shadows > 1 )
	n_shadows = 1; // only one mirror for now*/
	//if( false )
	for(RQShadowplanesIter iter = c_queue->renderQueue_shadowplanes->begin();iter != c_queue->renderQueue_shadowplanes->end();++iter) {
		// determine visibility of shadowplane
		Entity *mirror = *iter;
		ObjectData *d_obj = dynamic_cast<ObjectData *>(mirror->getObject()); // cast to the internal representation
		ShadowPlane *shadow = d_obj == NULL ? NULL : d_obj->getShadowPlane();
		// shadow shouldn't be NULL!
		if( !shadow->isReflection() || shadow->isRenderToTexture() ) {
			continue;
		}
		//continue;

		// do opaque objects' reflections
		{
			// mark out where to draw reflections
			genv->getRenderer()->clear(false, false, true);
			genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_REPLACE);
			genv->getRenderer()->enableColorMask(false);
			genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_NONE);

			genv->getRenderer()->pushMatrix();
			mirror->getMatrix()->load();
			mirror->render(this->genv, this->current_time, ObjectData::RENDER_STENCIL, false, false, NULL);
			genv->getRenderer()->popMatrix();

			genv->getRenderer()->enableColorMask(true);
			genv->getRenderer()->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_READWRITE);
			// only draw where stencil is; allow overdraw
			genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_DRAWATSTENCIL_OVERDRAW);
		}
		genv->getRenderer()->pushMatrix();
		const Plane *plane = shadow->getPlane();
		const double pl[4] = { -plane->normal.x, -plane->normal.y, -plane->normal.z, -plane->D };
		setClipPlane(pl);
		Vector3D point = plane->normal * plane->D;
		genv->getRenderer()->translate(point.x, point.y, point.z);
		//genv->getRenderer()->translate(0, 0.1, 0);
		genv->getRenderer()->scale(1.0f, -1.0f, 1.0f); // TODO: transformation is hardcoded atm
		genv->getRenderer()->translate(-point.x, -point.y, -point.z);
		genv->getRenderer()->setFrontFace(false); // TODO: fix for multiple recursions
		drawSector(sector, mirror, recursion+1);
		this->c_queue = queue; // need to reset!
		genv->getRenderer()->setFrontFace(true); // TODO: fix for multiple recursions
		unsetClipPlane();
		genv->getRenderer()->popMatrix();
		genv->getRenderer()->setStencilMode(Renderer::RENDERSTATE_STENCIL_NONE);
	}

	buildRenderQueue(sector);

	if( recursion > 0 ) {
		/*
		Shadow volumes disabled on recursion, due to conflicts with stencil buffer usage.
		Really needs using render to texture for recursion to work.
		Note that recursing 1 level works with shadow volumes, but it's a bit of a hack:
		the objects are drawn in ambient pass with the reflection stencil still enabled,
		and only after is it blown away. But when we have the lighting pass, due to drawing
		with z-equal, we will only draw where we are supposed to. This falls apart for 2
		levels of recursion!
		TODO: okay for render-to-texture?
		*/
		queue->renderQueue_shadowcasters->clear();
	}

	// only ambient lighting for now
	disableLights();

	drawMirrors();

	if( !use_shadow_volumes ) {
		doProjectiveShadows(sector);
	}

	// do opaque objects
	drawOpaque(recursion);

	// reenable fog for the rest
	//world->getFog()->draw();
	this->drawFog();

	// do transparent objects
	drawTransparent();

	delete queue;
	this->c_queue = NULL;
	_DEBUG_CHECK_ERROR_;
	V__EXIT;
}

/*void Graphics3D::setSpecular(const float specular[4]) {
	genv->getRenderer()->setSpecular(specular);
}

void Graphics3D::setSpecularOff() {
	genv->getRenderer()->setSpecularOff();
}*/

/* Dummy function - used when things aren't working too well..
*/
void test() {
	/*
	//glDisable(GL_CULL_FACE);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_COLOR);
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);						// Enables Depth Testing
	glEnable(GL_COLOR_MATERIAL);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	GLfloat light[4] = {1.0f,1.0f,1.0f,1.0f};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT,light);
	glLoadIdentity();
	glRotatef(180.0f,0.0f,1.0f,0.0f);
	glTranslatef(-3.0f,-1.8f,-3.0f);
	//glTranslatef(-3.0f,-5.0f,-3.0f);

	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE); // explosion effect

	glNormal3f(0.0f,0.0f,1.0f);
	//GLfloat c[4] = {0.0,1.0,0.0,1.0};
	//glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE,c);
	//glMaterialfv(GL_FRONT,GL_EMISSION,black);
	glColor4f(1.0f,0.0f,0.0f,1.0f);
	//glBegin(GL_QUADS);
	glBegin(GL_POLYGON);
	glVertex3f(0,4,0+16);
	glVertex3f(16,4,0+16);
	glVertex3f(16,4,16+16);
	glVertex3f(0,4,16+16);
	glEnd();

	//glBlendFunc(GL_ONE, GL_ZERO);
	//glDisable(GL_BLEND);
	*/
}

void Graphics3D::set3DMode(bool mode_3d) {
	genv->getRenderer()->enableCullFace(mode_3d);
	genv->getRenderer()->enableTexturing(mode_3d);
	genv->getRenderer()->enableLighting(mode_3d);
	genv->getRenderer()->setDepthBufferMode(mode_3d ? Renderer::RENDERSTATE_DEPTH_READWRITE : Renderer::RENDERSTATE_DEPTH_NONE);
	genv->getRenderer()->setColor4(255, 255, 255, 255); // make sure that color is reset, both on enter and exit
}

/* Render the frame!
*/
void Graphics3D::drawFrame() {
	//LOG("drawFrame\n");
	V__ENTER;
	_DEBUG_CHECK_ERROR_;

	const int frame_count = 20;
	if( frames == 0 ) {
		frame_time = clock();
	}
	else if( frames % frame_count == 0 ) {
		int new_frame_time = clock();
		int time = new_frame_time - frame_time;
		if(time != 0)
			fps = (frame_count*1000.0f)/((float)time);
		frame_time = new_frame_time;
	}
	frames++;
	this->current_time = clock();

	//test();
	//return;
	/* We disable the early exit with NULL world, as this somehow messes up the alt-tab behaviour on Direct3D9 with "Conquests" on fullscreen mode - seems to
	   be a problem with DrawPrimitiveUP calls being messed up. We should investigate getting rid of DrawPrimitiveUP calls to see if that helps, but in the
	   meantime we disable this early exit. (It's probably better to avoid exiting earlier anyway, for more consistent behaviour.)
	   The problem shows up both on Intel GMA (Windows XP), and NVIDIA 8600 (Windows 7). Problem occurs with both SDL and non-SDL.
	*/
	/*if( world == NULL ) {
		return;
	}*/
	if( world != NULL && Vision::isLocked() ) {
		// we allow Vision to be locked if no world is set (needed for Dungeon)
		Vision::setError(new VisionException(this, VisionException::V_LOCKED, "Graphics3D::drawFrame() not allowed when locked"));
	}
	//return; // test

	const int width = genv->getWidth();
	const int height = genv->getHeight();

	//hdr_enabled = false;
	if( this->hdr_enabled ) {
		//genv->getRenderer()->bindHdrFbo();
		this->hdr_fbo->bind();
		genv->getRenderer()->pushAttrib(true);
		this->setViewport();
		this->clearFrame();
	}

	_DEBUG_CHECK_ERROR_;
	genv->getRenderer()->switchToProjection();
	genv->getRenderer()->popMatrix();
	genv->getRenderer()->switchToModelview();

	_DEBUG_CHECK_ERROR_;

	this->set3DMode(true);

	_DEBUG_CHECK_ERROR_;

	genv->getRenderer()->enableMultisample(true); // GL_MULTISAMPLE_ARB has no effect on crappy ATI, but this is needed for other cards
	_DEBUG_CHECK_ERROR_;

	/*
	// disabled for now, as causes OpenGL error on non-NVIDIA cards - how to test if this hint is available?
	GLenum hints = GL_NICEST;
	glHint(GL_MULTISAMPLE_FILTER_HINT_NV, hints);
	*/

	_DEBUG_CHECK_ERROR_;

	if( world != NULL ) {
		// ambient light
		float col[4];
		//world->lightAmbient.floatArray(col);
		world->getAmbience(&col[0], &col[1], &col[2]);
		col[3] = 1.0f;
		genv->getRenderer()->setAmbient(col);
	}

	_DEBUG_CHECK_ERROR_;

	// do viewpoint
	genv->getRenderer()->loadIdentity();
	if( world != NULL ) {
		//world->viewPoint->multGLMatrixInverse();
		SceneGraphNode *vp = dynamic_cast<SceneGraphNode *>(world->getViewpoint()); // cast to internal representation
		vp->multGLMatrixInverse();
		// recurse up the hierarchy, if necessary
		SceneGraphNode *parent = dynamic_cast<SceneGraphNode *>(world->getViewpoint()->getParent()); // cast to internal representation
		while(parent != NULL) {
			parent->multGLMatrixInverse();

			//parent = parent->getParent();
			parent = dynamic_cast<SceneGraphNode *>(parent->getParent()); // cast to internal representation
		}
	}
	if( this->worldToEyeMatrix == NULL ) {
		this->worldToEyeMatrix = this->getRenderer()->createTransformationMatrix();
	}
	this->worldToEyeMatrix->save();

	if( world != NULL ) {
		drawSector(world->getSector(), NULL, 0);
	}

	if( genv->renderFunc != NULL ) {
		(*genv->renderFunc)();
	}

	// reset
	genv->getRenderer()->switchToProjection();
	genv->getRenderer()->pushMatrix();
	genv->getRenderer()->ortho2D(genv->getWidth(), genv->getHeight());
	genv->getRenderer()->switchToModelview();
	genv->getRenderer()->loadIdentity();

	this->set3DMode(false);

	if( this->hdr_enabled ) {
		genv->getRenderer()->popAttrib();
		genv->getRenderer()->unbindFramebuffer();
		genv->getRenderer()->enableTexturing(true);

		//genv->getRenderer()->enableHdrFbo();
		this->hdr_fbo->enableTexture();
		//bloom_enabled = false;

		if( bloom_enabled ) {
			this->std_shaders.blurw_pixel_shader->enable();
			this->getRenderer()->getActivePixelShader()->SetUniformParameter1f("tx_step_x", 1.0f/(float)width);

			// first of a two-phase process, so we must direct the output into the temp framebuffer
			this->temp_fbo->bind();
			genv->getRenderer()->pushAttrib(true);
			this->setViewport();
			this->clearFrame();
		}
		else {
			// we need a vertex shader, otherwise this doesn't work on Direct3D9 / NVIDIA 8600.
			this->std_shaders.hdr_vertex_shader->enable();
			this->std_shaders.hdr_pixel_shader->enable();
		}

		//this->blurw_pixel_shader->enable();
		//this->blurh_hdr_pixel_shader->enable();
		//this->c_pixel_shader->SetUniformParameter1f("tx_step_y", 1.0f/(float)height);

		float texcoord_data[] = {
			0.0f, 1.0f,
			1.0f, 1.0f,
			1.0f, 0.0f,
			0.0f, 0.0f
		};
		float texcoord_data_inv[] = {
			0.0f, 0.0f,
			1.0f, 0.0f,
			1.0f, 1.0f,
			0.0f, 1.0f
		};
		/*float vertex_data[] = {
			0.0f, 0.0f, 0.0f,
			(float)width, 0.0f, 0.0f,
			(float)width, (float)height, 0.0f,
			0.0f, (float)height, 0.0f
		};*/
		// the hdr vertex shader doesn't transform vertex, so we hand the transformed vertices in
		float vertex_data[] = {
			-1.0f, 1.0f, 0.0f,
			1.0f, 1.0f, 0.0f,
			1.0f, -1.0f, 0.0f,
			-1.0f, -1.0f, 0.0f
		};
		bool inverted = this->getRenderer()->isYTextureInverted();
		this->getRenderer()->render(VertexArray::DRAWINGMODE_QUADS, vertex_data, inverted ? texcoord_data_inv : texcoord_data, 4);

		this->getRenderer()->getActivePixelShader()->disable();

		if( bloom_enabled ) {
			genv->getRenderer()->popAttrib();
			genv->getRenderer()->unbindFramebuffer();

			// second phase - render from temp_fbo into screen
			this->temp_fbo->enableTexture();
			this->std_shaders.blurh_hdr_pixel_shader->enable();

			this->getRenderer()->getActivePixelShader()->SetUniformParameter1f("tx_step_y", 1.0f/(float)height);

			this->getRenderer()->render(VertexArray::DRAWINGMODE_QUADS, vertex_data, texcoord_data, 4);

			this->getRenderer()->getActivePixelShader()->disable();
		}

		this->getRenderer()->getActiveVertexShader()->disable();
		genv->getRenderer()->enableTexturing(false);

	}

	// disable so that 2D/text rendering is sharp
	genv->getRenderer()->enableMultisample(false); // GL_MULTISAMPLE_ARB has no effect on crappy ATI, but this is needed for other cards
	V__EXIT;
}

void Graphics3D::clearFrame() {
	V__ENTER;
	if( world != NULL ) {
		if( world->getSky() == NULL ) {
			if( world->getAtmosphere() ) {
				if( world->getClouds() != NULL ) {
					// since we'll be blending the atmosphere over the clouds, and the cloud layer doesn't fill the screen, we need a black background first!
					genv->getRenderer()->setClearColor(0.0,0.0,0.0,0.0);
					genv->getRenderer()->clear(true, true, false);
				}
				else {
					// atmosphere fills the screen
					genv->getRenderer()->clear(false, true, false);
				}
			}
			else {
				// no sky, no atmosphere: use user specified back colour
				const Color *backColor = world->getBackColor();
				genv->getRenderer()->setClearColor(backColor->getRf(),backColor->getGf(),backColor->getBf(),0.0);
				genv->getRenderer()->clear(true, true, false);
			}
		}
		else {
			genv->getRenderer()->clear(false, true, false);
		}
	}
	else {
		genv->getRenderer()->clear(true, true, false);
	}
	V__EXIT;
}

int Graphics3D::getRenderToTextureSize() const {
	int size = 512;
	// mustn't be greater than the window width/height, as we currently work by rendering to the window, then copying to a texture!
	while( size > genv->getWidth() )
		size /= 2;
	while( size > genv->getHeight() )
		size /= 2;
	return size;
}

/*Vector2D Graphics3D::getScreenSizeAtDist(float z) const {
	double H = 2.0 * tan( M_PI * fovy / 360.0 ) * z;
	//double W = H * this->width / this->height; // ?
	double W = H * genv->getWidth() / genv->getHeight(); // ?
	return Vector2D((float)W, (float)H);
}

Vector3D Graphics3D::getDirectionOfPoint(int x,int y) const {
	const int width = genv->getWidth();
	const int height = genv->getHeight();
	double sx = x - width / 2.0;
	double sy = height / 2.0 - y;
	double Z = genv->getRenderer()->getZNear();
	double H = 2.0 * tan( M_PI * fovy / 360.0 ) * Z;
	double Y = sy * H / height;
	double W = H * width / height; // ?
	double X = sx * W / width;
	Vector3D vec((float)X,(float)Y,(float)-Z);
	vec.normalise();
	return vec;
}

Vector3D Graphics3D::getDirectionOfRotatedPoint(Rotation3D *rot,int x,int y) const {
	Vector3D vec = this->getDirectionOfPoint(x,y);
	vec.rotateBy(rot);
	return vec;
}

void Graphics3D::eyeToScreen(int *x,int *y,const Vector3D *eye) const {
	const int width = genv->getWidth();
	const int height = genv->getHeight();
	double Z = genv->getRenderer()->getZNear();
	double scale = - Z / eye->z;
	double X = eye->x * scale;
	double Y = eye->y * scale;
	double H = tan( M_PI * fovy / 360.0 ) * Z;
	double W = H * width / height; // ?
	double sx = X * width / (2.0 * W);
	double sy = Y * height / (2.0 * H);
	*x = (int)sx + width / 2;
	*y = height/2 - (int)sy;

}*/

void Graphics3D::init() {
	_DEBUG_CHECK_ERROR_;

	genv->getRenderer()->setToDefaults();
	genv->getRenderer()->clear(true, true, true);
	this->set3DMode(false);

	if( genv->hasShaders() && genv->wantStdShaders() )
	{
		// ambient (always per vertex)
		this->std_shaders.vertex_shader_ambient = Shader::createShader(genv, "shaders/simple.cg", "main_ambient_vs", true);
		this->std_shaders.pixel_shader_ambient = Shader::createShader(genv, "shaders/simple.cg", "main_ambient_ps", false);

		// directional per vertex
		//this->vertex_shader_directional = Shader::createShader(genv, "shaders/simple.cg", "main_directional_vs", true);
		// directional per pixel
		this->std_shaders.vertex_shader_directional = Shader::createShader(genv, "shaders/simple.cg", "main_perpixel_vs", true);
		this->std_shaders.pixel_shader_directional = Shader::createShader(genv, "shaders/simple.cg", "main_directional_ps", false);

		// point per vertex
		//this->vertex_shader_point = Shader::createShader(genv, "shaders/simple.cg", "main_point_vs", true);
		// point per pixel
		//this->vertex_shader_point = Shader::createShader(genv, "shaders/simple.cg", "main_perpixel_vs", true);
		this->std_shaders.vertex_shader_point = this->std_shaders.vertex_shader_directional;
		this->std_shaders.pixel_shader_point = Shader::createShader(genv, "shaders/simple.cg", "main_point_ps", false);

		if( std_shaders.vertex_shader_ambient != NULL && std_shaders.pixel_shader_ambient != NULL &&
			std_shaders.vertex_shader_directional != NULL && std_shaders.pixel_shader_directional != NULL &&
			std_shaders.vertex_shader_point != NULL && std_shaders.pixel_shader_point != NULL ) {
				std_shaders.has_general = true;
		}
		else {
			// we need all of the general shaders to work together (we can't mix and match shaders, due to z-buffering issues etc)
			// also remember that some of these general shaders are required for some of the more advanced effects below
			LOG("failed to load one or more basic shaders - disabling all of them\n");
			delete std_shaders.vertex_shader_ambient;
			std_shaders.vertex_shader_ambient = NULL;
			delete std_shaders.pixel_shader_ambient;
			std_shaders.pixel_shader_ambient = NULL;
			delete std_shaders.vertex_shader_directional;
			std_shaders.vertex_shader_directional = NULL;
			delete std_shaders.pixel_shader_directional;
			std_shaders.pixel_shader_directional = NULL;
			delete std_shaders.vertex_shader_point;
			std_shaders.vertex_shader_point = NULL;
			delete std_shaders.pixel_shader_point;
			std_shaders.pixel_shader_point = NULL;
		}

		// bump shader
		if( std_shaders.has_general && this->genv->getVisionPrefs()->bumpmapping ) {
			// need general shaders for ambient
			this->std_shaders.bump_vertex_shader_directional = Shader::createShader(genv, "shaders/simple.cg", "main_perpixel_bump_vs", true);
			this->std_shaders.bump_pixel_shader_directional = Shader::createShader(genv, "shaders/simple.cg", "main_bump_directional_ps", false);
			this->std_shaders.bump_vertex_shader_point = this->std_shaders.bump_vertex_shader_directional;
			this->std_shaders.bump_pixel_shader_point = Shader::createShader(genv, "shaders/simple.cg", "main_bump_point_ps", false);

			if( std_shaders.bump_vertex_shader_directional != NULL && std_shaders.bump_pixel_shader_directional != NULL &&
				std_shaders.bump_vertex_shader_point != NULL && std_shaders.bump_pixel_shader_point != NULL ) {
					std_shaders.has_bump = true;
			}
		}
		//this->bump_pixel_shader_directional = this->bump_pixel_shader_point;

		// shadow shaders
		//this->vertex_shader_shadow_point = Shader::createShader(genv, "shaders/simple.cg", "main_shadow_point_vs", true);

		// water shaders
		if( this->genv->getVisionPrefs()->watereffects && this->genv->getVisionPrefs()->api != VISIONAPI_OPENGL ) {
			// full replacement set of shaders, so general shaders not required
			// n.b., disabled on OpenGL, as using the water shaders causes a driver crash on Intel GMA 3100 - see Vision Text (8, 5) (we also used to have problems on Intel GMA 950, when using power function for fresnel)
			this->std_shaders.water_vertex_shader_ambient = Shader::createShader(genv, "shaders/simple.cg", "main_refl_vs", true);
			this->std_shaders.water_pixel_shader_ambient = Shader::createShader(genv, "shaders/simple.cg", "main_refl_ps", false);
			this->std_shaders.water_vertex_shader_directional = this->std_shaders.water_vertex_shader_ambient;
			this->std_shaders.water_pixel_shader_directional = Shader::createShader(genv, "shaders/simple.cg", "main_refl_directional_ps", false);
			this->std_shaders.water_vertex_shader_point = this->std_shaders.water_vertex_shader_ambient;
			this->std_shaders.water_pixel_shader_point = Shader::createShader(genv, "shaders/simple.cg", "main_refl_point_ps", false);

			if( std_shaders.water_vertex_shader_ambient != NULL && std_shaders.water_pixel_shader_ambient != NULL &&
				std_shaders.water_vertex_shader_directional != NULL && std_shaders.water_pixel_shader_directional != NULL &&
				std_shaders.water_vertex_shader_point != NULL && std_shaders.water_pixel_shader_point != NULL ) {
					std_shaders.has_water = true;
			}
			else if( std_shaders.has_general ) {
				// try to load simple water shader instead
				// we need general shader here (for the vertex shader)
				this->std_shaders.water_simple_pixel_shader = Shader::createShader(genv, "shaders/simple.cg", "main_water_ps", false);
			}
		}
		else if( std_shaders.has_general ) {
			// but we do need general shader here (for the vertex shader)
			this->std_shaders.water_simple_pixel_shader = Shader::createShader(genv, "shaders/simple.cg", "main_water_ps", false);
		}

		// splatting shaders
		if( std_shaders.has_general && this->genv->getVisionPrefs()->terrainsplatting ) {
			// need general shaders here (for the vertex shader)
			this->std_shaders.splat_pixel_shader_ambient = Shader::createShader(genv, "shaders/simple.cg", "main_splat_ambient_ps", false);
			this->std_shaders.splat_pixel_shader_directional = Shader::createShader(genv, "shaders/simple.cg", "main_splat_directional_ps", false);
			this->std_shaders.splat_pixel_shader_point = Shader::createShader(genv, "shaders/simple.cg", "main_splat_point_ps", false);
			if( std_shaders.splat_pixel_shader_ambient != NULL &&
				std_shaders.splat_pixel_shader_directional != NULL &&
				std_shaders.splat_pixel_shader_point != NULL ) {
					std_shaders.has_splat = true;
			}
		}

		// hdr shader
		if( RendererInfo::framebufferObject ) {
			// don't need general shaders
			this->std_shaders.hdr_vertex_shader = Shader::createShader(genv, "shaders/effects.cg", "main_hdr_vs", true);
			this->std_shaders.hdr_pixel_shader = Shader::createShader(genv, "shaders/effects.cg", "main_hdr_ps", false);
			// TODO: fix blur shaders for D3D9 renderer:
			//this->std_shaders.blurw_pixel_shader = Shader::createShader(genv, "shaders/effects.cg", "main_blurw_ps", false);
			//this->std_shaders.blurh_hdr_pixel_shader = Shader::createShader(genv, "shaders/effects.cg", "main_blurh_hdr_ps", false);
		}

		// cloud shaders
		// don't need general shaders
		this->std_shaders.pixel_shader_clouds = Shader::createShader(genv, "shaders/atmosphere.cg", "main_clouds_ps", false);

		// atmosphere shaders
		if( this->genv->getVisionPrefs()->atmosphere ) {
			// don't need general shaders
			this->std_shaders.vertex_shader_skyfromatmos = Shader::createShader(genv, "shaders/atmosphere.cg", "main_vs", true);
			this->std_shaders.pixel_shader_skyfromatmos = Shader::createShader(genv, "shaders/atmosphere.cg", "main_ps", false);
			if( std_shaders.vertex_shader_skyfromatmos != NULL && std_shaders.pixel_shader_skyfromatmos != NULL ) {
				this->std_shaders.has_atmosphere = true;
			}
		}

		/*if( !genv->hasShaders() ) {
			// something must have gone wrong, so to avoid risk of incompatible/untested combinations of shaders, we turn off shaders altogether
			this->freeShaders();                        
		}*/
	}

	_DEBUG_CHECK_ERROR_;

}

/*void Graphics3D::getViewOrientation(Vector3D *right,Vector3D *up) const {
//TransformationMatrix mat;
TransformationMatrix *mat = this->genv->getRenderer()->createTransformationMatrix();
mat->save();
right->set((*mat)[0],(*mat)[4],(*mat)[8]);
up->set((*mat)[1],(*mat)[5],(*mat)[9]);
delete mat;
}

void Graphics3D::getViewDirection(Vector3D *dir) const {
TransformationMatrix *mat = this->genv->getRenderer()->createTransformationMatrix();
mat->save();
dir->set((*mat)[2],(*mat)[6],(*mat)[10]);
delete mat;
}*/

/*void Fog::draw() const {
	if(this->enabled) {
		if( this->fogtype == FOG_LINEAR ) {
			GraphicsEnvironment::getSingleton()->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_LINEAR, start, end, density);
		}
		else if( this->fogtype == FOG_EXP ) {
			GraphicsEnvironment::getSingleton()->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_EXP, start, end, density);
		}
		else if( this->fogtype == FOG_EXP2 ) {
			GraphicsEnvironment::getSingleton()->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_EXP2, start, end, density);
		}
		float fog[4] = {0.0f, 0.0f, 0.0f, 0.0f};
		this->floatArray(fog);
		GraphicsEnvironment::getSingleton()->getRenderer()->setFogColor(fog);
	}
	else {
		GraphicsEnvironment::getSingleton()->getRenderer()->setFogMode(Renderer::RENDERSTATE_FOG_NONE, 0.0f, 0.0f, 0.0f);
	}
}*/
