//---------------------------------------------------------------------------

#include <cassert>

#include "GLGraphicsEnvironment.h"
#include "GLRenderer.h"
#include "../GraphicsEngine.h"
#include "../Texture.h"

#include "../Win32/Win32InputHandler.h"

#include <cg/cggl.h>

#include "../wglext.h"

//---------------------------------------------------------------------------

// WGL_ARB_pixel_format
PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = NULL;

/*
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB = NULL;
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB = NULL;

PFNGLGENOCCLUSIONQUERIESNVPROC    glGenOcclusionQueriesNV    = NULL;
PFNGLDELETEOCCLUSIONQUERIESNVPROC glDeleteOcclusionQueriesNV = NULL;
PFNGLGETOCCLUSIONQUERYUIVNVPROC   glGetOcclusionQueryuivNV   = NULL;
PFNGLBEGINOCCLUSIONQUERYNVPROC    glBeginOcclusionQueryNV    = NULL;
PFNGLENDOCCLUSIONQUERYNVPROC      glEndOcclusionQueryNV      = NULL;

PFNGLLOCKARRAYSEXTPROC glLockArraysEXT = NULL;
PFNGLUNLOCKARRAYSEXTPROC glUnlockArraysEXT = NULL;

PFNGLGENBUFFERSARBPROC glGenBuffersARB = NULL;// VBO Name Generation Procedure
PFNGLBINDBUFFERARBPROC glBindBufferARB = NULL;// VBO Bind Procedure
PFNGLBUFFERDATAARBPROC glBufferDataARB = NULL;// VBO Data Loading Procedure
PFNGLBUFFERSUBDATAARBPROC glBufferSubDataARB = NULL;
PFNGLDELETEBUFFERSARBPROC glDeleteBuffersARB = NULL;// VBO Deletion Procedure

PFNGLDRAWRANGEELEMENTSPROC glDrawRangeElementsEXT = NULL;

PFNGLSTENCILOPSEPARATEATIPROC glStencilOpSeparateATI = NULL;
PFNGLSTENCILFUNCSEPARATEATIPROC glStencilFuncSeparateATI = NULL;

PFNGLACTIVESTENCILFACEEXTPROC glActiveStencilFaceEXT = NULL;

PFNGLBINDRENDERBUFFEREXTPROC glBindRenderbufferEXT = NULL;
PFNGLDELETERENDERBUFFERSEXTPROC glDeleteRenderbuffersEXT = NULL;
PFNGLGENRENDERBUFFERSEXTPROC glGenRenderbuffersEXT = NULL;
PFNGLRENDERBUFFERSTORAGEEXTPROC glRenderbufferStorageEXT = NULL;
PFNGLBINDFRAMEBUFFEREXTPROC glBindFramebufferEXT = NULL;
PFNGLDELETEFRAMEBUFFERSEXTPROC glDeleteFramebuffersEXT = NULL;
PFNGLGENFRAMEBUFFERSEXTPROC glGenFramebuffersEXT = NULL;
PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC glCheckFramebufferStatusEXT = NULL;
PFNGLFRAMEBUFFERTEXTURE2DEXTPROC glFramebufferTexture2DEXT = NULL;
PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbufferEXT = NULL;

*/

//const int bits = 16;
const int bits = 32;

//GLGraphicsEnvironment::GLGraphicsEnvironment(int width,int height,int multisample,bool fullscreen,bool hdr) : GraphicsEnvironment(width, height, multisample, fullscreen) {
//GLGraphicsEnvironment::GLGraphicsEnvironment(VisionPrefs *visionPrefs) : GraphicsEnvironment(visionPrefs->resolution_width, visionPrefs->resolution_height, visionPrefs->multisample, visionPrefs->fullscreen) {
GLGraphicsEnvironment::GLGraphicsEnvironment(VisionPrefs visionPrefs) : GraphicsEnvironment(visionPrefs), hWnd(NULL) {
	LOG("startup OpenGL GraphicsEnvironment\n");
	this->hInstance = GetModuleHandle(NULL);

	/*WNDCLASS wc;
	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style =  CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = ( WNDPROC )GLGraphicsEnvironment::WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = TEXT("Vision3_main");
	if( !RegisterClass( &wc ) ) {
		LOG("Failed to register class\n");
		return;
	}*/

	this->hDC = NULL;
	this->g = new GLGraphics2D(this);
	this->g->owned = true; // owned by this GraphicsEnvironment class
	this->g3 = new Graphics3D(this, visionPrefs.hdr);
	//this->renderer = new GLRenderer(this->g3);
	this->renderer = new GLRenderer(this);
	//this->inputHandler = new InputHandler();
	this->inputSystem = new InputSystem(new Win32MouseInputHandler(), new KeyboardInputHandler());
	this->bActive = false;

	this->hRC = NULL;
	this->cgContext = NULL;
	this->cgVertexProfile = CG_PROFILE_UNKNOWN;
	this->cgPixelProfile = CG_PROFILE_UNKNOWN;

	openScreen(true, false); // don't try for multisample the first time[, as we may be able to do it via HDR]
	initialiseOpenGL(true);

	this->created = true;
}

GLGraphicsEnvironment::~GLGraphicsEnvironment() {
	checkError("GraphicsEnvironment::~GraphicsEnvironment enter");
	// no need to call ChangeDisplaySettings
	/*if(visionPrefs.fullscreen) {
		ChangeDisplaySettings(NULL,0); // need to do this before deleting anything
		//ShowCursor(TRUE);
	}*/

	delete g;
	delete g3;
	delete renderer;
	delete inputSystem;

	if( cgContext != NULL ) {
		cgDestroyContext(cgContext);
		cgContext = NULL;
	}
	closeScreen();

	/*int i=0;
	for(i=0;i<n_ptr;i++) {
	if(ptr[i] == this) {
	ptr[i] = NULL;
	break;
	}
	}
	bool empty = true;
	for(i=0;i<n_ptr;i++) {
	if(ptr[i] != NULL) {
	empty = false;
	break;
	}
	}
	if(empty) {
	delete [] ptr;
	ptr = NULL;
	n_ptr = 0;
	}*/
}

void GLGraphicsEnvironment::openScreen(bool createContext, bool try_multisample) {
	LOG("open screen...\n");

	//Vision::initGLExtensions(hInstance);

	WNDCLASS wc;
	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style =  CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = ( WNDPROC )GLGraphicsEnvironment::WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = TEXT("Vision3_main");
	if( !RegisterClass( &wc ) ) {
		Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_OPEN_WINDOW,"Failed to register class"));
		return;
	}

	//const int multisample = 0;
	//const int multisample = bits > 16 ? 2 : 0; // multisampling seems to require 32 bits
	if( visionPrefs.fullscreen ) {
		setFullscreen();
	}

	DWORD dwExStyle;
	DWORD dwStyle;
	RECT WindowRect;

	WindowRect.left=(long)0;
	WindowRect.right=(long)visionPrefs.resolution_width;
	WindowRect.top=(long)0;
	WindowRect.bottom=(long)visionPrefs.resolution_height;

	if ( visionPrefs.fullscreen ) {
		dwExStyle=WS_EX_APPWINDOW;
		//dwStyle = NULL;
		dwStyle=WS_POPUP;
		//ShowCursor(FALSE);
	}
	else {
		dwExStyle=WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		dwStyle=WS_OVERLAPPEDWINDOW;
	}
	AdjustWindowRectEx(&WindowRect, dwStyle, FALSE, dwExStyle); // Adjust Window To True Requested Size

	LOG("about to create window\n");
	// create main window
	if( !( hWnd = CreateWindowEx( dwExStyle,		// Extended Style For The Window
		TEXT("Vision3_main"),						// Class Name
		TEXT("Vision 3 window"),	// Window Title
		WS_CLIPSIBLINGS |							// Required Window Style
		WS_CLIPCHILDREN |							// Required Window Style
		dwStyle,									// Selected Window Style
		0, 0,										// Window Position
		WindowRect.right-WindowRect.left,			// Calculate Adjusted Window Width
		WindowRect.bottom-WindowRect.top,			// Calculate Adjusted Window Height
		NULL,										// No Parent Window
		NULL,										// No Menu
		hInstance,									// Instance
		NULL ) ) )					 				// Don't Pass Anything To WM_CREATE
	{
		closeScreen();
		Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_OPEN_WINDOW,"Could not create window"));
		return;
	}
	LOG("created window OK\n");

	//Vision::initGLExtensions(hInstance);

	// get the device context (DC)
	hDC = GetDC( hWnd );
	if( hDC == NULL ) {
		closeScreen();
		Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_OPEN_WINDOW,"Could not obtain DC"));
		return;
	}

	/*const int colorBits = bits > 16 ? 24 : bits;
	const int alphaBits = bits > 16 ? 8 : 0;*/
	const int depthBits = 24;
	const int stencilBits = 8;
	PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	//pfd.cAlphaBits = alphaBits;
	//pfd.cColorBits = colorBits;
	pfd.cColorBits = bits;
	pfd.cDepthBits = depthBits;
	pfd.cStencilBits = stencilBits;
	//pfd.iLayerType = PFD_MAIN_PLANE;
	//pfd.dwLayerMask = PFD_MAIN_PLANE;

	multisample_ok = false;
	int format = 0;
	//if( wglChoosePixelFormatARB != NULL ) {
	if( wglChoosePixelFormatARB != NULL && RendererInfo::multiSample && visionPrefs.multisample > 0 &&
		try_multisample ) {
			// only use WGL if required
			LOG("using WGL ChoosePixelFormat\n");
			UINT nformats = 0;
			//if( Vision::multiSample && multisample > 0 )
			{
				LOG("enabling multisampling...\n");
				//const int multisample = 0;
				multisample_ok = true;
				int iattr[] = {
					WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
					WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
					WGL_DOUBLE_BUFFER_ARB, GL_TRUE,
					WGL_ACCELERATION_ARB, WGL_FULL_ACCELERATION_ARB,
					WGL_PIXEL_TYPE_ARB, WGL_TYPE_RGBA_ARB,
					//WGL_ALPHA_BITS_ARB, alphaBits,
					//WGL_COLOR_BITS_ARB, colorBits,
					WGL_COLOR_BITS_ARB, bits,
					WGL_DEPTH_BITS_ARB, depthBits,
					WGL_STENCIL_BITS_ARB, stencilBits,
					WGL_SAMPLE_BUFFERS_ARB, visionPrefs.multisample > 0 ? GL_TRUE : GL_FALSE,
					WGL_SAMPLES_ARB, visionPrefs.multisample,
					0
				};
				wglChoosePixelFormatARB(hDC, iattr, NULL, 1, &format, &nformats);
			}
			/*else {
			int iattr[] = {
			WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
			WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
			WGL_DOUBLE_BUFFER_ARB, GL_TRUE,
			WGL_ACCELERATION_ARB, WGL_FULL_ACCELERATION_ARB,
			WGL_PIXEL_TYPE_ARB, WGL_TYPE_RGBA_ARB,
			//WGL_ALPHA_BITS_ARB, alphaBits,
			//WGL_COLOR_BITS_ARB, colorBits,
			WGL_COLOR_BITS_ARB, bits,
			WGL_DEPTH_BITS_ARB, depthBits,
			WGL_STENCIL_BITS_ARB, stencilBits,
			// no multisample
			0
			};
			wglChoosePixelFormatARB(hDC, iattr, NULL, 1, &format, &nformats);
			}*/
			if( format == 0 ) {
				LOG("WGL ChoosePixelFormat failed, try standard as backup\n");
			}
	}
	if( format == 0 ) {
		// either WGL not (yet) available, or it failed
		LOG("using standard ChoosePixelFormat\n");
		multisample_ok = false;
		format = ChoosePixelFormat( hDC, &pfd );
	}
	if( format == 0 ) {
		closeScreen();
		Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_OPEN_WINDOW,"Failed to obtain pixel format"));
		return;
	}
	if( !SetPixelFormat( hDC, format, &pfd ) ) {
		closeScreen();
		Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_OPEN_WINDOW,"Failed to set pixel format"));
		return;
	}

	if(createContext) {
		// create and enable the render context (RC)
		hRC = wglCreateContext( hDC );
		if(hRC == NULL) {
			closeScreen();
			Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_INIT_3D,"Could not create OpenGL context"));
			return;
		}
		if(!wglMakeCurrent( hDC, hRC )) {
			closeScreen();
			Vision::setError(new VisionException(this,VisionException::V_FAILED_TO_INIT_3D,"Could not make OpenGL context current"));
			return;
		}
	}

	return;
}

void GLGraphicsEnvironment::closeScreen() {
	LOG("close screen...\n");
	checkError("GraphicsEnvironment::closeScreen enter");
	this->sound = NULL; // clear reference to block calls to it, as it may have been deleted by this stage

	wglMakeCurrent( NULL, NULL );
	// checkError can't be used from now on, as no GL context is current
	if(hRC != NULL) {
		wglDeleteContext( hRC ); // causes crash on Win2000 in Borland IDE (supposedly a Voodoo driver bug)
		hRC = NULL;
	}

	if(hWnd != NULL) {
		ReleaseDC( hWnd, hDC );
		// destroy the window explicitly
		DestroyWindow( hWnd );
		hWnd = NULL;
	}

	if (!UnregisterClassA("Vision3_main",hInstance)) {
		LOG("could not unregister class!\n");
		/*MessageBox(NULL,"Could Not Unregister Class.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);*/
		//hInstance = NULL;
	}
}

void GLGraphicsEnvironment::setFullscreen() {
	DEVMODE dmScreenSettings;
	memset(&dmScreenSettings,0,sizeof(dmScreenSettings));
	dmScreenSettings.dmSize = sizeof(dmScreenSettings);
	dmScreenSettings.dmPelsWidth = visionPrefs.resolution_width;
	dmScreenSettings.dmPelsHeight = visionPrefs.resolution_height;
	dmScreenSettings.dmBitsPerPel = bits;
	dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;
	// Try To Set Selected Mode And Get Results. CDS_FULLSCREEN gets rid of Start Bar.
	if( ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL ) {
		closeScreen();
		Vision::setError(new VisionException(this,VisionException::V_DISPLAYMODE_NOT_AVAILABLE,"The requested displaymode was not available"));
		return;
	}
}

void GLGraphicsEnvironment::checkError(const char *label) {
	if( !existsOrHasContext() ) {
		LOG("Tried to check for GL errors when no available GL context!\n");
		Vision::setError(new VisionException(NULL,VisionException::V_GL_ERROR,"Tried to check for GL errors when no available GL context"));
		return;
	}
	//return;
	// OpenGL error handling
	GLenum gl_err;
	while( gl_err = glGetError(), gl_err != GL_NO_ERROR ) {
		LOG("GL Error (%s): %d , %s\n", label, gl_err, gluErrorString(gl_err));
		Vision::setError(new VisionException(NULL,VisionException::V_GL_ERROR,"OpenGL reported an error"));
	}
}

void GLGraphicsEnvironment::cgErrorCallback(void) {
	CGerror error = cgGetError();
	LOG("CG ERROR %d : %s\n", error, cgGetErrorString(error));
	const char *lastError = cgGetLastErrorString(NULL);
	if( lastError != NULL ) {
		LOG("Last Error String: %s\n", lastError);
	}
	GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(GraphicsEnvironment::getSingleton());
	const char *lastListing = cgGetLastListing(genv->cgContext);
	if( lastListing != NULL ) {
		LOG("Last Listing String: %s\n", lastListing);
	}
	Vision::setError(new VisionException(NULL,VisionException::V_CG_ERROR,"CG reported an error"));
}

void GLGraphicsEnvironment::initCG() {
	// setup CG
	checkError("GraphicsEnvironment::initCG enter");

	cgSetErrorCallback(cgErrorCallback);

	cgContext = cgCreateContext();
	if( cgContext != NULL ) {
		LOG("Successfully initialised CG Context\n");

		cgVertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);
		if( cgVertexProfile == CG_PROFILE_UNKNOWN ) {
			LOG("### CG ERROR: Invalid vertex profile type\n");
			cgDestroyContext(cgContext);
			cgContext = NULL;
		}
		else {
			LOG("Successfully obtained vertex profile\n");
			cgGLSetOptimalOptions(cgVertexProfile);

			cgPixelProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
			if( cgPixelProfile == CG_PROFILE_UNKNOWN ) {
				LOG("### CG ERROR: Invalid pixel profile type\n");
				cgDestroyContext(cgContext);
				cgContext = NULL;
			}
			else {
				LOG("Successfully obtained pixel profile\n");
				cgGLSetOptimalOptions(cgPixelProfile);
			}
		}
	}
	else {
		LOG("### Failed to initialise CG Context\n");
	}

	//cgGLSetDebugMode(CG_FALSE);
	checkError("GraphicsEnvironment::initCG exit");
}

//bool GLGraphicsEnvironment::initialiseOpenGL(bool createContext,bool hdr) {
void GLGraphicsEnvironment::initialiseOpenGL(bool createContext) {
	LOG("GraphicsEnvironment::initialiseOpenGL(createContext=%d)\n", createContext);
	checkError("GraphicsEnvironment::initialiseOpenGL enter");
	char *gl_vendor = (char *)glGetString(GL_VENDOR);
	LOG("GL Vendor: %s\n", gl_vendor);
	char *gl_renderer = (char *)glGetString(GL_RENDERER);
	LOG("GL Renderer: %s\n", gl_renderer);
	char *gl_version = (char *)glGetString(GL_VERSION);
	LOG("GL Version: %s\n", gl_version);

	//RendererInfo::initGLExtensions(this);
	initGLExtensions();
	if( visionPrefs.shaders ) {
		initCG();
	}
	g3->init(); // also inits shaders

	//g3->setWantHDR(hdr);

	// we do this, as the resize code expects there to be a projection matrix (for the 3D viewing mode) on the stack
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glMatrixMode(GL_MODELVIEW);
	g3->ReSizeGLScene(renderer->getZNear(), renderer->getZFar(), visionPrefs.resolution_width, visionPrefs.resolution_height, renderer->getPersp(), fovy); // do this first, to try to enable HDR
	ASSERT(!multisample_ok); // shouldn't have been attempted yet
	/*if( g3->isHDREnabled() ) {
	multisample_ok = true; // no need to use multisampling, if doing it via HDR
	}*/
	if( wglChoosePixelFormatARB != NULL && RendererInfo::multiSample && visionPrefs.multisample > 0 && !multisample_ok
		&& !g3->isHDREnabled() // multisampling with HDR not supported
		) {
			// can now create a window supporting multisampling
			// closing screen also loses CG context?
			if( cgContext != NULL ) {
				g3->freeShaders();
				cgDestroyContext(cgContext);
				cgContext = NULL;
			}
			closeScreen();
			openScreen(createContext, true);
			// need to reinitialise
			initCG();
			g3->init(); // also inits shaders
			// again - as above - we do this, as the resize code expects there to be a projection matrix (for the 3D viewing mode) on the stack
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glMatrixMode(GL_MODELVIEW);
			g3->ReSizeGLScene(renderer->getZNear(), renderer->getZFar(), visionPrefs.resolution_width, visionPrefs.resolution_height, renderer->getPersp(), fovy); // need to reset viewport/perspective info
	}
	if( !multisample_ok && visionPrefs.multisample > 0 ) {
		LOG("failed to create screen with multisampling\n");
		visionPrefs.multisample = 0; // reset multisample to 0 if not available
	}

	checkError("GraphicsEnvironment::initialiseOpenGL exit");
	return;
}

void GLGraphicsEnvironment::setVisible(bool visible) {
	if( visible ) {
		ShowWindow(hWnd,SW_SHOW);
		SetForegroundWindow(hWnd);
		SetFocus(hWnd);
	}
	else
		ShowWindow(hWnd,SW_HIDE);
}

void GLGraphicsEnvironment::setWindowTitle(const char *title) {
	SetWindowTextA(hWnd, title);
}

void GLGraphicsEnvironment::swapBuffers() {
	this->renderer->endScene();
	SwapBuffers( hDC );
}

void GLGraphicsEnvironment::processEvents() {
	// most events handled by WndProc callback
	MSG msg;
	if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {
		// handle or dispatch messages
		if ( msg.message == WM_QUIT ) {
			request_quit = true;
		} else {
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
	}
}

void GLGraphicsEnvironment::waitForEvents() {
	// call Win32 API
	WaitMessage();
}

void GLGraphicsEnvironment::sleep(int time) {
	// call Win32 API
	Sleep(time);
}

bool GLGraphicsEnvironment::saveScreen(const char *filename,const char *type) {
	unsigned char *data = new unsigned char[visionPrefs.resolution_width*visionPrefs.resolution_height*3];
	memset(data,0,visionPrefs.resolution_width*visionPrefs.resolution_height*3);
	//glReadBuffer(GL_FRONT);
	glReadPixels(0,0,visionPrefs.resolution_width,visionPrefs.resolution_height,GL_RGB,GL_UNSIGNED_BYTE,data);
	/*Image2D *image = new Image2D();
	image->rgbdata = (rgb_struct *)data;
	image->w = width;
	image->h = height;*/
	Image2D *image = new Image2D(data, visionPrefs.resolution_width, visionPrefs.resolution_height);
	bool ok = image->save(filename,type);

	delete image;
	return ok;
}

Shader *GLGraphicsEnvironment::createShader(const char *file, const char *func, bool vertex_shader) {
	if( !this->hasShaders() )
		return NULL;
	//CGGLShader *shader = new CGGLShader(g3, static_cast<GLRenderer *>(g3->getRenderer()), vertex_shader);
	CGGLShader *shader = new CGGLShader(static_cast<GLRenderer *>(g3->getRenderer()), vertex_shader);
	LOG("Loading %s shader %d : %s / %s\n", vertex_shader?"vertex":"pixel", shader, file, func);
	shader->cgProfile = vertex_shader ? this->cgVertexProfile : this->cgPixelProfile;
	shader->cgProgram = cgCreateProgramFromFile(this->cgContext, CG_SOURCE, file, shader->cgProfile, func, 0);
	if( shader->cgProgram == NULL ) {
		LOG("### Failed to load %s shader:\n", vertex_shader?"vertex":"pixel");
		CGerror Error = cgGetError();
		LOG(cgGetErrorString(Error));
		LOG("\n");
		delete shader;
		//this->cgContext = NULL;
		return NULL;
	}
	cgGLLoadProgram(shader->cgProgram);
	shader->init();
	g3->addShader(shader);

	return shader;
}

void GLGraphicsEnvironment::initGLExtensions() {
	RendererInfo::init();
	GLRendererInfo::init();

	// first we need to create a dummy window, before we can use glGetString
	/*WNDCLASS wc;
	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = CS_OWNDC;
	wc.lpfnWndProc = ( WNDPROC )WndProc_dummy;
	wc.hInstance = hInstance;
	//wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
	//wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	//wc.hbrBackground = (HBRUSH)GetStockObject( BLACK_BRUSH );
	wc.lpszClassName = "Vision3_dummy";
	RegisterClass( &wc );

	HWND hwnd = CreateWindow("Vision3_dummy", "Vision3_dummy",
	WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
	0, 0, 32, 32, 0, 0, hInstance, 0);
	if( hwnd == NULL ) {
	Vision::setError(new VisionException(NULL,VisionException::V_FAILED_TO_OPEN_WINDOW,"Could not create dummy window"));
	return;
	}
	HDC hDC = GetDC(hwnd);

	PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.cColorBits = 16;
	pfd.cDepthBits = 15;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;

	int iFormat = ChoosePixelFormat( hDC, &pfd );
	SetPixelFormat( hDC, iFormat, &pfd );
	HGLRC hRC = wglCreateContext( hDC );
	if(hRC == NULL) {
	DestroyWindow(hwnd);
	UnregisterClass("Vision3_dummy", hInstance);
	Vision::setError(new VisionException(NULL,VisionException::V_FAILED_TO_INIT_OPENGL,"Could not create dummy OpenGL context"));
	return;
	}
	if(!wglMakeCurrent( hDC, hRC )) {
	DestroyWindow(hwnd);
	UnregisterClass("Vision3_dummy", hInstance);
	Vision::setError(new VisionException(NULL,VisionException::V_FAILED_TO_INIT_OPENGL,"Could not make dummy OpenGL context current"));
	return;
	}*/

	// now we can grab the extension info
	PFNWGLGETEXTENSIONSSTRINGARBPROC glGetExtensionsStringARB =
		(PFNWGLGETEXTENSIONSSTRINGARBPROC)wglGetProcAddress("wglGetExtensionsStringARB");
	const char *wgl_extension = NULL;
	if( glGetExtensionsStringARB ) {
		//wgl_extension = glGetExtensionsStringARB(this->getHDC());
		wgl_extension = glGetExtensionsStringARB(this->hDC);
	}

	char *extension = (char *)glGetString(GL_EXTENSIONS);
	if(extension == NULL) {
		/*DestroyWindow(hwnd);
		UnregisterClass("Vision3_dummy", hInstance);*/
		return;
	}
	LOG("Extensions: %s\n", extension);
	if( wgl_extension != NULL ) {
		LOG("WGL Extensions: %s\n", wgl_extension);
	}

	if(wgl_extension != NULL && strstr(wgl_extension, "WGL_ARB_pixel_format")) {
		GLRendererInfo::pixelFormat = true;
		wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC)wglGetProcAddress("wglChoosePixelFormatARB");
		//__wglewChoosePixelFormatARB(hdc, iattr, 0, 256, formats, &count);
	}
	LOG("WGL Pixel Format? %s\n", GLRendererInfo::pixelFormat?"T":"F");

	if(wgl_extension != NULL && strstr(wgl_extension, "WGL_ARB_multisample")) {
		RendererInfo::multiSample = true;
	}
	LOG("Multisample? %s\n", RendererInfo::multiSample?"T":"F");

	/*if(wgl_extension != NULL && strstr(wgl_extension, "WGL_ARB_pbuffer") && strstr(wgl_extension, "WGL_ARB_render_texture")) {
	pbuffer = true;
	wglCreatePbufferARB    = (PFNWGLCREATEPBUFFERARBPROC)wglGetProcAddress("wglCreatePbufferARB");
	wglGetPbufferDCARB     = (PFNWGLGETPBUFFERDCARBPROC)wglGetProcAddress("wglGetPbufferDCARB");
	wglReleasePbufferDCARB = (PFNWGLRELEASEPBUFFERDCARBPROC)wglGetProcAddress("wglReleasePbufferDCARB");
	wglDestroyPbufferARB   = (PFNWGLDESTROYPBUFFERARBPROC)wglGetProcAddress("wglDestroyPbufferARB");
	wglQueryPbufferARB     = (PFNWGLQUERYPBUFFERARBPROC)wglGetProcAddress("wglQueryPbufferARB");

	wglBindTexImageARB     = (PFNWGLBINDTEXIMAGEARBPROC)wglGetProcAddress("wglBindTexImageARB");
	wglReleaseTexImageARB  = (PFNWGLRELEASETEXIMAGEARBPROC)wglGetProcAddress("wglReleaseTexImageARB");
	wglSetPbufferAttribARB = (PFNWGLSETPBUFFERATTRIBARBPROC)wglGetProcAddress("wglSetPbufferAttribARB");
	}
	LOG("P Buffer? %s\n", pbuffer?"T":"F");*/

	glGetIntegerv(GL_MAX_LIGHTS, &RendererInfo::maxLights);
	LOG("Max Lights? %d\n", RendererInfo::maxLights);

	if(strstr(extension, "GL_ARB_multitexture")) {
		//glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &RendererInfo::maxTextureUnits);

		glMultiTexCoord2fARB = (PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2fARB");
		glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
		glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress("glClientActiveTextureARB");
	}
	//LOG("Max Texture Units? %d\n", RendererInfo::maxTextureUnits);

	if(strstr(extension, "GL_NV_occlusion_query")) {
		RendererInfo::hardwareOcclusion = true;
		glGenOcclusionQueriesNV = (PFNGLGENOCCLUSIONQUERIESNVPROC)wglGetProcAddress("glGenOcclusionQueriesNV");
		glDeleteOcclusionQueriesNV = (PFNGLDELETEOCCLUSIONQUERIESNVPROC)wglGetProcAddress("glDeleteOcclusionQueriesNV");
		glGetOcclusionQueryuivNV = (PFNGLGETOCCLUSIONQUERYUIVNVPROC)wglGetProcAddress("glGetOcclusionQueryuivNV");
		glBeginOcclusionQueryNV = (PFNGLBEGINOCCLUSIONQUERYNVPROC)wglGetProcAddress("glBeginOcclusionQueryNV");
		glEndOcclusionQueryNV = (PFNGLENDOCCLUSIONQUERYNVPROC)wglGetProcAddress("glEndOcclusionQueryNV");
	}
	LOG("Hardware Occlusion? %s\n", RendererInfo::hardwareOcclusion?"T":"F");

	if(strstr(extension, "GL_EXT_compiled_vertex_array")) {
		GLRendererInfo::compiledVAs = true;
		glLockArraysEXT = (PFNGLLOCKARRAYSEXTPROC)wglGetProcAddress("glLockArraysEXT");
		glUnlockArraysEXT = (PFNGLUNLOCKARRAYSEXTPROC)wglGetProcAddress("glUnlockArraysEXT");
	}
	LOG("Compiled Vertex Array? %s\n", GLRendererInfo::compiledVAs?"T":"F");

	//if(false) {
	if(strstr(extension, "GL_ARB_vertex_buffer_object")) {
		GLRendererInfo::vertexBufferObjects = true;
		glGenBuffersARB = (PFNGLGENBUFFERSARBPROC) wglGetProcAddress("glGenBuffersARB");
		glBindBufferARB = (PFNGLBINDBUFFERARBPROC) wglGetProcAddress("glBindBufferARB");
		glBufferDataARB = (PFNGLBUFFERDATAARBPROC) wglGetProcAddress("glBufferDataARB");
		glBufferSubDataARB = (PFNGLBUFFERSUBDATAARBPROC) wglGetProcAddress("glBufferSubDataARB");
		glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC) wglGetProcAddress("glDeleteBuffersARB");
	}
	LOG("VBOs? %s\n", GLRendererInfo::vertexBufferObjects?"T":"F");
	if( GLRendererInfo::vertexBufferObjects && !visionPrefs.vbo ) {
		GLRendererInfo::vertexBufferObjects = false;
		LOG(">>> Disabling VBOs\n");
	}

	//if(false) {
	if(strstr(extension, "GL_EXT_draw_range_elements")) {
		GLRendererInfo::drawRangeElements = true;
		glDrawRangeElementsEXT = (PFNGLDRAWRANGEELEMENTSPROC) wglGetProcAddress("glDrawRangeElementsEXT");
	}
	LOG("Draw Range Elements? %s\n", GLRendererInfo::drawRangeElements?"T":"F");

	if(strstr(extension, "GL_ATI_separate_stencil") && strstr(extension, "GL_EXT_stencil_wrap") ) {
		GLRendererInfo::separateStencilATI = true;
		glStencilOpSeparateATI = (PFNGLSTENCILOPSEPARATEATIPROC) wglGetProcAddress("glStencilOpSeparateATI");
		glStencilFuncSeparateATI = (PFNGLSTENCILFUNCSEPARATEATIPROC) wglGetProcAddress("glStencilFuncSeparateATI");
	}
	LOG("ATI two sided stencil? %s\n", GLRendererInfo::separateStencilATI?"T":"F");
	if( GLRendererInfo::separateStencilATI && !visionPrefs.twosidedstencil ) {
		GLRendererInfo::separateStencilATI = false;
		LOG(">>> Disabling ATI two sided stencil\n");
	}

	if(strstr(extension, "GL_EXT_stencil_two_side") && strstr(extension, "GL_EXT_stencil_wrap") ) {
		GLRendererInfo::separateStencil = true;
		glActiveStencilFaceEXT = (PFNGLACTIVESTENCILFACEEXTPROC) wglGetProcAddress("glActiveStencilFaceEXT");
	}
	LOG("Two sided stencil? %s\n", GLRendererInfo::separateStencil?"T":"F");
	if( GLRendererInfo::separateStencil && !visionPrefs.twosidedstencil ) {
		GLRendererInfo::separateStencil = false;
		LOG(">>> Disabling two sided stencil\n");
	}

	if(strstr(extension, "GL_EXT_framebuffer_object")) {
		RendererInfo::framebufferObject = true;
		glBindRenderbufferEXT = (PFNGLBINDRENDERBUFFEREXTPROC) wglGetProcAddress("glBindRenderbufferEXT");
		glDeleteRenderbuffersEXT = (PFNGLDELETERENDERBUFFERSEXTPROC) wglGetProcAddress("glDeleteRenderbuffersEXT");
		glGenRenderbuffersEXT = (PFNGLGENRENDERBUFFERSEXTPROC) wglGetProcAddress("glGenRenderbuffersEXT");
		glRenderbufferStorageEXT = (PFNGLRENDERBUFFERSTORAGEEXTPROC) wglGetProcAddress("glRenderbufferStorageEXT");
		glBindFramebufferEXT = (PFNGLBINDFRAMEBUFFEREXTPROC) wglGetProcAddress("glBindFramebufferEXT");
		glDeleteFramebuffersEXT = (PFNGLDELETEFRAMEBUFFERSEXTPROC) wglGetProcAddress("glDeleteFramebuffersEXT");
		glGenFramebuffersEXT = (PFNGLGENFRAMEBUFFERSEXTPROC) wglGetProcAddress("glGenFramebuffersEXT");

		glCheckFramebufferStatusEXT = (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC) wglGetProcAddress("glCheckFramebufferStatusEXT");
		glFramebufferTexture2DEXT = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC) wglGetProcAddress("glFramebufferTexture2DEXT");
		glFramebufferRenderbufferEXT = (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC) wglGetProcAddress("glFramebufferRenderbufferEXT");
	}
	LOG("FBOs? %s\n", RendererInfo::framebufferObject?"T":"F");

	/*DestroyWindow(hwnd);
	UnregisterClass("Vision3_dummy", hInstance);*/
}

LRESULT CALLBACK GLGraphicsEnvironment::WndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam ) {
	switch ( message ) {
		//case WM_ACTIVATEAPP:
		case WM_ACTIVATE:
			{
				//GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getGenv(hWnd));
				GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
				if(genv!=NULL)
					genv->bActive = ( LOWORD( wParam ) != WA_INACTIVE );
				if( genv != NULL && genv->visionPrefs.fullscreen ) {
					if( genv->bActive ) {
						ShowWindow(genv->getHWND(), SW_RESTORE);
						genv->setFullscreen();
					}
					else {
						ShowWindow(genv->getHWND(), SW_MINIMIZE);
						ChangeDisplaySettings(NULL, 0);
					}
				}
				/*if(genv!=NULL && genv->fullscreen) {
				if(genv->bActive) {
				DEVMODE dmScreenSettings;					// Device Mode
				memset(&dmScreenSettings,0,sizeof(dmScreenSettings));		// Makes Sure Memory's Cleared
				dmScreenSettings.dmSize=sizeof(dmScreenSettings);		// Size Of The Devmode Structure
				dmScreenSettings.dmPelsWidth	= genv->width;			// Selected Screen Width
				dmScreenSettings.dmPelsHeight	= genv->height;			// Selected Screen Height
				dmScreenSettings.dmBitsPerPel	= genv->bits;				// Selected Bits Per Pixel
				dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;
				// Try To Set Selected Mode And Get Results.  NOTE: CDS_FULLSCREEN Gets Rid Of Start Bar.
				ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN);
				ShowCursor(FALSE);						// Show Mouse Pointer

				}
				else {
				ChangeDisplaySettings(NULL,0);					// If So Switch Back To The Desktop
				ShowCursor(TRUE);						// Show Mouse Pointer

				//wglMakeCurrent( NULL, NULL );
				}
				}*/
				//bActive = wparam;
				return 0;
			}
		case WM_CREATE:
			return 0;
		case WM_CLOSE:
			{
				PostQuitMessage( 0 );
				return 0;
			}
		case WM_DESTROY:
			return 0;
		case WM_KEYDOWN:
			{
				//InputHandler::keys[wParam] = true;
				//GraphicsEnvironment *genv = getGenv(hWnd);
				GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->keys[wParam] = true;
				genv->getInputSystem()->getKeyboard()->setKey(wParam, true);
				/*if(wParam == VK_ESCAPE)
				PostQuitMessage( 0 );*/
				return 0;
			}
		case WM_KEYUP:
			{
				//InputHandler::keys[wParam] = false;
				//GraphicsEnvironment *genv = getGenv(hWnd);
				GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->keys[wParam] = false;
				genv->getInputSystem()->getKeyboard()->setKey(wParam, false);
				return 0;
			}
		case WM_LBUTTONDOWN:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
				//LOG(">>>leftmousedown\n");
				//genv->getInputHandler()->mouseLChange(true);
				genv->getInputSystem()->getMouse()->mouseLChange(true);
				break;
			}
		case WM_MBUTTONDOWN:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->mouseMChange(true);
				genv->getInputSystem()->getMouse()->mouseMChange(true);
				break;
			}
		case WM_RBUTTONDOWN:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->mouseRChange(true);
				genv->getInputSystem()->getMouse()->mouseRChange(true);
				break;
			}
		case WM_LBUTTONUP:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
				//LOG(">>>leftmouseup\n");
				//genv->getInputHandler()->mouseLChange(false);
				genv->getInputSystem()->getMouse()->mouseLChange(false);
				break;
			}
		case WM_MBUTTONUP:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->mouseMChange(false);
				genv->getInputSystem()->getMouse()->mouseMChange(false);
				break;
			}
		case WM_RBUTTONUP:
			{
				//GraphicsEnvironment *genv = getGenv(hWnd);
				GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
				//genv->getInputHandler()->mouseRChange(false);
				genv->getInputSystem()->getMouse()->mouseRChange(false);
				break;
			}
		case WM_SIZE:
			{
				/*GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getGenv(hWnd));
				//GraphicsEnvironment::ReSizeGLScene(LOWORD(lParam),HIWORD(lParam));  // LoWord=Width, HiWord=Height
				if(genv!=NULL)
					genv->resize(LOWORD(lParam),HIWORD(lParam));  // LoWord=Width, HiWord=Height
				return 0;*/
				// no longer supported
				// if re-enabled, watch for bug where this gets called, from CloseScreen, if quitting from a full-screen OpenGL application (when we haven't alt-tabbed)
				break;
			}
			/*default:
			return DefWindowProc( hWnd, message, wParam, lParam );*/
	}

	if( message != NULL ) {
		//GraphicsEnvironment *genv = getGenv(hWnd);
		GLGraphicsEnvironment *genv = static_cast<GLGraphicsEnvironment *>(getSingleton());
		if( genv->getSound() != NULL ) {
			genv->getSound()->processEvents(message, wParam, lParam);
		}
	}

	return DefWindowProc( hWnd, message, wParam, lParam );
}
