#pragma once

#include <vector>
using std::vector;
#include <set>
using std::set;
#include <string>
using std::string;

#ifndef _WIN32
	#define stricmp strcasecmp
	#define strnicmp strncasecmp
#endif

#include "VisionShared.h"

#ifndef LOG
#define LOG V_log
#endif
/*inline void V_log(const char *text, ...) {
	logText(text, ...);
}*/

#ifdef _DEBUG

#include <ctime>

extern int V__tracelevel;

#define V__ENTER                                         \
	V__tracelevel++;                                     \
	int __time_start = clock();

#define V__EXIT                                          \
	V__tracelevel--;                                     \
	int __time_diff = clock() - __time_start;            \
	{                                                    \
		V__logtime(__FILE__, __FUNCTION__, __time_diff); \
	}
/*	if( __time_diff > 0 )                                \*/

#else

#define V__ENTER
#define V__EXIT

#endif

void V__logtime(const char *file, const char *func, int time);

class VisionObject {
protected:
	//char name[V_MAX_NAME+1];
	string name;
public:
	//bool kill_later;
	V_TAG_t tag;
	int deleteLevel;
	bool owned; // if true, this object shouldn't be freed when flushing memory, as it's owned by another VisionObject

	VisionObject();
	virtual ~VisionObject();
	//virtual void update(float time) {}
	virtual V_CLASS_t getClass() const=0;
	virtual bool isSubClass(V_CLASS_t v_class)=0;
	void setName(const char *name);
	const char *getName() const;
	unsigned int getNameLength() const {
		return this->name.length();
	}
	virtual int memUsage()=0;
	virtual bool check() const;
	virtual void prepare() {} // used by the ObjectData class and its subclasses
	/*virtual void kill() {
	this->kill_later = true;
	}*/
};

//typedef vector<World *> VWorldsCollection;
//typedef set<World *> VWorldsCollection;
//typedef VWorldsCollection::iterator VWorldsIter;

typedef vector<VisionObject *> VSafedeleteCollection;
typedef VSafedeleteCollection::iterator VSafedeleteIter;

class Vision {
	//static bool locked;
	static int lock_level;

	static vector<VisionObject *> tags;
	//static VWorldsCollection worlds;
	static VSafedeleteCollection safedeletequeue;
	static bool any_kills;

	static int default_persistence;

public:
	//static VisionException *error;

	static void initialise(const char *application_name);
	static const char *getApplicationName();
	static char *getApplicationFilename(const char *name);
	static void flushAll();
	static void flush(int deleteLevel);
	//static void doKills();
	static void lock();
	static void unlock(bool reprepare);
	static void unlock() {
		unlock(true);
	}
	static bool isLocked();
	/*DLL static void safeDeleteImage(Image2D *image);
	DLL static void safeDeleteObjectData(ObjectData *obj);*/
	static void cleanup();
	static int addTag(VisionObject *ptr);
	static VisionObject *ptrFromTag(V_TAG_t tag);
	static int getNTags();
	static void removeTag(V_TAG_t tag);
	//static void addWorld(World *world);
	//static void removeWorld(World *world);
	//static VisionException *getError();
	static void setError(VisionException *ve);
	static void update(int time);
	static void safeDelete(VisionObject *vo) {
		safedeletequeue.push_back(vo);
	}
	static void registerKill() {
		any_kills = true;
	}
	static void setDefaultPersistence(int level) {
		LOG("Vision::setDefaultPersistence(%d)\n", level);
		default_persistence = level;
	}
	static int getDefaultPersistence() {
		return default_persistence;
	}
	static void check();
	static void setPersistenceAll(int level);
	static void debugMemUsage();
};

void createThreads(int n_threads, VI_ThreadFunction **function, void **data);
