//---------------------------------------------------------------------------
#include <cassert>
#include <ctime>

#include "VisionIface.h"

#include "VisionPrefs.h"

#include "GraphicsEnvironment.h"
#if defined(V_USE_SDL) || defined(__linux)
#include "OpenGL/SDL_GLGraphicsEnvironment.h"
#endif

#ifdef _WIN32
#ifdef V_USE_SDL
#include "DirectX/SDL_D3D9GraphicsEnvironment.h"
#else
#include "OpenGL/GLGraphicsEnvironment.h"
#include "DirectX/D3D9GraphicsEnvironment.h"
#endif
#endif

#include "Sound.h"

#if defined(V_USE_SDL) || defined(__linux)
#elif _WIN32
#include "DirectX/DXSound.h"
#endif

#include "GraphicsEngine.h"
#include "Terrain.h"

#include "Renderer.h"

#include "Panel.h"

#include "InputHandler.h"

#include "World.h"
#include "Entity.h"
#include "Object3D.h"

#include "Texture.h"

//---------------------------------------------------------------------------

/*VI_GraphicsEnvironment *V_GRAPHICSENV_get_ptr(V_TAG_t v_genv) {
	GraphicsEnvironment *genv = static_cast<GraphicsEnvironment *>(Vision::ptrFromTag(v_genv));
	return genv;
}

VI_Panel *V_PANEL_get_ptr(V_TAG_t v_panel) {
	Panel *panel = static_cast<Panel *>(Vision::ptrFromTag(v_panel));
	return panel;
}

VI_World *V_WORLD_get_ptr(V_TAG_t v_world) {
	World *world = static_cast<World *>(Vision::ptrFromTag(v_world));
	return world;
}

VI_Sky *V_SKY_get_ptr(V_TAG_t v_sky) {
	SkyBox *sky = static_cast<SkyBox *>(Vision::ptrFromTag(v_sky));
	return sky;
}

VI_Font *V_FONT_get_ptr(V_TAG_t v_font) {
	FontBuffer *font = static_cast<FontBuffer *>(Vision::ptrFromTag(v_font));
	return font;
}

VI_Image *V_IMAGE_get_ptr(V_TAG_t v_image) {
	Image2D *image = static_cast<Image2D *>(Vision::ptrFromTag(v_image));
	return image;
}

VI_Entity *V_ENTITY_get_ptr(V_TAG_t v_entity) {
	VI_Entity *node = dynamic_cast<VI_Entity *>(Vision::ptrFromTag(v_entity));
	return node;
}

VI_Object *V_OBJECTDATA_get_ptr(V_TAG_t v_object) {
	ObjectData *object = static_cast<ObjectData *>(Vision::ptrFromTag(v_object));
	return object;
}

VI_VObject *V_VOBJECT_get_ptr(V_TAG_t v_vobject) {
	Object3D *vobject = static_cast<Object3D *>(Vision::ptrFromTag(v_vobject));
	return vobject;
}

VI_Terrain *V_TERRAIN_get_ptr(V_TAG_t v_terrain) {
	TerrainEngine *terrain = static_cast<TerrainEngine *>(Vision::ptrFromTag(v_terrain));
	return terrain;
}

VI_Graphics2D *V_GRAPHICS2D_get_ptr(V_TAG_t v_g2d) {
	Graphics2D *g2d = static_cast<Graphics2D *>(Vision::ptrFromTag(v_g2d));
	return g2d;
}*/

/*class DummyObject : public VisionObject {
public:
	DummyObject() : VisionObject() {
	}
	virtual ~DummyObject() {
	}
	virtual V_CLASS_t getClass() {
		return V_CLASS_UNKNOWN;
	}
	virtual bool isSubClass(V_CLASS_t v_class) {
		return false;
	}
	virtual int memUsage() {
		return 0;
	}
};*/

void VI_addtag() {
	//Vision::addTag(NULL);
	//new SkyBox();
	//new DummyObject();
	//int size = sizeof(DummyObject);
	VI_log("VI_addtag()\n");
	int size = 65536;
	new char[size];
}

void VI_initialise(const char *application_name) {
	Vision::initialise(application_name);
}

char *VI_getApplicationFilename(const char *name) {
	char *filename = Vision::getApplicationFilename(name);
	return filename;
}

void VI_cleanup() {
	Vision::cleanup();
}

void VI_check() {
	Vision::check();
}

void VI_set_persistence_all(int level) {
	Vision::setPersistenceAll(level);
}

void VI_set_default_persistence(int level) {
	Vision::setDefaultPersistence(level);
}

void VI_debug_mem_usage_all() {
	Vision::debugMemUsage();
}

void VI_lock() {
	Vision::lock();
}

void VI_unlock() {
	Vision::unlock();
}

bool VI_islocked() {
	return Vision::isLocked();
}

void VI_flush(int level) {
	Vision::flush(level);
}

void VI_flush_all() {
	Vision::flushAll();
}

void VI_update(int time) {
	Vision::update(time);
}

bool VI_keydown(int v_keycode) {
	return KEY_DOWN(v_keycode);
}

bool VI_keyany() {
	//return InputHandler::keyAny();
	GraphicsEnvironment *ge = GraphicsEnvironment::getSingleton();
	//InputHandler *ih = ge->getInputHandler();
	KeyboardInputHandler *ih = ge->getInputSystem()->getKeyboard();
	return ih->keyAny();
}

bool VI_keypressed(int v_keycode) {
	GraphicsEnvironment *ge = GraphicsEnvironment::getSingleton();
	KeyboardInputHandler *ih = ge->getInputSystem()->getKeyboard();
	return ih->wasPressed(v_keycode);
}

void VI_createThreads(int n_threads, VI_ThreadFunction **function, void **data) {
	createThreads(n_threads, function, data);
}

void VI_movementSimpleXY(VI_SceneGraphNode *node,int time) {
	SceneGraphNode *d_node = dynamic_cast<SceneGraphNode *>(node); // cast to the internal representation
	static Movement movement;
	movement.inputXY(d_node->getPoint(), time);
}

void VI_movementSimpleXYhover(VI_SceneGraphNode *node,int time) {
	SceneGraphNode *d_node = dynamic_cast<SceneGraphNode *>(node); // cast to the internal representation
	static Movement movement;
	movement.inputXYHover(d_node->getPoint(), time);
}

/*#ifdef _WIN32
VI_Sound *VI_Sound::create(HWND hwnd) {
	LOG("VI_Sound::create(%d)\n",hwnd);

	Sound *sound = NULL;
	sound = new DXSound(hwnd);
	if( !sound->isInitialised() ) {
		LOG("    Error: Failed to initialise sound\n");
		delete sound;
		return NULL;
	}
	return sound;
}
#endif*/

VI_Sound *VI_Sound::create() {
	LOG("VI_Sound::create()\n");

	Sound *sound = NULL;
#if defined(V_USE_SDL) || defined(__linux)
	sound = new SDL_Sound();
#elif _WIN32
	sound = new DXSound();
#endif

	if( !sound->isInitialised() ) {
		LOG("    Error: Failed to initialise sound\n");
		delete sound;
		return NULL;
	}
	return sound;
}

void VI_GraphicsEnvironment::render(VI_GraphicsEnvironment_Update update) {
	this->setVisible(true);

	static bool quit = false;

	quit = false;
	bool fin = false;
	int t_start = clock();
	// TODO: make more OS-independent; and needs to call handleInput()?
	while( !quit && !fin ) {
		// check for messages
#ifdef _WIN32
		MSG msg;
		if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {
			// handle or dispatch messages
			if ( msg.message == WM_QUIT ) {
				quit = true;
			} else {
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		}
		else
#endif
		{
			if( this->isActive() ) {
				//int t_start = clock();
				t_start = clock();
				this->drawFrame();
				this->swapBuffers();
				this->handleInput();
				int time_diff = clock() - t_start;
				//t_start = clock();

				VI_update(time_diff);
				if( !quit && !fin && update != NULL )
					fin = !update(time_diff);
			}
#ifdef _WIN32
			else {
				WaitMessage();
			}
#endif
		}
	}
}

VI_HAnimationInfo *VI_HAnimationInfo::create(const char *name, float start, float end) {
	AnimationInfo *animationinfo = new AnimationInfo();
	animationinfo->setName(name);
	animationinfo->start = start;
	animationinfo->end = end;
	return animationinfo;
}

VI_Mesh *VI_Mesh::create(const char *name) {
	return create(name, 1);
}

VI_Mesh *VI_Mesh::create(const char *name, int n_frames) {
	Mesh *obj = new Mesh(n_frames);
	obj->setName(name);
	if( !Vision::isLocked() ) {
		obj->prepare();
	}
	return obj;
}

/*VI_VObject *VI_VObject::create() {
	return new Object3D();
}*/

VI_VObject *VI_VObject::create(const char *name) {
	return create(name, 1);
}

VI_VObject *VI_VObject::create(const char *name, int n_frames) {
	Object3D *obj = new Object3D(n_frames);
	obj->setName(name);
	if( !Vision::isLocked() ) {
		obj->prepare();
	}
	return obj;
}

VI_Particlesystem *VI_Particlesystem::createSnowstorm(int max_particles,VI_Texture *texture,float width,float height,float depth) {
	Texture *d_texture = static_cast<Texture *>(texture); // cast to the internal representation
	Snowstorm *ss = new Snowstorm(max_particles, d_texture, width, height, depth);
	if( !Vision::isLocked() ) {
		ss->prepare();
	}
	return ss;
}

VI_GeneralParticlesystem *VI_GeneralParticlesystem::create(int max_particles) {
	NWNParticleSystem *ps = new NWNParticleSystem(max_particles);
	if( !Vision::isLocked() ) {
		ps->prepare();
	}
	return ps;
}

/*VI_VObject *VI_Modeller::createGrid(float w,float d,unsigned short nx,unsigned short ny,bool mirror,bool share) {
	LOG("VI_createGrid(%f,%f,%d,%d,%d,%d)\n",w,d,nx,ny,mirror,share);
	VI_VObject *obj = Modeller::createGrid(w,d,nx,ny,mirror,share);

	return obj;
}

VI_VObject *VI_Modeller::createCuboid(float w,float h,float d) {
	VI_VObject *obj = Modeller::createCuboid(w, h, d);

	return obj;
}

VI_VObject *VI_Modeller::createSphere(float r,unsigned short d) {
	return Modeller::createSphere(r, d);
}

void VI_Modeller::deformAttractor(VI_Object *obj, Vector3D centre, float radius, float scale) {
	return Modeller::deformAttractor(obj, centre, radius, scale);
}*/

VI_GraphicsEnvironment *VI_createGraphicsEnvironment(int width,int height,int multisample,bool fullscreen,bool hdr) {
	VisionPrefs visionPrefs; // fill defaults in constructor
	visionPrefs.resolution_width = width;
	visionPrefs.resolution_height = height;
	visionPrefs.fullscreen = fullscreen;
	visionPrefs.multisample = multisample;
	visionPrefs.hdr = hdr;
	return VI_createGraphicsEnvironment(&visionPrefs);
}

VI_GraphicsEnvironment *VI_createGraphicsEnvironment(VisionPrefs *visionPrefs) {
	GraphicsEnvironment *ge = NULL;
	try {
#ifdef _WIN32
		if( visionPrefs->api == VISIONAPI_OPENGL ) {
#ifdef V_USE_SDL
			ge = new SDL_GLGraphicsEnvironment(*visionPrefs);
#else
			ge = new GLGraphicsEnvironment(*visionPrefs);
#endif
		}
		else {
#ifdef V_USE_SDL
			ge = new SDL_D3D9GraphicsEnvironment(*visionPrefs);
#else
			ge = new D3D9GraphicsEnvironment(*visionPrefs);
#endif
		}
#else
		ge = new SDL_GLGraphicsEnvironment(*visionPrefs);
#endif
	}
	catch(VisionException *ve) {
		delete ve;
		return NULL;
	}
	ASSERT( ge != NULL );
	ASSERT( ge->isCreated() ); // any errors should have been thrown as an exception now!
	/*if(!ge->isCreated()) {
		// error!
		delete ge;
		return NULL;
	}*/
	return ge;
}

VI_World *VI_createWorld() {
	World *world = new World();
	return world;
}

VI_Sky *VI_createSky(const char *filename,const char *ext) {
	SkyBox *sky = new SkyBox();
	if( !sky->init(filename,ext) ) {
		delete sky;
		return NULL;
	}
	return sky;
}

VI_Sky *VI_createSky1(const char *filename) {
	SkyBox *sky = new SkyBox();
	if( !sky->init1(filename) ) {
		delete sky;
		return NULL;
	}
	return sky;
}

VI_Sky *VI_createSky1(VI_Texture *texture) {
	Texture *d_texture = static_cast<Texture *>(texture); // cast to the internal representation
	SkyBox *sky = new SkyBox();
	sky->init1(d_texture);
	return sky;
}

VI_Clouds *VI_createClouds() {
	Clouds *clouds = new Clouds();
	return clouds;
}

VI_Atmosphere *VI_createAtmosphere(VI_Entity *light_entity) {
	if( GraphicsEnvironment::getSingleton()->getGraphics3D()->hasAtmosphereShaders() == NULL ) {
		return NULL; // not supported
	}

	Entity *d_light_entity = static_cast<Entity *>(light_entity); // cast to the internal representation
	Atmosphere *atmosphere = new Atmosphere(d_light_entity);
	return atmosphere;
}

VI_Terrain *VI_createTerrain(VI_GraphicsEnvironment *genv,int width,int depth,float xscale,float zscale,float hscale) {
	LOG("VI_createTerrain(%d,%d,%d,%f,%f,%f)\n",genv,width,depth,xscale,zscale,hscale);
	GraphicsEnvironment *d_genv = static_cast<GraphicsEnvironment *>(genv); // cast to the internal representation
	TerrainEngine *terrain = new TerrainEngine(d_genv->getGraphics3D(),xscale,zscale,hscale);
	terrain->create(width, depth);
	// no need to prepare(), as done automatically by create()
	return terrain;
}

VI_Terrain *VI_createTerrainRandom(VI_GraphicsEnvironment *genv,int width,int depth,float xscale,float zscale,float hscale,const int base_heights[4],int min_height,int max_height,float roughness,int seed) {
	LOG("VI_createTerrainRandom(%d,%d,%d,%f,%f,%f)\n",genv,width,depth,xscale,zscale,hscale);
	GraphicsEnvironment *d_genv = static_cast<GraphicsEnvironment *>(genv); // cast to the internal representation
	TerrainEngine *terrain = new TerrainEngine(d_genv->getGraphics3D(),xscale,zscale,hscale);
	terrain->createRandom(width, depth, base_heights, min_height, max_height, roughness, seed);
	// no need to prepare(), as done automatically by createRandom()
	return terrain;
}

VI_Image *VI_createImage(const char *filename) {
	Image2D *image = Image2D::loadImage(filename);
	return image;
}

VI_Image *VI_createImage(int w,int h) {
	Image2D *image = new Image2D(w, h);
	return image;
}

VI_Image *VI_Image::createNoise(int w,int h) {
	const unsigned char filter_max[3] = {255, 0, 0};
	const unsigned char filter_min[3] = {0, 0, 0};
	Image2D *image = Image2D::createNoise(w, h, 3.2f, 3.2f, filter_max, filter_min, V_NOISEMODE_PERLIN, 1);
	return image;
}

VI_Image *VI_Image::createNoise(int w,int h,float scale_u,float scale_v,const unsigned char filter_max[3],const unsigned char filter_min[3],V_NOISEMODE_t noisemode,int n_iterations) {
	Image2D *image = Image2D::createNoise(w, h, scale_u, scale_v, filter_max, filter_min, noisemode, n_iterations);
	return image;
}

VI_Image *VI_Image::createRadial(int w,int h,const unsigned char filter_max[3],const unsigned char filter_min[3],bool squared) {
	Image2D *image = Image2D::createRadial(w, h, filter_max, filter_min, squared);
	return image;
}

VI_Image *VI_createImageNoise(int w,int h) {
	//Image2D *image = Image2D::createNoise(w, h);
	const unsigned char filter_max[3] = {255, 0, 0};
	const unsigned char filter_min[3] = {0, 0, 0};
	Image2D *image = Image2D::createNoise(w, h, 3.2f, 3.2f, filter_max, filter_min, V_NOISEMODE_PERLIN, 1);
	return image;
}

VI_Image *VI_createImageNoise(int w,int h,float scale,const unsigned char filter_max[3],const unsigned char filter_min[3],V_NOISEMODE_t noisemode,int n_iterations) {
	Image2D *image = Image2D::createNoise(w, h, scale, scale, filter_max, filter_min, noisemode, n_iterations);
	return image;
}

VI_Image *VI_createImageNoise(int w,int h,float scale_u,float scale_v,const unsigned char filter_max[3],const unsigned char filter_min[3],V_NOISEMODE_t noisemode,int n_iterations) {
	Image2D *image = Image2D::createNoise(w, h, scale_u, scale_v, filter_max, filter_min, noisemode, n_iterations);
	return image;
}

VI_Image *VI_createImageRadial(int w,int h,const unsigned char filter_max[3],const unsigned char filter_min[3],bool squared) {
	Image2D *image = Image2D::createRadial(w, h, filter_max, filter_min, squared);
	return image;
}


VI_Texture *VI_createTexture(const char *filename) {
	Texture *tx = GraphicsEnvironment::getSingleton()->getRenderer()->loadTexture(filename);
	return tx;
}

VI_Texture *VI_createTextureWithMask(const char *filename,unsigned char r,unsigned char g,unsigned char b) {
	//rgb_struct maskcol = { r, g, b };
	unsigned char maskcol[] = {r, g, b};
	Texture *tx = GraphicsEnvironment::getSingleton()->getRenderer()->loadTextureWithMask(filename, maskcol);
	return tx;
}

VI_Texture *VI_createTexture(VI_Image *image) {
	Texture *tx = GraphicsEnvironment::getSingleton()->getRenderer()->createTexture(image);
	return tx;
}

VI_Texture *VI_createTextureWithMask(VI_Image *image,unsigned char r,unsigned char g,unsigned char b) {
	unsigned char maskcol[] = {r, g, b};
	Texture *tx = GraphicsEnvironment::getSingleton()->getRenderer()->createTextureWithMask(image, maskcol);
	return tx;
}

VI_Billboard *VI_loadBillboardWithMask(const char *filename,float width,float height,unsigned char r,unsigned char g,unsigned char b,int n_billboards) {
	LOG("V_BILLBOARD_load_with_mask(%s,%f,%f,%d,%d,%d,%d)\n",filename,width,height,r,g,b,n_billboards);

	//rgb_struct maskcol = { r, g, b };
	unsigned char maskcol[] = {r, g, b};
	Texture *tx = GraphicsEnvironment::getSingleton()->getRenderer()->loadTextureWithMask(filename, maskcol);
	if( tx == NULL ) {
		return NULL;
	}

	BillBoard *bb = new BillBoard(tx, width, height, false, n_billboards);
	if( !Vision::isLocked() ) {
		bb->prepare();
	}
	return bb;
}

VI_Entity *VI_createLight(VI_Light **light,float aR,float aG,float aB,float dR,float dG,float dB,float x,float y,float z) {
	Light *l = new Light(aR,aG,aB,dR,dG,dB);
	if( !Vision::isLocked() ) {
		l->prepare();
	}
	*light = l;
	Entity *obj = new Entity(l);
	Vector3D v(x, y, z);
	obj->setPosition(&v);
	return obj;
}

VI_Entity *VI_createExplosion(VI_Explosion **explosion,VI_Texture *texture,float aR,float aG,float aB,float dR,float dG,float dB,float x,float y,float z) {
	Texture *d_texture = static_cast<Texture *>(texture); // cast to the internal representation
	Explosion *ex = new Explosion(d_texture,aR,aG,aB,dR,dG,dB);
	if( !Vision::isLocked() ) {
		ex->prepare();
	}
	*explosion = ex;
	Entity *obj = new Entity(ex);
	Vector3D v(x, y, z);
	obj->setPosition(&v);
	return obj;
}

/*VI_Entity *VI_createEntity() {
	return new Entity();
}*/

VI_SceneGraphNode *VI_createSceneGraphNode() {
	return new SceneGraphNode();
}

VI_Entity *VI_createEntity(VI_Object *obj) {
	return new Entity(obj);
}

VI_ShadowPlane *VI_createShadowPlane(Vector3D normal,float D,bool shadow,bool reflection) {
	Plane plane;
	plane.normal = normal;
	plane.D = D;
	ShadowPlane *sp = new ShadowPlane(&plane, shadow, reflection, false);
	return sp;
}

VI_Panel *VI_createPanel() {
	Panel *panel = new Panel();
	return panel;
}

VI_Button *VI_createButton(const char *text,const VI_Font *font) {
	const FontBuffer *d_font = static_cast<const FontBuffer *>(font); // cast to the internal representation
	Button *button = new Button(text, d_font);
	return button;
}

VI_ImageButton *VI_createImageButton(VI_Texture *texture) {
	ImageButton *imagebutton = new ImageButton(texture);
	return imagebutton;
}

VI_ImageButton *VI_createImageButton(VI_Texture *texture, int width, int height) {
	ImageButton *imagebutton = new ImageButton(texture, width, height);
	return imagebutton;
}

VI_Cycle *VI_createCycle(char **entries,const VI_Font *font) {
	const FontBuffer *d_font = static_cast<const FontBuffer *>(font); // cast to the internal representation
	Cycle *cycle = new Cycle(entries, d_font);
	return cycle;
}

VI_Cycle *VI_createCycle(const VI_Font *font) {
	const FontBuffer *d_font = static_cast<const FontBuffer *>(font); // cast to the internal representation
	Cycle *cycle = new Cycle(NULL, d_font);
	return cycle;
}

VI_Listbox *VI_createListbox(int width,int height,char **entries,const VI_Font *font) {
	const FontBuffer *d_font = static_cast<const FontBuffer *>(font); // cast to the internal representation
	Listbox *listbox = new Listbox(width, height, entries, d_font);
	return listbox;
}

VI_Listbox *VI_createListbox(int width,int height,const std::string *entries,int n_entries,const VI_Font *font) {
	const FontBuffer *d_font = static_cast<const FontBuffer *>(font); // cast to the internal representation
	Listbox *listbox = new Listbox(width, height, entries, n_entries, d_font);
	return listbox;
}

VI_Stringgadget *VI_createStringgadget(const VI_Font *font,int width) {
	const FontBuffer *d_font = static_cast<const FontBuffer *>(font); // cast to the internal representation
	StringGadget *stringgadget = new StringGadget(d_font, width);
	return stringgadget;
}

VI_Textfield *VI_createTextfield(const VI_Font *font,int width,int height) {
	const FontBuffer *d_font = static_cast<const FontBuffer *>(font); // cast to the internal representation
	TextField *textfield = new TextField(d_font, width, height);
	return textfield;
}
