#pragma once

#include <cmath>

#include <vector>
using std::vector;
#include <algorithm>
using std::find;

class Matrix;
class Rotation3D;

float V_cosd(float theta);
float V_sind(float theta);

class Vector2D {
public:
	float x, y;
	Vector2D() : x(0.0), y(0.0) {
	}
	Vector2D(const Vector2D *v) : x(v->x), y(v->y) {
	}
	Vector2D(const float x,const float y) : x(x), y(y) {
	}
	void set(const float x,const float y) {
		this->x = x;
		this->y = y;
	}
	float &operator[] (const int i) {
		return *((&x) + i);
	}
	const float &operator[] (const int i) const {
		return *((&x) + i);
	}
	const bool operator== (const Vector2D& v) const {
		return (v.x==x && v.y==y);
	}
	const bool operator!= (const Vector2D& v) const {
		return !(v.x==x && v.y==y);
		//return !(v == *this);
	}
	/*const*/ Vector2D operator+ () const {
		return Vector2D(x,y);
		//return (*this);
	}
	/*const*/ Vector2D operator- () const {
		return Vector2D(-x,-y);
	}
	const Vector2D& operator= (const Vector2D& v) {
		x = v.x;
		y = v.y;
		return *this;
	}
	/*const*/ Vector2D& operator+= (const Vector2D& v) {
		x+=v.x;
		y+=v.y;
		return *this;
	}
	/*const*/ Vector2D& operator-= (const Vector2D& v) {
		x-=v.x;
		y-=v.y;
		return *this;
	}
	/*const*/ Vector2D& operator*= (const float& s) {
		x*=s;
		y*=s;
		return *this;
	}
	/*const*/ Vector2D& operator/= (const float& s) {
		const float r = 1 / s;
		x *= r;
		y *= r;
		return *this;
	}
	/*const*/ Vector2D operator+ (const Vector2D& v) const {
		return Vector2D(x + v.x, y + v.y);
	}
	/*const*/ Vector2D operator- (const Vector2D& v) const {
		return Vector2D(x - v.x, y - v.y);
	}
	/*const*/ Vector2D operator* (const float& s) const {
		return Vector2D(x*s,y*s);
	}
	/*friend inline const Vector3D operator* (const float& s,const Vector3D& v) {
	return v * s;
	}*/
	/*const*/ Vector2D operator/ (float s) const {
		s = 1/s;
		return Vector2D(s*x,s*y);
	}

	void scale(const float s) {
		this->x *= s;
		this->y *= s;
	}
	void scale(const float sx,const float sy) {
		this->x*=sx;
		this->y*=sy;
	}
	void translate(const float tx,const float ty) {
		this->x+=tx;
		this->y+=ty;
	}
	void add(const Vector2D *v) {
		this->x+=v->x;
		this->y+=v->y;
	}
	void subtract(const Vector2D *v) {
		this->x-=v->x;
		this->y-=v->y;
	}
	const float square() const {
		return (x*x + y*y);
	}
	const float magnitude() const {
		return sqrt( x*x + y*y );
	}
	const float dot(const Vector2D *v) const {
		return ( x * v->x + y * v->y);
	}
	const float operator%(const Vector2D& v) const {
		return ( x * v.x + y * v.y);
	}
	void normalise() {
		float mag = magnitude();
		x = x/mag;
		y = y/mag;
	}
	void normaliseSafe() {
		float mag = magnitude();
		if(mag==0)
			return;
		x = x/mag;
		y = y/mag;
	}
	/*const*/ Vector2D unit() const {
		float mag = magnitude();
		return Vector2D(x / mag, y / mag);
		//return (*this) / magnitude();
	}
	bool isZero() const {
		return ( x == 0 && y == 0 );
	}
	bool isZero(float tol) const {
		float mag = magnitude();
		return mag <= tol;
	}
	bool isEqual(const Vector2D *that, float tol) const {
		Vector2D diff = *this - *that;
		float mag = diff.magnitude();
		return mag <= tol;
	}
};

class Vector3D {
public:
	float x,y,z;
	static Vector3D forwards;
	static Vector3D backwards;
	static Vector3D left;
	static Vector3D right;
	static Vector3D upwards;
	static Vector3D downwards;
	Vector3D() : x(0.0), y(0.0), z(0.0) {
	}
	Vector3D(const Vector3D *v) : x(v->x), y(v->y), z(v->z) {
	}
	Vector3D(const float x,const float y,const float z) : x(x), y(y), z(z) {
	}
	void set(const float x,const float y,const float z) {
		this->x=x;
		this->y=y;
		this->z=z;
	}
	float &operator[] (const int i) {
		return *((&x) + i);
	}
	const float &operator[] (const int i) const {
		return *((&x) + i);
	}
	const bool operator== (const Vector3D& v) const {
		return (v.x==x && v.y==y && v.z==z);
	}
	const bool operator!= (const Vector3D& v) const {
		return !(v.x==x && v.y==y && v.z==z);
		//return !(v == *this);
	}
	/*const*/ Vector3D operator+ () const {
		return Vector3D(x,y,z);
		//return (*this);
	}
	/*const*/ Vector3D operator- () const {
		return Vector3D(-x,-y,-z);
	}
	const Vector3D& operator= (const Vector3D& v) {
		x = v.x;
		y = v.y;
		z = v.z;
		return *this;
	}
	/*const*/ Vector3D& operator+= (const Vector3D& v) {
		x+=v.x;
		y+=v.y;
		z+=v.z;
		return *this;
	}
	/*const*/ Vector3D& operator-= (const Vector3D& v) {
		x-=v.x;
		y-=v.y;
		z-=v.z;
		return *this;
	}
	/*const*/ Vector3D& operator*= (const float& s) {
		x*=s;
		y*=s;
		z*=s;
		return *this;
	}
	/*const*/ Vector3D& operator/= (const float& s) {
		const float r = 1 / s;
		x *= r;
		y *= r;
		z *= r;
		return *this;
	}
	/*const*/ Vector3D operator+ (const Vector3D& v) const {
		return Vector3D(x + v.x, y + v.y, z + v.z);
	}
	/*const*/ Vector3D operator- (const Vector3D& v) const {
		return Vector3D(x - v.x, y - v.y, z - v.z);
	}
	/*const*/ Vector3D operator* (const float& s) const {
		return Vector3D(x*s,y*s,z*s);
	}
	/*friend inline const Vector3D operator* (const float& s,const Vector3D& v) {
	return v * s;
	}*/
	/*const*/ Vector3D operator/ (float s) const {
		s = 1/s;
		return Vector3D(s*x,s*y,s*z);
	}

	void scale(const float s) {
		this->x *= s;
		this->y *= s;
		this->z *= s;
	}
	void scale(const float sx,const float sy,const float sz) {
		this->x*=sx;
		this->y*=sy;
		this->z*=sz;
	}
	void translate(const float tx,const float ty,const float tz) {
		this->x+=tx;
		this->y+=ty;
		this->z+=tz;
	}
	void add(const Vector3D *v) {
		this->x+=v->x;
		this->y+=v->y;
		this->z+=v->z;
	}
	void subtract(const Vector3D *v) {
		this->x-=v->x;
		this->y-=v->y;
		this->z-=v->z;
	}
	const float square() const {
		return (x*x + y*y + z*z);
	}
	const float magnitude() const {
		return sqrt( x*x + y*y + z*z );
	}
	const float dot(const Vector3D *v) const {
		return ( x * v->x + y * v->y + z * v->z);
	}
	const float operator%(const Vector3D& v) const {
		return ( x * v.x + y * v.y + z * v.z);
	}
	void cross(const Vector3D *v) {
		float nx = this->y * v->z - this->z * v->y;
		float ny = -this->x * v->z + this->z * v->x;
		float nz = this->x * v->y - this->y * v->x;
		this->x=-nx;
		this->y=-ny;
		this->z=-nz;
	}
	/*const*/ Vector3D operator^(const Vector3D& v) const {
		return Vector3D(- y * v.z + z * v.y,
			x * v.z - z * v.x,
			- x * v.y + y * v.x);
	}
	void normalise() {
		float mag = magnitude();
		x = x/mag;
		y = y/mag;
		z = z/mag;
	}
	void normaliseSafe() {
		float mag = magnitude();
		if(mag==0)
			return;
		x = x/mag;
		y = y/mag;
		z = z/mag;
	}
	/*const*/ Vector3D unit() const {
		float mag = magnitude();
		return Vector3D(x / mag,y / mag,z / mag);
		//return (*this) / magnitude();
	}
	bool isZero() const {
		return ( x == 0 && y == 0 && z == 0 );
	}
	bool isZero(float tol) const {
		float mag = magnitude();
		return mag <= tol;
	}
	bool isEqual(const Vector3D *that, float tol) const {
		Vector3D diff = *this - *that;
		float mag = diff.magnitude();
		return mag <= tol;
	}

	void rotateBy(const Rotation3D *r);
	void rotateBy(float x,float y,float z);
	void dropOnLine(const Vector3D *o,const Vector3D *n);
	void parallelComp(const Vector3D *n);
	void perpComp(const Vector3D *n);
	float distFromLine(const Vector3D *o,const Vector3D *n) const;
	float distFromLineSq(const Vector3D *o,const Vector3D *n) const;
	Vector3D closestOnLine(const Vector3D *a,const Vector3D *b) const;
	bool pointInPolygon(const Vector3D *points,int n_points) const;
	Vector3D closestPointOnPolygon(const Vector3D *points,int n_points) const;
	static bool intersectSweptSphereWithPlane(bool *full,Vector3D *intersection1,Vector3D *intersection2,const Vector3D *a,const Vector3D *b,float r,const Vector3D *n,float D);
};

class Sphere {
public:
	Vector3D centre;
	float radius;
	float radiusSquared;

	Sphere() {
		this->radius = 0.0;
		this->radiusSquared = 0.0;
	}
	void setRadiusSquared(float radiusSquared) {
		this->radiusSquared = radiusSquared;
		this->radius = sqrt(radiusSquared);
	}
	void setRadius(float radius) {
		this->radius = radius;
		this->radiusSquared = radius*radius;
	}
};

class Box {
	//bool available; // false if AABB not in use
	Vector3D corners[8];
public:

	Box() {
		//this->available = false;
		for(int i=0;i<8;i++) {
			corners[i].set(0,0,0);
		}
	}
	void setAABB(const Vector3D extremes[2]) {
		//this->available = true;
		corners[0] = extremes[0];
		corners[1].set(extremes[1].x, extremes[0].y, extremes[0].z);
		corners[2].set(extremes[1].x, extremes[1].y, extremes[0].z);
		corners[3].set(extremes[0].x, extremes[1].y, extremes[0].z);
		corners[4].set(extremes[0].x, extremes[0].y, extremes[1].z);
		corners[5].set(extremes[1].x, extremes[0].y, extremes[1].z);
		corners[6].set(extremes[0].x, extremes[1].y, extremes[1].z);
		corners[7] = extremes[1];
	}
	Vector3D getCorner(int i) const {
		return corners[i];
	}
	void setCorner(int i, Vector3D corner) {
		corners[i] = corner;
	}
	Vector3D calculateCentre() const {
		Vector3D centre;
		for(int i=0;i<8;i++) {
			centre += corners[i];
		}
		centre /= 8;
		return centre;
	}
	Vector3D getX() const {
		return ( corners[1] - corners[0] ) * 0.5;
	}
	Vector3D getY() const {
		return ( corners[3] - corners[0] ) * 0.5;
	}
	Vector3D getZ() const {
		return ( corners[4] - corners[0] ) * 0.5;
	}
	bool collisionLine(float *intersection, const Vector3D *pos, const Vector3D *dir) const;
};

class Quaternion {
public:
	float w, x, y, z;
	Quaternion() : w(0), x(0), y(0), z(0) {
	}
	Quaternion(const float w,const float x,const float y,const float z) : w(w), x(x), y(y), z(z) {
	}
	Quaternion(Matrix *matrix) {
		set(matrix);
	}
	~Quaternion() {
	}

	void set(const float w,const float x,const float y,const float z) {
		this->w = w;
		this->x = x;
		this->y = y;
		this->z = z;
	}
	void setEuler(float rx,float ry,float rz);
	void set(Matrix *matrix);
	void setAngleAxis(float ax,float ay,float az,float angle);

	void convertToEuler(float *rx,float *ry,float *rz) const;
	void convertToMatrix(Matrix *matrix) const;
	void convertToMatrixInverse(Matrix *matrix) const;
	template<class T>
	void convertToGLMatrix(T glmat[16]) const {
		T xx = x * x;
		T xy = x * y;
		T xz = x * z;
		T xw = x * w;
		T yy = y * y;
		T yz = y * z;
		T yw = y * w;
		T zz = z * z;
		T zw = z * w;

		glmat[0] = 1 - 2 * ( yy + zz );
		glmat[4] = 2 * ( xy - zw );
		glmat[8] = 2 * ( xz + yw );

		glmat[1] = 2 * ( xy + zw );
		glmat[5] = 1 - 2 * ( xx + zz );
		glmat[9] = 2 * ( yz - xw );

		glmat[2] = 2 * ( xz - yw );
		glmat[6] = 2 * ( yz + xw );
		glmat[10] = 1 - 2 * ( xx + yy );

		glmat[3] = 0;
		glmat[7] = 0;
		glmat[11] = 0;

		glmat[12] = 0;
		glmat[13] = 0;
		glmat[14] = 0;

		glmat[15] = 1;
	}
	template<class T>
	void convertInverseToGLMatrix(T glmat[16]) const {
		T xx = x * x;
		T xy = x * y;
		T xz = x * z;
		T xw = x * w;
		T yy = y * y;
		T yz = y * z;
		T yw = y * w;
		T zz = z * z;
		T zw = z * w;

		glmat[0] = 1 - 2 * ( yy + zz );
		glmat[1] = 2 * ( xy - zw );
		glmat[2] = 2 * ( xz + yw );

		glmat[4] = 2 * ( xy + zw );
		glmat[5] = 1 - 2 * ( xx + zz );
		glmat[6] = 2 * ( yz - xw );

		glmat[8] = 2 * ( xz - yw );
		glmat[9] = 2 * ( yz + xw );
		glmat[10] = 1 - 2 * ( xx + yy );

		glmat[3] = 0;
		glmat[7] = 0;
		glmat[11] = 0;

		glmat[12] = 0;
		glmat[13] = 0;
		glmat[14] = 0;

		glmat[15] = 1;
	}
	//void convertToGLMatrix(GLfloat glmat[16]);
	//void convertInverseToGLMatrix(GLfloat glmat[16]);
	Vector3D getHeading() const;
	Vector3D getUp() const;

	float square() const {
		return( w*w + x*x + y*y + z*z );
	}
	float magnitude() const {
		return sqrt( w*w + x*x + y*y + z*z );
	}
	float dot(const Quaternion& q) const;
	Quaternion slerp(const Quaternion& q, float t) const;
	Quaternion operator+ () const {
		return Quaternion(w,x,y,z);
	}
	Quaternion operator- () const {
		return Quaternion(-w,-x,-y,-z);
	}
	Quaternion operator+ (const Quaternion& q) const {
		return Quaternion(w + q.w, x + q.x, y + q.y, z + q.z);
	}
	Quaternion operator- (const Quaternion& q) const {
		return Quaternion(w - q.w, x - q.x, y - q.y, z - q.z);
	}
	Quaternion operator*(const Quaternion& q) const;
};

class Rotation3D {
	//float x,y,z;
	Quaternion quat;

	void check() {
		/*while( x < 0 )
		x += 360;
		while( x > 360 )
		x -= 360;
		while( y < 0 )
		y += 360;
		while( y > 360 )
		y -= 360;
		while( z < 0 )
		z += 360;
		while( z > 360 )
		z -= 360;*/
	}
public:
	Rotation3D();
	Rotation3D(float x, float y, float z);

	void rotateTo(const Rotation3D *rot) {
		this->quat = rot->quat;
	}
	void rotate(float x,float y,float z);
	void rotateLocal(float x,float y,float z);
	void rotateToEuler(float x,float y,float z);
	/*void rotateLocalX(float theta);
	void rotateLocalY(float theta);*/
	void getEuler(float *x,float *y,float *z) const;
	void getMatrix(Matrix *matrix) const;
	void getMatrixInverse(Matrix *matrix) const;
	Vector3D getHeading() const {
		return quat.getHeading();
	}
	Vector3D getUp() const {
		return quat.getUp();
	}
	/*void multGLMatrix();
	void multGLMatrixTranspose();*/
	void rotateToIdentity() {
		quat.set(1,0,0,0);
	}
	Quaternion *getQuaternion() {
		return &quat;
	}
	const Quaternion *getQuaternion() const {
		return &quat;
	}
	Rotation3D slerp(Rotation3D *rot, float t) const {
		Quaternion res_q = this->quat.slerp(rot->quat, t);
		Rotation3D res;
		res.quat = res_q;
		return res;
	}
};

class Plane {
public:
	Vector3D normal;
	float D;
	Plane() {
		normal.set(0.0,0.0,0.0);
		D = 0.0;
	}
	float dist(Vector3D *p) const {
		return normal.dot(p) - D;
	}
};

class Matrix {
private:
	float m[3][3]; // [col][row]
public:
	Matrix() {
		setZero();
	}
	void setRotationMatrix(const Rotation3D *r);
	void setRotationMatrixInverse(const Rotation3D *r);
	//void setRotationMatrix(float rx,float ry,float rz);
	//void setAntiRotationMatrix(Rotation3D *r);
	void transformVector(Vector3D *v);
	void setElement(int row,int col,float value) {
		this->m[col][row] = value;
	}
	float getElement(int row,int col) const {
		return this->m[col][row];
	}
	void setZero() {
		for(int i=0;i<3;i++) {
			for(int j=0;j<3;j++) {
				this->m[i][j] = 0.0;
			}
		}
	}
	void setIdentity() {
		for(int i=0;i<3;i++) {
			for(int j=0;j<3;j++) {
				this->m[i][j] = (i==j) ? 1.0f : 0.0f;
			}
		}
	}
	void setOpenGL(const float v[16]) {
		this->m[0][0] = v[0];
		this->m[0][1] = v[1];
		this->m[0][2] = v[2];
		this->m[1][0] = v[4];
		this->m[1][1] = v[5];
		this->m[1][2] = v[6];
		this->m[2][0] = v[8];
		this->m[2][1] = v[9];
		this->m[2][2] = v[10];
	}
	void setOpenGLTranspose(const float v[16]) {
		this->m[0][0] = v[0];
		this->m[1][0] = v[1];
		this->m[2][0] = v[2];
		this->m[0][1] = v[4];
		this->m[1][1] = v[5];
		this->m[2][1] = v[6];
		this->m[0][2] = v[8];
		this->m[1][2] = v[9];
		this->m[2][2] = v[10];
	}
	float trace() const {
		return ( this->m[0][0] + this->m[1][1] + this->m[2][2] );
	}
};

class Matrix4d {
	float m[4][4]; // column, row
public:
	Matrix4d() { setZero(); }
	Matrix4d(const float glmatrix[16]) {
		for(int i=0;i<4;i++) {
			for(int j=0;j<4;j++) {
				m[i][j] = glmatrix[4*i+j];
			}
		}
	}

	void setElement(int row,int col,float value) {
		this->m[col][row] = value;
	}
	float getElement(int row,int col) const {
		return this->m[col][row];
	}
	void transformVector(Vector3D *v);
	Matrix4d * multiply(const Matrix4d *that);
	void setZero() {
		for(int i=0;i<4;i++) {
			for(int j=0;j<4;j++) {
				m[i][j]=0;
			}
		}
	}
	void setIdentity() {
		for(int i=0;i<4;i++) {
			for(int j=0;j<4;j++) {
				if(i==j)
					m[i][j] = 1.0f;
				else
					m[i][j] = 0;
			}
		}
	}
	static Matrix4d * createIdentity() {
		Matrix4d * mat = new Matrix4d();
		mat->setIdentity();
		return mat;
	}
	void setScale(float x,float y,float z) {
		setZero();
		m[0][0]= x;
		m[1][1]= y;
		m[2][2]= z;
		m[3][3]= 1.0f;
	}
	static Matrix4d * createScale(float x,float y,float z) {
		Matrix4d * mat = new Matrix4d();
		mat->setScale(x,y,z);
		return mat;
	}
	void setTranslation(float x,float y,float z) {
		setIdentity();
		m[3][0] = x;
		m[3][1] = y;
		m[3][2] = z;
		m[3][3] = 1.0f;
	}
	static Matrix4d * createTranslation(float x,float y,float z) {
		Matrix4d * mat = new Matrix4d();
		mat->setTranslation(x,y,z);
		return mat;
	}
	void setRotation(float x,float y,float z,float angle);
	static Matrix4d * createRotation(float x,float y,float z,float angle) {
		Matrix4d * mat = new Matrix4d();
		mat->setRotation(x,y,z,angle);
		return mat;
	}
};

void VI_initPerlin();
float VI_perlin_noise1(float arg);
float VI_perlin_noise2(float vec[2]);
float VI_perlin_noise3(float vec[3]);

template<class T>
bool remove_vec(vector<T> *vec,const T& value) {
	for(unsigned int i=0;i<vec->size();i++) {
		if( vec->at(i) == value ) {
			vec->erase(vec->begin() + i);
			return true;
		}
	}
	return false;
}

template<class T>
bool contains_vec(vector<T> *vec,const T& value) {
	if( find(vec->begin(), vec->end(), value) == vec->end() ) {
		return false;
	}
	return true;
}

/*template<class T>
bool addIfAbsent_vec(vector<T> *vec,const T& value) {
	if( find(vec->begin(), vec->end(), value) == vec->end() ) {
		vec->push_back(value);
		return true;
	}
	return false;
}*/
