#define _CRT_SECURE_NO_WARNINGS

#include "game.h"
#include "civilization.h"
#include "city.h"
#include "buildable.h"
#include "infowindow.h"

#include <ctime>
#include <cstdio>
#include <cerrno>
#include <iomanip>

#if _WIN32
#include <io.h>
#include <direct.h>
#define access _access
#endif

#if __linux
#include <sys/stat.h> // for mkdir
#include <dirent.h>
#endif

using std::swap;

#include <sstream>
using std::stringstream;

#include <queue>
using std::priority_queue;

#include "../Vision/VisionPrefs.h"

//bool view_3d = true;
//const bool view_3d = false;

//const float def_vp_height_c = 8.0f;
//const float def_vp_height_c = 12.0f;
const float def_view3d_scale = 1.5f; // scaling of default zoom level, compared with 2d (it's useful to set the 3d zoom level to being more zoomed out, due to the angle of the viewpoint)

const float fade_duration_c = 1.0f;

//const bool CHEAT = true;
//const bool SPY = true;
//const bool RANDOMSEED = false;

const bool CHEAT = false;
const bool SPY = false;
const bool RANDOMSEED = true;

//const bool SAVE_FIRST_TURN = false;
const bool SAVE_FIRST_TURN = true;

//const char savegames_dirname[] = "conquests_savegames";
char *savegames_dirname = NULL;
const char ext[] = ".sav";
//const char filter[] = "Save Game Files\0*.sav\0";

Game *game_g = NULL;
//MainGamestate *game = NULL;
MainGamestate *MainGamestate::mainGamestate = NULL;
OptionsGamestate *OptionsGamestate::optionsGamestate = NULL;

const int versionMajor = 0;
const int versionMinor = 11;

const int blink_speed_c = (int)(CLOCKS_PER_SEC  * 0.5);
const int fire_effect_duration_c = (int)(CLOCKS_PER_SEC * 2.0);

int frames = 0;
const int frame_delay = 10;
bool quit = false;

VI_Texture *MapSquare::mapTextures[MapSquare::N_TYPES];
VI_Texture *MapSquare::mapOverlays[MapSquare::N_TYPES][n_map_overlays_c];
//VI_Texture *MapSquare::roadTextures[9];
VI_Texture *MapSquare::combinedRoadTextures[256];
VI_Texture *MapSquare::combinedRailwayTextures[256];

const float xscale = 1.0f;
const float zscale = 1.0f;

const int cursor_width_c = 32;
const int cursor_height_c = 32;

static const char *cursor_arrow[] = {
  "X                               ",
  "XX                              ",
  "X.X                             ",
  "X..X                            ",
  "X...X                           ",
  "X....X                          ",
  "X.....X                         ",
  "X......X                        ",
  "X.......X                       ",
  "X........X                      ",
  "X.....XXXXX                     ",
  "X..X..X                         ",
  "X.X X..X                        ",
  "XX  X..X                        ",
  "X    X..X                       ",
  "     X..X                       ",
  "      X..X                      ",
  "      X..X                      ",
  "       XX                       ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "0,0"
};

static const char *cursor_wait[] = {
  "XXXXXXXXXXX                     ",
  "X.........X                     ",
  "X.........X                     ",
  " X.......X                      ",
  " X.......X                      ",
  " X.......X                      ",
  "  X.....X                       ",
  "   X...X                        ",
  "    X.X                         ",
  "   X...X                        ",
  "  X.....X                       ",
  " X.......X                      ",
  " X.......X                      ",
  " X.......X                      ",
  "X.........X                     ",
  "X.........X                     ",
  "XXXXXXXXXXX                     ",
  "                                ",
  "                                ", // base
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "0,0"
};

static const char *cursor_target[] = {
  "     X                          ",
  "    X.X                         ",
  "    X.X                         ",
  "    X.X                         ",
  " XXXX.XXXX                      ",
  "X.........X                     ",
  " XXXX.XXXX                      ",
  "    X.X                         ",
  "    X.X                         ",
  "    X.X                         ",
  "     X                          ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ", // base
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "5,5"
};

/*static const char *cursor_target[] = {
  "     X                          ",
  "     X                          ",
  "     X                          ",
  "     X                          ",
  "     X                          ",
  "XXXXXXXXXXX                     ",
  "     X                          ",
  "     X                          ",
  "     X                          ",
  "     X                          ",
  "     X                          ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ", // base
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "5,5"
};*/

static const char *cursor_block[] = {
  "         XX                     ",
  "   XXXXXX.X                     ",
  "  X....X.X                      ",
  " X....X.XX                      ",
  " X...X.X.X                      ",
  " X..X.X..X                      ",
  " X.X.X...X                      ",
  " XX.X....X                      ",
  " X.X....X                       ",
  "X.XXXXXX                        ",
  "XX                              ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ", // base
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "5,5"
};

/*
static const char *cursor_[] = {
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ", // base
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "                                ",
  "0,0"
};
*/

void Game::setCursor(const char *image[]) {
	unsigned char data[4*cursor_width_c] = "";
	unsigned char mask[4*cursor_height_c] = "";

	for(int i=-1,row=0;row<cursor_height_c;row++) {
		for(int col=0;col<cursor_width_c;col++) {
			if ( col % 8 ) {
				data[i] <<= 1;
				mask[i] <<= 1;
			}
			else {
				i++;
				data[i] = mask[i] = 0;
			}
			switch (image[row][col]) {
				case 'X':
					data[i] |= 0x01;
					mask[i] |= 0x01;
					break;
				case '.':
					mask[i] |= 0x01;
					break;
				case ' ':
					break;
			}
		}
	}
	int hot_x = -1, hot_y = -1;
	sscanf(image[cursor_height_c], "%d,%d", &hot_x, &hot_y);
	this->genv->setCursor(data, mask, cursor_width_c, cursor_height_c, hot_x, hot_y);
	//return SDL_CreateCursor(data, mask, 32, 32, hot_x, hot_y);
}

void initButton(VI_Button *button) {
	button->setForeground(0.0f, 0.0f, 0.0f);
	button->setBackground(246.0f/255.0f, 172/255.0f, 100.0f/255.0f);
}

// Scripting functions

MainGamestate *AIInterface::mainGamestate = NULL;
Civilization *AIInterface::ai_civilization = NULL;

const Map::Distance *AIInterface::calculateDistanceMap(int *width, int *height, const Unit *unit, int xpos, int ypos) {
	const Map::Distance *dists = mainGamestate->getMap()->calculateDistanceMap(ai_civilization, unit, true, true, xpos, ypos);
	*width = mainGamestate->getMap()->getWidth();
	*height = mainGamestate->getMap()->getHeight();
	return dists;
}

int AIInterface::getYear() {
	return mainGamestate->getYear();
}

vector<Civilization *> AIInterface::getCivilizations(GetcivilizationsType type) {
	// n.b., we return civilizations not yet made contact - this is okay, as the AI may still see their cities or units
	vector<Civilization *> civilizations;
	for(int i=0;i<mainGamestate->getNCivilizations();i++) {
		Civilization *civ = mainGamestate->getCivilization(i);
		if( civ == ai_civilization )
			continue;
		if( civ->isDead() )
			continue;
		const Relationship *relationship = mainGamestate->findRelationship(ai_civilization, civ);
		if( type == GETCIVILIZATIONS_WAR && relationship->getStatus() != Relationship::STATUS_WAR )
			continue;
		if( type == GETCIVILIZATIONS_PEACE && relationship->getStatus() != Relationship::STATUS_PEACE )
			continue;
		civilizations.push_back(civ);
	}
	return civilizations;
}

vector<City *> AIInterface::getCities(Civilization *civilization) {
	vector<City *> cities;
	/*const Relationship *relationship = mainGamestate->findRelationship(ai_civilization, civilization);
	ASSERT( relationship->madeContact() );*/
	for(int i=0;i<civilization->getNCities();i++) {
		City *city = civilization->getCity(i);
		Pos2D city_pos = city->getPos();
		if( !ai_civilization->isExplored(city_pos.x, city_pos.y) )
			continue;
		cities.push_back(city);
	}
	return cities;
}

vector<Unit *> AIInterface::getUnits(Civilization *civilization) {
	vector<Unit *> units;
	for(int i=0;i<civilization->getNUnits();i++) {
		Unit *unit = civilization->getUnit(i);
		Pos2D unit_pos = unit->getPos();
		if( !ai_civilization->isExplored(unit_pos.x, unit_pos.y) )
			continue;
		units.push_back(unit);
	}
	return units;
}

const Relationship *AIInterface::findRelationship(Civilization *civilization) {
	const Relationship *relationship = mainGamestate->findRelationship(civilization, ai_civilization);
	return relationship;
}

bool AIInterface::hasAdjacentUnexplored(int xpos, int ypos) {
	for(int iy=ypos-1;iy<=ypos+1;iy++) {
		for(int ix=xpos-1;ix<=xpos+1;ix++) {
			if( !mainGamestate->getMap()->isValid(ix, iy) )
				continue;
			if( ix == xpos && iy == ypos ) {
				continue;
			}
			if( !ai_civilization->isExplored(ix, iy) ) {
				return true;
			}
		}
	}
	return false;
}

bool AIInterface::hasAdjacentExplored(int xpos, int ypos) {
	for(int iy=ypos-1;iy<=ypos+1;iy++) {
		for(int ix=xpos-1;ix<=xpos+1;ix++) {
			if( !mainGamestate->getMap()->isValid(ix, iy) )
				continue;
			if( ix == xpos && iy == ypos ) {
				continue;
			}
			if( ai_civilization->isExplored(ix, iy) ) {
				return true;
			}
		}
	}
	return false;
}

vector<Pos2D> AIInterface::getAdjacentSquares(int xpos, int ypos, bool include_centre) {
	vector<Pos2D> squares;
	for(int iy=ypos-1;iy<=ypos+1;iy++) {
		for(int ix=xpos-1;ix<=xpos+1;ix++) {
			if( !mainGamestate->getMap()->isValid(ix, iy) )
				continue;
			if( !include_centre && ix == xpos && iy == ypos ) {
				continue;
			}
			if( !ai_civilization->isExplored(ix, iy) ) {
				continue;
			}
			int x = ix, y = iy;
			mainGamestate->getMap()->reduceToBase(&x, &y);
			squares.push_back(Pos2D(x, y));
		}
	}
	return squares;
}

vector<Pos2D> AIInterface::getCitySquares(int xpos, int ypos, bool include_centre) {
	vector<Pos2D> squares;
	for(int i=0;i<City::getNCitySquares();i++) {
		if( !include_centre && i == City::getCentreCitySquareIndex() ) {
			continue;
		}
		Pos2D city_pos = City::getCitySquare(mainGamestate->getMap(), i, Pos2D(xpos, ypos));
		if( !mainGamestate->getMap()->isValid(city_pos.x, city_pos.y) ) {
			continue;
		}
		if( !ai_civilization->isExplored(city_pos.x, city_pos.y) ) {
			continue;
		}
		squares.push_back(city_pos);
	}
	return squares;
}

MapSquare::Type AIInterface::getTerrain(int xpos, int ypos) {
	if( ai_civilization->isExplored(xpos, ypos) ) {
		return mainGamestate->getMap()->getSquare(xpos, ypos)->getType();
	}
	return MapSquare::TYPE_UNKNOWN;
}

MapSquare::Road AIInterface::getRoad(int xpos, int ypos) {
	if( ai_civilization->isExplored(xpos, ypos) ) {
		return mainGamestate->getMap()->getSquare(xpos, ypos)->getRoad();
	}
	return MapSquare::ROAD_UNKNOWN;
}

const Civilization *AIInterface::getTerritory(int xpos, int ypos) {
	if( ai_civilization->isExplored(xpos, ypos) ) {
		return mainGamestate->getMap()->getSquare(xpos, ypos)->getTerritory();
	}
	return NULL;
}

bool AIInterface::canBuildCity(int xpos, int ypos) {
	bool can_build = false;
	if( ai_civilization->isExplored(xpos, ypos) ) {
		// TODO: still cheating here, as nearby squares might not be explored
		can_build = mainGamestate->getMap()->getSquare(xpos, ypos)->canBuildCity();
	}
	return can_build;
}

bool AIInterface::givesRoadBonus(int xpos, int ypos) {
	if( ai_civilization->isExplored(xpos, ypos) ) {
		return mainGamestate->getMap()->getSquare(xpos, ypos)->givesRoadBonus();
	}
	return false;
}

bool AIInterface::canBeImproved(int xpos, int ypos) {
	if( ai_civilization->isExplored(xpos, ypos) ) {
		return mainGamestate->getMap()->getSquare(xpos, ypos)->canBeImproved(ai_civilization);
	}
	return false;
}

bool AIInterface::hasEnemies(int xpos, int ypos) {
	if( ai_civilization->isExplored(xpos, ypos) ) {
		return mainGamestate->getMap()->getSquare(xpos, ypos)->hasEnemies(ai_civilization);
	}
	return false;
}

const vector<Unit *> *AIInterface::getUnits(int xpos, int ypos) {
	if( ai_civilization->isExplored(xpos, ypos) ) {
		const vector<Unit *> *units = mainGamestate->getMap()->findUnitsAt(xpos, ypos);
		return units;
	}
	return NULL;
}

Unit *AIInterface::getBestDefender(int xpos, int ypos) {
	if( ai_civilization->isExplored(xpos, ypos) ) {
		return mainGamestate->getMap()->getSquare(xpos, ypos)->getBestDefender();
	}
	return NULL;
}

City *AIInterface::getCity(int xpos, int ypos) {
	if( ai_civilization->isExplored(xpos, ypos) ) {
		return mainGamestate->getMap()->findCity(xpos, ypos);
	}
	return NULL;
}

vector<const Buildable *> AIInterface::cityGetBuildables(const City *city, GetbuildablesType type) {
	vector<const Buildable *> buildables;
	if( city->getCivilization() != ai_civilization ) {
		return buildables;
	}
	for(int i=0;i<game_g->getGameData()->getNBuildables();i++) {
		const Buildable *buildable = game_g->getGameData()->getBuildable(i);
		if( type == GETBUILDABLES_IMPROVEMENTS && buildable->getType() != Buildable::TYPE_IMPROVEMENT )
			continue;
		if( type == GETBUILDABLES_UNITS && buildable->getType() != Buildable::TYPE_UNIT )
			continue;
		if( city->canBuild(buildable) ) {
			buildables.push_back(buildable);
		}
	}
	return buildables;
}

const Buildable *AIInterface::cityGetBuildable(const City *city) {
	if( city->getCivilization() != ai_civilization ) {
		return NULL;
	}
	return city->getBuildable();
}

bool AIInterface::canBomb(const UnitTemplate *unit_template, const City *source_city, Pos2D target, bool check_launch_capability) {
	if( check_launch_capability && !source_city->canLaunch(unit_template) ) {
		return false;
	}
	return mainGamestate->canBomb(unit_template, source_city->getPos(), target);
}

void luafunc_returnPointer(lua_State *ls, void *ptr) {
	if( ptr == NULL )
		lua_pushnil(ls);
	else
		lua_pushlightuserdata(ls, ptr);
}

template<class T>
void luafunc_returnArrayPointers(lua_State *ls, const vector<T *> *array) {
	// open/create table
	lua_newtable(ls);
	int top = lua_gettop(ls);

	int table_size = 0;
	if( array != NULL ) {
		table_size = array->size();
		for(int i=0;i<array->size();i++) {
			T *t = array->at(i);
			// new cell
			lua_pushnumber(ls, i+1); // lua arrays start at 1
			lua_pushlightuserdata(ls, t);
			lua_settable(ls, top); // insert the new cell (and pop index/value off stack)
		}
	}

	// close table
	lua_pushliteral(ls, "n");
	lua_pushnumber(ls, table_size); // number of cells
	lua_settable(ls, top);
}

template<class T>
void luafunc_returnArrayConstPointers(lua_State *ls, const vector<const T *> *array) {
	// open/create table
	lua_newtable(ls);
	int top = lua_gettop(ls);

	int table_size = 0;
	if( array != NULL ) {
		table_size = array->size();
		for(int i=0;i<array->size();i++) {
			// need const cast to return UnitTemplate to Lua! Lua won't be able modify it anyway
			T *t = const_cast<T *>(array->at(i));
			// new cell
			lua_pushnumber(ls, i+1); // lua arrays start at 1
			lua_pushlightuserdata(ls, t);
			lua_settable(ls, top); // insert the new cell (and pop index/value off stack)
		}
	}

	// close table
	lua_pushliteral(ls, "n");
	lua_pushnumber(ls, table_size); // number of cells
	lua_settable(ls, top);
}

void luafunc_returnArrayInts(lua_State *ls, const vector<int> *array) {
	// open/create table
	lua_newtable(ls);
	int top = lua_gettop(ls);

	int table_size = 0;
	if( array != NULL ) {
		table_size = array->size();
		for(int i=0;i<array->size();i++) {
			int t = array->at(i);
			// new cell
			lua_pushnumber(ls, i+1); // lua arrays start at 1
			lua_pushinteger(ls, t);
			lua_settable(ls, top); // insert the new cell (and pop index/value off stack)
		}
	}

	// close table
	lua_pushliteral(ls, "n");
	lua_pushnumber(ls, table_size); // number of cells
	lua_settable(ls, top);
}

/* Adds a new technology to the game engine.
 * Args: name (string), base cost (int), requires1 (string), requires2 (string), age (int), ai_weight(int)
 * Returns: pointer to the new technology
 */
int lua_addNewTechnology(lua_State *ls) {
	//VI_log("lua_addNewTechnology()\n");
	//this->addTechnology( technology = new Technology("Agriculture", 100, NULL, NULL, Technology::AGE_ANCIENT, 20) );
	int n_args = lua_gettop(ls);
	if( n_args != 6 ) {
		VI_log("lua_addNewTechnology: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_addNewTechnology: incorrect number of arguments");
		lua_error(ls);
	}
	const char *name = NULL;
	int base_cost = 0;
	const Technology *requires1 = NULL;
	const Technology *requires2 = NULL;
	Technology::Age age = Technology::AGE_ANCIENT;
	int ai_weight = 0;

	int arg_index = 1;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_addNewTechnology: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		name = lua_tostring(ls, arg_index);
	}
	arg_index++;

	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_addNewTechnology: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		base_cost = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_addNewTechnology: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		const char *tech_name = lua_tostring(ls, arg_index);
		if( *tech_name != '\0' ) {
			requires1 = game_g->findTechnology(tech_name);
			if( requires1 == NULL ) {
				VI_log("lua_addNewTechnology: requires1 unknown tech %s\n", tech_name);
				lua_error(ls);
			}
		}
	}
	arg_index++;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_addNewTechnology: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		const char *tech_name = lua_tostring(ls, arg_index);
		if( *tech_name != '\0' ) {
			requires2 = game_g->findTechnology(tech_name);
			if( requires2 == NULL ) {
				VI_log("lua_addNewTechnology: requires2 unknown tech %s\n", tech_name);
				lua_error(ls);
			}
		}
	}
	arg_index++;

	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_addNewTechnology: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		age = static_cast<Technology::Age>(lua_tointeger(ls, arg_index));
	}
	arg_index++;

	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_addNewTechnology: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ai_weight = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	VI_log("lua_addNewTechnology: %s, %d, %s, %s, %d, %d\n", name, base_cost, requires1==NULL?"NULL":requires1->getName(), requires2==NULL?"NULL":requires2->getName(), age, ai_weight);
	Technology *technology = new Technology(name, base_cost, requires1, requires2, age, ai_weight);
	game_g->addTechnology(technology);

	lua_pushlightuserdata(ls, technology);
	return 1;
}

/* Finds a technology by name.
 * Args: name (string)
 * Returns: pointer to the technology
 */
/*int lua_findTechnology(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_findTechnology: incorrect number of arguments\n");
		lua_error(ls);
	}

	const Technology *technology = NULL;
	int arg_index = 1;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_findTechnology: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		const char *tech_name = lua_tostring(ls, arg_index);
		technology = game_g->findTechnology(tech_name);
	}
	arg_index++;

	lua_pushlightuserdata(ls, technology);
	return 1;
}*/

/* Sets the info text for a technology.
 * Args: technology (pointer), info (string)
 * Returns: nothing
 */
int lua_technologySetInfo(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_technologySetInfo: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_technologySetInfo: incorrect number of arguments");
		lua_error(ls);
	}

	Technology *technology = NULL;
	const char *info = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_technologySetInfo: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		technology = static_cast<Technology *>(ptr);
	}
	arg_index++;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_technologySetInfo: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		info = lua_tostring(ls, arg_index);
	}
	arg_index++;

	technology->setInfo(info);

	return 0;
}

/* Sets the ai hint enum for a technology.
 * Args: technology (pointer), aiHint (int)
 * Returns: nothing
 */
int lua_technologySetAIHint(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_technologySetAIHint: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_technologySetAIHint: incorrect number of arguments");
		lua_error(ls);
	}

	Technology *technology = NULL;
	Technology::AIHint aiHint = Technology::AIHINT_NONE;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_technologySetAIHint: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		technology = static_cast<Technology *>(ptr);
	}
	arg_index++;

	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_technologySetAIHint: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		aiHint = static_cast<Technology::AIHint>(lua_tointeger(ls, arg_index));
	}
	arg_index++;

	technology->setAIHint(aiHint);

	return 0;
}

/* Adds a new race to the game engine.
 * Args: name (string), adjective name (string), color r (int), color g (int), color b (int), image (string)
 * Returns: pointer to the new race
 */
int lua_addNewRace(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 6 ) {
		VI_log("lua_addNewRace: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_addNewRace: incorrect number of arguments");
		lua_error(ls);
	}
	const char *name = NULL;
	const char *name_adjective = NULL;
	unsigned char rgb[] = {0, 0, 0};
	const char *image = NULL;

	int arg_index = 1;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_addNewRace: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		name = lua_tostring(ls, arg_index);
	}
	arg_index++;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_addNewRace: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		name_adjective = lua_tostring(ls, arg_index);
	}
	arg_index++;

	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_addNewRace: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		int col = lua_tointeger(ls, arg_index);
		if( col < 0 || col > 255 ) {
			VI_log("lua_addNewRace: argument %d , number not valid range\n", arg_index);
		}
		rgb[0] = static_cast<unsigned char>(col);
	}
	arg_index++;

	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_addNewRace: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		int col = lua_tointeger(ls, arg_index);
		if( col < 0 || col > 255 ) {
			VI_log("lua_addNewRace: argument %d , number not valid range\n", arg_index);
		}
		rgb[1] = static_cast<unsigned char>(col);
	}
	arg_index++;

	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_addNewRace: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		int col = lua_tointeger(ls, arg_index);
		if( col < 0 || col > 255 ) {
			VI_log("lua_addNewRace: argument %d , number not valid range\n", arg_index);
		}
		rgb[2] = static_cast<unsigned char>(col);
	}
	arg_index++;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_addNewRace: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		image = lua_tostring(ls, arg_index);
	}
	arg_index++;

	VI_log("lua_addNewRace: %s, %s, %d, %d, %d, %s\n", name, name_adjective, rgb[0], rgb[1], rgb[2], image);
	Race *race = new Race(name, name_adjective, rgb, image);
	// constructor automatically adds race to the game engine

	lua_pushlightuserdata(ls, race);
	return 1;
}

/* Sets the starting technology for a race.
 * Args: race (pointer), technology (string)
 * Returns: nothing
 */
int lua_raceSetStartingTechnology(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_raceSetStartingTechnology: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_raceSetStartingTechnology: incorrect number of arguments");
		lua_error(ls);
	}

	Race *race = NULL;
	const Technology *technology = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_raceSetStartingTechnology: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		race = static_cast<Race *>(ptr);
	}
	arg_index++;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_raceSetStartingTechnology: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		const char *tech_name = lua_tostring(ls, arg_index);
		if( *tech_name != '\0' ) {
			technology = game_g->findTechnology(tech_name);
			if( technology == NULL ) {
				VI_log("lua_raceSetStartingTechnology: requires2 unknown tech %s\n", tech_name);
				lua_error(ls);
			}
		}
	}
	arg_index++;

	race->setStartingTechnology(technology);

	return 0;
}

/* Adds a city name for a race.
 * Args: race (pointer), name (string)
 * Returns: nothing
 */
int lua_raceAddCityName(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_raceAddCityName: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_raceAddCityName: incorrect number of arguments");
		lua_error(ls);
	}

	Race *race = NULL;
	const char *name = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_raceAddCityName: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		race = static_cast<Race *>(ptr);
	}
	arg_index++;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_raceAddCityName: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		name = lua_tostring(ls, arg_index);
	}
	arg_index++;

	race->addCityName(name);

	return 0;
}

/* Finds a buildable by name.
 * Args: name (string)
 * Returns: pointer to the buildable
 */
int lua_findBuildable(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_findBuildable: incorrect number of arguments\n");
		lua_error(ls);
	}

	const Buildable *buildable = NULL;
	int arg_index = 1;

	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_findBuildable: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		const char *buildable_name = lua_tostring(ls, arg_index);
		buildable = game_g->findBuildable(buildable_name);
	}
	arg_index++;

	lua_pushlightuserdata(ls, const_cast<Buildable *>(buildable));
	return 1;
}

/* Returns a buildable's name, type, cost
 * Args: buildable (pointer)
 * Returns: name (string), type BUILDABLETYPE_* (int), cost (int)
 */
int lua_buildableGetInfo(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_buildableGetInfo: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_buildableGetInfo: incorrect number of arguments");
		lua_error(ls);
	}

	const char *name = NULL;
	Buildable::Type type = Buildable::TYPE_IMPROVEMENT;
	int cost = 0;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_buildableGetInfo: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		Buildable *buildable = static_cast<Buildable *>(ptr);
		UnitTemplate *test = static_cast<UnitTemplate *>(ptr);
		name = buildable->getName();
		type = buildable->getType();
		cost = buildable->getCost();
	}
	arg_index++;

	lua_pushstring(ls, name);
	lua_pushinteger(ls, (int)type);
	lua_pushinteger(ls, cost);
	return 3;
}

/* Returns an improvement's ai_weight, ai_hint
 * Args: improvement (pointer)
 * Returns: ai_weight (int), ai_hint AIHINT_* (int)
 */
int lua_improvementGetInfo(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_improvementGetInfo: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_improvementGetInfo: incorrect number of arguments");
		lua_error(ls);
	}

	int ai_weight = 0;
	Improvement::AIHint ai_hint = Improvement::AIHINT_NONE;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_improvementGetInfo: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		const Improvement *improvement = static_cast<const Improvement *>(ptr);
		ai_weight = improvement->getAIWeight();
		ai_hint = improvement->getAIHint();
	}
	arg_index++;

	lua_pushinteger(ls, ai_weight);
	lua_pushinteger(ls, static_cast<int>(ai_hint));
	return 2;
}

/* Returns a unit template's types.
 * Args: unit_template (pointer)
 * Returns: is_foot (boolean), is_air (boolean), is_missile (boolean)
 */
int lua_unittemplateType(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_unittemplateType: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_unittemplateType: incorrect number of arguments");
		lua_error(ls);
	}

	bool is_foot = false, is_air = false, is_missile = false;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_unittemplateType: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		UnitTemplate *unit_template = static_cast<UnitTemplate *>(ptr);
		is_foot = unit_template->isFoot();
		is_air = unit_template->isAir();
		is_missile = unit_template->isMissile();
	}
	arg_index++;

	lua_pushboolean(ls, is_foot);
	lua_pushboolean(ls, is_air);
	lua_pushboolean(ls, is_missile);
	return 3;
}

/* Returns a unit template's stats.
 * Args: unit_template (pointer)
 * Returns: attack (int), defence (int), moves (int), cost (int)
 */
int lua_unittemplateStats(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_unittemplateStats: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_unittemplateStats: incorrect number of arguments");
		lua_error(ls);
	}

	int attack = 0, defence = 0, moves = 0, cost = 0;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_unittemplateStats: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		UnitTemplate *unit_template = static_cast<UnitTemplate *>(ptr);
		attack = unit_template->getAttack();
		defence = unit_template->getDefence();
		moves = unit_template->getMoves();
		cost = unit_template->getCost();
	}
	arg_index++;

	lua_pushinteger(ls, attack);
	lua_pushinteger(ls, defence);
	lua_pushinteger(ls, moves);
	lua_pushinteger(ls, cost);
	return 4;
}

/* Returns a unit template's air stats.
 * Args: unit_template (pointer)
 * Returns: air_attack (int), air_defence (int), bombard_power (int), bombard (int), air_range (int), nucleartype NUCLEARTYPE_* (int)
 */
int lua_unittemplateAirStats(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_unittemplateAirStats: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_unittemplateAirStats: incorrect number of arguments");
		lua_error(ls);
	}

	int air_attack = 0, air_defence = 0, bombard_power = 0, bombard = 0, air_range = 0;
	UnitTemplate::NuclearType nucleartype = UnitTemplate::NUCLEARTYPE_NONE;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_unittemplateAirStats: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		UnitTemplate *unit_template = static_cast<UnitTemplate *>(ptr);
		air_attack = unit_template->getAirAttack();
		air_defence = unit_template->getAirDefence();
		bombard_power = unit_template->getBombardPower();
		bombard = unit_template->getBombard();
		air_range = unit_template->getAirRange();
		nucleartype = unit_template->getNuclearType();
	}
	arg_index++;

	lua_pushinteger(ls, air_attack);
	lua_pushinteger(ls, air_defence);
	lua_pushinteger(ls, bombard_power);
	lua_pushinteger(ls, bombard);
	lua_pushinteger(ls, air_range);
	lua_pushinteger(ls, static_cast<int>(nucleartype));
	return 6;
}

/* Returns a unit template's abilities.
 * Args: unit_template (pointer)
 * Returns: can_build_roads (boolean), can_build_city (boolean), can_bomb (boolean)
 */
int lua_unittemplateAbility(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_unittemplateAbility: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_unittemplateAbility: incorrect number of arguments");
		lua_error(ls);
	}

	bool can_build_roads = false, can_build_city = false, can_bomb = false;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_unittemplateAbility: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		UnitTemplate *unit_template = static_cast<UnitTemplate *>(ptr);
		can_build_roads = unit_template->canBuildRoads();
		can_build_city = unit_template->canBuildCity();
		can_bomb = unit_template->canBomb();
	}
	arg_index++;

	lua_pushboolean(ls, can_build_roads);
	lua_pushboolean(ls, can_build_city);
	lua_pushboolean(ls, can_bomb);
	return 3;
}

/* Returns a unit's template.
 * Args: unit (pointer)
 * Returns: unit_template (pointer)
 */
int lua_unitGetTemplate(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_unitGetTemplate: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_unitGetTemplate: incorrect number of arguments");
		lua_error(ls);
	}

	const UnitTemplate *unit_template = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_unitGetTemplate: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		Unit *unit = static_cast<Unit *>(ptr);
		unit_template = unit->getTemplate();
	}
	arg_index++;

	// need const cast to return UnitTemplate to Lua! Lua won't be able modify it anyway
	lua_pushlightuserdata(ls, const_cast<UnitTemplate *>(unit_template));
	return 1;
}

/* Returns a unit's civilization.
 * Args: unit (pointer)
 * Returns: civilization (pointer)
 */
int lua_unitGetCivilization(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_unitGetCivilization: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_unitGetCivilization: incorrect number of arguments");
		lua_error(ls);
	}

	Civilization *civilization = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_unitGetCivilization: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		Unit *unit = static_cast<Unit *>(ptr);
		civilization = unit->getCivilization();
	}
	arg_index++;

	lua_pushlightuserdata(ls, civilization);
	return 1;
}

/* Returns a unit's position.
 * Args: unit (pointer)
 * Returns: xpos (int), ypos (int)
 */
int lua_unitGetPos(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_unitGetPos: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_unitGetPos: incorrect number of arguments");
		lua_error(ls);
	}

	Pos2D pos;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_unitGetPos: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		Unit *unit = static_cast<Unit *>(ptr);
		pos = unit->getPos();
	}
	arg_index++;

	lua_pushinteger(ls, pos.x);
	lua_pushinteger(ls, pos.y);
	return 2;
}

/* Returns a civilization's name.
 * Args: civilization (pointer)
 * Returns: name (string)
 */
int lua_civilizationGetName(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_civilizationGetName: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_civilizationGetName: incorrect number of arguments");
		lua_error(ls);
	}

	const char *name = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_civilizationGetName: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		const Civilization *civilization = static_cast<const Civilization *>(ptr);
		name = civilization->getName();
	}
	arg_index++;

	lua_pushstring(ls, name);
	return 1;
}

/* Returns a city's civilization.
 * Args: city (pointer)
 * Returns: civilization (pointer)
 */
int lua_cityGetCivilization(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_cityGetCivilization: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityGetCivilization: incorrect number of arguments");
		lua_error(ls);
	}

	Civilization *civilization = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityGetPos: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		const City *city = static_cast<const City *>(ptr);
		civilization = city->getCivilization();
	}
	arg_index++;

	lua_pushlightuserdata(ls, civilization);
	return 1;
}

/* Returns a city's position.
 * Args: city (pointer)
 * Returns: xpos (int), ypos (int)
 */
int lua_cityGetPos(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_cityGetPos: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityGetPos: incorrect number of arguments");
		lua_error(ls);
	}

	Pos2D pos;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityGetPos: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		const City *city = static_cast<const City *>(ptr);
		pos = city->getPos();
	}
	arg_index++;

	lua_pushinteger(ls, pos.x);
	lua_pushinteger(ls, pos.y);
	return 2;
}

/* Returns a city's name.
 * Args: city (pointer)
 * Returns: name (string)
 */
int lua_cityGetName(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_cityGetName: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityGetName: incorrect number of arguments");
		lua_error(ls);
	}

	const char *name = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityGetName: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		const City *city = static_cast<const City *>(ptr);
		name = city->getName();
		//VI_log("lua_cityGetName: %s\n", city->getName());
	}
	arg_index++;

	lua_pushstring(ls, name);
	return 1;
}

/* Returns a city's size.
 * Args: city (pointer)
 * Returns: size (int)
 */
int lua_cityGetPublicInfo(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_cityGetPublicInfo: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityGetPublicInfo: incorrect number of arguments");
		lua_error(ls);
	}

	int size = 0;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityGetPublicInfo: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		const City *city = static_cast<const City *>(ptr);
		size = city->getSize();
	}
	arg_index++;

	lua_pushinteger(ls, size);
	return 1;
}

/* Returns whether "capital" (first city), production, science, sea travel range, air travel range.
 * is_capital is always false, production, science sea_travel_range, air_travel_range always 0, for cities that don't belong to the AI.
 * Args: city (pointer)
 * Returns: is_capital (boolean), production (int), science (int), sea_travel_range (int), air_travel_range (int)
 */
int lua_cityGetInfo(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_cityGetInfo: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityGetInfo: incorrect number of arguments");
		lua_error(ls);
	}

	bool is_capital = false;
	int production = 0, science = 0, sea_travel_range = 0, air_travel_range = 0;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityGetInfo: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		const City *city = static_cast<const City *>(ptr);
		if( city->getCivilization() == AIInterface::getAICivilization() ) {
			is_capital = city == AIInterface::getAICivilization()->getCity(0);
			production = city->getProduction();
			science = city->getScience();
			sea_travel_range = city->getTravelRange(false);
			air_travel_range = city->getTravelRange(true);
		}
	}
	arg_index++;

	lua_pushboolean(ls, is_capital);
	lua_pushinteger(ls, production);
	lua_pushinteger(ls, science);
	lua_pushinteger(ls, sea_travel_range);
	lua_pushinteger(ls, air_travel_range);
	return 5;
}

/* Returns sea travel range, air travel range.
 * sea_travel_range, air_travel_range is always 0 for cities that don't belong to the AI.
 * Args: city (pointer)
 * Returns: sea_travel_range (int), air_travel_range (int)
 */
int lua_cityGetRanges(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_cityGetRanges: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityGetRanges: incorrect number of arguments");
		lua_error(ls);
	}

	int sea_travel_range = 0, air_travel_range = 0;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityGetRanges: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		const City *city = static_cast<const City *>(ptr);
		if( city->getCivilization() == AIInterface::getAICivilization() ) {
			sea_travel_range = city->getTravelRange(false);
			air_travel_range = city->getTravelRange(true);
		}
	}
	arg_index++;

	lua_pushinteger(ls, sea_travel_range);
	lua_pushinteger(ls, air_travel_range);
	return 2;
}

/* Returns whether a city can build something.
 * can_build is always false for cities that don't belong to the AI.
 * Args: city (pointer), buildable (pointer)
 * Returns: can_build (boolean)
 */
int lua_cityCanBuild(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_cityCanBuild: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityCanBuild: incorrect number of arguments");
		lua_error(ls);
	}

	const City *city = NULL;
	const Buildable *buildable = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityCanBuild: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		city = static_cast<const City *>(ptr);
	}
	arg_index++;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityCanBuild: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		//const char *buildable_name = lua_tostring(ls, arg_index);
		//buildable = static_cast<const Buildable *>(lua_touserdata(ls, arg_index));
		void *ptr = lua_touserdata(ls, arg_index);
		buildable = static_cast<const Buildable *>(ptr);
	}
	arg_index++;

	bool can_build = city->getCivilization() == AIInterface::getAICivilization() && city->canBuild(buildable);
	lua_pushboolean(ls, can_build);
	return 1;
}

/* Returns whether a city can launch an air unit type.
 * can_launch is always false for cities that don't belong to the AI.
 * Args: city (pointer), unit_template (pointer)
 * Returns: can_launch (boolean)
 */
int lua_cityCanLaunch(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_cityCanLaunch: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityCanLaunch: incorrect number of arguments");
		lua_error(ls);
	}

	const City *city = NULL;
	const UnitTemplate *unit_template = NULL;

	int arg_index = 1;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityCanLaunch: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		city = static_cast<const City *>(ptr);
	}
	arg_index++;

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_cityCanLaunch: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		unit_template = static_cast<const UnitTemplate *>(ptr);
	}
	arg_index++;

	bool can_launch = city->getCivilization() == AIInterface::getAICivilization() && city->canLaunch(unit_template);
	lua_pushboolean(ls, can_launch);
	return 1;
}

/* Returns whether a city can be built at this location.
 * (Always returns false if square not explored.)
 * Args: xpos (int), ypos (int)
 * Returns: can_build_city (boolean)
 */
int lua_canBuildCity(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_canBuildCity: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_canBuildCity: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_canBuildCity: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_canBuildCity: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	try {
		bool can_build = AIInterface::canBuildCity(xpos, ypos);
		lua_pushboolean(ls, can_build);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_canBuildCity: caught error");
		lua_error(ls);
	}

	return 1;
}

/* Returns whether a square can have improvements (roads, railways) built at this location.
 * (Always returns false if square not explored.)
 * Args: xpos (int), ypos (int)
 * Returns: can_be_improved (boolean)
 */
int lua_canBeImproved(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_canBeImproved: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_canBuildCity: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_canBeImproved: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_canBeImproved: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	try {
		bool can_be_improved = AIInterface::canBeImproved(xpos, ypos);
		lua_pushboolean(ls, can_be_improved);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_canBeImproved: caught error");
		lua_error(ls);
	}

	return 1;
}

/* Returns whether the square is explored.
 * Args: xpos (int), ypos (int)
 * Returns: explored (boolean)
 */
int lua_isExplored(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_isExplored: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_isExplored: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_isExplored: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_isExplored: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	bool is_explored = AIInterface::getAICivilization()->isExplored(xpos, ypos);
	lua_pushboolean(ls, is_explored);
	return 1;
}

/* Returns true if the square is unexplored, but at least one adjacent square is explored.
 * Args: xpos (int), ypos (int)
 * Returns: explorable (boolean)
 */
int lua_isExplorable(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_isExplorable: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_isExplorable: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_isExplorable: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_isExplorable: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	bool is_explorable = false;
	if( !AIInterface::getAICivilization()->isExplored(xpos, ypos) ) {
		is_explorable = AIInterface::hasAdjacentExplored(xpos, ypos);
	}
	lua_pushboolean(ls, is_explorable);
	return 1;
}

/* Returns the terrain at this square.
 * Returns TYPE_UNKNOWN if not explored.
 * Args: xpos (int), ypos (int)
 * Returns: terrain TYPE_* (int)
 */
int lua_getTerrain(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_getTerrain: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getTerrain: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getTerrain: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getTerrain: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	MapSquare::Type terrain = AIInterface::getTerrain(xpos, ypos);
	lua_pushinteger(ls, static_cast<int>(terrain));
	return 1;
}

/* Returns the road at this square.
 * Returns ROAD_UNKNOWN if not explored.
 * Args: xpos (int), ypos (int)
 * Returns: road ROAD_* (int)
 */
int lua_getRoad(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_getRoad: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getRoad: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getRoad: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getRoad: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	MapSquare::Road road = AIInterface::getRoad(xpos, ypos);
	lua_pushinteger(ls, static_cast<int>(road));
	return 1;
}

/* Returns whether there's an adjacent unexplored square - i.e., moving to this square will uncover it.
 * Args: xpos (int), ypos (int)
 * Returns: adjacent_unexplored (boolean)
 */
int lua_hasAdjacentUnexplored(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_hasAdjacentUnexplored: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_hasAdjacentUnexplored: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_hasAdjacentUnexplored: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_hasAdjacentUnexplored: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	bool adjacent_unexplored = false;
	try {
		adjacent_unexplored = AIInterface::hasAdjacentUnexplored(xpos, ypos);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_hasAdjacentUnexplored: caught error");
		lua_error(ls);
	}

	lua_pushboolean(ls, adjacent_unexplored);
	return 1;
}

/* Returns whether there are enemy units at a square. This means those belonging to a civilization that the AI is at war with; and only for units with attack greater than 0.
 * Args: xpos (int), ypos (int)
 * (Always returns false if square not explored.)
 * Returns: has_enemies (boolean)
 */
int lua_hasEnemies(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_hasEnemies: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_hasEnemies: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_hasEnemies: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_hasEnemies: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	bool has_enemies = false;
	try {
		has_enemies = AIInterface::hasEnemies(xpos, ypos);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_hasEnemies: caught error");
		lua_error(ls);
	}

	lua_pushboolean(ls, has_enemies);
	return 1;
}

/* Returns the units at a square.
 * Args: xpos (int), ypos (int)
 * (Always returns no units if square not explored.)
 * Returns: units (table of pointers)
 */
int lua_getUnitsAt(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_getUnitsAt: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getUnitsAt: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getUnitsAt: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getUnitsAt: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	try {
		const vector<Unit *> *units = AIInterface::getUnits(xpos, ypos);
		/*
		// open/create table
		lua_newtable(ls);
		int top = lua_gettop(ls);

		int table_size = 0;
		if( units != NULL ) {
			table_size = units->size();
			for(int i=0;i<units->size();i++) {
				Unit *unit = units->at(i);
				// new cell
				lua_pushnumber(ls, i+1); // lua arrays start at 1
				lua_pushlightuserdata(ls, unit);
				lua_settable(ls, top); // insert the new cell (and pop index/value off stack)
			}
		}

		// close table
		lua_pushliteral(ls, "n");
		lua_pushnumber(ls, table_size); // number of cells
		lua_settable(ls, top);*/
		luafunc_returnArrayPointers(ls, units);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getUnitsAt: caught error");
		lua_error(ls);
	}

	return 1;
}

/* Returns the best defender at a square.
 * Args: xpos (int), ypos (int)
 * (Always returns NULL if square not explored, or no units.)
 * Returns: unit (pointer)
 */
int lua_getBestDefenderAt(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_getBestDefenderAt: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getBestDefenderAt: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getBestDefenderAt: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getBestDefenderAt: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	Unit *unit = NULL;
	try {
		unit = AIInterface::getBestDefender(xpos, ypos);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getBestDefenderAt: caught error");
		lua_error(ls);
	}

	lua_pushlightuserdata(ls, unit);
	return 1;
}

/* Returns the city at a square.
 * (Always returns NULL if square not explored.)
 * Args: xpos (int), ypos (int)
 * Returns: city (pointer)
 */
int lua_getCityAt(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_getCityAt: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getCityAt: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getCityAt: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getCityAt: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	City *city = NULL;
	try {
		city = AIInterface::getCity(xpos, ypos);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getCityAt: caught error");
		lua_error(ls);
	}

	//VI_log("lua_getCityAt returns: %d\n", city);
	luafunc_returnPointer(ls, city);
	return 1;
}

/* Returns a list of explored adjacent squares.
 * Args: xpos (int), ypos (int), include_centre (boolean)
 * Returns: xposes (table of pointers), yposes (table of pointers)
 */
int lua_getAdjacentSquares(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 3 ) {
		VI_log("lua_getAdjacentSquares: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getAdjacentSquares: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getAdjacentSquares: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getAdjacentSquares: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	bool include_centre = false;
	if( !lua_isboolean(ls, arg_index) ) {
		VI_log("lua_getAdjacentSquares: argument %d not a boolean\n", arg_index);
		lua_error(ls);
	}
	else {
		include_centre = lua_toboolean(ls, arg_index);
	}
	arg_index++;

	vector<Pos2D> adjacent_squares = AIInterface::getAdjacentSquares(xpos, ypos, include_centre);
	vector<int> xposes;
	vector<int> yposes;
	for(int i=0;i<adjacent_squares.size();i++) {
		Pos2D pos = adjacent_squares.at(i);
		xposes.push_back( pos.x );
		yposes.push_back( pos.y );
	}
	luafunc_returnArrayInts(ls, &xposes);
	luafunc_returnArrayInts(ls, &yposes);
	return 2;
}

/* Returns a list of explored city squares.
 * Args: xpos (int), ypos (int), include_centre (boolean)
 * Returns: xposes (table of pointers), yposes (table of pointers)
 */
int lua_getCitySquares(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 3 ) {
		VI_log("lua_getCitySquares: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getCitySquares: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getCitySquares: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getCitySquares: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	bool include_centre = false;
	if( !lua_isboolean(ls, arg_index) ) {
		VI_log("lua_getCitySquares: argument %d not a boolean\n", arg_index);
		lua_error(ls);
	}
	else {
		include_centre = lua_toboolean(ls, arg_index);
	}
	arg_index++;

	vector<Pos2D> city_squares = AIInterface::getCitySquares(xpos, ypos, include_centre);
	vector<int> xposes;
	vector<int> yposes;
	for(int i=0;i<city_squares.size();i++) {
		Pos2D pos = city_squares.at(i);
		xposes.push_back( pos.x );
		yposes.push_back( pos.y );
	}
	luafunc_returnArrayInts(ls, &xposes);
	luafunc_returnArrayInts(ls, &yposes);
	return 2;
}

/* Returns whether the AI has discovered all technologies.
 * Args: none
 * Returns: answer (boolean)
 */
int lua_haveAllTechnology(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 0 ) {
		VI_log("lua_haveAllTechnology: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_haveAllTechnology: incorrect number of arguments");
		lua_error(ls);
	}

	bool answer = AIInterface::getAICivilization()->hasAllTechnology();
	lua_pushboolean(ls, answer);

	return 1;
}

/* Returns the list of the civilizations.
 * Args: type GETCIVILIZATIONS_* (int)
 * Returns: civilizations (table of pointers)
 */
int lua_getCivilizations(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_getCivilizations: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getCivilizations: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	AIInterface::GetcivilizationsType type = AIInterface::GETCIVILIZATIONS_ALL;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getCivilizations: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		type = static_cast<AIInterface::GetcivilizationsType>(lua_tointeger(ls, arg_index));
	}
	arg_index++;

	try {
		const vector<Civilization *> civilizations = AIInterface::getCivilizations(type);
		luafunc_returnArrayPointers(ls, &civilizations);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getCivilizations: caught error");
		lua_error(ls);
	}

	return 1;
}

/* Returns the list of the civilization's cities
 * Args: civilization (pointer)
 * Returns: cities (table of pointers)
 */
int lua_getCities(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_getCities: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getCities: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	Civilization *civilization = NULL;
	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_getCities: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		civilization = static_cast<Civilization *>(ptr);
	}
	arg_index++;

	try {
		const vector<City *> cities = AIInterface::getCities(civilization);
		luafunc_returnArrayPointers(ls, &cities);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getCities: caught error");
		lua_error(ls);
	}

	return 1;
}

/* Returns the list of the AI's cities
 * Args: none
 * Returns: cities (table of pointers)
 */
int lua_getMyCities(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 0 ) {
		VI_log("lua_getMyCities: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getMyCities: incorrect number of arguments");
		lua_error(ls);
	}

	try {
		const vector<City *> cities = AIInterface::getCities(AIInterface::getAICivilization());
		luafunc_returnArrayPointers(ls, &cities);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getMyCities: caught error");
		lua_error(ls);
	}

	return 1;
}

/* Returns the list of the civilization's units
 * Args: civilization (pointer)
 * Returns: units (table of pointers)
 */
int lua_getUnits(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_getUnits: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getUnits: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	Civilization *civilization = NULL;
	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_getUnits: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		civilization = static_cast<Civilization *>(ptr);
	}
	arg_index++;

	try {
		const vector<Unit *> units = AIInterface::getUnits(civilization);
		luafunc_returnArrayPointers(ls, &units);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getUnits: caught error");
		lua_error(ls);
	}

	return 1;
}

/* Returns the list of the AI's units
 * Args: none
 * Returns: units (table of pointers)
 */
int lua_getMyUnits(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 0 ) {
		VI_log("lua_getMyUnits: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getMyUnits: incorrect number of arguments");
		lua_error(ls);
	}

	try {
		const vector<Unit *> units = AIInterface::getUnits(AIInterface::getAICivilization());
		luafunc_returnArrayPointers(ls, &units);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getMyUnits: caught error");
		lua_error(ls);
	}

	return 1;
}

/* Returns relationship info for a civilization
 * Args: civilization (pointer)
 * Returns: made_contact (boolean), at_war (boolean) (at_war is false if not made contact)
 */
int lua_getRelationshipInfo(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_getRelationshipInfo: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getRelationshipInfo: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	Civilization *civilization = NULL;
	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_getRelationshipInfo: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		civilization = static_cast<Civilization *>(ptr);
	}
	arg_index++;

	bool made_contact = false, at_war = false;
	try {
		const Relationship *relationship = AIInterface::findRelationship(civilization);
		made_contact = relationship->madeContact();
		at_war = relationship->getStatus() != Relationship::STATUS_PEACE;
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getRelationshipInfo: caught error");
		lua_error(ls);
	}

	lua_pushboolean(ls, made_contact);
	lua_pushboolean(ls, at_war);
	return 2;
}

/* Returns the buildables that a city can build.
 * (Always returns empty array if city doesn't belong to the AI.)
 * Args: city (pointer), type GETBUILDABLES_* (int)
 * Returns: buildables (table of pointers)
 */
int lua_cityGetBuildables(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 2 ) {
		VI_log("lua_cityGetBuildables: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityGetBuildables: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	const City *city = NULL;
	AIInterface::GetbuildablesType type = AIInterface::GETBUILDABLES_ALL;

	if( !lua_isuserdata(ls, arg_index) ) {
		VI_log("lua_cityGetBuildables: argument %d not userdata\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		city = static_cast<const City *>(ptr);
	}
	arg_index++;

	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_cityGetBuildables: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		int value = lua_tonumber(ls, arg_index);
		type = static_cast<AIInterface::GetbuildablesType>(value);
	}
	arg_index++;

	try {
		vector<const Buildable *> buildables = AIInterface::cityGetBuildables(city, type);
		luafunc_returnArrayConstPointers(ls, &buildables);
		/*
		// open/create table
		lua_newtable(ls);
		int top = lua_gettop(ls);

		int table_size = 0;
		table_size = buildables.size();
		for(int i=0;i<buildables.size();i++) {
			const Buildable *buildable = buildables.at(i);
			// new cell
			lua_pushnumber(ls, i+1); // lua arrays start at 1
			// need const cast to return UnitTemplate to Lua! Lua won't be able modify it anyway
			lua_pushlightuserdata(ls, const_cast<Buildable *>(buildable));
			lua_settable(ls, top); // insert the new cell (and pop index/value off stack)
		}

		// close table
		lua_pushliteral(ls, "n");
		lua_pushnumber(ls, table_size); // number of cells
		lua_settable(ls, top);
		*/
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_cityGetBuildables: caught error");
		lua_error(ls);
	}

	return 1;
}

/* Returns the buildable that a city is currently building.
 * (Always returns NULL if not building anything, or city doesn't belong to the AI.)
 * Args: city (pointer)
 * Returns: buildable (pointer)
 */
int lua_cityGetBuildable(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_cityGetBuildable: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_cityGetBuildable: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	const City *city = NULL;

	if( !lua_isuserdata(ls, arg_index) ) {
		VI_log("lua_cityGetBuildable: argument %d not userdata\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		city = static_cast<const City *>(ptr);
	}
	arg_index++;

	const Buildable *buildable = NULL;
	try {
		buildable = AIInterface::cityGetBuildable(city);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_cityGetBuildable: caught error");
		lua_error(ls);
	}

	// need const cast to return UnitTemplate to Lua! Lua won't be able modify it anyway
	//lua_pushlightuserdata(ls, const_cast<Buildable *>(buildable));
	luafunc_returnPointer(ls, const_cast<Buildable *>(buildable));
	return 1;
}

/* Returns the game year
 * Args: none
 * Returns: year (int)
 */
int lua_getYear(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 0 ) {
		VI_log("lua_getYear: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getYear: incorrect number of arguments");
		lua_error(ls);
	}

	int year = -1;

	try {
		year = AIInterface::getYear();
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_getYear: caught error");
		lua_error(ls);
	}

	lua_pushinteger(ls, year);
	return 1;
}

/* Returns whether an air unit can bomb a square.
 * Args: unit_template (pointer), city (pointer), target_xpos (int), target_ypos (int), check_launch_capability (boolean)
 * Returns: can_bomb (boolean)
 */
int lua_canBomb(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 5 ) {
		VI_log("lua_canBomb: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_canBomb: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	const UnitTemplate *unit_template = NULL;
	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_canBomb: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		unit_template = static_cast<const UnitTemplate *>(ptr);
	}
	arg_index++;

	const City *city = NULL;

	if( !lua_isuserdata(ls, arg_index) ) {
		VI_log("lua_canBomb: argument %d not userdata\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		city = static_cast<const City *>(ptr);
	}
	arg_index++;

	int target_xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_canBomb: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		target_xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int target_ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_canBomb: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		target_ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;


	bool check_launch_capability = false;
	if( !lua_isboolean(ls, arg_index) ) {
		VI_log("lua_canBomb: argument %d not a boolean\n", arg_index);
		lua_error(ls);
	}
	else {
		check_launch_capability = lua_toboolean(ls, arg_index);
	}
	arg_index++;

	bool can_bomb = false;

	try {
		can_bomb = AIInterface::canBomb(unit_template, city, Pos2D(target_xpos, target_ypos), check_launch_capability);
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_canBomb: caught error");
		lua_error(ls);
	}

	lua_pushboolean(ls, can_bomb);
	return 1;
}

/* Calculates distance map.
 * Args: unit (pointer), xpos (int), ypos (int)
 * unit may be NULL, to indicate a general land unit.
 * Returns: costs (table of ints), squares (table of ints), width (int), height (int)
 */
int lua_calculateDistanceMap(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 3 ) {
		VI_log("lua_calculateDistanceMap: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_calculateDistanceMap: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	const Unit *unit = NULL;
	if( !lua_isnil(ls, arg_index) && !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_calculateDistanceMap: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		unit = static_cast<const Unit *>(ptr);
	}
	arg_index++;

	int xpos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_calculateDistanceMap: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		xpos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int ypos = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_calculateDistanceMap: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		ypos = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	const Map::Distance *dists = NULL;
	int width = 0, height = 0;

	try {
		dists = AIInterface::calculateDistanceMap(&width, &height, unit, xpos, ypos);
		//dists = new Map::Distance[width*height];
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_calculateDistanceMap: caught error");
		lua_error(ls);
	}

	/*int time_s = clock();
	//VI_log(">>> return table\n");
	vector<int> costs;
	vector<int> squares;
	for(int i=0;i<width*height;i++) {
		costs.push_back( dists[i].cost );
		squares.push_back( dists[i].squares );
	}
	delete [] dists;
	luafunc_returnArrayInts(ls, &costs);
	luafunc_returnArrayInts(ls, &squares);*/

	/*static int total_time = 0;
	int time = clock() - time_s;
	total_time += time;
	VI_log("<<< %d : %d\n", time, total_time);*/

	// need const cast for Lua!
	lua_pushlightuserdata(ls, const_cast<Map::Distance *>(dists));

	lua_pushinteger(ls, width);
	lua_pushinteger(ls, height);
	//return 4;
	return 3;
}

/* Returns the distance cost.
 * Args: dists (pointer), width (int), x (int), y (int)
 * Returns: distance cost (int)
 */
int lua_getDistanceCost(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 4 ) {
		VI_log("lua_getDistanceCost: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_getDistanceCost: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	const Map::Distance *dists = NULL;
	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_getDistanceCost: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		dists = static_cast<const Map::Distance *>(ptr);
	}
	arg_index++;

	int width = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getDistanceCost: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		width = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int x = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getDistanceCost: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		x = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int y = 0;
	if( !lua_isnumber(ls, arg_index) ) {
		VI_log("lua_getDistanceCost: argument %d not a number\n", arg_index);
		lua_error(ls);
	}
	else {
		y = lua_tointeger(ls, arg_index);
	}
	arg_index++;

	int dist = dists[y * width + x].cost;
	lua_pushinteger(ls, dist);
	return 1;
}

/* Automates a worker unit.
 * Args: unit (pointer), costs (table of ints), squares (table of ints)
 * Returns: found_target (boolean), dist (int), candidates_x (table of ints), candidates_y (table of ints)
 */
int lua_automateWorker(lua_State *ls) {
	int n_args = lua_gettop(ls);
	//if( n_args != 3 ) {
	if( n_args != 2 ) {
		VI_log("lua_automateWorker: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_automateWorker: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	const Unit *unit = NULL;
	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_automateWorker: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		unit = static_cast<const Unit *>(ptr);
	}
	arg_index++;

	int time_s = clock();
	//Map::Distance *dists = NULL;
	const Map::Distance *dists = NULL;
	/*int size = 0;

	if( !lua_istable(ls, arg_index) ) {
		VI_log("lua_automateWorker: argument %d not a table\n", arg_index);
		lua_error(ls);
	}
	else {
		size = lua_objlen(ls, arg_index);
		dists = new Map::Distance[size];
		for(int i=0;i<size;i++) {
			lua_pushinteger(ls, i+1);
			lua_gettable(ls, arg_index);
			if( !lua_isnumber(ls, -1) ) {
				VI_log("lua_automateWorker: argument %d table index %d not a number\n", arg_index, i+1);
				break;
			}
			dists[i].cost = lua_tointeger(ls, -1);
			lua_pop(ls, 1); 
		}
	}
	arg_index++;

	if( !lua_istable(ls, arg_index) ) {
		VI_log("lua_automateWorker: argument %d not a table\n", arg_index);
		lua_error(ls);
	}
	else {
		if( size != lua_objlen(ls, arg_index) ) {
			VI_log("lua_automateWorker: argument %d unexpected length of table\n", arg_index);
			lua_error(ls);
		}
		for(int i=0;i<size;i++) {
			lua_pushinteger(ls, i+1);
			lua_gettable(ls, arg_index);
			if( !lua_isnumber(ls, -1) ) {
				VI_log("lua_automateWorker: argument %d table index %d not a number\n", arg_index, i+1);
				break;
			}
			dists[i].squares = lua_tointeger(ls, -1);
			lua_pop(ls, 1); 
		}
	}
	arg_index++;*/

	if( !lua_islightuserdata(ls, arg_index) ) {
		VI_log("lua_automateWorker: argument %d not a pointer\n", arg_index);
		lua_error(ls);
	}
	else {
		void *ptr = lua_touserdata(ls, arg_index);
		dists = static_cast<const Map::Distance *>(ptr);
	}
	arg_index++;

	/*static int total_time = 0;
	int time = clock() - time_s;
	total_time += time;
	VI_log("<<< %d : %d\n", time, total_time);*/

	bool found_target = false;
	int dist = 0;
	vector<int> candidates_x;
	vector<int> candidates_y;
	try {
		vector<Pos2D> candidates;
		found_target = unit->automateBuildRoads(&dist, &candidates, dists);
		for(vector<Pos2D>::const_iterator iter = candidates.begin(); iter != candidates.end(); ++iter) {
			Pos2D pos = *iter;
			candidates_x.push_back(pos.x);
			candidates_y.push_back(pos.y);
		}
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_automateWorker: caught error");
		lua_error(ls);
	}

	//delete [] dists;

	lua_pushboolean(ls, found_target);
	lua_pushinteger(ls, dist);
	luafunc_returnArrayInts(ls, &candidates_x);
	luafunc_returnArrayInts(ls, &candidates_y);

	return 4;
}

/* Logs the text to Conquests's log file.
 * Args: text (string)
 * Returns: nothing
 */
int lua_log(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 1 ) {
		VI_log("lua_log: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_log: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	const char *text = NULL;
	if( !lua_isstring(ls, arg_index) ) {
		VI_log("lua_log: argument %d not a string\n", arg_index);
		lua_error(ls);
	}
	else {
		text = lua_tostring(ls, arg_index);
	}
	arg_index++;

	try {
		VI_log("Lua: ");
		VI_log(text);
		VI_log("\n");
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_log: caught error");
		lua_error(ls);
	}

	return 0;
}

/* 
 * Args: 
 * Returns: nothing
 */
int lua_(lua_State *ls) {
	int n_args = lua_gettop(ls);
	if( n_args != 0 ) {
		VI_log("lua_: incorrect number of arguments\n");
		lua_pushstring(ls, "lua_: incorrect number of arguments");
		lua_error(ls);
	}

	int arg_index = 1;

	try {
	}
	catch(const char *error) {
		VI_log("error: %s\n", error);
		lua_pushstring(ls, "lua_: caught error");
		lua_error(ls);
	}

	return 0;
}

// remember to only return what a civ can see!

void Game::registerScripts() {
	VI_log("register lua functions...\n");
	lua_register(ls, "conquests_addNewTechnology", lua_addNewTechnology);
	lua_register(ls, "conquests_technologySetInfo", lua_technologySetInfo);
	lua_register(ls, "conquests_technologySetAIHint", lua_technologySetAIHint);
	lua_register(ls, "conquests_addNewRace", lua_addNewRace);
	lua_register(ls, "conquests_raceSetStartingTechnology", lua_raceSetStartingTechnology);
	lua_register(ls, "conquests_raceAddCityName", lua_raceAddCityName);
	lua_register(ls, "conquests_findBuildable", lua_findBuildable);
	lua_register(ls, "conquests_buildableGetInfo", lua_buildableGetInfo);
	lua_register(ls, "conquests_improvementGetInfo", lua_improvementGetInfo);
	lua_register(ls, "conquests_unittemplateType", lua_unittemplateType);
	lua_register(ls, "conquests_unittemplateStats", lua_unittemplateStats);
	lua_register(ls, "conquests_unittemplateAirStats", lua_unittemplateAirStats);
	lua_register(ls, "conquests_unittemplateAbility", lua_unittemplateAbility);
	lua_register(ls, "conquests_unitGetTemplate", lua_unitGetTemplate);
	lua_register(ls, "conquests_unitGetCivilization", lua_unitGetCivilization);
	lua_register(ls, "conquests_unitGetPos", lua_unitGetPos);
	lua_register(ls, "conquests_civilizationGetName", lua_civilizationGetName);
	lua_register(ls, "conquests_cityGetCivilization", lua_cityGetCivilization);
	lua_register(ls, "conquests_cityGetPos", lua_cityGetPos);
	lua_register(ls, "conquests_cityGetName", lua_cityGetName);
	lua_register(ls, "conquests_cityGetPublicInfo", lua_cityGetPublicInfo);
	lua_register(ls, "conquests_cityGetInfo", lua_cityGetInfo);
	lua_register(ls, "conquests_cityGetRanges", lua_cityGetRanges);
	lua_register(ls, "conquests_cityCanBuild", lua_cityCanBuild);
	lua_register(ls, "conquests_cityCanLaunch", lua_cityCanLaunch);
	lua_register(ls, "conquests_isExplored", lua_isExplored);
	lua_register(ls, "conquests_isExplorable", lua_isExplorable);
	lua_register(ls, "conquests_getTerrain", lua_getTerrain);
	lua_register(ls, "conquests_getRoad", lua_getRoad);
	lua_register(ls, "conquests_canBuildCity", lua_canBuildCity);
	lua_register(ls, "conquests_canBeImproved", lua_canBeImproved);
	lua_register(ls, "conquests_hasAdjacentUnexplored", lua_hasAdjacentUnexplored);
	lua_register(ls, "conquests_hasEnemies", lua_hasEnemies);
	lua_register(ls, "conquests_getUnitsAt", lua_getUnitsAt);
	lua_register(ls, "conquests_getBestDefenderAt", lua_getBestDefenderAt);
	lua_register(ls, "conquests_getCityAt", lua_getCityAt);
	lua_register(ls, "conquests_getAdjacentSquares", lua_getAdjacentSquares);
	lua_register(ls, "conquests_getCitySquares", lua_getCitySquares);
	lua_register(ls, "conquests_haveAllTechnology", lua_haveAllTechnology);
	lua_register(ls, "conquests_getCivilizations", lua_getCivilizations);
	lua_register(ls, "conquests_getCities", lua_getCities);
	lua_register(ls, "conquests_getMyCities", lua_getMyCities);
	lua_register(ls, "conquests_getUnits", lua_getUnits);
	lua_register(ls, "conquests_getMyUnits", lua_getMyUnits);
	lua_register(ls, "conquests_getRelationshipInfo", lua_getRelationshipInfo);
	lua_register(ls, "conquests_cityGetBuildables", lua_cityGetBuildables);
	lua_register(ls, "conquests_cityGetBuildable", lua_cityGetBuildable);
	lua_register(ls, "conquests_calculateDistanceMap", lua_calculateDistanceMap);
	lua_register(ls, "conquests_getDistanceCost", lua_getDistanceCost);
	lua_register(ls, "conquests_automateWorker", lua_automateWorker);
	lua_register(ls, "conquests_getYear", lua_getYear);
	lua_register(ls, "conquests_canBomb", lua_canBomb);
	lua_register(ls, "conquests_log", lua_log);

}

#if 0
bool Game::runScript(const char *script, const char *function) {
	/*int lua_result = luaL_dofile(ls, script);
	if( lua_result != 0 ) {
		VI_log("Failed to do Lua script %s, error %d\n", script, lua_result);
		VI_log("Lua error: %s\n", lua_tostring(ls, -1));
		return false;
	}*/
	// load the file
	int lua_result = luaL_loadfile(ls, script);
	if( lua_result != 0 ) {
		VI_log("Failed to load Lua script %s, error %d\n", script, lua_result);
		VI_log("Lua error: %s\n", lua_tostring(ls, -1));
		return false;
	}
	// run the main (global)
	lua_result = lua_pcall(ls, 0, LUA_MULTRET, 0);
	if( lua_result != 0 ) {
		VI_log("Failed to run Lua script %s, error %d\n", script, lua_result);
		VI_log("Lua error: %s\n", lua_tostring(ls, -1));
		return false;
	}
	if( function != NULL ) {
		pushScriptFunction(function);
		return runScriptFunction(0);
	}
	return true;
}
#endif

/* Loads the script into memory, and runs the globals.
 */
bool Game::loadScript(const char *script) {
	/*int lua_result = luaL_dofile(ls, script);
	if( lua_result != 0 ) {
		VI_log("Failed to do Lua script %s, error %d\n", script, lua_result);
		VI_log("Lua error: %s\n", lua_tostring(ls, -1));
		return false;
	}*/
	// load the file
	int lua_result = luaL_loadfile(ls, script);
	if( lua_result != 0 ) {
		VI_log("Failed to load Lua script %s, error %d\n", script, lua_result);
		VI_log("Lua error: %s\n", lua_tostring(ls, -1));
		return false;
	}
	// run the main (global)
	lua_result = lua_pcall(ls, 0, LUA_MULTRET, 0);
	if( lua_result != 0 ) {
		VI_log("Failed to run Lua script %s, error %d\n", script, lua_result);
		VI_log("Lua error: %s\n", lua_tostring(ls, -1));
		return false;
	}
	return true;
}
	
void Game::pushScriptFunction(const char *function) {
	lua_getfield(ls, LUA_GLOBALSINDEX, function); // function to be called
}

bool Game::runScriptFunction(int n_args) {
	// run the function previously pushed by pushScriptFunction
	int lua_result = lua_pcall(ls, n_args, LUA_MULTRET, 0);
	if( lua_result != 0 ) {
		VI_log("Failed to run Lua function, error %d\n", lua_result);
		VI_log("Lua error: %s\n", lua_tostring(ls, -1));
		return false;
	}
	return true;
}

/*bool MainGamestate::runScript(const char *script) {
	AIInterface::setMainGamestate(this);
	bool ok = game_g->runScript(script);
	AIInterface::setMainGamestate(NULL);
	return ok;
}*/

// end scripting functions

MapSquare::MapSquare() : mainGamestate(NULL), pos(-1, -1), type(TYPE_GRASSLAND), road(ROAD_NONE), progress_road(0), city(NULL), territory(NULL), target(false) {
}

MapSquare::~MapSquare() {
}

void MapSquare::validateType(Type type) {
	int i_type = (int)type;
	ASSERT( i_type >= 0 && i_type < N_TYPES );
}

int MapSquare::getRoadCost() const {
	switch( type ) {
		case TYPE_OCEAN:
			ASSERT( false );
			break;
		case TYPE_GRASSLAND:
			return 3;
		case TYPE_FOREST:
		case TYPE_DESERT:
		case TYPE_ARTIC:
			return 4;
		case TYPE_HILLS:
			return 5;
		case TYPE_MOUNTAINS:
			return 6;
	}
	ASSERT(false);
	return -1;
}

VI_Texture *MapSquare::getOverlayTexture(int index) const {
	ASSERT( index >= 0 && index < n_map_overlays_c );
	return mapOverlays[type][index];
}

bool MapSquare::drawAboveRoad() const {
	return type == TYPE_FOREST;
}

void MapSquare::setProgressRoad(int progress_road) {
	ASSERT( progress_road < this->getRoadCost() );
	this->progress_road = progress_road;
}

bool MapSquare::buildRoad(const Unit *unit) {
	ASSERT( this->road == ROAD_NONE );
	int required_progress = this->getRoadCost();
	ASSERT( this->progress_road < required_progress );
	int increase = 1;
	if( unit->getCivilization()->hasBonus(Civilization::BONUS_WORKERS) )
		increase = 2;
	this->progress_road += increase;
	if( this->progress_road >= required_progress ) {
		this->road = ROAD_BASIC;
		this->progress_road = 0;
		return true;
	}
	return false;
}

bool MapSquare::buildRailways(const Unit *unit) {
	ASSERT( this->road == ROAD_BASIC );
	int required_progress = this->getRoadCost();
	ASSERT( this->progress_road < required_progress );
	int increase = 1;
	if( unit->getCivilization()->hasBonus(Civilization::BONUS_WORKERS) )
		increase = 2;
	this->progress_road += increase;
	if( this->progress_road >= required_progress ) {
		this->road = ROAD_RAILWAYS;
		this->progress_road = 0;
		return true;
	}
	return false;
}

/*bool MapSquare::hasWorkers() const {
	for(vector<Unit *>::const_iterator iter = units.begin(); iter != units.end(); ++iter) {
		const Unit *unit = *iter;
		if( unit->getStatus() == Unit::STATUS_BUILDING_ROAD ) {
			return true;
		}
	}
	return false;
}*/

bool MapSquare::canBeImproved(const Civilization *civilization) const {
	if( this->isLand() ) {
		if( this->road == ROAD_NONE )
			return true;
		else if( this->road == ROAD_BASIC ) {
			const Technology *tech = game_g->getGameData()->findTechnology("Steam Power");
			if( civilization->hasTechnology(tech) )
				return true;
			return false;
		}
	}
	return false;
}

bool MapSquare::canMoveTo(const UnitTemplate *unit_template) const {
	// for now, unit_template is ignored, but if it's ever implemented, we need to allow it to be NULL, which means to consider a generic land unit
	bool ok = true;
	switch( type ) {
		case TYPE_OCEAN:
		case TYPE_MOUNTAINS:
			ok = false;
			break;
		case TYPE_GRASSLAND:
		case TYPE_DESERT:
		case TYPE_ARTIC:
		case TYPE_FOREST:
		case TYPE_HILLS:
			ok = true;
			break;
		default:
			ASSERT(false);
	}
	return ok;
}

const char *MapSquare::getName() const {
	switch( type ) {
		case TYPE_OCEAN:
			return "Ocean";
		case TYPE_GRASSLAND:
			return "Grassland";
		case TYPE_FOREST:
			return "Forest";
		case TYPE_HILLS:
			return "Hills";
		case TYPE_MOUNTAINS:
			return "Mountains";
		case TYPE_DESERT:
			return "Desert";
		case TYPE_ARTIC:
			return "Artic";
	}
	ASSERT(false);
	return NULL;
}

Rational MapSquare::getMovementCost(const Civilization *civilization) const {
	/*if( this->road == ROAD_BASIC ) {
		return Rational(1, road_move_scale_c);
		//return 0;
	}
	else if( this->road == ROAD_RAILWAYS ) {
		return 0;
	}*/
	if( this->road != ROAD_NONE ) {
		bool ok = false;
		if( this->territory == NULL || civilization == this->territory ) {
			ok = true;
		}
		else {
			const Relationship *relationship = mainGamestate->findRelationship(this->territory, civilization);
			if( relationship->getStatus() == Relationship::STATUS_PEACE ) {
				ok = true;
			}
		}

		if( ok ) {
			if( this->road == ROAD_BASIC ) {
				return Rational(1, road_move_scale_c);
			}
			else if( this->road == ROAD_RAILWAYS ) {
				return 0;
			}
		}
	}
	switch( type ) {
		case TYPE_OCEAN:
			return 1;
		case TYPE_GRASSLAND:
			return 1;
		case TYPE_FOREST:
			return 2;
		case TYPE_HILLS:
			return 2;
		case TYPE_DESERT:
			return 1;
		case TYPE_ARTIC:
			return 2;
		case TYPE_MOUNTAINS:
			return 3;
		default:
			ASSERT(false);
	}
	return 0;
}

int MapSquare::getScaledMovementCost(const Civilization *civilization) const {
	Rational mp_r = this->getMovementCost(civilization);
	mp_r *= road_move_scale_c;
	int mp, mp_num, mp_den;
	mp_r.value(&mp, &mp_num, &mp_den);
	ASSERT( mp_num == 0 );
	return mp;
}

bool MapSquare::hasForeignUnits(const Civilization *civ) const {
	if( units.size() > 0 && units.at(0)->getCivilization() != civ ) {
		return true;
	}
	else if( city != NULL && city->getCivilization() != civ ) {
		return true;
	}
	return false;
}

bool MapSquare::hasEnemies(const Civilization *civ) const {
	if( units.size() > 0 && units.at(0)->getCivilization() != civ ) {
		const Relationship *relationship = mainGamestate->findRelationship(units.at(0)->getCivilization(), civ);
		if( relationship->getStatus() == Relationship::STATUS_WAR ) {
			for(vector<Unit *>::const_iterator iter = units.begin(); iter != units.end(); ++iter) {
				const Unit *unit = *iter;
				if( unit->getTemplate()->getAttack() > 0 ) {
					return true;
				}
			}
		}
	}
	return false;
}

Unit *MapSquare::getBestDefender() const {
	Unit *defender = NULL;
	int best_defence = 0;
	for(vector<Unit *>::const_iterator iter = units.begin(); iter != units.end(); ++iter) {
		Unit *this_unit = *iter;
		int defence = this_unit->getTemplate()->getDefence();
		// find best defender - for equal defences, choose one with lower attack (so only the weaker unit is put at risk of being destroyed)
		if( defender == NULL || defence > best_defence || ( defence == best_defence && this_unit->getTemplate()->getAttack() < defender->getTemplate()->getAttack() ) ) {
			defender = this_unit;
			best_defence = defence;
		}
	}
	return defender;
}

void MapSquare::removeUnit(const Unit *unit) {
	vector<Unit *>::iterator iter = find(units.begin(), units.end(), unit);
	ASSERT( iter != units.end() );
	units.erase(iter);
}

void MapSquare::loadTextures() {
	for(int i=0;i<N_TYPES;i++) {
		mapTextures[i] = NULL;
		for(int j=0;j<n_map_overlays_c;j++)
			mapOverlays[i][j] = NULL;
	}
	/*for(int i=0;i<9;i++) {
		roadTextures[i] = NULL;
	}*/
	for(int i=0;i<256;i++) {
		combinedRoadTextures[i] = NULL;
		combinedRailwayTextures[i] = NULL;
	}

	VI_Image *image = NULL;

	//mapTextures[TYPE_OCEAN] = VI_createTexture("conquests/data/gfx/mapsquare_ocean.jpg");
	//mapTextures[TYPE_OCEAN] = VI_createTexture("conquests/data/gfx/mapsquare_ocean.png");
	unsigned char filter_max_ocean[3] = {180, 215, 240};
	unsigned char filter_min_ocean[3] = {60, 95, 115};
	image = VI_createImageNoise(64, 64, 4.0f, filter_max_ocean, filter_min_ocean, V_NOISEMODE_PERLIN, 4);
	mapTextures[TYPE_OCEAN] = VI_createTexture(image);

	//mapTextures[TYPE_GRASSLAND] = VI_createTexture("conquests/data/gfx/mapsquare_grassland.jpg");
	//mapTextures[TYPE_GRASSLAND] = VI_createTexture("conquests/data/gfx/mapsquare_grassland.png");
	unsigned char filter_max_grassland[3] = {110, 135, 90};
	unsigned char filter_min_grassland[3] = {50, 75, 40};
	image = VI_createImageNoise(64, 64, 8.0f, filter_max_grassland, filter_min_grassland, V_NOISEMODE_PERLIN, 4);
	mapTextures[TYPE_GRASSLAND] = VI_createTexture(image);

	mapTextures[TYPE_FOREST] = mapTextures[TYPE_GRASSLAND]; // uses same base
	mapTextures[TYPE_HILLS] = mapTextures[TYPE_GRASSLAND]; // uses same base
	mapTextures[TYPE_MOUNTAINS] = mapTextures[TYPE_GRASSLAND]; // uses same base

	//mapTextures[TYPE_DESERT] = VI_createTexture("conquests/data/gfx/mapsquare_desert.jpg");
	//mapTextures[TYPE_DESERT] = VI_createTexture("conquests/data/gfx/mapsquare_desert.png");
	unsigned char filter_max_desert[3] = {225, 190, 160};
	unsigned char filter_min_desert[3] = {185, 150, 120};
	image = VI_createImageNoise(64, 64, 8.0f, filter_max_desert, filter_min_desert, V_NOISEMODE_PERLIN, 4);
	mapTextures[TYPE_DESERT] = VI_createTexture(image);

	//mapTextures[TYPE_ARTIC] = VI_createTexture("conquests/data/gfx/mapsquare_artic.jpg");
	//mapTextures[TYPE_ARTIC] = VI_createTexture("conquests/data/gfx/mapsquare_artic.png");
	unsigned char filter_max_artic[3] = {250, 250, 255};
	unsigned char filter_min_artic[3] = {205, 200, 210};
	image = VI_createImageNoise(64, 64, 8.0f, filter_max_artic, filter_min_artic, V_NOISEMODE_PERLIN, 4);
	mapTextures[TYPE_ARTIC] = VI_createTexture(image);

	mapOverlays[TYPE_FOREST][0] = VI_createTexture("conquests/data/gfx/mapoverlay_forest.png");
	mapOverlays[TYPE_HILLS][0] = VI_createTexture("conquests/data/gfx/mapoverlay_hills.png");
	mapOverlays[TYPE_MOUNTAINS][0] = VI_createTexture("conquests/data/gfx/mapoverlay_mountains.png");
	mapOverlays[TYPE_FOREST][1] = VI_createTexture("conquests/data/gfx/mapoverlay_forest1.png");
	mapOverlays[TYPE_HILLS][1] = VI_createTexture("conquests/data/gfx/mapoverlay_hills1.png");
	mapOverlays[TYPE_MOUNTAINS][1] = VI_createTexture("conquests/data/gfx/mapoverlay_mountains1.png");

	/*for(int i=0;i<9;i++) {
		stringstream name;
		name << "conquests/data/gfx/roads_" << i << ".png";
		roadTextures[i] = VI_createTexture(name.str().c_str());
	}*/
	VI_Image **roadImages = new VI_Image *[9];
	for(int i=0;i<9;i++) {
		stringstream name;
		name << "conquests/data/gfx/roads_" << i << ".png";
		roadImages[i] = VI_createImage(name.str().c_str());
	}

	for(int i=0;i<256;i++) {
		VI_Image *image = VI_createImage(64, 64);
		for(int y=0;y<image->getHeight();y++) {
			for(int x=0;x<image->getWidth();x++) {
				image->putRGB(x, y, 255, 0, 255);
			}
		}
		bool any = false;
		for(int j=0;j<8;j++) {
			int mask = 1 << j;
			int bit = i & mask;
			if( bit > 0 ) {
				any = true;
				image->merge(roadImages[j]);
			}
		}
		if( !any ) {
			image->merge(roadImages[8]);
		}
		//combinedRoadTextures[i] = VI_createTexture(image);
		combinedRoadTextures[i] = VI_createTextureWithMask(image, 255, 0, 255);
	}

	VI_Image **railwayImages = new VI_Image *[9];
	for(int i=0;i<9;i++) {
		stringstream name;
		name << "conquests/data/gfx/railways_" << i << ".png";
		railwayImages[i] = VI_createImage(name.str().c_str());
	}

	for(int i=0;i<256;i++) {
		VI_Image *image = VI_createImage(64, 64);
		for(int y=0;y<image->getHeight();y++) {
			for(int x=0;x<image->getWidth();x++) {
				image->putRGB(x, y, 255, 0, 255);
			}
		}
		bool any = false;
		for(int j=0;j<8;j++) {
			int mask = 1 << j;
			int bit = i & mask;
			if( bit > 0 ) {
				any = true;
				image->merge(railwayImages[j]);
			}
		}
		if( !any ) {
			image->merge(railwayImages[8]);
		}
		combinedRailwayTextures[i] = VI_createTextureWithMask(image, 255, 0, 255);
	}

	/*for(int i=0;i<9;i++) {
		roadTextures[i] = VI_createTexture(roadImages[i]);
	}*/
}

Map::Map(MainGamestate *mainGamestate) : mainGamestate(mainGamestate), width(0), height(0), topology(TOPOLOGY_UNDEFINED), squares(NULL), splat_alpha(NULL) {
}

Map::Map(MainGamestate *mainGamestate, int width, int height) : mainGamestate(mainGamestate), width(width), height(height), topology(TOPOLOGY_FLAT), splat_alpha(NULL) {
	init();
}

Map::~Map() {
	free();
}

void Map::init() {
	this->squares = new MapSquare[width*height];
	for(int y=0;y<height;y++) {
		for(int x=0;x<width;x++) {
			this->getSquare(x, y)->setMaingameState(mainGamestate);
			this->getSquare(x, y)->setPos(Pos2D(x, y));
		}
	}
}

void Map::free() {
	if( this->squares != NULL ) {
		delete [] this->squares;
		this->squares = NULL;
	}
	this->width = 0;
	this->height = 0;
	if( splat_alpha ) {
		delete splat_alpha;
		splat_alpha = NULL;
	}
}

void Map::createAlphamap() {
	//ASSERT( game->splat_pixel_shader_ambient != NULL );
	ASSERT( this->splat_alpha == NULL );
	VI_Image *alphamap = VI_createImage(width, height);
	for(int y=0;y<height;y++) {
		for(int x=0;x<width;x++) {
			const MapSquare *square = this->getSquare(x, y);
			if( square->isLand() ) {
				if( square->getType() == MapSquare::TYPE_DESERT )
					alphamap->putRGB(x, y, 0, 255, 0);
				else if( square->getType() == MapSquare::TYPE_ARTIC )
					alphamap->putRGB(x, y, 0, 0, 255);
				else
					//alphamap->putRGB(x, y, 255, 0, 0);
					alphamap->putRGB(x, y, 0, 0, 0);
			}
			else {
				//alphamap->putRGB(x, y, 0, 0, 0);
				alphamap->putRGB(x, y, 255, 0, 0);
			}
		}
	}
	alphamap->makePowerOf2();
	Renderer *renderer = game_g->getGraphicsEnvironment()->getRenderer();
	// so that the splat_alpha doesn't get flushed, as it's created before we close the OptionsGamestate, and we delete in manually ourselves anyway (TODO: can this be improved?)
	VI_set_default_persistence(-1);
	this->splat_alpha = renderer->createTexture(alphamap);
	VI_set_default_persistence(0);
}

bool Map::createRandom() {
	VI_log("create random map: %d x %d\n", width, height);
	ASSERT( width > 0 && height > 0 );
	unsigned char filter_max[] = {255, 0, 0};
	unsigned char filter_min[] = {0, 0, 0};
	VI_initPerlin(); // reset for random seed
	VI_Image *image = VI_createImageNoise(width, height, 5.0f, filter_max, filter_min, V_NOISEMODE_PERLIN, 10);
	//VI_Image *image = VI_createImageNoise(width, height, 1.0f, filter_max, filter_min, V_NOISEMODE_PERLIN, 10);

	// scale
	unsigned char min_r = 255, max_r = 0;
	for(int y=0;y<height;y++) {
		for(int x=0;x<width;x++) {
			unsigned char r = 0, g = 0, b = 0;
			image->getRGB(&r, &g, &b, x, y);
			if( r < min_r )
				min_r = r;
			if( r > max_r )
				max_r = r;
		}
	}
	float scale_r = 255.0f / ( ((float)max_r) - ((float)min_r) );
	for(int y=0;y<height;y++) {
		for(int x=0;x<width;x++) {
			unsigned char r = 0, g = 0, b = 0;
			image->getRGB(&r, &g, &b, x, y);
			float fr = r;
			fr -= (float)min_r;
			fr *= scale_r;
			if( fr < 0.0f )
				fr = 0.0f;
			else if( fr > 255.0f )
				fr = 255.0f;
			r = (unsigned char)fr;
			image->putRGB(x, y, r, g, b);
		}
	}

	const int forest_coverage = 10;
	const int land_coverage = 48;
	const int artic_extent = 0.10 * height;
	const int desert_extent = 0.05 * height;
	this->setAllTo(MapSquare::TYPE_OCEAN);
	int sx = 0, ex = 1;
	if( topology == TOPOLOGY_FLAT ) {
		// avoid land on boundaries
		sx = 1;
		ex = width-1;
	}
	else if( topology == TOPOLOGY_CYLINDER ) {
		sx  = 0;
		ex = width;
	}
	else {
		ASSERT( false );
	}
	for(int y=1;y<height-1;y++) {
		//VI_log("#");
		for(int x=sx;x<ex;x++) {
			unsigned char r = 0, g = 0, b = 0;
			image->getRGB(&r, &g, &b, x, y);
			//VI_log("%d,", r);
			ASSERT( g == 0 && b == 0 );
			MapSquare::Type type = MapSquare::TYPE_OCEAN;
			float fr = ((float)r) / 255.0f;
			/*if( r < 127 ) {
				type = MapSquare::TYPE_OCEAN;
			}
			else if( r < 175 ) {
				type = MapSquare::TYPE_GRASSLAND;
			}
			else if( r < 185 ) {
				type = MapSquare::TYPE_HILLS;
			}
			else {
				type = MapSquare::TYPE_MOUNTAINS;
			}*/
			if( 100.0f * fr <= 100.0f - land_coverage ) {
				type = MapSquare::TYPE_OCEAN;
			}
			else {
				float scaled_fr = ( 100.0f * fr - 100.0f + land_coverage ) / land_coverage;
				/*if( scaled_fr < 0.375f )
					type = MapSquare::TYPE_GRASSLAND;
				else if( scaled_fr < 0.45f )
					type = MapSquare::TYPE_HILLS;
				else
					type = MapSquare::TYPE_MOUNTAINS;*/
				if( scaled_fr < 0.55f )
					type = MapSquare::TYPE_GRASSLAND;
				else if( scaled_fr < 0.8f )
					type = MapSquare::TYPE_HILLS;
				else
					type = MapSquare::TYPE_MOUNTAINS;
				int flip_chance = 20;
				if( rand() % flip_chance == 0 ) {
					if( type == MapSquare::TYPE_GRASSLAND )
						type = rand() % 3 <= 1 ? MapSquare::TYPE_HILLS : MapSquare::TYPE_MOUNTAINS;
					else
						type = MapSquare::TYPE_GRASSLAND;
				}
			}

			if( type == MapSquare::TYPE_GRASSLAND ) {
				int forest = rand() % 100;
				if( forest < forest_coverage )
					type = MapSquare::TYPE_FOREST;
				else {
					const int var = 3;
					int ty = y + ( rand() % (2*var+1) ) - var;
					if( ty < artic_extent || height-1-ty < artic_extent ) {
						type = MapSquare::TYPE_ARTIC;
					}
					else if( ty >= height/2 - desert_extent && ty < height/2 + desert_extent ) {
						type = MapSquare::TYPE_DESERT;
					}
				}
			}
			{
				// test
				//type = ( x % 2 == y % 2 ) ? MapSquare::TYPE_GRASSLAND : MapSquare::TYPE_OCEAN;
			}
			this->getSquare(x, y)->setType(type);
		}
		//VI_log("\n");
	}

	delete image;
	this->createAlphamap();
	return true;
}

bool Map::readFile(const char *filename) {
	free();

	FILE *file = fopen(filename, "r");
	if( file == NULL )
		return false;

	const int BUFSIZE = 1024;
	char buffer[BUFSIZE+1];
	if( fgets(buffer, BUFSIZE, file) == NULL )
		return false;
	this->width = atoi(buffer);
	if( fgets(buffer, BUFSIZE, file) == NULL )
		return false;
	this->height = atoi(buffer);
	if( this->width <= 0 || this->height <= 0 )
		return false;

	if( fgets(buffer, BUFSIZE, file) == NULL )
		return false;
	int i_topology = atoi(buffer);
	if( i_topology != (int)TOPOLOGY_FLAT && i_topology != (int)TOPOLOGY_CYLINDER ) {
		VI_log("unknown topology type\n");
		return false;
	}
	this->topology = (Map::Topology)i_topology;

	init();
	for(int i=0;i<width*height;i++) {
		this->squares[i].setType(MapSquare::TYPE_OCEAN);
	}

	char *line = new char[width+2];
	for(int y=0;y<height;y++) {
		if( fgets(line, width+2, file) == NULL ) {
			delete [] line;
			delete [] this->squares;
			this->squares = NULL;
			this->width = 0;
			this->height = 0;
			return false;
		}
		for(int x=0;x<width;x++) {
			MapSquare::Type type = static_cast<MapSquare::Type>(line[x] - '0');
			/*if( type != MapSquare::TYPE_OCEAN && type != MapSquare::TYPE_HILLS )
			type = MapSquare::TYPE_GRASSLAND;*/
			this->getSquare(x, y)->setType(type);
			/*if( type == MapSquare::TYPE_GRASSLAND ) {
			this->getSquare(x, y)->setRoad(MapSquare::ROAD_BASIC);
			}*/
		}
	}

	this->createAlphamap();
	fclose(file);
	return true;
}

void Map::reduceScreenPosToBase(int *x, int *y) const {
	if( topology == TOPOLOGY_CYLINDER ) {
		int screen_w = game_g->getGraphicsEnvironment()->getWidth();
		int whole_w = this->width * mainGamestate->getSqWidth();
		while( *x < 0 )
			*x += whole_w;
		while( *x >= screen_w )
			*x -= whole_w;
	}
}

City *Map::findCity(int x, int y) const {
	/*for(int i=0;i<game->getNCivilizations();i++) {
	const Civilization *civilization = game->getCivilization(i);
	for(int j=0;j<civilization->getNCities();j++) {
	City *city = civilization->getCity(j);
	if( city->getX() == x && city->getY() == y )
	return city;
	}
	}
	return NULL;*/
	City *city = this->getSquare(x, y)->getCity();
	if( city != NULL ) {
		this->reduceToBase(&x, &y);
		ASSERT( city->getX() == x && city->getY() == y );
	}
	return city;
}

const vector <Unit *> *Map::findUnitsAt(int x, int y) const {
	//VI_log("Map::findUnitsAt(%d, %d)\n", x, y);
	/*vector<Unit *> *units = new vector<Unit *>();
	for(int i=0;i<game->getNCivilizations();i++) {
	const Civilization *civilization = game->getCivilization(i);
	for(int j=0;j<civilization->getNUnits();j++) {
	Unit *unit = civilization->getUnit(j);
	if( unit->getX() == x && unit->getY() == y )
	units->push_back(unit);
	}
	}
	return units;*/
	//const vector<Unit *> *units = this->getSquare(x, y)->getUnits();
	const vector<Unit *> *units = this->getSquare(x, y)->getUnits();
	if( units->size() > 0 ) {
		const Unit *unit = units->at(0);
		this->reduceToBase(&x, &y);
		ASSERT( unit->getX() == x && unit->getY() == y );
	}
	return units;
}

Civilization *Map::findCivilizationAt(Pos2D pos) const {
	// civilization isn't owned by this class, so okay to return non-const in const member function
	const MapSquare *square = this->getSquare(pos);
	const City *city = square->getCity();
	if( city != NULL )
		return city->getCivilization();
	const vector<Unit *> *units = square->getUnits();
	if( units->size() > 0 )
		return units->at(0)->getCivilization();
	return NULL;
}

bool MapSquare::canBuildCity() const {
	//if( this->isLand() )
	//	return false;
	if( type != TYPE_GRASSLAND && type != TYPE_FOREST && type != TYPE_HILLS &&
		type != TYPE_DESERT && type != TYPE_ARTIC ) {
			return false;
	}
	for(int i=0;i<mainGamestate->getNCivilizations();i++) {
		const Civilization *civilization = mainGamestate->getCivilization(i);
		for(int j=0;j<civilization->getNCities();j++) {
			const City *city = civilization->getCity(j);
			int dist_x = (city->getX() - pos.x);
			int dist_y = (city->getY() - pos.y);
			this->mainGamestate->getMap()->reduceDiff(&dist_x, &dist_y);
			dist_x = abs(dist_x);
			dist_y = abs(dist_y);
			const int min_gap_c = 4;
			//if( dist_x <= min_gap_c && dist_y <= min_gap_c ) {
			//	return false;
			//}
			if( dist_x > min_gap_c || dist_y > min_gap_c ) {
				// okay
			}
			else if( ( dist_x >= 3 && dist_y >= 4 ) || ( dist_x >= 4 && dist_y >= 3 ) ) {
				// still okay
			}
			else {
				return false;
			}
		}
	}
	return true;
}

/*bool Map::canBuildCity(int x, int y) const {
const MapSquare *square = getSquare(x, y);
//if( square->isLand() )
//	return false;
MapSquare::Type squareType = square->getType();
if( squareType != MapSquare::TYPE_GRASSLAND && squareType != MapSquare::TYPE_FOREST && squareType != MapSquare::TYPE_HILLS &&
squareType != MapSquare::TYPE_DESERT && squareType != MapSquare::TYPE_ARTIC ) {
return false;
}
for(int i=0;i<game->getNCivilizations();i++) {
const Civilization *civilization = game->getCivilization(i);
for(int j=0;j<civilization->getNCities();j++) {
const City *city = civilization->getCity(j);
int dist_x = abs(city->getX() - x);
int dist_y = abs(city->getY() - y);
const int min_gap_c = 4;
//if( dist_x <= min_gap_c && dist_y <= min_gap_c ) {
//	return false;
//}
if( dist_x > min_gap_c || dist_y > min_gap_c ) {
// okay
}
else if( ( dist_x >= 3 && dist_y >= 4 ) || ( dist_x >= 4 && dist_y >= 3 ) ) {
// still okay
}
else {
return false;
}
}
}
return true;
}*/

Pos2D Map::findRandomFreeSquare() const {
	const bool mindist = 2;
	vector<Pos2D> freesquares;
	for(int y=0;y<height;y++) {
		for(int x=0;x<width;x++) {
			const MapSquare *square = this->getSquare(x, y);
			if( square->getType() == MapSquare::TYPE_GRASSLAND ) {
				bool ok = true;
				for(int cy=y-mindist;cy<=y+mindist && ok;cy++) {
					for(int cx=x-mindist;cx<=x+mindist && ok;cx++) {
						if( this->isValid(cx, cy) ) {
							const MapSquare *this_square = this->getSquare(cx, cy);
							if( this_square->getCity() != NULL || this_square->getUnits()->size() > 0 ) {
								ok = false;
							}
						}
					}
				}
				if( ok ) {
					Pos2D pos(x, y);
					freesquares.push_back(pos);
				}
			}
		}
	}
	ASSERT( freesquares.size() > 0 );
	int index = rand() % freesquares.size();
	Pos2D pos = freesquares.at(index);
	return pos;
}

bool Map::isOnOrAdjacent(Pos2D pos, MapSquare::Type type) const {
	for(int y=pos.y-1;y<=pos.y+1;y++) {
		for(int x=pos.x-1;x<=pos.x+1;x++) {
			if( this->isValid(x, y) ) {
				const MapSquare *square = this->getSquare(x, y);
				if( square->getType() == type ) {
					return true;
				}
			}
		}
	}
	return false;
}

bool Map::isWithinCityRadius(Pos2D pos, MapSquare::Type type, bool include_centre) const {
	for(int i=0;i<City::getNCitySquares();i++) {
		if( !include_centre && i == City::getCentreCitySquareIndex() ) {
			continue;
		}
		Pos2D this_pos = City::getCitySquare(this, i, pos);
		if( this->isValid(this_pos.x, this_pos.y) ) {
			const MapSquare *square = this->getSquare(this_pos);
			if( square->getType() == type ) {
				return true;
			}
		}
	}
	return false;
}

void Map::ensureNearbyTerrain(Pos2D pos, MapSquare::Type type) {
	ASSERT( type != MapSquare::TYPE_OCEAN );
	for(int y=pos.y-1;y<=pos.y+1;y++) {
		for(int x=pos.x-1;x<=pos.x+1;x++) {
			if( this->isValid(x, y) ) {
				MapSquare *square = this->getSquare(x, y);
				if( square->getType() == type ) {
					// already present
					return;
				}
			}
		}
	}
	for(int y=pos.y-1;y<=pos.y+1;y++) {
		for(int x=pos.x-1;x<=pos.x+1;x++) {
			if( this->isValid(x, y) ) {
				MapSquare *square = this->getSquare(x, y);
				if( square->isLand() && square->getType() != type ) {
					square->setType(type);
					return;
				}
			}
		}
	}
}

/*void Map::calculateDistanceMap_processQueueItem(int indx, const Distance *dists, Distance *temp, vector<int> *queue, bool reached_target, const Civilization *civilization, const Unit *unit, bool forwards, int ex, int ey, bool by_sea) const {
}*/

#if 0
Map::Distance *Map::calculateDistanceMap(const Civilization *civilization, const Unit *unit, bool include_travel, bool forwards, int sx, int sy, int ex, int ey, bool by_sea) {
	// if unit is NULL, assume general land unit, unless by_sea is true
	ASSERT( civilization != NULL );
	//VI_log("Map::calculateDistanceMap from %d, %d\n", sx, sy);
	// N.B. - the AI cheats here, in that we also know the distance map for unexplored area.
	// To be fairer, we should estimate the movement cost always as (1*3) for unexplored squares?
	int time_s = clock();
	//const int MAX_VALUE = 32767;
	Distance *dists = new Distance[width*height];
	Distance *temp = new Distance[width*height];
	/*for(int i=0;i<width*height;i++)
		dists[i] = -1;*/
	//int *queue = new int[width*height];
	vector<int> queue;
	queue.push_back(sy * width + sx);
	int c_queue = 0;
	//int n_queue = 1;

	dists[sy * width + sx].cost = 0;
	dists[sy * width + sx].squares = 0;

	bool reached_target = false;

	if( include_travel ) {
		const City *city = this->findCity(sx, sy);
		if( city != NULL ) {
			// TODO: cache target squares from City
			city->initTravel(false, unit == NULL ? NULL : unit->getTemplate());
			for(int i=0;i<width*height && !reached_target;i++) {
				if( squares[i].isTarget() ) {
					dists[i].cost = 1;
					dists[i].squares = 1;
					//queue[n_queue++] = i;
					queue.push_back(i);
					if( i == ey * width + ex ) {
						reached_target = true;
					}
				}
			}
			this->resetTargets();
		}
	}

	while( !reached_target ) {
		for(int i=0;i<width*height;i++)
			temp[i] = dists[i];

		// process items in queue
		///int old_n_queue = n_queue;
		int old_n_queue = queue.size();
		for(int i=c_queue;i<old_n_queue && !reached_target;i++) {
			int indx = queue[i];
			//this->calculateDistanceMap_processQueueItem(indx, dists, temp, &queue, reached_target, civilization, unit, forwards, ex, ey, by_sea);

			int cost = dists[indx].cost;
			int squares = dists[indx].squares;
			ASSERT( cost != -1 );
			ASSERT( squares != -1 );
			int x = indx % width;
			int y = indx / width;
			//LOG("process queue: %d : %d ( %d, %d) = %d\n", i, indx, x, y, cost);
			const MapSquare *from_square = this->getSquare(x, y);

			for(int cy=y-1;cy<=y+1 && !(reached_target);cy++) {
				for(int cx=x-1;cx<=x+1 && !(reached_target);cx++) {

					if( cx == x && cy == y )
						continue;

					if( !isValid(cx,cy) )
						continue;

					//if( dists[cy*width + cx] == -1 )
					{
						const MapSquare *square = this->getSquare(cx, cy);
						bool valid_move = false;
						if( by_sea )
							valid_move = !square->isLand();
						else if( unit == NULL ) {
							// general land unit
							valid_move = square->canMoveTo(NULL);
						}
						else {
							// also take into account other units there, and if we can attack
							valid_move = unit->canMoveTo( Pos2D(cx, cy) );
						}
						/*if( ( unit == NULL && ( square->isLand() == !by_sea ) ) ||
						( unit != NULL && unit->canMoveTo( Pos2D(cx, cy) ) ) ) {*/
						if( valid_move ) {
							int mp = forwards ? square->getScaledMovementCost(civilization) : from_square->getScaledMovementCost(civilization);
							if( unit == NULL )
								mp = min(mp,road_move_scale_c);
							else
								mp = min(mp,road_move_scale_c*unit->getTemplate()->getMoves());
							int val = cost + mp;
							//ASSERT( val <= MAX_VALUE );

							// add to queue if not already present

							bool cheaper = false;
							int c_val = temp[cy * width + cx].cost;
							if( c_val == -1 || val <= c_val ) {
								if( c_val == -1 || val < c_val ) {
									temp[cy * width + cx].cost = val;
									cheaper = true;
								}
								int c_squares = temp[cy * width + cx].squares;
								if( c_squares == -1 || val < c_val || squares+1 < c_squares ) { // if val < c_val (i.e., we've found a route with lower cost), we should always reset the number of squares
									temp[cy * width + cx].squares = squares+1;
									cheaper = true;
								}
							}
							if( cheaper ) {
								queue.push_back(cy*width + cx);
							}
							// TODO:
							/*if( cx == ex && cy == ey ) {
							*reached_target = true;
							}*/
							/*if( stop_x != -1 && stop_y != -1 ) {
							int diffx = abs( cx - stop_x );
							int diffy = abs( cy - stop_y );
							if( diffx <= 1 && diffy <= 1 ) {
							check_stop = true;
							}
							}*/
						}
					}

				}
			}
		}

		c_queue = old_n_queue;
		if( c_queue == queue.size() ) {
			break;
		}

		swap(temp, dists);
	}

	//VI_log("### %d\n", queue.size());
	delete [] temp;
	//delete [] queue;
	//VI_log("### Map::calculateDistanceMap %d : %d\n", unit, clock() - time_s);
	return dists;
}
#endif

// new, Dijkstra's
//#if 0

class DistanceComparison {
public:
	bool operator() (const Map::Distance &lhs, const Map::Distance &rhs) const {
		// return iff rhs < lhs
		/*if( lhs.cost == rhs.cost ) {
			if( lhs.squares == rhs.squares )
				return false; // equal (also includes cost==-1 case)
			return rhs.squares < lhs.squares;
		}
		else if( lhs.cost == -1 )
			return true;
		else if( rhs.cost == -1 )
			return false;
		return rhs.cost < lhs.cost;*/
		// don't have to worry about cost == -1 cases
		if( lhs.cost == rhs.cost ) {
			return rhs.squares < lhs.squares;
		}
		return rhs.cost < lhs.cost;
	}
};

Map::Distance *Map::calculateDistanceMap(const Civilization *civilization, const Unit *unit, bool include_travel, bool forwards, int sx, int sy, int ex, int ey, bool by_sea, bool by_travel) {
	// if unit is NULL, assume general land unit (that's capable of attack), unless by_sea is true
	// if by_travel is true, then we ignore whether the squares are explored or not (as we'll be revealing squares within the travel distance)
	ASSERT( civilization != NULL );
	//VI_log("Map::calculateDistanceMap from %d, %d\n", sx, sy);
	int time_s = clock();
	//VI_log("Map::calculateDistanceMap() enter\n");
	int wh = width*height;
	Distance *dists = new Distance[wh];
	bool *done_squares = new bool[wh];
	for(int i=0;i<wh;i++) {
		dists[i].indx = i;
		done_squares[i] = false;
	}

	dists[sy * width + sx].cost = 0;
	dists[sy * width + sx].squares = 0;

	priority_queue<Distance, vector<Distance>, DistanceComparison> queue;
	queue.push(dists[sy * width + sx]);

	bool reached_target = false;

	if( include_travel ) {
		ASSERT( !by_sea ); // when counting by sea, shouldn't be including travelling
		ASSERT( forwards ); // can only cope with forwards calculations
		const City *city = this->findCity(sx, sy);
		if( city != NULL ) {
			// TODO: cache target squares from City
			city->initTravel(false, unit == NULL ? NULL : unit->getTemplate());
			for(int i=0;i<width*height && !reached_target;i++) {
				if( squares[i].isTarget() ) {
					ASSERT( dists[i].cost == -1 ); // shouldn't already be in the queue!
					//dists[i].cost = 1;
					dists[i].cost = unit == NULL ? road_move_scale_c : road_move_scale_c*unit->getTemplate()->getMoves();
					dists[i].squares = 1;
					queue.push(dists[i]);
					if( i == ey * width + ex ) {
						reached_target = true;
					}
				}
			}
			this->resetTargets();
		}
	}

	//while( !reached_target && !queue.empty() ) {
	while( !queue.empty() ) {
		/*{
			priority_queue<Distance, vector<Distance>, DistanceComparison> queue_copy(queue);
			int count = 0;
			while( !queue_copy.empty() ) {
				const Distance dist = queue_copy.top();
				VI_log("### %d : %d : %d, %d\n", count, dist.indx, dist.cost, dist.squares);
				queue_copy.pop();
				count++;
			}
		}*/
		// find current cheapest
		const Distance cheapest_dist = queue.top();
		queue.pop();
		int cheapest = cheapest_dist.indx;
		//VI_log(">>> queue size %d\n", queue.size());

		int x = cheapest % width;
		int y = cheapest / width;
		const MapSquare *from_square = this->getSquareBase(x, y);
		// if there's an enemy unit here, we can't pass through it this turn, so it should be considered an "end"
		const vector<Unit *> *units_at_sq = from_square->getUnits();
		if( forwards && units_at_sq->size() > 0 && units_at_sq->at(0)->getCivilization() != civilization ) {
			// skip
		}
		// check that we've got the "cheapest" version
		else if( !done_squares[cheapest] ) {
			//VI_log("### %d : (%d,%d): %d, %d : %d\n", cheapest, cheapest % width, cheapest / width, cheapest_dist.cost, cheapest_dist.squares, done_squares[cheapest]);
			done_squares[cheapest] = true;

			// assert in sync with dists array:
			ASSERT( cheapest_dist.cost == dists[cheapest].cost );
			ASSERT( cheapest_dist.squares == dists[cheapest].squares );

			// process this square
			int cost = dists[cheapest].cost;
			/*if( cost == -1 ) {
				// done all squares that can be visited
				break;
			}*/
			//ASSERT( cost != -1 );
			int dists_squares = dists[cheapest].squares;
			//ASSERT( dists_squaressquares != -1 );

			for(int cy=y-1;cy<=y+1;cy++) {
				for(int cx=x-1;cx<=x+1;cx++) {
					// n.b., this code is performance critical!
					// see tests TEST_DISTANCES_0, TEST_DISTANCES_1

					if( cx == x && cy == y )
						continue;

					if( !isValid(cx,cy) )
						continue;

					/*if( !by_travel && !civilization->isExplored(cx, cy) ) {
						//VI_log("ping\n");
						continue;
					}*/

					int base_cx = cx, base_cy = cy;
					reduceToBase(&base_cx, &base_cy); // need to call this, so that we access the correct entry in dists array

					// now use *Base functions for performance

					//int indx = base_cy*width+base_cx;
					if( !by_travel && !civilization->isExploredBase(base_cx, base_cy) ) {
					//if( !by_travel && !civilization->isExploredBase(indx) ) {
						continue;
					}

					//const MapSquare *square = this->getSquareBase(base_cx, base_cy);
					//const MapSquare *square = &squares[indx]; // for performance
					const MapSquare *square = &squares[base_cy*width+base_cx]; // for performance
					bool valid_move = false;
					if( by_sea )
						valid_move = !square->isLand();
					else if( unit == NULL ) {
						// general land unit
						valid_move = square->canMoveTo(NULL);
					}
					else {
						// also take into account other units there, and if we can attack
						valid_move = unit->canMoveTo( Pos2D(base_cx, base_cy) );
					}
					if( !by_sea && !forwards ) {
						// if searching backwards, we shouldn't ever move to a square with another civ's unit, as we won't ever be able to move from that square in this go
						const vector<Unit *> *units_at_nsq = square->getUnits();
						if( units_at_nsq->size() > 0 && units_at_nsq->at(0)->getCivilization() != civilization ) {
							valid_move = false;
						}
					}
					if( valid_move ) {
						int mp = forwards ? square->getScaledMovementCost(civilization) : from_square->getScaledMovementCost(civilization);
						if( unit == NULL )
							mp = min(mp,road_move_scale_c);
						else
							mp = min(mp,road_move_scale_c*unit->getTemplate()->getMoves());
						int val = cost + mp;

						int c_val = dists[base_cy * width + base_cx].cost;
						bool changed = false;
						if( c_val == -1 || val <= c_val ) {
							if( c_val == -1 || val < c_val ) {
								dists[base_cy * width + base_cx].cost = val;
								changed = true;
							}
							int c_squares = dists[base_cy * width + base_cx].squares;
							if( c_squares == -1 || val < c_val || dists_squares+1 < c_squares ) { // if val < c_val (i.e., we've found a route with lower cost), we should always reset the number of squares
								dists[base_cy * width + base_cx].squares = dists_squares+1;
								changed = true;
							}
						}
						if( changed ) {
							// update with cheaper version
							queue.push(dists[base_cy * width + base_cx]);
						}
					}
				}
			}
		}
	}

	delete [] done_squares;
	
	/*static int total_time = 0;
	int time = clock() - time_s;
	total_time += time;
	static int count = 0;
	count++;
	//VI_log("Map::calculateDistanceMap() exit %d : %d\n", unit, clock() - time_s);
	VI_log("### %d Map::calculateDistanceMap %d : %d, %d\n", count, unit, time, total_time);
	//VI_log(">>> ### dist: %d\n", dists[8 * this->getWidth() + 4].cost);
	*/

	return dists;
}
//#endif

#if 0
// OLD, buggy
Map::Distance *Map::calculateDistanceMap(const Unit *unit, bool include_travel, bool forwards, int sx, int sy, int ex, int ey, bool by_sea) {
	//VI_log("Map::calculateDistanceMap from %d, %d\n", sx, sy);
	// N.B. - the AI cheats here, in that we also know the distance map for unexplored area.
	// To be fairer, we should estimate the movement cost always as (1*3) for unexplored squares?
	int time_s = clock();
	// if unit is NULL, assume general land unit, unless by_sea is true
	const int MAX_VALUE = 32767;
	Distance *dists = new Distance[width*height];
	Distance *temp = new Distance[width*height];
	/*for(int i=0;i<width*height;i++)
		dists[i] = -1;*/
	int *queue = new int[width*height];
	queue[0] = sy * width + sx;
	int c_queue = 0;
	int n_queue = 1;

	dists[sy * width + sx].cost = 0;
	dists[sy * width + sx].squares = 0;

	bool reached_target = false;

	if( include_travel ) {
		const City *city = this->findCity(sx, sy);
		if( city != NULL ) {
			// TODO: cache target squares from City
			city->initTravel(false, unit == NULL ? NULL : unit->getTemplate());
			for(int i=0;i<width*height && !reached_target;i++) {
				if( squares[i].isTarget() ) {
					dists[i].cost = 1;
					dists[i].squares = 1;
					queue[n_queue++] = i;
					if( i == ey * width + ex ) {
						reached_target = true;
					}
				}
			}
			this->resetTargets();
		}
	}

	while( !reached_target ) {
		for(int i=0;i<width*height;i++)
			temp[i] = dists[i];

		// process items in queue
		int old_n_queue = n_queue;
		for(int i=c_queue;i<old_n_queue && !reached_target;i++) {
			int indx = queue[i];
			int cost = dists[indx].cost;
			int squares = dists[indx].squares;
			ASSERT( cost != -1 );
			ASSERT( squares != -1 );
			int x = indx % width;
			int y = indx / width;
			//LOG("process queue: %d : %d ( %d, %d) = %d\n", i, indx, x, y, cost);
			const MapSquare *from_square = this->getSquare(x, y);

			for(int cy=y-1;cy<=y+1 && !reached_target;cy++) {
				for(int cx=x-1;cx<=x+1 && !reached_target;cx++) {

					if( cx == x && cy == y )
						continue;

					if( !isValid(cx,cy) )
						continue;

					//if( dists[cy*width + cx] == -1 )
					{
						const MapSquare *square = this->getSquare(cx, cy);
						bool valid_move = false;
						if( by_sea )
							valid_move = !square->isLand();
						else if( unit == NULL ) {
							// general land unit
							valid_move = square->canMoveTo(NULL);
						}
						else {
							// also take into account other units there, and if we can attack
							valid_move = unit->canMoveTo( Pos2D(cx, cy) );
						}
						/*if( ( unit == NULL && ( square->isLand() == !by_sea ) ) ||
						( unit != NULL && unit->canMoveTo( Pos2D(cx, cy) ) ) ) {*/
						if( valid_move ) {
							int mp = forwards ? square->getScaledMovementCost() : from_square->getScaledMovementCost();
							if( unit == NULL )
								mp = min(mp,road_move_scale_c);
							else
								mp = min(mp,road_move_scale_c*unit->getTemplate()->getMoves());
							int val = cost + mp;
							ASSERT( val <= MAX_VALUE );

							// add to queue if not already present
							if( temp[cy * width + cx].cost == -1 ) {
								//LOG("queue[%d] = %d : ( %d, %d ) = %d\n", n_queue, cy*width + cx, cx, cy, val);
								ASSERT( n_queue < width*height );
								/*for(int j=0;j<n_queue;j++) {
								if( queue[j] == cy*width + cx ) {
								LOG("### repeat: %d\n", j);
								ASSERT(false);
								}
								}*/
								queue[n_queue++] = cy*width + cx;
							}

							int c_val = temp[cy * width + cx].cost;
							if( c_val == -1 || val <= c_val ) {
								/*if( c_val != -1 && val < c_val ) {
									VI_log("    overwrote (%d,%d) : old cost %d, new cost %d\n", cx, cy, c_val, val);
								}*/
								temp[cy * width + cx].cost = val;
								int c_squares = temp[cy * width + cx].squares;
								if( c_squares == -1 || val < c_val || squares+1 <= c_squares ) // if val < c_val (i.e., we've found a route with lower cost), we should always reset the number of squares
									temp[cy * width + cx].squares = squares+1;
							}
							if( cx == ex && cy == ey ) {
								reached_target = true;
							}
							/*if( stop_x != -1 && stop_y != -1 ) {
							int diffx = abs( cx - stop_x );
							int diffy = abs( cy - stop_y );
							if( diffx <= 1 && diffy <= 1 ) {
							check_stop = true;
							}
							}*/
						}
					}

				}
			}
		}
		// print out new info:
		//if( sx == 53 && sy == 10 ) {
		/*if( true ) {
			VI_log("queue expanded from %d to %d\n", old_n_queue, n_queue);
			for(int i=old_n_queue;i<n_queue;i++) {
				int indx = queue[i];
				int x = indx % width;
				int y = indx / width;
				VI_log("%d: (%d,%d) : cost %d, squares %d\n", i, x, y, temp[indx].cost, temp[indx].squares);
			}
		}*/
		//LOG("===\n");

		c_queue = old_n_queue;
		if( c_queue == n_queue ) {
			break;
		}

		swap(temp, dists);
		/*if( check_stop && !done && dists[stop_y * width + stop_x] != -1 ) {
		done = true;
		for(int dir=0;dir<8 && done;dir++) {
		int cx = 0, cy = 0;
		if( !this->translateDirection(&cx, &cy, stop_x, stop_y, dir) ) {
		if( dists[cy * width + cx] == -1 && unit->validSquare(cx,cy) ) {
		// valid adjacent square that hasn't been checked
		done = false;
		}
		}
		}
		}*/
	}

	delete [] temp;
	delete [] queue;
	//VI_log("### Map::calculateDistanceMap %d : %d\n", unit, clock() - time_s);
	return dists;
}
#endif

void Map::resetTargets() {
	for(int i=0;i<width*height;i++) {
		this->squares[i].setTarget(false);
	}
}

#if 0
short *Map::calculateDistanceMap(const Unit *unit, int sx, int sy, int ex, int ey) const {
	// N.B. - the AI cheats here, in that we also know the distance map for unexplored area.
	// To be fairer, we should estimate the movement cost always as (1*3) for unexplored squares?
	int time_s = clock();
	// if unit is NULL, assume general land unit
	short *dists = new short[width*height];
	short *temp = new short[width*height];
	for(int i=0;i<width*height;i++)
		dists[i] = -1;

	dists[sy * width + sx] = 0;
	bool done = false;
	bool reached_target = false;
	while( !done && !reached_target ) {
		done = true;
		for(int i=0;i<width*height;i++)
			temp[i] = -1;
		//bool check_stop = false;
		for(int y=0;y<height && !reached_target;y++) {
			for(int x=0;x<width && !reached_target;x++) {
				int cost = dists[y * width + x];
				if(cost==-1) {
					/*if(unit->validLocation(this,x,y))
					done = false;*/
					continue;
				}
				for(int cy=y-1;cy<=y+1 && !reached_target;cy++) {
					for(int cx=x-1;cx<=x+1 && !reached_target;cx++) {

						if( !isValid(cx,cy) )
							continue;

						if( dists[cy*width + cx] == -1 ) {
							MapSquare *square = this->getSquare(cx, cy);
							if( ( unit == NULL && square->isLand() ) ||
								( unit != NULL && unit->canMoveTo( square ) ) ) {
									int mp = square->getScaledMovementCost();
									if( unit == NULL )
										mp = min(mp,road_move_scale_c);
									else
										mp = min(mp,road_move_scale_c*unit->getTemplate()->getMoves());
									temp[cy * width + cx] = cost + mp;
									if( cx == ex && cy == ey ) {
										reached_target = true;
									}
									done = false;
									/*if( stop_x != -1 && stop_y != -1 ) {
									int diffx = abs( cx - stop_x );
									int diffy = abs( cy - stop_y );
									if( diffx <= 1 && diffy <= 1 ) {
									check_stop = true;
									}
									}*/
							}
						}

					}
				}
				temp[y * width + x] = cost;
			}
		}
		/*short *t = temp;
		temp = dists;
		dists = t;*/
		swap(temp, dists);
		/*if( check_stop && !done && dists[stop_y * width + stop_x] != -1 ) {
		done = true;
		for(int dir=0;dir<8 && done;dir++) {
		int cx = 0, cy = 0;
		if( !this->translateDirection(&cx, &cy, stop_x, stop_y, dir) ) {
		if( dists[cy * width + cx] == -1 && unit->validSquare(cx,cy) ) {
		// valid adjacent square that hasn't been checked
		done = false;
		}
		}
		}
		}*/
	}

	delete [] temp;
	VI_log("### Map::calculateDistanceMap %d : %d\n", unit, clock() - time_s);
	return dists;
}
#endif

CityWindow::CityWindow(MainGamestate *mainGamestate, City *city) :
mainGamestate(mainGamestate), city(city), floatingInfoWindow(NULL),
okayButton(NULL), destroyButton(NULL), buildableListbox(NULL),
cached_power_per_turn(NULL) {
	/*if( InfoWindow::getActiveWindow() != NULL ) {
	// delete existing one
	delete InfoWindow::getActiveWindow();
	}
	*/

	this->cached_power_per_turn = city->getPowerPerTurn();

	this->panel = VI_createPanel();
	this->panel->setPaintFunc(paintPanel, this);
	//const int w = 480, h = 480;
	const int panel_x = 128, panel_y = 32;
	int w = game_g->getGraphicsEnvironment()->getWidth() - 2 * panel_x;
	//int w = game_g->getGraphicsEnvironment()->getWidth() - panel_x - 64;
	w = max(w, 496);
	int h = game_g->getGraphicsEnvironment()->getHeight() - panel_y - 16;
	h = min(h, 480);
	this->panel->setSize(w, h);
	this->panel->setBackground(1.0f, 1.0f, 1.0f, 0.9f);
	this->panel->setTexture(mainGamestate->getCityPanelTexture());

	okayButton = VI_createButton("Okay", game_g->getFont());
	okayButton->setUserData(this);
	okayButton->setAction(action);
	okayButton->setKeyShortcut(V_RETURN);
	this->panel->add(okayButton, 16, h - 32);

	destroyButton = VI_createButton("Destroy City", game_g->getFont());
	destroyButton->setUserData(this);
	destroyButton->setAction(action);
	this->panel->add(destroyButton, okayButton->getWidth() + 32, h - 32);

	vector<const Buildable *> *buildables = new vector<const Buildable *>();
	for(int i=0;i<mainGamestate->getNBuildables();i++) {
		const Buildable *buildable = mainGamestate->getBuildable(i);
		if( city->canBuild(buildable) ) {
			buildables->push_back(buildable);
		}
	}
	// convert to names
	vector<string> buildable_names;
	for(vector<const Buildable *>::const_iterator iter = buildables->begin();iter != buildables->end(); ++iter) {
		const Buildable *buildable = *iter;
		//const char *name = buildable->getName();
		stringstream name;
		name << buildable->getName();
		if( buildable->getType() == Buildable::TYPE_UNIT ) {
			const UnitTemplate *unit_template = static_cast<const UnitTemplate *>(buildable);
			/*name << " [";
			name << unit_template->getAttack();
			name << "/";
			name << unit_template->getDefence();
			name << "/";
			name << unit_template->getMoves();
			name << "]";*/
			name << " " << MainGamestate::getUnitStatsString(unit_template, true);
		}
		name << " (";
		name << buildable->getCost();
		name << ")";
		buildable_names.push_back(name.str());
	}
	//delete buildables;
	int ypos = 112;
	if( buildable_names.size() > 0 ) {
		buildableListbox = VI_createListbox(256, 128, &*buildable_names.begin(), buildable_names.size(), game_g->getFont());
		buildableListbox->setHasInputFocus(true);
		buildableListbox->setUserData(buildables);
		buildableListbox->setForeground(0.0f, 0.0f, 0.0f);
		//buildableListbox->setPopupText("Select an item that you wish to build, then click \"Change Build\" below.", game->getFont());
		buildableListbox->setActive(0);
		this->panel->add(buildableListbox, 16, ypos);

		changeBuildButton = VI_createButton("Change Build", game_g->getFont());
		changeBuildButton->setUserData(this);
		changeBuildButton->setAction(action);
		changeBuildButton->setPopupText("Change the current build item to the item selected in the list above.", game_g->getFont());
		this->panel->add(changeBuildButton, 16, ypos + 136);

		infoBuildButton = VI_createButton("Info.", game_g->getFont());
		infoBuildButton->setUserData(this);
		infoBuildButton->setAction(action);
		infoBuildButton->setPopupText("Show information about the item selected in the list above.", game_g->getFont());
		this->panel->add(infoBuildButton, 120, ypos + 136);
	}
	else {
		buildableListbox = NULL;
		changeBuildButton = NULL;
		infoBuildButton = NULL;
	}

	if( city->getNImprovements() > 0 ) {
		vector<string> improvement_names;
		improvement_names.push_back("Improvements:");
		for(int i=0;i<city->getNImprovements();i++) {
			const Improvement *improvement = city->getImprovement(i);
			stringstream name;
			name << improvement->getName();
			improvement_names.push_back(name.str());
		}
		//improvementsListbox = VI_createListbox(256, 224, &*improvement_names.begin(), improvement_names.size(), game_g->getFont());
		improvementsListbox = VI_createListbox(w - 336, 224, &*improvement_names.begin(), improvement_names.size(), game_g->getFont());
		improvementsListbox->setForeground(0.0f, 0.0f, 0.0f);
		this->panel->add(improvementsListbox, 320, 16);
	}
	else {
		improvementsListbox = NULL;
	}

	const vector<Unit *> *units = mainGamestate->getMap()->findUnitsAt(city->getX(), city->getY());
	vector<Unit *> *units_copy = new vector<Unit *>(); // hack as setUserData only accepts non-const pointer
	for(vector<Unit *>::const_iterator iter = units->begin(); iter != units->end(); ++iter) {
		Unit *unit = *iter;
		units_copy->push_back(unit);
	}
	// convert to names
	vector<string> unit_names = MainGamestate::getUnitNames(units);
	//ypos += 168;
	ypos += 160;
	if( unit_names.size() > 0 ) {
		//int listbox_h = h - 352;
		//int listbox_h = h - 344;
		int listbox_h = h - 336;
		//int listbox_h = h - 320;
		unitsListbox = VI_createListbox(w - 32, listbox_h, &*unit_names.begin(), unit_names.size(), game_g->getFont());
		// don't give input focus
		unitsListbox->setForeground(0.0f, 0.0f, 0.0f);
		unitsListbox->setUserData(units_copy);
		unitsListbox->setActive(0);
		this->panel->add(unitsListbox, 16, ypos);

		int xpos = 16;
		activateUnitButton = VI_createButton("Activate Unit", game_g->getFont());
		activateUnitButton->setUserData(this);
		activateUnitButton->setAction(action);
		activateUnitButton->setPopupText("Activate (and unfortify) the unit selected in the list above.", game_g->getFont());
		//this->panel->add(activateUnitButton, 16, ypos + 136);
		this->panel->add(activateUnitButton, xpos, ypos + listbox_h + 8);
		xpos += activateUnitButton->getWidth() + 8;

		activateAllUnitsButton = VI_createButton("Activate All Units", game_g->getFont());
		activateAllUnitsButton->setUserData(this);
		activateAllUnitsButton->setAction(action);
		activateAllUnitsButton->setPopupText("Activate (and unfortify) all the units in the list above.", game_g->getFont());
		//this->panel->add(activateAllUnitsButton, 112, ypos + 136);
		//this->panel->add(activateAllUnitsButton, 112, ypos + listbox_h + 8);
		this->panel->add(activateAllUnitsButton, xpos, ypos + listbox_h + 8);
		xpos += activateAllUnitsButton->getWidth() + 8;

		fortifyAllUnitsButton = VI_createButton("Fortify All Units", game_g->getFont());
		fortifyAllUnitsButton->setUserData(this);
		fortifyAllUnitsButton->setAction(action);
		fortifyAllUnitsButton->setPopupText("Fortify all the units in the list above.", game_g->getFont());
		this->panel->add(fortifyAllUnitsButton, xpos, ypos + listbox_h + 8);
		xpos += fortifyAllUnitsButton->getWidth() + 8;

	}
	else {
		unitsListbox = NULL;
		activateUnitButton = NULL;
		activateAllUnitsButton = NULL;
		fortifyAllUnitsButton = NULL;
	}

	mainGamestate->getPanel()->add(this->panel, panel_x, panel_y);

	//game->lockGUI();
	game_g->getGraphicsEnvironment()->setModalPanel( this->panel );
}

CityWindow::~CityWindow() {
	if( this->floatingInfoWindow != NULL ) {
		delete floatingInfoWindow;
	}
	/*if( InfoWindow::getActiveWindow() != NULL ) {
		delete InfoWindow::getActiveWindow();
	}*/
	if( this->buildableListbox != NULL ) {
		vector<const Buildable *> *buildables = static_cast<vector<const Buildable *> *>(this->buildableListbox->getUserData());
		delete buildables;
	}
	if( this->unitsListbox != NULL ) {
		vector<Unit *> *units = static_cast<vector<Unit *> *>(this->unitsListbox->getUserData());
		delete units;
	}

	mainGamestate->getPanel()->remove(this->panel);
	delete this->panel;
	delete this->okayButton;
	delete this->destroyButton;
	delete this->buildableListbox;
	delete this->changeBuildButton;
	delete this->infoBuildButton;
	delete this->unitsListbox;
	delete this->activateUnitButton;
	delete this->activateAllUnitsButton;
	delete this->fortifyAllUnitsButton;
	// TODO: delete child GUI elements recursively?
	//game->unlockGUI();
	//game->getGraphicsEnvironment()->setModalPanel(NULL);
}

void CityWindow::paintPanel(VI_Graphics2D *g2d, void *data) {
	CityWindow *cityWindow = static_cast<CityWindow *>(data);
	const City *city = cityWindow->city;
	g2d->setColor3i(0, 0, 0);
	int xpos = 16, ypos = 16;
	const int ydiff = 16;
	game_g->getFont()->writeTextExt(xpos, ypos, "%s", city->getName()); ypos += ydiff;
	game_g->getFont()->writeTextExt(xpos, ypos, "Size: %d", city->getSize()); ypos += ydiff;
	game_g->getFont()->writeTextExt(xpos, ypos, "Population: %d", city->getPopulation()); ypos += ydiff;
	game_g->getFont()->writeTextExt(xpos, ypos, "(Growth: %d / %d)", city->getProgressPopulation(), city->getPopulationCost()); ypos += ydiff;
	const Buildable *buildable = city->getBuildable();
	if( buildable == NULL ) {
		g2d->setColor3i(255, 0, 0);
		game_g->getFont()->writeText(xpos, ypos, "Building: Nothing"); ypos+= ydiff;
		g2d->setColor3i(0, 0, 0);
	}
	else {
		game_g->getFont()->writeTextExt(xpos, ypos, "Building: %s", buildable->getName()); ypos += ydiff;
		game_g->getFont()->writeTextExt(xpos, ypos, "(Built: %d / %d)", city->getProgressBuildable(), buildable->getCost());
	}

	xpos = 160;
	ypos = 16;
	game_g->getFont()->writeTextExt(xpos, ypos, "Production: %d", city->getProduction()); ypos += ydiff;
	game_g->getFont()->writeTextExt(xpos, ypos, "Science: %d", city->getScience()); ypos += ydiff;
	game_g->getFont()->writeTextExt(xpos, ypos, "Power Per Turn: %d", city->getPowerPerTurn()); ypos += ydiff;
	if( city->isResisting() ) {
		g2d->setColor3i(255, 0, 0);
		game_g->getFont()->writeTextExt(xpos, ypos, "RESISTANCE: %d%%", city->getResistance());
		g2d->setColor3i(0, 0, 0);
	}
	ypos += ydiff;

	/*xpos = 320;
	ypos = 16;
	for(int i=0;i<city->getNImprovements();i++) {
		const Improvement *improvement = city->getImprovement(i);
		if( i == 0 ) {
			game_g->getFont()->writeText(xpos, ypos, "Improvements:"); ypos+= ydiff;
		}
		game_g->getFont()->writeTextExt(xpos, ypos, "%s", improvement->getName()); ypos+= ydiff;
	}
	ypos+= ydiff;*/
}

void CityWindow::action(VI_Panel *source) {
	CityWindow *cityWindow = static_cast<CityWindow *>(source->getUserData());
	cityWindow->mainGamestate->playSound(MainGamestate::SOUND_BLIP);
	if( source == cityWindow->okayButton ) {
		MainGamestate *mainGamestate = cityWindow->mainGamestate;
		delete cityWindow;
		game_g->getGraphicsEnvironment()->flushInput(); // needed to avoid problem where the mouse click is registered after the window closes - is there a better way to do this?
		return;
	}
	else if( source == cityWindow->destroyButton ) {
		MainGamestate *mainGamestate = cityWindow->mainGamestate;
		stringstream str;
		str << "Are you sure you wish to destroy the city of " << cityWindow->city->getName() << "?";
		if( InfoWindow::confirm(game_g->getGraphicsEnvironment(), mainGamestate->getPanel(), game_g->getFont(), game_g->getPanelTexture(), str.str().c_str(), "Yes, destroy the city", "No") ) {
			City *city = cityWindow->city;
			mainGamestate->addFireEffect(city->getPos());
			// need to also destroy any air units
			const vector<Unit *> *units = mainGamestate->getMap()->findUnitsAt(city->getPos());
			// n.b. - count backwards as the units vector is modified when we delete units!
			for(int i=units->size()-1;i>=0;i--) {
				const Unit *unit = units->at(i);
				if( unit->getTemplate()->isAir() ) {
					if( mainGamestate->getActiveUnit() == unit ) {
						mainGamestate->setActiveUnit(NULL);
					}
					delete unit;
				}
			}
			delete cityWindow;
			delete city;
		}
		return;
	}
	else if( source == cityWindow->changeBuildButton ) {
		int active = cityWindow->buildableListbox->getActive();
		if( active != -1 ) {
			vector<const Buildable *> *buildables = static_cast<vector<const Buildable *> *>(cityWindow->buildableListbox->getUserData());
			const Buildable *buildable = buildables->at(active);
			bool ok = true;
			int current_progress = cityWindow->city->getProgressBuildable();
			if( buildable != cityWindow->city->getBuildable() && current_progress > 0 ) {
				stringstream text;
				text << "Are you sure you wish to change production and waste " << current_progress << " resources?";
				if( !InfoWindow::confirm(game_g->getGraphicsEnvironment(), game_g->getGamestate()->getPanel(), game_g->getFont(), game_g->getPanelTexture(), text.str(), "Yes, Change Production", "Cancel") ) {
					ok = false;
				}
			}
			if( ok ) {
				cityWindow->city->setBuildable(buildable);
			}
		}
	}
	else if( source == cityWindow->infoBuildButton ) {
		int active = cityWindow->buildableListbox->getActive();
		if( active != -1 ) {
			vector<const Buildable *> *buildables = static_cast<vector<const Buildable *> *>(cityWindow->buildableListbox->getUserData());
			const Buildable *buildable = buildables->at(active);
			stringstream text;
			text << buildable->getName();
			text << ":\n\n";
			if( buildable->getType() == Buildable::TYPE_IMPROVEMENT ) {
				const Improvement *improvement = static_cast<const Improvement *>(buildable);
				text << "City Improvement\n";
				text << "Cost : " << buildable->getCost() << "\n";
				if( improvement->getPowerPerTurn() > 0 ) {
					text << "Power Per Turn : " << improvement->getPowerPerTurn() << "\n";
				}
				bool any_allows = false;
				for(int i=0;i<game_g->getGameData()->getNBuildables();i++) {
					const Buildable *b = game_g->getGameData()->getBuildable(i);
					if( b->getRequiresImprovement() == buildable ) {
						if( !any_allows ) {
							any_allows = true;
							text << "Allows :\n";
						}
						text << "    " << b->getName() << "\n";
					}
				}
				if( buildable->getReplacedBy() != NULL ) {
					text << "Replaced by : " << buildable->getReplacedBy()->getName() << "\n";
				}
				if( buildable->getObsoletedBy() != NULL ) {
					text << "Made obsolete by : " << buildable->getObsoletedBy()->getName() << "\n";
				}
			}
			else if( buildable->getType() == Buildable::TYPE_UNIT ) {
				const UnitTemplate *unit_template = static_cast<const UnitTemplate *>(buildable);
				text << "Unit\n";
				text << "Cost : " << buildable->getCost() << "\n";
				if( unit_template->isAir() ) {
					text << "Attack : " << unit_template->getAirAttack() << "\n";
					text << "Defence : " << unit_template->getAirDefence() << "\n";
				}
				else {
					text << "Attack : " << unit_template->getAttack() << "\n";
					text << "Defence : " << unit_template->getDefence() << "\n";
				}
				if( unit_template->getBombard() > 0 ) {
					text << "Bombard : " << unit_template->getBombard() << "%\n";
					text << "Bombard Power : " << unit_template->getBombardPower() << "\n";
				}
				if( unit_template->getAirRange() > 0 ) {
					text << "Air Range : " << unit_template->getAirRange() << "\n";
				}
				text << "Moves : " << unit_template->getMoves() << "\n";
				text << "Visibility Range : " << unit_template->getVisibilityRange() << "\n";
				if( unit_template->getReplacedBy() != NULL ) {
					text << "Replaced by : " << unit_template->getReplacedBy()->getName() << "\n";
				}
				if( unit_template->getObsoletedBy() != NULL ) {
					text << "Made obsolete by : " << unit_template->getObsoletedBy()->getName() << "\n";
				}
				bool nl = false;
				if( unit_template->isFoot() ) {
					if( !nl )
						text << "\n";
					text << "Foot unit.\n";
				}
				if( unit_template->canUpgrade() ) {
					if( !nl )
						text << "\n";
					text << "This unit can upgrade.\n";
				}
				if( unit_template->canBuildCity() ) {
					if( !nl )
						text << "\n";
					nl = true;
					text << "This unit can build cities.\n";
				}
				if( unit_template->canBuildRoads() ) {
					if( !nl )
						text << "\n";
					nl = true;
					text << "This unit can build roads/railways.\n";
				}
			}
			else {
				ASSERT( false );
			}
			text << "\n";
			text << buildable->getInfo();
			text << "\n";
			cityWindow->createInfoWindow(text.str());
		}
	}
	else if( source == cityWindow->activateUnitButton && cityWindow->city->getCivilization() == cityWindow->mainGamestate->getPlayer() ) {
		// need check for player in case of debug mode
		int active = cityWindow->unitsListbox->getActive();
		if( active != -1 ) {
			vector<Unit *> *units = static_cast<vector<Unit *> *>(cityWindow->unitsListbox->getUserData());
			Unit *unit = units->at(active);
			if( cityWindow->mainGamestate->getPhase() == MainGamestate::PHASE_PLAYERTURN && unit->hasMovesLeft() ) {
				cityWindow->mainGamestate->activateUnit(unit, false);
			}
			else {
				// still activate (for next turn?)
				//unit->setStatus(Unit::STATUS_NORMAL);
				cityWindow->mainGamestate->wakeupUnit(unit);
			}
			vector<string> unit_names = MainGamestate::getUnitNames(units);
			cityWindow->unitsListbox->reset(&*unit_names.begin(), unit_names.size());
			//cityWindow->unitsListbox->setActive(active);
		}
	}
	else if( source == cityWindow->activateAllUnitsButton && cityWindow->city->getCivilization() == cityWindow->mainGamestate->getPlayer() ) {
		// need check for player in case of debug mode
		vector<Unit *> *units = static_cast<vector<Unit *> *>(cityWindow->unitsListbox->getUserData());
		for(vector<Unit *>::iterator iter = units->begin(); iter != units->end(); ++iter) {
			Unit *unit = *iter;
			if( cityWindow->mainGamestate->getPhase() == MainGamestate::PHASE_PLAYERTURN && unit->hasMovesLeft() ) {
				cityWindow->mainGamestate->activateUnit(unit, false);
			}
			else {
				// still activate (for next turn?)
				//unit->setStatus(Unit::STATUS_NORMAL);
				cityWindow->mainGamestate->wakeupUnit(unit);
			}
			vector<string> unit_names = MainGamestate::getUnitNames(units);
			cityWindow->unitsListbox->reset(&*unit_names.begin(), unit_names.size());
		}
	}
	else if( source == cityWindow->fortifyAllUnitsButton && cityWindow->city->getCivilization() == cityWindow->mainGamestate->getPlayer() ) {
		// need check for player in case of debug mode
		vector<Unit *> *units = static_cast<vector<Unit *> *>(cityWindow->unitsListbox->getUserData());
		for(vector<Unit *>::iterator iter = units->begin(); iter != units->end(); ++iter) {
			Unit *unit = *iter;
			unit->setStatus(Unit::STATUS_FORTIFIED);
			if( unit == cityWindow->mainGamestate->getActiveUnit() ) {
				cityWindow->mainGamestate->setActiveUnit(NULL);
			}
			vector<string> unit_names = MainGamestate::getUnitNames(units);
			cityWindow->unitsListbox->reset(&*unit_names.begin(), unit_names.size());
		}
	}
}

void CityWindow::actionInfoWindow(VI_Panel *source) {
	void *ptr = source->getUserData();
	CityWindow *cityWindow = static_cast<CityWindow *>(ptr);
	cityWindow->mainGamestate->playSound(MainGamestate::SOUND_BLIP);
	ASSERT( cityWindow->floatingInfoWindow != NULL );
	delete cityWindow->floatingInfoWindow;
	cityWindow->floatingInfoWindow = NULL;
	game_g->getGraphicsEnvironment()->flushInput(); // needed to avoid problem where the mouse click is registered after the window closes - is there a better way to do this?
}

void CityWindow::createInfoWindow(string text) {
	if( this->floatingInfoWindow != NULL ) {
		// delete existing one
		delete floatingInfoWindow;
	}
	/*if( InfoWindow::getActiveWindow() != NULL ) {
		// delete existing one
		delete InfoWindow::getActiveWindow();
	}*/
	string okay = "Okay";
	int window_x = 312;
	int window_w = 200;
	this->floatingInfoWindow = new InfoWindow(game_g->getGraphicsEnvironment(), this->panel, text, game_g->getFont(), game_g->getPanelTexture(), &okay, 1, window_x, 16, window_w, 320);
	this->floatingInfoWindow->getButton(0)->setAction(actionInfoWindow);
	this->floatingInfoWindow->getButton(0)->setUserData(this);
}

Effect::Effect(VI_Entity *node_2d, VI_Entity *node_3d, int xpos, int ypos, int duration) : node_2d(node_2d), node_3d(node_3d), xpos(xpos), ypos(ypos) {
	time_e = clock() + duration;
}

void MainGamestate::actionInfoWindow(VI_Panel *source) {
	mainGamestate->playSound(SOUND_BLIP);
	ASSERT( mainGamestate->infoWindow != NULL );
	delete mainGamestate->infoWindow;
	mainGamestate->infoWindow = NULL;
	game_g->getGraphicsEnvironment()->flushInput(); // needed to avoid problem where the mouse click is registered after the window closes - is there a better way to do this?
}

void MainGamestate::createInfoWindow(string text) {
	/*if( InfoWindow::getActiveWindow() != NULL ) {
		// delete existing one
		delete InfoWindow::getActiveWindow();
	}*/
	if( infoWindow != NULL ) {
		// delete existing one
		delete infoWindow;
	}
	string okay = "Close Help Window";
	const int wid = 200;
	infoWindow = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text, game_g->getFont(), game_g->getPanelTexture(), &okay, 1, game_g->getGraphicsEnvironment()->getWidth() - wid - 16, 16, wid, 320);
	this->infoWindow->getButton(0)->setAction(actionInfoWindow);
}

string MainGamestate::getUnitStatsString(const UnitTemplate *unit_template, bool include_moves) {
	stringstream text;
	if( unit_template->isAir() ) {
		text << "[" << unit_template->getAirAttack() << "/" << unit_template->getAirDefence() << "/" << unit_template->getAirRange() << "/" << unit_template->getBombard() << "%";
		if( include_moves ) {
			text << "/" << unit_template->getMoves();
		}
		text << "]";
	}
	else {
		text << "[" << unit_template->getAttack() << "/" << unit_template->getDefence();
		if( include_moves ) {
			text << "/" << unit_template->getMoves();
		}
		text << "]";
	}
	return text.str();
}

string MainGamestate::getUnitName(const Unit *unit) {
	const UnitTemplate *unit_template = unit->getTemplate();
	stringstream name;
	name << unit->getCivilization()->getNameAdjective();
	name << " ";
	name << unit_template->getName();
	name << " (Moves: ";
	Rational moves_left = unit->movesLeft();
	int moves_left_integer, moves_left_num, moves_left_den;
	moves_left.value(&moves_left_integer, &moves_left_num, &moves_left_den);
	ASSERT( moves_left_num == 0 || moves_left_den == road_move_scale_c );
	if( moves_left_num == 0 ) {
		name << moves_left_integer;
	}
	else if( moves_left_integer == 0 ) {
		name << moves_left_num << "/" << moves_left_den;
	}
	else {
		name << moves_left_integer << " " << moves_left_num << "/" << moves_left_den;
	}
	name << " out of " << unit_template->getMoves() << ") ";
	name << getUnitStatsString(unit_template, false);
	if( unit->getStatus() == Unit::STATUS_NORMAL ) {
		// skip
	}
	else if( unit->getStatus() == Unit::STATUS_FORTIFIED ) {
		name << " [FORTIFIED]";
	}
	else if( unit->getStatus() == Unit::STATUS_BUILDING_ROAD ) {
		name << " [BUILDING ROAD]";
	}
	else if( unit->getStatus() == Unit::STATUS_BUILDING_RAILWAYS ) {
		name << " [BUILDING RAILWAYS]";
	}
	else {
		ASSERT(false);
	}
	if( unit->isAutomated() ) {
		Pos2D goto_target;
		if( unit->hasGotoTarget(&goto_target) ) {
			name << " [GOTO " << goto_target.x << "," << goto_target.y << "]";
		}
		else {
			name << " [AUTOMATED]";
		}
	}
	return name.str();
}

vector<string> MainGamestate::getUnitNames(const vector<Unit *> *units) {
	vector<string> unit_names;
	for(vector<Unit *>::const_iterator iter = units->begin();iter != units->end(); ++iter) {
		const Unit *unit = *iter;
		string name = getUnitName(unit);
		unit_names.push_back(name.c_str());
	}
	return unit_names;
}

void GameData::addTechnology(Technology *technology) {
	technologies.push_back(technology);
}

void GameData::addBuildable(Buildable *buildable) {
	buildables.push_back(buildable);
	if( buildable->getType() == Buildable::TYPE_IMPROVEMENT ) {
		const Improvement *improvement = static_cast<const Improvement *>(buildable);
		improvements.push_back(improvement);
	}
	else if( buildable->getType() == Buildable::TYPE_UNIT ) {
		const UnitTemplate *unit_template = static_cast<const UnitTemplate *>(buildable);
		unit_templates.push_back(unit_template);
	}
	else {
		ASSERT(false);
	}
}

int MainGamestate::getSqWidth() const {
	return this->scale_width * this->def_sq_width;
}

int MainGamestate::getSqHeight() const {
	return this->scale_height * this->def_sq_height;
}

// given (fractional) map position x, y, return the screen coords
Pos2D MainGamestate::getScreenPos(float x, float y) const {
	ASSERT( !view_3d );
	// must be floats, not ints, for smooth movement of units to work!
	//this->map->reduceToBase(&x, &y);
	int sq_width = this->getSqWidth();
	int sq_height = this->getSqHeight();
	//Pos2D pos(game->sq_width * (x - map_pos_x), game->sq_height * ( y - map_pos_y ));
	int off_x = (int)(map_pos_x * sq_width);
	int off_y = (int)(map_pos_y * sq_height);
	// take into account wrap around
	Pos2D pos(sq_width * x - off_x, sq_height * y - off_y);
	this->map->reduceScreenPosToBase(&pos.x, &pos.y);
	return pos;
}

// given screen coords x, y, return the (fractional) map position
void MainGamestate::getMapPosf(float *mx, float *my, int x, int y) const {
	if( view_3d ) {
		VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
		Vector3D view_dir = game_g->getGraphicsEnvironment()->getDirectionOfPoint(vp, x, y);
		Vector3D pos = vp->getPosition();
		double lambda = - pos.y / view_dir.y;
		ASSERT( lambda >= 0.0 );
		pos.x += (float)(lambda * view_dir.x);
		pos.z += (float)(lambda * view_dir.z);
		pos.x /= xscale;
		pos.z /= zscale;
		*mx = pos.x;
		*my = pos.z;
	}
	else {
		int sq_width = this->getSqWidth();
		int sq_height = this->getSqHeight();
		int off_x = (int)(map_pos_x * sq_width);
		int off_y = (int)(map_pos_y * sq_height);
		*mx = ((float)(x+off_x))/(float)sq_width;
		*my = ((float)(y+off_y))/(float)sq_height;
		//this->map->reduceToBase(mx, my);
	}
}

Pos2D MainGamestate::getMapPos(int x, int y) const {
	/*int sq_width = this->getSqWidth();
	int sq_height = this->getSqHeight();
	int off_x = (int)(map_pos_x * sq_width);
	int off_y = (int)(map_pos_y * sq_height);
	Pos2D pos(((float)(x+off_x))/(float)sq_width, ((float)(y+off_y))/(float)sq_height);
	return pos;*/
	float mx = 0.0f, my = 0.0f;
	getMapPosf(&mx, &my, x, y);
	Pos2D pos(mx, my);
	return pos;
}

void MainGamestate::switchView() {
	if( view_3d ) {
		float mx0 = 0.0f, mx1 = 0.0f, my = 0.0f;
		getMapPosf(&mx0, &my, 0, 0);
		getMapPosf(&mx1, &my, game_g->getGraphicsEnvironment()->getWidth()-1, 0);

		view_3d = false;

		VI_Terrain *terrain = game_g->getGraphicsEnvironment()->getWorld()->getTerrain();
		ASSERT( terrain != NULL );
		delete terrain;
		game_g->getGraphicsEnvironment()->getWorld()->setTerrain(NULL);

		// sort out position
		this->map_pos_x = mx0;
		this->map_pos_y = my;

		// sort out zoom
		float width = mx1 - mx0;
		// g_width / ( def_sq_width * scale_width ) == width
		this->scale_width = game_g->getGraphicsEnvironment()->getWidth() / ( width * this->def_sq_width );
		this->scale_height = this->scale_width;
	}
	else {
		float o_mx = 0.0f, o_my = 0.0f;
		getMapPosf(&o_mx, &o_my, 0, 0);
		//getMapPosf(&o_mx, &o_my, game_g->getGraphicsEnvironment()->getWidth()/2, game_g->getGraphicsEnvironment()->getHeight()/2);

		view_3d = true;

		this->generateTerrain();

		// sort out zoom
		float prev_width = game_g->getGraphicsEnvironment()->getWidth() / this->getSqWidth();
		reset3DZoom(prev_width);

		// sort out position
		float n_mx = 0.0f, n_my = 0.0f;
		getMapPosf(&n_mx, &n_my, 0, 0);
		VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
		Vector3D vp_pos = vp->getPosition();
		vp_pos.x += o_mx - n_mx;
		vp_pos.z += o_my - n_my;
		vp->setPosition(&vp_pos);

	}
	this->checkMapView();
}

/*void Game::lockGUI() {
ASSERT( !mainGUILocked );
VI_log("Game::lockGUI\n");
mainGUILocked = true;
}

void Game::unlockGUI() {
ASSERT( mainGUILocked );
VI_log("Game::unlockGUI\n");
mainGUILocked = false;
}*/

void MainGamestate::beginTurn() {
	VI_log("Begin turn\n");
	// called at start of new turn
	for(vector<Civilization *>::iterator iter = civilizations.begin(); iter != civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		for(int i=0;i<civilization->getNUnits();i++) {
			Unit *unit = civilization->getUnit(i);
			unit->reset();
		}
	}
	moved_units_this_turn = false;
	year++;
	VI_log("Year is now %d\n", year);
	phase = PHASE_PLAYERTURN;
	if( !test_mode && ( year > 1 || SAVE_FIRST_TURN ) ) {
		string autofilename = savegames_dirname;
		autofilename += "/auto.sav";
		string auto2filename = savegames_dirname;
		auto2filename += "/auto2.sav";
		// either of these two functions may be expected to fail if files don't exist
		if( remove(auto2filename.c_str()) != 0 ) {
			VI_log("failed to remove old autosave2 file - error: %d\n", errno);
		}
		if( rename(autofilename.c_str(), auto2filename.c_str()) != 0 ) {
			VI_log("failed to rename old autosave file - error: %d\n", errno);
		}
		this->save(autofilename.c_str(), true);
		/*char *full_autofilename = VI_getApplicationFilename(autofilename.c_str());
		char *full_auto2filename = VI_getApplicationFilename(auto2filename.c_str());
		if( remove(full_auto2filename) != 0 ) {
			VI_log("failed to remove old autosave2 file - error: %d\n", errno);
		}
		if( rename(full_autofilename, full_auto2filename) != 0 ) {
			VI_log("failed to rename old autosave file - error: %d\n", errno);
		}
		this->save(full_autofilename, true);
		delete [] full_autofilename;
		delete [] full_auto2filename;*/
	}

	// currently called per-frame anyway - but needed here for testing in "silent" mode
	this->calculateTerritory();
}

void MainGamestate::logEndturnTime(const char *text) {
	int time_new = clock();
	VI_log("### %s: %d\n", text, time_new - time_store);
	time_store = time_new;
}

void MainGamestate::endPlayerTurn() {
	VI_log("End player's turn\n");

	// test:
	/*AIInterface::setMainGamestate(this);
	AIInterface::setAICivilization(this->getCivilization(1));
	game_g->runScript("conquests/data/scripts/ai.lua", "test");
	AIInterface::setMainGamestate(NULL);
	AIInterface::setAICivilization(NULL);*/

	time_store = time_endplayerturn_s = clock();
	active_unit = NULL;
	updateUnitButtons();
	setInputMode(INPUTMODE_NORMAL);
	phase = PHASE_CPUTURN;
	//time_playerturnend = clock();
	if( infoWindow != NULL ) {
		delete infoWindow;
		infoWindow = NULL;
	}

	// consider declaring war, or making peace
	int permille_chance_war_per_turn_c = 15;
	if( this->aiaggression == AIAGGRESSION_PEACEFUL )
		permille_chance_war_per_turn_c /= 2;
	const int permille_chance_peace_per_turn_c = 20; // only for between AIs
	for(vector<Civilization *>::const_iterator iter1=civilizations.begin();iter1 != civilizations.end(); ++iter1) {
		Civilization *civ1 = *iter1;
		if( civ1 == player )
			continue;
		if( civ1->isDead() )
			continue;
		for(vector<Civilization *>::const_iterator iter2=civilizations.begin();iter2 != civilizations.end(); ++iter2) {
			Civilization *civ2 = *iter2;
			if( civ1 == civ2 )
				continue;
			if( civ2->isDead() )
				continue;
			Relationship *relationship = this->findRelationship(civ1, civ2);
			if( !relationship->madeContact() ) {
				continue;
			}
			bool inform_player = false;
			if( civ2 == player ) {
				inform_player = true;
			}
			else if( !player->isDead() ) {
				const Relationship *r = this->findRelationship(civ1, player);
				if( r->madeContact() )
					inform_player = true;
				else {
					r = this->findRelationship(civ2, player);
					if( r->madeContact() )
						inform_player = true;
				}
			}
			int dummy = 0;
			int civ1_s = civ1->getMilitary(&dummy);
			int civ2_s = civ2->getMilitary(&dummy);
			int civ2_n = civ2->getNuclear();
			if( relationship->getStatus() == Relationship::STATUS_WAR ) {
				// we have to take the maximum of from the point of view of both civs
				int min_turns_war1 = this->getMinTurnsWar(civ1, civ2);
				int min_turns_war2 = this->getMinTurnsWar(civ2, civ1);
				int min_turns_war = max(min_turns_war1, min_turns_war2);
				VI_log("%s vs %s: min_turns_war: %d, %d -> %d\n", civ1->getName(), civ2->getName(), min_turns_war1, min_turns_war2, min_turns_war);
				if( civ2 != player && year >= relationship->getYearDeclaredWar() + min_turns_war ) {
					//if( civ2 != player ) {
					// consider making peace
					int chance_make_peace = permille_chance_peace_per_turn_c;
					if( civ2_n >= 2 ) {
						chance_make_peace *= 10;
					}
					else if( civ1_s > 1.5 * civ2_s ) {
						chance_make_peace /= 2;
					}
					else if( 1.5 * civ1_s < civ2_s ) {
						chance_make_peace *= 2;
					}
					int prob = rand() % 1000;
					//prob = 0;
					VI_log("peace? rolled a %d ; chance_make_peace %d\n", prob, chance_make_peace);
					if( prob < chance_make_peace ) {
						relationship->setStatus(Relationship::STATUS_PEACE);
						VI_log("%s makes peace with %s\n", civ1->getName(), civ2->getName());
						if( inform_player ) {
							stringstream text;
							text << "The war between the ";
							text << civ1->getName();
							text << " and the ";
							text << civ2->getName();
							text << " is over.";
							InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture());
							window->doModal();
							delete window;
						}
					}
				}
			}
			else {
				int chance_declare_war = permille_chance_war_per_turn_c;
				if( civ2_n >= 2 ) {
					chance_declare_war /= 5;
				}
				else if( civ1_s > 1.5 * civ2_s ) {
					chance_declare_war *= 2;
				}
				else if( 1.5 * civ1_s < civ2_s ) {
					chance_declare_war /= 2;
				}
				int prob = rand() % 1000;
				//prob = 0;
				VI_log("%s vs %s: war? rolled a %d ; chance_declare_war %d\n", civ1->getName(), civ2->getName(), prob, chance_declare_war);
				if( prob < chance_declare_war ) {
					relationship->setStatus(Relationship::STATUS_WAR);
					VI_log("%s declare war on %s\n", civ1->getName(), civ2->getName());
					if( inform_player ) {
						stringstream text;
						text << civ1->getName();
						if( civ2 == player ) {
							text << " have declared war on us!";
						}
						else {
							text << " declare war on the ";
							text << civ2->getName();
							text << "!";
						}
						InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture());
						window->doModal();
						delete window;
					}
				}
			}
		}
	}

	// exchange technologies, share maps?
	for(vector<Civilization *>::const_iterator iter1=civilizations.begin();iter1 != civilizations.end(); ++iter1) {
		Civilization *civ1 = *iter1;
		if( civ1 == player )
			continue;
		if( civ1->isDead() )
			continue;
		for(vector<Civilization *>::const_iterator iter2=iter1+1;iter2 != civilizations.end(); ++iter2) {
			Civilization *civ2 = *iter2;
			if( civ2 == player )
				continue;
			if( civ2->isDead() )
				continue;
			Relationship *relationship = this->findRelationship(civ1, civ2);
			if( !relationship->madeContact() ) {
				continue;
			}
			if( relationship->getStatus() == Relationship::STATUS_PEACE ) {
				bool done = false;
				while( !done ) {
					vector<const Technology *> civ1_distinct_technologies;
					vector<const Technology *> civ2_distinct_technologies;
					Civilization::getDistinctTechnologies(this, &civ1_distinct_technologies, &civ2_distinct_technologies, civ1, civ2);
					if( civ1_distinct_technologies.size() > 0 && civ2_distinct_technologies.size() > 0 ) {
						// exchange
						int choose1 = rand() % civ2_distinct_technologies.size();
						const Technology *tech1 = civ2_distinct_technologies.at( choose1 );
						civ1->addTechnology( tech1 );
						int choose2 = rand() % civ1_distinct_technologies.size();
						const Technology *tech2 = civ1_distinct_technologies.at( choose2 );
						civ2->addTechnology( tech2 );
						VI_log("civs exchange technology: %s take %s; %s take %s\n", civ1->getName(), tech1->getName(), civ2->getName(), tech2->getName());
					}
					else {
						done = true;
					}
				}
				if( civ1->canShareMaps(civ2) && civ1->compareMaps(civ2) ) {
					civ1->shareMaps(civ2);
					VI_log("civs share maps: %s and %s\n", civ1->getName(), civ2->getName());
				}
			}
		}
	}

	// reset AI that needs doing
	for(vector<Civilization *>::iterator iter = civilizations.begin(); iter != civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		civilization->resetAI();
	}
	// do city AI here
	// not strictly accurate, as all players have their cities updated, then they move, but will do for now
	for(vector<Civilization *>::iterator iter = civilizations.begin(); iter != civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		if( civilization == player ) {
			continue;
		}
		for(int i=0;i<civilization->getNCities();i++) {
			City *city = civilization->getCity(i);
			city->doAI();
		}
	}
	for(vector<Civilization *>::iterator iter = civilizations.begin(); iter != civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		if( civilization == player ) {
			continue;
		}
		for(int i=0;i<civilization->getNUnits();i++) {
			Unit *unit = civilization->getUnit(i);
			//unit->setAIStartPos();
			unit->setAIMoveCount();
			unit->setAIDest();
		}
	}
	logEndturnTime("city based AI");
	VI_log("Done city-based AI\n");
}

void MainGamestate::endTurn() {
	logEndturnTime("unit based AI");
	// called when all players' turns end
	VI_log("End turn\n");
	ASSERT( active_unit == NULL );
	// set all objects that need updating
	for(vector<Civilization *>::iterator iter = civilizations.begin(); iter != civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		for(int i=0;i<civilization->getNCities();i++) {
			City *city = civilization->getCity(i);
			//city->update();
			city->setNeedsUpdate();
		}
		for(int i=0;i<civilization->getNUnits();i++) {
			Unit *unit = civilization->getUnit(i);
			unit->update();
		}
		civilization->update();
	}
	//VI_log("### updating all cities and units: %d", clock() - time_endturn_s);
	logEndturnTime("updating all cities and units");
	phase = PHASE_ENDTURN;
}

void MainGamestate::updateUnitButtons() const {
	bool enabled = this->active_unit != NULL;
	this->waitButton->setVisible(enabled);
	this->skipTurnButton->setVisible(enabled);
	this->fortifyButton->setVisible(enabled);
	//this->buildCityButton->setVisible(enabled && active_unit->getTemplate()->canBuildCity() && map->canBuildCity(active_unit->getX(), active_unit->getY()));
	this->buildCityButton->setVisible(enabled && active_unit->getTemplate()->canBuildCity() && map->getSquare(active_unit->getX(), active_unit->getY())->canBuildCity());
	this->buildRoadButton->setVisible(enabled && active_unit->getTemplate()->canBuildRoads() && map->getSquare(active_unit->getX(), active_unit->getY())->getRoad() == MapSquare::ROAD_NONE);
	const Technology *tech = game_g->getGameData()->findTechnology("Steam Power");
	this->buildRailwaysButton->setVisible(enabled && active_unit->getTemplate()->canBuildRoads() && map->getSquare(active_unit->getX(), active_unit->getY())->getRoad() == MapSquare::ROAD_BASIC && player->hasTechnology(tech));
	this->automateButton->setVisible(enabled && active_unit->getTemplate()->canBuildRoads());
	this->gotoButton->setVisible(enabled && !active_unit->getTemplate()->isAir());

	bool can_travel = false;
	bool can_bomb = false;
	//if( enabled && !active_unit->getTemplate()->isAir() ) {
	if( enabled ) {
		const City *city = map->findCity(active_unit->getPos());
		//if( city != NULL && city->canTravel(active_unit->getTemplate()) ) {
		/*if( city != NULL && city->canTravel(by_air) ) {
			can_travel = true;
		}*/
		if( city != NULL ) {
			int sea_range = 0;
			int air_range = 0;
			city->getTravelRanges(&sea_range, &air_range, active_unit->getTemplate());
			if( sea_range > 0 || air_range > 0 ) {
				can_travel = true;
			}
		}
		if( active_unit->getTemplate()->isAir() ) {
			ASSERT( city != NULL );
			if( active_unit->getTemplate()->canBomb() && city->canLaunch(active_unit->getTemplate()) ) {
				can_bomb = true;
			}
		}
	}
	this->travelButton->setVisible(can_travel);
	//bool can_air_raid = enabled && active_unit->getTemplate()->isAir() && ( active_unit->getTemplate()->getBombard() > 0 || active_unit->getTemplate()->getNuclearType() != UnitTemplate::NUCLEARTYPE_NONE );
	this->airRaidButton->setVisible( can_bomb && active_unit->hasMovesLeft() );
}

void MainGamestate::activateUnit(Unit *unit, bool centre) {
	VI_log("MainGamestate::activateUnit\n");
	ASSERT( unit->getCivilization() == player );
	ASSERT( phase == PHASE_PLAYERTURN );
	ASSERT( unit->hasMovesLeft() );
	unit->setStatus(Unit::STATUS_NORMAL);
	unit->setAutomated(false); // also unsets goto
	blink_time_start = clock();
	active_unit = unit;
	updateUnitButtons();
	cache_active_unit = active_unit;
	moved_units_this_turn = true;
	if( centre ) {
		centreUnit();
	}
	bool created_window = false;
	if( active_unit->getTemplate()->canBuildCity() ) {
		if( !help_buildcity ) {
			this->createInfoWindow("Settlers can be used to build cities. Press 'B', or click 'Build City'.\n\nYou can also move units around, using the arrow keys, the numeric keypad, or right clicking on an adjacent square.");
			created_window = true;
			help_buildcity = true;
		}
	}
	else if( active_unit->getTemplate()->canBuildRoads() ) {
		if( !help_buildroad ) {
			this->createInfoWindow("This unit can be used to build roads. Press 'R', or click 'Build Road'.");
			created_window = true;
			help_buildroad = true;
		}
	}
	else if( active_unit->getTemplate()->canBomb() ) {
		if( !help_bomb ) {
			this->createInfoWindow("This unit can bomb enemy cities! Press 'B', or click 'Bomb', then select one of the highlighted red squares (an enemy city within the unit's range).");
			created_window = true;
			help_bomb = true;
		}
	}

	/*if( !created_window && InfoWindow::getActiveWindow() != NULL ) {
		delete InfoWindow::getActiveWindow();
	}*/
	if( !created_window && infoWindow != NULL ) {
		delete infoWindow;
		infoWindow = NULL;
	}
	//VI_addtag();
}

void MainGamestate::wakeupUnit(Unit *unit) {
	// makes the unit available - e.g., unfortifying, unautomating - without making it the active unit
	ASSERT( unit->getCivilization() == player );
	unit->setStatus(Unit::STATUS_NORMAL);
	unit->setAutomated(false); // also unsets goto
}

void MainGamestate::home() {
	if( player->getNCities() > 0 ) {
		const City *city = player->getCity(0);
		centre(city->getPos());
	}
	else if( player->getNUnits() > 0 ) {
		const Unit *unit = player->getUnit(0);
		centre(unit->getPos());
	}
}

void MainGamestate::setInputMode(InputMode inputMode) {
	if( this->inputMode != inputMode ) {
		this->inputMode = inputMode;
/*#ifdef _WIN32
		HCURSOR cursor = NULL;
		if( inputMode == INPUTMODE_TRAVEL || inputMode == INPUTMODE_BOMBARD )
			cursor = LoadCursor( NULL, IDC_CROSS );
		else {
			cursor = LoadCursor( NULL, IDC_ARROW );
		}
		SetCursor(cursor);
		SetClassLong(game_g->getGraphicsEnvironment()->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
		if( inputMode == INPUTMODE_TRAVEL || inputMode == INPUTMODE_BOMBARD )
			game_g->setCursor(cursor_target);
		else {
			game_g->setCursor(cursor_arrow);
		}
		if( inputMode == INPUTMODE_TRAVEL || inputMode == INPUTMODE_BOMBARD ) {
		}
		else {
			map->resetTargets();
		}
	}
}

void MainGamestate::centre(Pos2D pos) {
	this->centre(pos.x + 0.5f, pos.y + 0.5f);
}

void MainGamestate::centre(float mx, float my) {
	if( view_3d ) {
		VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
		Vector3D pos = vp->getPosition();
		Vector3D dir = game_g->getGraphicsEnvironment()->getDirectionOfPoint(vp, game_g->getGraphicsEnvironment()->getWidth()/2, game_g->getGraphicsEnvironment()->getHeight()/2);
		ASSERT( dir.y < 0.0f );
		double scale = pos.y / dir.y;
		pos.set(mx, 0.0f, my);
		pos += dir * scale;
		vp->setPosition(&pos);
	}
	else {
		/*int n_x = ( game_g->getGraphicsEnvironment()->getWidth() / this->getSqWidth() ) + 1;
		int n_y = ( game_g->getGraphicsEnvironment()->getHeight() / this->getSqHeight() ) + 1;*/
		float n_x = ( game_g->getGraphicsEnvironment()->getWidth() / (float)this->getSqWidth() );
		float n_y = ( game_g->getGraphicsEnvironment()->getHeight() / (float)this->getSqHeight() );
		mx -= n_x/2;
		my -= n_y/2;
		/*if( mx < 0 )
			mx = 0;
		if( my < 0 )
			my = 0;*/
		map_pos_x = mx;
		map_pos_y = my;
	}

	checkMapView();
}

void MainGamestate::centreUnit() {
	if( active_unit != NULL ) {
		Pos2D pos = active_unit->getPos();
		centre(pos);
	}
}

void MainGamestate::addFireEffect(Pos2D pos) {
	Effect *effect = new Effect(this->fireEntity_2d, this->fireEntity_3d, pos.x, pos.y, fire_effect_duration_c);
	this->effects.push_back(effect);
}

struct AskTechnologyInfo {
	VI_Textfield *textfieldInfo;
	const vector<const Technology *> *player_technologies;
	AskTechnologyInfo() : textfieldInfo(NULL), player_technologies(NULL) {
	}
};

void MainGamestate::askTechnologyAction(VI_Panel *source) {
	/*VI_Textfield *textfieldInfo = static_cast<VI_Textfield *>(source->getUserData());
	ASSERT( textfieldInfo != NULL );*/
	AskTechnologyInfo *askTechnologyInfo = static_cast<AskTechnologyInfo *>(source->getUserData());
	ASSERT( askTechnologyInfo != NULL );
	VI_Textfield *textfieldInfo = askTechnologyInfo->textfieldInfo;
	ASSERT( textfieldInfo != NULL );
	const vector<const Technology *> *player_technologies = askTechnologyInfo->player_technologies;
	ASSERT( player_technologies != NULL );
	VI_Listbox *listbox = dynamic_cast<VI_Listbox *>(source);
	int active = listbox->getActive();
	ASSERT( active >= 0 && active < player_technologies->size() );
	const Technology *technology = player_technologies->at(active);
	textfieldInfo->clear();
	textfieldInfo->addText( technology->getName() );
	textfieldInfo->addText( ":\n\n" );

	bool any_units = false;
	for(int i=0;i<mainGamestate->getNUnitTemplates();i++) {
		const UnitTemplate *unit_template = mainGamestate->getUnitTemplate(i);
		if( unit_template->getRequiresTechnology() == technology ) {
			if( !any_units ) {
				textfieldInfo->addText("Units:\n");
			}
			textfieldInfo->addText(unit_template->getName());
			textfieldInfo->addText("\n");
			any_units = true;
		}
	}
	if( any_units ) {
		textfieldInfo->addText("\n");
	}

	bool any_improvements = false;
	for(int i=0;i<mainGamestate->getNImprovements();i++) {
		const Improvement *improvement = mainGamestate->getImprovement(i);
		if( improvement->getRequiresTechnology() == technology ) {
			if( !any_improvements ) {
				textfieldInfo->addText("Improvements:\n");
			}
			textfieldInfo->addText(improvement->getName());
			textfieldInfo->addText("\n");
			any_improvements = true;
		}
	}
	if( any_improvements ) {
		textfieldInfo->addText("\n");
	}

	bool any_technology = false;
	for(int i=0;i<mainGamestate->getNTechnologies();i++) {
		const Technology *tech = mainGamestate->getTechnology(i);
		if( tech->getRequires1() == technology || tech->getRequires2() == technology ) {
			if( !any_technology ) {
				textfieldInfo->addText("Technology:\n");
			}
			textfieldInfo->addText(tech->getName());
			textfieldInfo->addText("\n");
			any_technology = true;
		}
	}
	if( any_technology ) {
		textfieldInfo->addText("\n");
	}

	bool any_obsoletes = false;
	for(int i=0;i<mainGamestate->getNBuildables();i++) {
		const Buildable *buildable = mainGamestate->getBuildable(i);
		if( buildable->getObsoletedBy() == technology ) {
			if( !any_obsoletes ) {
				textfieldInfo->addText("Makes Obsolete:\n");
			}
			textfieldInfo->addText(buildable->getName());
			textfieldInfo->addText("\n");
			any_obsoletes = true;
		}
	}
	if( any_obsoletes ) {
		textfieldInfo->addText("\n");
	}

	textfieldInfo->addText( technology->getInfo() );
}

void MainGamestate::askTechnology() {
	int science = player->getTotalScience();
	vector<const Technology *> *player_technologies = new vector<const Technology *>();
	/*for(vector<Technology *>::const_iterator iter = technologies.begin(); iter != technologies.end(); ++iter) {
		const Technology *technology = *iter;*/
	for(int i=0;i<this->getNTechnologies();i++) {
		const Technology *technology = this->getTechnology(i);
		if( player->canResearch(technology) ) {
			player_technologies->push_back(technology);
		}
	}
	if( player_technologies->size() == 0 ) {
		delete player_technologies;
		return;
	}

	// convert to names
	vector<string> technology_names;
	for(vector<const Technology *>::const_iterator iter = player_technologies->begin();iter != player_technologies->end(); ++iter) {
		const Technology *technology = *iter;
		/*const char *name = technology->getName();
		technology_names.push_back(name);*/
		stringstream name;
		name << technology->getName();
		name << " (";
		int n_turns = (int)ceil( ((double)technology->getCost(this)) / (double)science );
		name << n_turns;
		name << " turns)";
		technology_names.push_back(name.str().c_str());
	}

	string okay = "Okay";
	//const int window_h_c = 512;
	int window_h_c = game_g->getGraphicsEnvironment()->getHeight() - infowindow_def_y_c - 32;
	InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), &okay, 1, infowindow_def_x_c, infowindow_def_y_c, infowindow_def_w_c, window_h_c);
	VI_Panel *windowPanel = window->getPanel();

	VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), infowindow_def_w_c - 32, 64);
	int age = (int)player->getAge();
	if( age >= (int)Technology::AGE_INDUSTRIAL )
		textfield->addText("What should our scientists be researching?");
	else
		textfield->addText("What should our wise men be researching?");
	textfield->setForeground(1.0f, 1.0f, 1.0f);
	windowPanel->add(textfield, 16, 16);

	//VI_Listbox *listbox = VI_createListbox(infowindow_def_w_c - 32, window_h_c - 128, &*technology_names.begin(), technology_names.size(), this->getFont());
	VI_Listbox *listbox = VI_createListbox(infowindow_def_w_c/2 - 16, window_h_c - 128, &*technology_names.begin(), technology_names.size(), game_g->getFont());
	listbox->setHasInputFocus(true);
	listbox->setForeground(1.0f, 1.0f, 1.0f);
	listbox->setActive(0);
	windowPanel->add(listbox, 16, 80);

	VI_Textfield *textfieldInfo = VI_createTextfield(game_g->getFont(), infowindow_def_w_c/2 - 32, window_h_c - 128);
	textfieldInfo->setForeground(1.0f, 1.0f, 1.0f);
	windowPanel->add(textfieldInfo, infowindow_def_w_c/2 + 16, 80);
	//textfieldInfo->addText("blah blah");

	AskTechnologyInfo askTechnologyInfo;
	askTechnologyInfo.textfieldInfo = textfieldInfo;
	askTechnologyInfo.player_technologies = player_technologies;
	listbox->setUserData(&askTechnologyInfo);
	listbox->setAction(askTechnologyAction);
	askTechnologyAction(listbox);

	window->doModal();

	int active = listbox->getActive();

	delete window;
	delete textfield;
	delete listbox;

	//int active = InfoWindow::select("What should our wise men be researching?", "Okay", "", 512, &technology_names);

	ASSERT( active >= 0 && active < player_technologies->size() );
	player->setTechnology( player_technologies->at( active ) );
	delete player_technologies;
}

void MainGamestate::calculateTerritory() {
	for(int y=0;y<this->map->getHeight();y++) {
		for(int x=0;x<this->map->getWidth();x++) {
			MapSquare *square = this->map->getSquare(x, y);
			square->setTerritory(NULL);
		}
	}
	for(vector<Civilization *>::iterator iter = this->civilizations.begin(); iter != this->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		for(int i=0;i<civilization->getNCities();i++) {
			const City *city = civilization->getCity(i);
			for(int j=0;j<city->getNCitySquares();j++) {
				Pos2D pos = city->getCitySquare(j);
				if( this->map->isValid(pos.x, pos.y) ) {
					MapSquare *square = this->map->getSquare(pos);
					if( square->getTerritory() != NULL ) {
						VI_log("### OVERLAPPING TERRITORY at %d, %d\n", pos.x, pos.y);
						ASSERT( square->getTerritory() == NULL ); // shouldn't have overlapping territory!
					}
					square->setTerritory(civilization);
				}
			}
		}
	}
}

void MainGamestate::renderQuad(Pos2D pos, bool flat, const unsigned char *rgba) {
	renderQuad(pos.x, pos.y, flat, rgba);
}

void MainGamestate::renderQuad(float pos_x, float pos_z, bool flat, const unsigned char *rgba) {
	Renderer *renderer = game_g->getGraphicsEnvironment()->getRenderer();
	const float texcoord_data[] = {
		0.0f, 0.0f,
		1.0f, 0.0f,
		1.0f, 1.0f,
		0.0f, 1.0f,
	};
	const unsigned char rgba_white[] = {255, 255, 255, 255};
	if( rgba == NULL )
		rgba = rgba_white;
	if( flat ) {
		const float vertex_data[] = {
			pos_x*xscale, 0.0f, (pos_z+1)*zscale,
			(pos_x+1)*xscale, 0.0f, (pos_z+1)*zscale,
			(pos_x+1)*xscale, 0.0f, pos_z*zscale,
			pos_x*xscale, 0.0f, pos_z*zscale
		};
		renderer->render(VertexArray::DRAWINGMODE_QUADS, vertex_data, texcoord_data, 4, rgba, true, 4);
	}
	else {
		/*Vector3D right, up;
		//for(int i=0;i<100;i++) {
		renderer->getViewOrientation(&right, &up); // TODO don't repeatedly recalculate with each function call
		*/
		//}
		//const float y_off = 0.8f;
		const float y_off = 1.0f;
		/*const float vertex_data[] = {
			pos.x*xscale, 0.0f, (pos.y+y_off)*zscale,
			(pos.x+1)*xscale, 0.0f, (pos.y+y_off)*zscale,
			(pos.x+1)*xscale, 1.0f, (pos.y+y_off)*zscale,
			pos.x*xscale, 1.0f, (pos.y+y_off)*zscale
		};*/
		Vector3D bottom(pos_x*xscale, 0.0f, (pos_z+y_off)*zscale);
		Vector3D top = bottom + mainGamestate->view_up;
		const float vertex_data[] = {
			bottom.x, bottom.y, bottom.z,
			bottom.x + xscale, bottom.y, bottom.z,
			top.x + xscale, top.y, top.z,
			top.x, top.y, top.z
		};
		renderer->render(VertexArray::DRAWINGMODE_QUADS, vertex_data, texcoord_data, 4, rgba, true, 4);
	}
}

void MainGamestate::renderFunc() {
	if( mainGamestate->map == NULL ) {
		return;
	}
	if( !mainGamestate->view_3d ) {
		return;
	}

	if( !mainGamestate->init_view_dirs ) {
		// must be done from within the 3D loop!
		game_g->getGraphicsEnvironment()->getRenderer()->getViewOrientation(&mainGamestate->view_right, &mainGamestate->view_up);
		mainGamestate->init_view_dirs = true;
	}

	const float eps = 1.0e-5;

	Renderer *renderer = game_g->getGraphicsEnvironment()->getRenderer();
	renderer->setDepthBufferMode(Renderer::RENDERSTATE_DEPTH_NONE);
	renderer->enableLighting(false);
	renderer->enableTexturing(false);
	renderer->setColor4(255, 255, 255, 255);

	Vector3D pos[3];
	mainGamestate->getMapPosf(&pos[0].x, &pos[0].z, 0, 0);
	mainGamestate->getMapPosf(&pos[1].x, &pos[1].z, game_g->getGraphicsEnvironment()->getWidth()-1, 0);
	mainGamestate->getMapPosf(&pos[2].x, &pos[2].z, 0, game_g->getGraphicsEnvironment()->getHeight()-1);
	int start_x = (int)floor(pos[0].x);
	int start_y = (int)floor(pos[0].z);
	int end_x = (int)ceil(pos[1].x);
	int end_y = (int)ceil(pos[2].z);
	start_x = max(0, start_x);
	start_y = max(0, start_y);
	if( mainGamestate->map->getTopology() == Map::TOPOLOGY_FLAT ) {
		end_x = min(mainGamestate->getMap()->getWidth()-1, end_x);
	}
	end_y = min(mainGamestate->getMap()->getHeight()-1, end_y);

	//VI_log("%d, %d to %d\n", start_x, start_y, end_x);

	// calculate territory (TODO: don't recalculate every frame!)
	mainGamestate->calculateTerritory();

	if( mainGamestate->grid ) {
		renderer->enableMultisample(false);
		// grid
		unsigned char rgba[] = {0, 0, 0, 255};
		//for(int x=1;x<mainGamestate->getMap()->getWidth();x++) {
		for(int x=start_x+1;x<=end_x;x++) {
			/*Vector3D pos0(x*xscale, 0.0f, 0*zscale);
			Vector3D pos1(x*xscale, 0.0f, mainGamestate->getMap()->getHeight()*zscale);*/
			Vector3D pos0(x*xscale, 0.0f, start_y*zscale);
			Vector3D pos1(x*xscale, 0.0f, (end_y+1)*zscale);
			renderer->draw(&pos0, &pos1, rgba);
		}
		//for(int y=1;y<mainGamestate->getMap()->getHeight();y++) {
		for(int y=start_y+1;y<=end_y;y++) {
			/*Vector3D pos0(0*xscale, 0.0f, y*zscale);
			Vector3D pos1(mainGamestate->getMap()->getWidth()*xscale, 0.0f, y*zscale);*/
			Vector3D pos0(start_x*xscale, 0.0f, y*zscale);
			Vector3D pos1((end_x+1)*xscale, 0.0f, y*zscale);
			renderer->draw(&pos0, &pos1, rgba);
		}
		renderer->enableMultisample(true);
	}

	// draw other map features
	//for(int cy=0;cy<mainGamestate->getMap()->getHeight();cy++) {
	//	for(int cx=0;cx<mainGamestate->getMap()->getWidth();cx++) {
	for(int cy=start_y;cy<=end_y;cy++) {
		for(int cx=start_x;cx<=end_x;cx++) {
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(cx, cy) ) {
				continue;
			}
			Pos2D pos(cx, cy);
			const MapSquare *square = mainGamestate->map->getSquare(cx, cy);
			//VI_Texture *map_texture = game->map->getSquare(cx, cy)->getTexture();
			//map_texture->draw(pos.x, pos.y + game->sq_height); // add sq_height due to texture coord location
			//map_texture->draw(game->sq_width * (x - frac_x), game->sq_height * ( y + 1 - frac_y ));
			//int overlay_index = ( cx % 31 + 7 * cy ) % 2;
			const int overlay_indices_lookup[] = { 0, 1, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0 };
			const int n_overlay_indices_lookup_c = sizeof(overlay_indices_lookup) / sizeof(overlay_indices_lookup[0]);
			int base_cx = cx, base_cy = cy;
			mainGamestate->map->reduceToBase(&base_cx, &base_cy);
			int overlay_index = overlay_indices_lookup[( base_cx + base_cy ) % n_overlay_indices_lookup_c];
			VI_Texture *mapoverlay_texture = square->getOverlayTexture(overlay_index);
			if( mapoverlay_texture != NULL && !square->drawAboveRoad() ) {
				renderer->enableTexturing(true);
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
				mapoverlay_texture->enable();
				renderQuad(pos, true);
				renderer->enableTexturing(false);
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
			}

			// draw roads
			if( square->getRoad() != MapSquare::ROAD_NONE ) {
				int index = 0;
				for(int c=0;c<8;c++) {
					int dx = 0, dy = 0;
					getDir(&dx, &dy, c);
					int tx = cx + dx;
					int ty = cy + dy;
					//if( tx < 0 || tx >= mainGamestate->map->getWidth() || ty < 0 || ty >= mainGamestate->map->getHeight() ) {
					if( !mainGamestate->map->isValid(tx, ty) ) {
						continue;
					}
					if( mainGamestate->map->getSquare(tx, ty)->getRoad() != MapSquare::ROAD_NONE ) {
						int mask = 1 << c;
						index += mask;
					}
				}
				ASSERT( index < 256 );
				VI_Texture *texture = NULL;
				if( square->getRoad() == MapSquare::ROAD_BASIC ) {
					texture = MapSquare::getRoadTexture(index);
				}
				else if( square->getRoad() == MapSquare::ROAD_RAILWAYS ) {
					texture = MapSquare::getRailwayTexture(index);
				}
				else {
					ASSERT(false);
				}
				renderer->enableTexturing(true);
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
				texture->enable();
				renderQuad(pos, true);
				renderer->enableTexturing(false);
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
			}

			if( mapoverlay_texture != NULL && square->drawAboveRoad() ) {
				renderer->enableTexturing(true);
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
				mapoverlay_texture->enable();
				renderQuad(pos, false);
				renderer->enableTexturing(false);
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
			}

			// targets / territory borders
			const Civilization *civilization = square->getTerritory();
			if( square->isTarget() ) {
				unsigned char rgba[] = {127, 0, 0, 127};
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
				renderQuad(pos, true, rgba);
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
				//g2d->drawRect(pos.x, pos.y, sq_width, sq_height);
			}
			else if( civilization != NULL ) {
				unsigned char rgba[] = {0, 0, 0, 255};
				const unsigned char *civ_rgb = civilization->getColor();
				const float px = 1.0f / 128.0f;
				for(int i=0;i<3;i++)
					rgba[i] = civ_rgb[i];
				if( mainGamestate->map->isValid(cx, cy-1) && mainGamestate->map->getSquare(cx, cy-1)->getTerritory() != civilization ) {
					// top
					Vector3D pos0(cx*xscale, 0.0f, cy*zscale);
					Vector3D pos1((cx+1)*xscale, 0.0f, cy*zscale);
					renderer->draw(&pos0, &pos1, rgba);
				}
				if( mainGamestate->map->isValid(cx, cy+1) && mainGamestate->map->getSquare(cx, cy+1)->getTerritory() != civilization ) {
					// bottom
					Vector3D pos0(cx*xscale, 0.0f, (cy+1-px)*zscale);
					Vector3D pos1((cx+1)*xscale, 0.0f, (cy+1-px)*zscale);
					renderer->draw(&pos0, &pos1, rgba);
				}
				if( mainGamestate->map->isValid(cx-1, cy) && mainGamestate->map->getSquare(cx-1, cy)->getTerritory() != civilization ) {
					// left
					Vector3D pos0(cx*xscale, 0.0f, cy*zscale);
					Vector3D pos1(cx*xscale, 0.0f, (cy+1)*zscale);
					renderer->draw(&pos0, &pos1, rgba);
				}
				if( mainGamestate->map->isValid(cx+1, cy) && mainGamestate->map->getSquare(cx+1, cy)->getTerritory() != civilization ) {
					// right
					Vector3D pos0((cx+1-px)*xscale, 0.0f, cy*zscale);
					Vector3D pos1((cx+1-px)*xscale, 0.0f, (cy+1)*zscale);
					renderer->draw(&pos0, &pos1, rgba);
				}
			}
		}
	}

	// Settler city square
	if( mainGamestate->active_unit != NULL && !mainGamestate->isGUILocked() && mainGamestate->active_unit->getTemplate()->canBuildCity() && mainGamestate->map->getSquare(mainGamestate->active_unit->getPos())->canBuildCity() ) {
		int time = clock() - mainGamestate->blink_time_start;
		if( ((int)( time / blink_speed_c )) % 2 != 0 ) {
			//game->genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_ADDITIVE);
			renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
			//renderer->setColor4(255, 255, 255, 31);
			unsigned char settler_square_rgba[] = {255, 255, 255, 31};
			for(int i=0;i<City::getNCitySquares();i++) {
				Pos2D pos = City::getCitySquare(mainGamestate->getMap(), i, mainGamestate->active_unit->getPos());
				mainGamestate->getMap()->reduceToBase(&pos.x, &pos.y, start_x, start_y);
				if( mainGamestate->map->isValid(pos.x, pos.y) ) {
					renderQuad(pos, true, settler_square_rgba);
				}
			}
			renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
		}
	}

	// initialise drawn flag
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		for(int i=0;i<civilization->getNUnits();i++) {
			Unit *unit = civilization->getUnit(i);
			unit->setMark(false);
		}
	}

	// draw cities
	renderer->enableTexturing(true);
	renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
	//VI_Texture *city_texture = mainGamestate->cityTexture;
	//VI_Texture *city_texture = mainGamestate->cityModernTexture;
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		VI_Texture *city_texture = NULL;
		if( civilization->getNCities() > 0 ) {
			int age = (int)civilization->getAge();
			city_texture = age >= (int)Technology::AGE_MODERN ? mainGamestate->cityModernTexture : mainGamestate->cityTexture;
			city_texture->enable();
		}
		for(int i=0;i<civilization->getNCities();i++) {
			const City *city = civilization->getCity(i);
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(city->getX(), city->getY()) ) {
				continue;
			}
			Pos2D pos = city->getPos();
			mainGamestate->getMap()->reduceToBase(&pos.x, &pos.y, start_x, start_y);
			if( pos.x < start_x || pos.x > end_x || pos.y < start_y || pos.y > end_y ) {
				continue;
			}
			renderQuad(pos, false);
			// text is drawn later, so that it appears on top

			const vector<Unit *> *units = mainGamestate->map->findUnitsAt(city->getX(), city->getY());
			for(vector<Unit *>::const_iterator iter = units->begin(); iter != units->end(); ++iter) {
				Unit *unit = *iter;
				if( unit != mainGamestate->active_unit ) {
					// don't draw non-active units that are located at a city
					unit->setMark(true);
				}
			}
		}
	}
	renderer->enableTexturing(false);
	renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);

	// draw units
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		const unsigned char *rgb = civilization->getColor();
		for(int i=0;i<civilization->getNUnits();i++) {
			Unit *unit = civilization->getUnit(i);
			if( unit->isMarked() ) {
				// either already drawn, or shouldn't draw
				continue;
			}
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(unit->getX(), unit->getY()) ) {
				continue;
			}
			Pos2D pos = unit->getPos();
			mainGamestate->getMap()->reduceToBase(&pos.x, &pos.y, start_x, start_y);
			if( pos.x < start_x || pos.x > end_x || pos.y < start_y || pos.y > end_y ) {
				continue;
			}
			if( unit->isMoving() ) {
				continue;
			}

			// find other units at this square
			vector<Unit *> units;
			const vector<Unit *> *units_at = mainGamestate->getMap()->findUnitsAt(pos);
			for(vector<Unit *>::const_iterator iter = units_at->begin(); iter != units_at->end(); ++iter) {
				Unit *unit2 = *iter;
				// n.b., don't include units already marked (units at a city other than the active_unit)
				if( !unit2->isMarked() ) {
					units.push_back(unit2);
					unit2->setMark(true); // so we don't deal with this unit later on
				}
			}
			// either find active unit, moving unit, or unit with best defence if active or moving unit isn't at this square
			Unit *draw_unit = NULL;
			for(vector<Unit *>::const_iterator iter_unit = units.begin();iter_unit != units.end();++iter_unit) {
				Unit *unit2 = *iter_unit;
				if( unit2->isMoving() ) {
					// don't draw moving units here
				}
				else if( unit2 == mainGamestate->active_unit ) {
					draw_unit = unit2;
					break;
				}
				else if( draw_unit == NULL || unit2->getTemplate()->getDefence() > draw_unit->getTemplate()->getDefence()
					|| ( unit2->getTemplate()->getDefence() == draw_unit->getTemplate()->getDefence() && unit2->getTemplate()->getAttack() > draw_unit->getTemplate()->getAttack() )
					) {
						draw_unit = unit2;
				}
			}

			unit = draw_unit;
			int time = clock() - mainGamestate->blink_time_start;
			if( unit == mainGamestate->active_unit && !mainGamestate->isGUILocked() && ((int)( time / blink_speed_c )) % 2 == 0 ) {
				continue;
			}
			// draw colour bar
			const float bar_frac = 0.25;
			unsigned char rgba[] = {rgb[0], rgb[1], rgb[2], 255};
			Vector3D bottom(pos.x*xscale, 0.0f, (pos.y+1.0f)*zscale);
			Vector3D top = bottom + mainGamestate->view_up*bar_frac;
			const float vertex_data[] = {
				bottom.x, bottom.y, bottom.z,
				bottom.x + xscale, bottom.y, bottom.z,
				top.x + xscale, top.y, top.z,
				top.x, top.y, top.z
			};
			renderer->render(VertexArray::DRAWINGMODE_QUADS, vertex_data, NULL, 4, rgba, true, 4);
			unsigned char rgba_black[] = {0, 0, 0, 255};
			renderer->render(VertexArray::DRAWINGMODE_LINE_LOOP, vertex_data, NULL, 4, rgba_black, true, 4);
			if( units.size() > 1 ) {
				// draw indicator for multiple units
				const float px = 1.0f/32.0f;
				float indicator_h = (units.size() - 1) * px;
				indicator_h = min(indicator_h, bar_frac - px);
				bottom = top + mainGamestate->view_up * px;
				top = bottom + mainGamestate->view_up * indicator_h;
				const float vertex_data[] = {
					bottom.x, bottom.y, bottom.z,
					bottom.x + xscale*px, bottom.y, bottom.z,
					top.x + xscale*px, top.y, top.z,
					top.x, top.y, top.z
				};
				unsigned char rgba_white[] = {255, 255, 255, 255};
				renderer->render(VertexArray::DRAWINGMODE_QUADS, vertex_data, NULL, 4, rgba_white, true, 4);
			}
			// draw unit texture
			renderer->enableTexturing(true);
			renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
			unit->getTemplate()->getTexture()->enable();
			renderQuad(pos, false);
			renderer->enableTexturing(false);
			renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
		}
	}

	// draw moving units
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		for(int i=0;i<civilization->getNUnits();i++) {
			Unit *unit = civilization->getUnit(i);
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(unit->getX(), unit->getY()) ) {
				continue;
			}
			if( unit->isMoving() ) {
				ASSERT( civilization == mainGamestate->player );
				float mx = 0.0f, my = 0.0f;
				unit->getMovingPosition(&mx, &my);
				mainGamestate->getMap()->reduceToBase(&mx, &my, (float)start_x, (float)start_y);
				renderer->enableTexturing(true);
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
				unit->getTemplate()->getTexture()->enable();
				renderQuad(mx, my, false);
				renderer->enableTexturing(false);
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
			}
		}
	}

	// draw smoke
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		if( civilization->getNCities() > 0 ) {
			// no smoke for modern cities
			int age = (int)civilization->getAge();
			if( age >= (int)Technology::AGE_MODERN )
				continue;
		}
		for(int i=0;i<civilization->getNCities();i++) {
			const City *city = civilization->getCity(i);
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(city->getX(), city->getY()) ) {
				continue;
			}
			Pos2D pos = city->getPos();
			mainGamestate->getMap()->reduceToBase(&pos.x, &pos.y, start_x, start_y);
			// N.B. Slightly modified test, so that smoke doesn't suddenly appear/disappear off the top or left hand side of the screen.
			if( pos.x < start_x-1 || pos.x > end_x || pos.y < start_y-1 || pos.y > end_y ) {
				continue;
			}

			//if( mainGamestate->scale_width >= 0.6f-eps && mainGamestate->scale_height >= 0.6f-eps )
			{
				renderer->enableTexturing(true);
				game_g->getGraphicsEnvironment()->render(mainGamestate->smokeEntity_3d, 0, (pos.x + 0.5f)*xscale, 0.5f, (pos.y + 0.5f)*zscale);
				renderer->enableTexturing(false); // need to disable texturing again!
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE); // need to disable blending again!
				renderer->enableLighting(false); // need to disable lighting again!
			}
		}
	}

	// draw effects
	int time = clock();
	//for(vector<Effect *>::iterator iter = mainGamestate->effects.begin(); iter != mainGamestate->effects.end(); ) {
	for(int i=0;i<mainGamestate->effects.size();) {
		Effect *effect = mainGamestate->effects.at(i);
		if( time >= effect->time_e ) {
			// delete effect
			delete effect;
			mainGamestate->effects.erase( mainGamestate->effects.begin() + i );
			continue;
		}
		else {
			if( mainGamestate->player == NULL || mainGamestate->player->isExplored(effect->xpos, effect->ypos) ) {
				Pos2D pos(effect->xpos, effect->ypos);
				renderer->enableTexturing(true);
				game_g->getGraphicsEnvironment()->render(effect->node_3d, 0, (pos.x + 0.5f)*xscale, 0.5f, (pos.y + 0.5f)*zscale);
				renderer->enableTexturing(false); // need to disable texturing again!
				renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE); // need to disable blending again!
				renderer->enableLighting(false); // need to disable lighting again!
			}
			i++;
		}
	}
}

void MainGamestate::paintPanel(VI_Graphics2D *g2d, void *data) {
	if( mainGamestate->map == NULL ) {
		return;
	}
	if( mainGamestate->view_3d ) {
		Vector3D pos[3];
		mainGamestate->getMapPosf(&pos[0].x, &pos[0].z, 0, 0);
		mainGamestate->getMapPosf(&pos[1].x, &pos[1].z, game_g->getGraphicsEnvironment()->getWidth()-1, 0);
		mainGamestate->getMapPosf(&pos[2].x, &pos[2].z, 0, game_g->getGraphicsEnvironment()->getHeight()-1);
		int start_x = (int)floor(pos[0].x);
		int start_y = (int)floor(pos[0].z);
		int end_x = (int)ceil(pos[1].x);
		int end_y = (int)ceil(pos[2].z);
		start_x = max(0, start_x);
		start_y = max(0, start_y);
		if( mainGamestate->map->getTopology() == Map::TOPOLOGY_FLAT ) {
			end_x = min(mainGamestate->getMap()->getWidth()-1, end_x);
		}
		end_y = min(mainGamestate->getMap()->getHeight()-1, end_y);

		// draw city text
		for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
			Civilization *civilization = *iter;
			for(int i=0;i<civilization->getNCities();i++) {
				const City *city = civilization->getCity(i);
				if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(city->getX(), city->getY()) ) {
					continue;
				}
				Pos2D pos = city->getPos();
				mainGamestate->getMap()->reduceToBase(&pos.x, &pos.y, start_x, start_y);
				// N.B. Slightly modified test, so that text doesn't suddenly appear/disappear off the top or left hand side of the screen.
				if( pos.x < start_x-1 || pos.x > end_x || pos.y < start_y-1 || pos.y > end_y ) {
					continue;
				}

				Vector3D pos3d(pos.x*xscale, 0.0f, (pos.y + 1.0f)*zscale);
				Pos2D screen_pos;
				game_g->getGraphicsEnvironment()->worldToScreen(&screen_pos.x, &screen_pos.y, &pos3d);
				if( screen_pos.x + 160 < 0 || screen_pos.x >= game_g->getGraphicsEnvironment()->getWidth() || screen_pos.y + 32 < 0 || screen_pos.y >= game_g->getGraphicsEnvironment()->getHeight() ) {
					continue;
				}
				game_g->getFont()->writeTextExt(screen_pos.x - 0, screen_pos.y, "%s (%d)", city->getName(), city->getSize());
				//if( mainGamestate->scale_width >= 0.4f-eps && mainGamestate->scale_height >= 0.4f-eps )
				{
					game_g->getFont()->writeTextExt(screen_pos.x - 0, screen_pos.y + 16, "(%s)", city->getCivilization()->getName());
					if( SPY || city->getCivilization() == mainGamestate->player ) {
						if( city->getBuildable() != NULL ) {
							game_g->getFont()->writeTextExt(screen_pos.x - 0, screen_pos.y + 32, "%s (%d / %d)", city->getBuildable()->getName(), city->getProgressBuildable(), city->getBuildable()->getCost());
						}
					}
				}
			}
		}

		mainGamestate->drawGUIText(g2d);

		return;
	}
	const float eps = 1.0e-5;

	int sq_width = mainGamestate->getSqWidth();
	int sq_height = mainGamestate->getSqHeight();

	const int n_x = ( game_g->getGraphicsEnvironment()->getWidth() / sq_width ) + 2;
	const int n_y = ( game_g->getGraphicsEnvironment()->getHeight() / sq_height ) + 2;
	/*int off_x = game->map_pos.x;
	int off_y = game->map_pos.y;*/
	const int off_x = (int)mainGamestate->map_pos_x;
	const int off_y = (int)mainGamestate->map_pos_y;
	// draw map
	{
		/*float frac_x = game->map_pos_x - off_x;
		float frac_y = game->map_pos_y - off_y;*/
		// draw base map
	//#if 0
		float vertex_data[12] = {
			0.0f, 0.0f, 0.0f,
			0.0f, 0.0f, 0.0f,
			0.0f, 0.0f, 0.0f,
			0.0f, 0.0f, 0.0f
		};
		/*float texcoord_data[] = {
		0.0f, 0.0f,
		0.0f, 1.0f,
		1.0f, 1.0f,
		1.0f, 0.0f
		};*/
		float texcoord_data[8];
		Renderer *renderer = game_g->getGraphicsEnvironment()->getRenderer();
		//VI_Texture *splat_alpha = NULL;
		VI_Texture *splat_alpha = mainGamestate->map->getAlphamap();
		int splat_alpha_w = -1;
		int splat_alpha_h = -1;
		bool texture_splatting = game_g->getSplatPixelShaderAmbient() != NULL && game_g->getSplatVertexShaderAmbient() != NULL;
		if( texture_splatting ) {
			// note: The vertex shader is trivial, but we need to use a vertex shader with the pixel shader, instead of using fixed function pipline.
			// This is needed for fix a problem on some graphics cards (e.g., NVIDIA 6100, 8600) on Windows Vista, 7, where the pixel shader fails to render under Direct3D9, giving the error:
			// "ps_3_0 shader can only be used with vs_3_0+ shader (SWVP or HWVP), or transformed vertices".
			game_g->getSplatVertexShaderAmbient()->enable();
			game_g->getSplatPixelShaderAmbient()->enable();

			game_g->getSplatVertexShaderAmbient()->setMatrices(); // needed for vertex transformation to work

			splat_alpha_w = splat_alpha->getWidth();
			splat_alpha_h = splat_alpha->getHeight();
			game_g->getSplatPixelShaderAmbient()->SetUniformParameter1f("splat_detail_scale_w", (float)mainGamestate->map->getWidth());
			game_g->getSplatPixelShaderAmbient()->SetUniformParameter1f("splat_detail_scale_h", (float)mainGamestate->map->getHeight());
		}
		for(int y=0;y<n_y;y++) {
			int cy = off_y + y;
			if( cy >= mainGamestate->map->getHeight() ) {
				break;
			}
			for(int x=0;x<n_x;x++) {
				int cx = off_x + x;
				if( mainGamestate->map->getTopology() == Map::TOPOLOGY_FLAT && cx >= mainGamestate->map->getWidth() ) {
					break;
				}
				if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(cx, cy) ) {
					// player may be NULL in test mode, in which case show everything
					continue;
				}
				Pos2D pos = mainGamestate->getScreenPos(cx, cy);
				//VI_Texture *map_texture = game->map->getSquare(cx, cy)->getTexture();
				//map_texture->draw(pos.x, pos.y + game->sq_height); // add sq_height due to texture coord location
				vertex_data[0] = (float)pos.x;
				vertex_data[1] = (float)pos.y;
				vertex_data[3] = (float)pos.x;
				vertex_data[4] = (float)(pos.y + sq_height);
				vertex_data[6] = (float)(pos.x + sq_width);
				vertex_data[7] = (float)(pos.y + sq_height);
				vertex_data[9] = (float)(pos.x + sq_width);
				vertex_data[10] = (float)pos.y;
				texcoord_data[0] = (float)cx;
				texcoord_data[1] = (float)cy;
				texcoord_data[2] = (float)cx;
				texcoord_data[3] = (float)cy+1.0f;
				texcoord_data[4] = (float)cx+1.0f;
				texcoord_data[5] = (float)cy+1.0f;
				texcoord_data[6] = (float)cx+1.0f;
				texcoord_data[7] = (float)cy;
				for(int i=0;i<8 && texture_splatting;i+=2) {
					texcoord_data[i] /= (float)mainGamestate->map->getWidth();
				}
				for(int i=1;i<8 && texture_splatting;i+=2) {
					texcoord_data[i] /= (float)mainGamestate->map->getHeight();
				}
				renderer->enableTexturing(true);

				if( texture_splatting ) {
					//VI_Texture *map_texture = MapSquare::getTexture(MapSquare::TYPE_OCEAN);
					VI_Texture *map_texture = MapSquare::getTexture(MapSquare::TYPE_GRASSLAND);
					map_texture->enable(); // base texture
					renderer->activeTextureUnit(1);
					splat_alpha->enable(); // alpha texture

					renderer->activeTextureUnit(2);
					//map_texture = MapSquare::getTexture(MapSquare::TYPE_GRASSLAND);
					map_texture = MapSquare::getTexture(MapSquare::TYPE_OCEAN);
					map_texture->enable();

					renderer->activeTextureUnit(3);
					map_texture = MapSquare::getTexture(MapSquare::TYPE_DESERT);
					map_texture->enable();

					renderer->activeTextureUnit(4);
					map_texture = MapSquare::getTexture(MapSquare::TYPE_ARTIC);
					map_texture->enable();

					renderer->activeTextureUnit(0);
				}
				else {
					VI_Texture *map_texture = mainGamestate->map->getSquare(cx, cy)->getTexture();
					map_texture->enable();
				}
				renderer->render(VertexArray::DRAWINGMODE_QUADS, vertex_data, texcoord_data, 4);
				renderer->enableTexturing(false);
			}
		}
		if( texture_splatting ) {
			game_g->getSplatPixelShaderAmbient()->disable();
			game_g->getSplatVertexShaderAmbient()->disable();
			//delete splat_alpha;
		}
	}
//#endif
	// calculate territory (TODO: don't recalculate every frame!)
	//for(int iter=0;iter<10;iter++)
	{
		mainGamestate->calculateTerritory();
	}

	if( mainGamestate->grid ) {
		// grid
		g2d->setColor3i(0, 0, 0);
		for(int x=1;x<n_x;x++) {
			int cx = off_x + x;
			if( mainGamestate->map->getTopology() == Map::TOPOLOGY_FLAT && cx >= mainGamestate->map->getWidth() ) {
				break;
			}
			Pos2D pos = mainGamestate->getScreenPos(cx, 0);
			g2d->draw(pos.x, 0, pos.x, game_g->getGraphicsEnvironment()->getHeight()-1);
		}
		for(int y=1;y<n_y;y++) {
			int cy = off_y + y;
			if( cy >= mainGamestate->map->getHeight() ) {
				break;
			}
			Pos2D pos = mainGamestate->getScreenPos(0, cy);
			g2d->draw(0, pos.y, game_g->getGraphicsEnvironment()->getWidth()-1, pos.y);
		}
	}

	// draw other map features
	for(int y=0;y<n_y;y++) {
		int cy = off_y + y;
		if( cy >= mainGamestate->map->getHeight() ) {
			break;
		}
		for(int x=0;x<n_x;x++) {
			int cx = off_x + x;
			if( mainGamestate->map->getTopology() == Map::TOPOLOGY_FLAT && cx >= mainGamestate->map->getWidth() ) {
				break;
			}
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(cx, cy) ) {
				continue;
			}
			Pos2D pos = mainGamestate->getScreenPos(cx, cy);
			const MapSquare *square = mainGamestate->map->getSquare(cx, cy);
			//VI_Texture *map_texture = game->map->getSquare(cx, cy)->getTexture();
			//map_texture->draw(pos.x, pos.y + game->sq_height); // add sq_height due to texture coord location
			//map_texture->draw(game->sq_width * (x - frac_x), game->sq_height * ( y + 1 - frac_y ));
			//int overlay_index = ( cx % 31 + 7 * cy ) % 2;
			const int overlay_indices_lookup[] = { 0, 1, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0 };
			const int n_overlay_indices_lookup_c = sizeof(overlay_indices_lookup) / sizeof(overlay_indices_lookup[0]);
			int base_cx = cx, base_cy = cy;
			mainGamestate->map->reduceToBase(&base_cx, &base_cy);
			int overlay_index = overlay_indices_lookup[( base_cx + base_cy ) % n_overlay_indices_lookup_c];
			VI_Texture *mapoverlay_texture = square->getOverlayTexture(overlay_index);
			if( mapoverlay_texture != NULL && !square->drawAboveRoad() ) {
				mapoverlay_texture->draw(pos.x, pos.y + sq_height, sq_width, sq_height);
			}

			// draw roads
			if( square->getRoad() != MapSquare::ROAD_NONE ) {
				/*bool any = false;
				for(int c=0;c<8;c++) {
					int dx = 0, dy = 0;
					getDir(&dx, &dy, c);
					int tx = cx + dx;
					int ty = cy + dy;
					if( tx < 0 || tx >= game->map->getWidth() || ty < 0 || ty >= game->map->getHeight() ) {
						continue;
					}
					if( game->map->getSquare(tx, ty)->getRoad() != MapSquare::ROAD_NONE ) {
						MapSquare::getRoadTexture(c)->draw(pos.x, pos.y + game->sq_height);
						any = true;
					}
				}
				if( !any ) {
					MapSquare::getRoadTexture(8)->draw(pos.x, pos.y + game->sq_height);
				}*/
				int index = 0;
				for(int c=0;c<8;c++) {
					int dx = 0, dy = 0;
					getDir(&dx, &dy, c);
					int tx = cx + dx;
					int ty = cy + dy;
					//if( tx < 0 || tx >= mainGamestate->map->getWidth() || ty < 0 || ty >= mainGamestate->map->getHeight() ) {
					if( !mainGamestate->map->isValid(tx, ty) ) {
						continue;
					}
					if( mainGamestate->map->getSquare(tx, ty)->getRoad() != MapSquare::ROAD_NONE ) {
						int mask = 1 << c;
						index += mask;
					}
				}
				ASSERT( index < 256 );
				if( square->getRoad() == MapSquare::ROAD_BASIC ) {
					MapSquare::getRoadTexture(index)->draw(pos.x, pos.y + sq_height, sq_width, sq_height);
				}
				else if( square->getRoad() == MapSquare::ROAD_RAILWAYS )
					MapSquare::getRailwayTexture(index)->draw(pos.x, pos.y + sq_height, sq_width, sq_height);
				else {
					ASSERT(false);
				}
			}

			if( mapoverlay_texture != NULL && square->drawAboveRoad() ) {
				mapoverlay_texture->draw(pos.x, pos.y + sq_height, sq_width, sq_height);
			}

			/*if( game->grid ) {
				// grid
				if( x > 0 || y > 0 )
					g2d->setColor3i(0, 0, 0);
				if( x > 0 )
					g2d->draw(pos.x, pos.y, pos.x, pos.y + game->sq_height-1);
				if( y > 0 )
					g2d->draw(pos.x, pos.y, pos.x + game->sq_width-1, pos.y);
			}*/

			// targets / territory borders
			const Civilization *civilization = square->getTerritory();
			if( square->isTarget() ) {
				g2d->setColor3i(127, 0, 0);
				g2d->drawRect(pos.x, pos.y, sq_width, sq_height);
			}
			else if( civilization != NULL ) {
				const unsigned char *rgb = civilization->getColor();
				g2d->setColor3i(rgb[0], rgb[1], rgb[2]);
				if( mainGamestate->map->isValid(cx, cy-1) && mainGamestate->map->getSquare(cx, cy-1)->getTerritory() != civilization ) {
					g2d->draw(pos.x, pos.y, pos.x + sq_width-1, pos.y);
				}
				if( mainGamestate->map->isValid(cx, cy+1) && mainGamestate->map->getSquare(cx, cy+1)->getTerritory() != civilization ) {
					g2d->draw(pos.x, pos.y + sq_height-1, pos.x + sq_width-1, pos.y + sq_height-1);
				}
				if( mainGamestate->map->isValid(cx-1, cy) && mainGamestate->map->getSquare(cx-1, cy)->getTerritory() != civilization ) {
					g2d->draw(pos.x, pos.y, pos.x, pos.y + sq_height-1);
				}
				if( mainGamestate->map->isValid(cx+1, cy) && mainGamestate->map->getSquare(cx+1, cy)->getTerritory() != civilization ) {
					g2d->draw(pos.x + sq_width-1, pos.y, pos.x + sq_width-1, pos.y + sq_height-1);
				}
			}
		}
	}

	// Settler city square
	if( mainGamestate->active_unit != NULL && !mainGamestate->isGUILocked() && mainGamestate->active_unit->getTemplate()->canBuildCity() && mainGamestate->map->getSquare(mainGamestate->active_unit->getPos())->canBuildCity() ) {
		int time = clock() - mainGamestate->blink_time_start;
		if( ((int)( time / blink_speed_c )) % 2 != 0 ) {
			//game->genv->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_ADDITIVE);
			game_g->getGraphicsEnvironment()->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
			g2d->setColor4i(255, 255, 255, 31);
			for(int i=0;i<City::getNCitySquares();i++) {
				Pos2D pos = City::getCitySquare(mainGamestate->getMap(), i, mainGamestate->active_unit->getPos());
				if( mainGamestate->map->isValid(pos.x, pos.y) ) {
					Pos2D screen_pos = mainGamestate->getScreenPos(pos.x, pos.y);
					g2d->fillRect(screen_pos.x, screen_pos.y, sq_width, sq_height);
				}
			}
			game_g->getGraphicsEnvironment()->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
		}
	}
	//return;

	// initialise drawn flag
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		for(int i=0;i<civilization->getNUnits();i++) {
			Unit *unit = civilization->getUnit(i);
			unit->setMark(false);
		}
	}

	// draw cities
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		VI_Texture *city_texture = NULL;
		if( civilization->getNCities() > 0 ) {
			int age = (int)civilization->getAge();
			city_texture = age >= (int)Technology::AGE_MODERN ? mainGamestate->cityModernTexture : mainGamestate->cityTexture;
		}
		for(int i=0;i<civilization->getNCities();i++) {
			const City *city = civilization->getCity(i);
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(city->getX(), city->getY()) ) {
				continue;
			}
			Pos2D pos = city->getPos();
			mainGamestate->getMap()->reduceToBase(&pos.x, &pos.y, off_x, off_y);
			if( pos.x < off_x || pos.x >= off_x + n_x || pos.y < off_y || pos.y >= off_y + n_y ) {
				continue;
			}
			Pos2D screen_pos = mainGamestate->getScreenPos(pos.x, pos.y);
			//VI_Texture *city_texture = mainGamestate->cityTexture;
			//VI_Texture *city_texture = mainGamestate->cityModernTexture;
			city_texture->draw(screen_pos.x, screen_pos.y + sq_height, sq_width, sq_height);
			// text is drawn later, so that it appears on top

			const vector<Unit *> *units = mainGamestate->map->findUnitsAt(city->getX(), city->getY());
			for(vector<Unit *>::const_iterator iter = units->begin(); iter != units->end(); ++iter) {
				Unit *unit = *iter;
				if( unit != mainGamestate->active_unit ) {
					// don't draw non-active units that are located at a city
					unit->setMark(true);
				}
			}
		}
	}

	// draw units
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		const unsigned char *rgb = civilization->getColor();
		for(int i=0;i<civilization->getNUnits();i++) {
			Unit *unit = civilization->getUnit(i);
			if( unit->isMarked() ) {
				// either already drawn, or shouldn't draw
				continue;
			}
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(unit->getX(), unit->getY()) ) {
				continue;
			}
			Pos2D pos = unit->getPos();
			mainGamestate->getMap()->reduceToBase(&pos.x, &pos.y, off_x, off_y);
			if( pos.x < off_x || pos.x >= off_x + n_x || pos.y < off_y || pos.y >= off_y + n_y ) {
				continue;
			}
			if( unit->isMoving() ) {
				continue;
			}

			// find other units at this square
			/*vector<Unit *> units;
			units.push_back(civilization->getUnit(i));
			for(int j=i+1;j<civilization->getNUnits();j++) {
				Unit *unit2 = civilization->getUnit(j);
				// n.b., don't include units already marked (units at a city other than the active_unit)
				if( !unit2->isMarked() && unit2->getPos() == unit->getPos() ) {
					units.push_back(unit2);
					unit2->setMark(true); // so we don't deal with this unit later on
				}
			}*/
			vector<Unit *> units;
			const vector<Unit *> *units_at = mainGamestate->getMap()->findUnitsAt(pos);
			for(vector<Unit *>::const_iterator iter = units_at->begin(); iter != units_at->end(); ++iter) {
				Unit *unit2 = *iter;
				// n.b., don't include units already marked (units at a city other than the active_unit)
				if( !unit2->isMarked() ) {
					units.push_back(unit2);
					unit2->setMark(true); // so we don't deal with this unit later on
				}
			}
			// either find active unit, moving unit, or unit with best defence if active or moving unit isn't at this square
			Unit *draw_unit = NULL;
			for(vector<Unit *>::const_iterator iter_unit = units.begin();iter_unit != units.end();++iter_unit) {
				Unit *unit2 = *iter_unit;
				if( unit2->isMoving() ) {
					// don't draw moving units here
				}
				else if( unit2 == mainGamestate->active_unit ) {
					draw_unit = unit2;
					break;
				}
				/*else if( unit2->isMoving() ) {
				draw_unit = unit2;
				break;
				}*/
				else if( draw_unit == NULL || unit2->getTemplate()->getDefence() > draw_unit->getTemplate()->getDefence()
					|| ( unit2->getTemplate()->getDefence() == draw_unit->getTemplate()->getDefence() && unit2->getTemplate()->getAttack() > draw_unit->getTemplate()->getAttack() )
					) {
						draw_unit = unit2;
				}
			}

			unit = draw_unit;
			Pos2D screen_pos = mainGamestate->getScreenPos(pos.x, pos.y);
			int time = clock() - mainGamestate->blink_time_start;
			if( unit == mainGamestate->active_unit && !mainGamestate->isGUILocked() && ((int)( time / blink_speed_c )) % 2 == 0 ) {
				continue;
			}
			//VI_log("%d units\n", units.size());
			// draw colour bar
			int rect_w = sq_width, rect_h = sq_height*0.25;
			rect_h = max(1, rect_h);
			int rect_x = screen_pos.x, rect_y = screen_pos.y + sq_height - 1 - rect_h;
			g2d->setColor3i(rgb[0], rgb[1], rgb[2]);
			g2d->fillRect(rect_x, rect_y, rect_w, rect_h);
			g2d->setColor3i(0, 0, 0);
			g2d->drawRect(rect_x, rect_y, rect_w, rect_h);
			if( units.size() > 1 ) {
				// draw indicator for multiple units
				g2d->setColor3i(255, 255, 255);
				int indicator_h = units.size() - 1;
				indicator_h = min(indicator_h, rect_y - screen_pos.y - 1); // avoid going too high
				g2d->drawRect(screen_pos.x, rect_y - indicator_h - 1, 1, indicator_h);
			}
			// draw unit texture
			unit->getTemplate()->getTexture()->draw(screen_pos.x, screen_pos.y + sq_height, sq_width, sq_height);
			/*g2d->setColor3i(255, 255, 255);
			renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
			g2d->setTexture(unit->getTemplate()->getTexture());
			g2d->fillRect(pos.x, pos.y, mainGamestate->sq_width, mainGamestate->sq_height);
			g2d->setTexture(NULL);
			renderer->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);*/
		}
	}

	// draw moving units
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		const unsigned char *rgb = civilization->getColor();
		for(int i=0;i<civilization->getNUnits();i++) {
			Unit *unit = civilization->getUnit(i);
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(unit->getX(), unit->getY()) ) {
				continue;
			}
			if( unit->isMoving() ) {
				ASSERT( civilization == mainGamestate->player );
				float mx = 0.0f, my = 0.0f;
				unit->getMovingPosition(&mx, &my);
				Pos2D pos = mainGamestate->getScreenPos(mx, my);
				/*int rect_w = game->sq_width, rect_h = 16;
				int rect_x = pos.x, rect_y = pos.y + game->sq_height - 1 - rect_h;
				g2d->setColor3i(rgb[0], rgb[1], rgb[2]);
				g2d->fillRect(rect_x, rect_y, rect_w, rect_h);
				g2d->setColor3i(0, 0, 0);
				g2d->drawRect(rect_x, rect_y, rect_w, rect_h);*/
				unit->getTemplate()->getTexture()->draw(pos.x, pos.y + sq_height, sq_width, sq_height);
			}
		}
	}

	// draw city text and smoke
	g2d->setColor3i(255, 255, 255);
	for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		//const unsigned char *rgb = civilization->getColor();
		//g2d->setColor3i(rgb[0], rgb[1], rgb[2]);
		int age = -1;
		if( civilization->getNCities() > 0 ) {
			age = (int)civilization->getAge();
		}
		for(int i=0;i<civilization->getNCities();i++) {
			const City *city = civilization->getCity(i);
			if( mainGamestate->player != NULL && !mainGamestate->player->isExplored(city->getX(), city->getY()) ) {
				continue;
			}
			Pos2D pos = city->getPos();
			mainGamestate->getMap()->reduceToBase(&pos.x, &pos.y, off_x, off_y);
			// N.B. Slightly modified test, so that text doesn't suddenly appear/disappear off the top or left hand side of the screen.
			if( pos.x < off_x-1 || pos.x >= off_x + n_x || pos.y < off_y-1 || pos.y >= off_y + n_y ) {
				continue;
			}
			Pos2D screen_pos = mainGamestate->getScreenPos(pos.x, pos.y);

			//g2d->setColor3i(rgb[0], rgb[1], rgb[2]);
			//g2d->fillRect(screen_pos.x - 16, screen_pos.y + mainGamestate->sq_height, 16, 16);

			//g2d->setColor3i(255, 255, 255);
			game_g->getFont()->writeTextExt(screen_pos.x - 0, screen_pos.y + sq_height, "%s (%d)", city->getName(), city->getSize());
			//game->getFont()->writeTextExt(screen_pos.x - 0, screen_pos.y + game->sq_height + 16, "Size: %d", city->getSize());
			if( mainGamestate->scale_width >= 0.4f-eps && mainGamestate->scale_height >= 0.4f-eps ) {
				game_g->getFont()->writeTextExt(screen_pos.x - 0, screen_pos.y + sq_height + 16, "(%s)", city->getCivilization()->getName());
				if( SPY || city->getCivilization() == mainGamestate->player ) {
					if( city->getBuildable() != NULL ) {
						game_g->getFont()->writeTextExt(screen_pos.x - 0, screen_pos.y + sq_height + 32, "%s (%d / %d)", city->getBuildable()->getName(), city->getProgressBuildable(), city->getBuildable()->getCost());
					}
				}
			}

			if( age < (int)Technology::AGE_MODERN && mainGamestate->scale_width >= 0.6f-eps && mainGamestate->scale_height >= 0.6f-eps ) {
				game_g->getGraphicsEnvironment()->getRenderer()->enableTexturing(true); // need to enable texturing again (OpenGL)
				game_g->getGraphicsEnvironment()->render(mainGamestate->smokeEntity_2d, 0, screen_pos.x + 0.5f*sq_width, screen_pos.y + 16.0f, 0.0f);
				//game->smokeTexture->draw(screen_pos.x + 0.5f*game->sq_width, screen_pos.y + 16.0f);
				game_g->getGraphicsEnvironment()->getRenderer()->enableTexturing(false); // need to disable texturing again!
				game_g->getGraphicsEnvironment()->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE); // need to disable blending again!
				game_g->getGraphicsEnvironment()->getRenderer()->enableLighting(false); // need to disable lighting again!
				game_g->getGraphicsEnvironment()->getRenderer()->enableCullFace(false); // need to disable cull face again!
			}
		}
	}

	// draw effects
	int time = clock();
	//for(vector<Effect *>::iterator iter = mainGamestate->effects.begin(); iter != mainGamestate->effects.end(); ) {
	for(int i=0;i<mainGamestate->effects.size();) {
		Effect *effect = mainGamestate->effects.at(i);
		if( time >= effect->time_e ) {
			// delete effect
			delete effect;
			mainGamestate->effects.erase( mainGamestate->effects.begin() + i );
			continue;
		}
		else {
			if( mainGamestate->player == NULL || mainGamestate->player->isExplored(effect->xpos, effect->ypos) ) {
				Pos2D pos = mainGamestate->getScreenPos(effect->xpos, effect->ypos);
				game_g->getGraphicsEnvironment()->getRenderer()->enableTexturing(true); // need to enable texturing again (OpenGL)
				game_g->getGraphicsEnvironment()->render(effect->node_2d, 0, pos.x + 0.5f*sq_width, pos.y + 16.0f, 0.0f);
				game_g->getGraphicsEnvironment()->getRenderer()->enableTexturing(false); // need to disable texturing again!
				game_g->getGraphicsEnvironment()->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE); // need to disable blending again!
				game_g->getGraphicsEnvironment()->getRenderer()->enableLighting(false); // need to disable lighting again!
			}
			i++;
		}
	}
//#endif
	mainGamestate->drawGUIText(g2d);
}

void MainGamestate::drawGUIText(VI_Graphics2D *g2d) const {
	const int frame_count = 50;
	static int panel_t = 0;
	static double fps = 0;
	if(frames % frame_count==0) {
		int new_t = clock();
		int time = new_t - panel_t;
		if(time != 0)
			fps = (frame_count*CLOCKS_PER_SEC)/((double)time);
		panel_t = new_t;
	}
	/*g2d->setColor3i(255, 255, 255);
	game_g->getFont()->writeTextExt(120, 16, "%.3f", fps);
	return; // TEST
	*/
	int xpos = 16, ypos = 16;
	//const int ydiff = 16;
	const int ydiff = game_g->getFont()->getHeight();
	const int ydiff_small = game_g->getSmallFont()->getHeight();
	//const int max_units_c = 10;
	const int max_units_c = 8;
	const int alpha_c = 95;
	if( this->player != NULL ) {
		//int total_y_size = 8 * ydiff + 32;
		int total_y_size = 7 * ydiff + 32;
		game_g->getGraphicsEnvironment()->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
		g2d->setColor4i(0, 0, 0, alpha_c);
		g2d->fillRect(xpos, ypos, 300, total_y_size);
		game_g->getGraphicsEnvironment()->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
		g2d->setColor4i(255, 255, 255, 255);
		game_g->getFont()->writeTextExt(xpos, ypos, "%s, Year %d", this->player->getName(), this->year); ypos += ydiff;
	}
	else {
		ypos += ydiff;
	}

	g2d->setColor3i(255, 255, 255);
	//game_g->getFont()->writeTextExt(120, 16, "%.3f", fps);
	game_g->getFont()->writeTextExt(8, game_g->getGraphicsEnvironment()->getHeight() - ydiff, "%.3f", fps);
	game_g->getFont()->writeTextExt(game_g->getGraphicsEnvironment()->getWidth() - 48, game_g->getGraphicsEnvironment()->getHeight() - ydiff, "v%d.%d", versionMajor, versionMinor);
	if( this->player != NULL ) {
		//game_g->getFont()->writeTextExt(xpos, ypos, "Year: %d", this->year); ypos += ydiff;
		/*int time_s = clock();
		const int n_iter = 10000;
		for(int i=0;i<n_iter;i++) {
			this->player->getTotalPopulation();
		}
		int time = clock() - time_s;
		VI_log("time taken %d for %d : %f\n", time, n_iter, ((float)time)/(float)n_iter);*/
		game_g->getFont()->writeTextExt(xpos, ypos, "Population: %d", this->player->getTotalPopulation()); ypos += ydiff; // n.b., testing suggests time taken for getTotalPopulation() is neglible, so fine to call every frame
		game_g->getFont()->writeTextExt(xpos, ypos, "Power: %d", this->player->getPower()); ypos += ydiff;
		const Technology *player_technology = this->player->getTechnology();
		if( player_technology != NULL ) {
			game_g->getFont()->writeTextExt(xpos, ypos, "Researching: %s (%d / %d)", player_technology->getName(), this->player->getProgressTechnology(), player_technology->getCost(this));
		}
		ypos += ydiff;
	}
	else {
		ypos += ydiff;
		ypos += ydiff;
		ypos += ydiff;
	}

	if( this->active_unit != NULL ) {
		ypos += 32;
		ASSERT( this->phase == PHASE_PLAYERTURN );
		const UnitTemplate *unit_template = this->getActiveUnit()->getTemplate();
		string unit_info = getUnitStatsString(unit_template, false);
		game_g->getFont()->writeTextExt(xpos, ypos, "Unit: %s %s", unit_template->getName(), unit_info.c_str()); ypos += ydiff;
		Rational moves_left = this->getActiveUnit()->movesLeft();
		int moves_left_integer, moves_left_num, moves_left_den;
		moves_left.value(&moves_left_integer, &moves_left_num, &moves_left_den);
		ASSERT( moves_left_num == 0 || moves_left_den == road_move_scale_c );
		if( moves_left_num == 0 ) {
			game_g->getFont()->writeTextExt(xpos, ypos, "Moves: %d out of %d", moves_left_integer, unit_template->getMoves()); ypos += ydiff;
		}
		else if( moves_left_integer == 0 ) {
			game_g->getFont()->writeTextExt(xpos, ypos, "Moves: %d/%d out of %d", moves_left_num, moves_left_den, unit_template->getMoves()); ypos += ydiff;
		}
		else {
			game_g->getFont()->writeTextExt(xpos, ypos, "Moves: %d %d/%d out of %d", moves_left_integer, moves_left_num, moves_left_den, unit_template->getMoves()); ypos += ydiff;
		}
		game_g->getFont()->writeTextExt(xpos, ypos, "%s", this->map->getSquare(this->getActiveUnit()->getX(), this->getActiveUnit()->getY())->getName()); ypos += ydiff;
		ypos += ydiff;

		const vector<Unit *> *units = this->map->findUnitsAt(this->getActiveUnit()->getX(), this->getActiveUnit()->getY());
		if( units->size() > 1 ) {
			int n_units = units->size() - 1;
			if( n_units > max_units_c ) {
				n_units = max_units_c + 1;
			}
			//int total_y_size = (n_units+1) * ydiff;
			int total_y_size = (n_units+1) * ydiff_small;
			game_g->getGraphicsEnvironment()->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
			g2d->setColor4i(0, 0, 0, alpha_c);
			//g2d->fillRect(xpos, ypos, 500, total_y_size);
			g2d->fillRect(xpos, ypos, 416, total_y_size);
			game_g->getGraphicsEnvironment()->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
			g2d->setColor4i(255, 255, 255, 255);

			//game_g->getFont()->writeText(xpos, ypos, "Other Units:"); ypos += ydiff;
			game_g->getSmallFont()->writeText(xpos, ypos, "Other Units:"); ypos += ydiff_small;
		}
		int count = 0;
		for(vector<Unit *>::const_iterator iter = units->begin(); iter != units->end(); ++iter ) {
			const Unit *unit = *iter;
			if( unit != this->getActiveUnit() ) {
				if( count == max_units_c ) {
					//game_g->getFont()->writeText(xpos, ypos, "more..."); ypos += ydiff;
					game_g->getSmallFont()->writeText(xpos, ypos, "more..."); ypos += ydiff_small;
					break;
				}
				string unit_name = this->getUnitName(unit);
				//game_g->getFont()->writeText(xpos, ypos, unit_name.c_str(), unit_name.length()); ypos += ydiff;
				game_g->getSmallFont()->writeText(xpos, ypos, unit_name.c_str(), unit_name.length()); ypos += ydiff_small;
				count++;
			}
		}
	}
	//else if( game->phase != PHASE_PLAYERTURN && !game->mainGUILocked ) {
	else if( this->phase != PHASE_PLAYERTURN && !this->isGUILocked() ) {
		int time = clock() - this->time_endplayerturn_s;
		if( ((int)( time / blink_speed_c )) % 2 == 0 ) {
			game_g->getFont()->writeText(xpos, ypos, "Thinking...");
		}
		ypos += ydiff;
	}
	//else if( !this->moved_units_this_turn && this->phase == PHASE_PLAYERTURN && !this->isGUILocked() ) {
	else if( !this->moved_units_this_turn && this->phase == PHASE_PLAYERTURN && !this->isGUILocked() && mainGamestate->endturnButton->isVisible() ) {
		// endturnButton check is so that this doesn't display when moving automated units
		// There's no active unit, we haven't moved units, we're in the player turn phase, and GUI isn't locked...
		ASSERT( this->active_unit == NULL );
		ypos += 32;
		game_g->getFont()->writeText(xpos, ypos, "Press RETURN or click End Turn"); ypos += ydiff;
	}

	/*if( game->phase != PHASE_PLAYERTURN ) {
	game->endturnButton->setVisible(false);
	}
	else {
	game->endturnButton->setVisible(true);
	}*/
	/*bool buttons_visible = game->phase == PHASE_PLAYERTURN;
	game->endturnButton->setVisible(buttons_visible);
	game->newgameButton->setVisible(buttons_visible);
	game->loadgameButton->setVisible(buttons_visible);
	game->savegameButton->setVisible(buttons_visible);
	game->citiesAdvisorButton->setVisible(buttons_visible);
	game->unitsAdvisorButton->setVisible(buttons_visible);
	game->civsAdvisorButton->setVisible(buttons_visible);
	game->techsAdvisorButton->setVisible(buttons_visible);*/
	/*int poly_count = 0, vertex_count = 0, distinct_vertex_count = 0, render_calls_count = 0;
	game->genv->getFrameStats(&poly_count, &vertex_count, &distinct_vertex_count, &render_calls_count);
	game->font->writeTextExt(140,32,"POLYS: %d",poly_count);
	game->font->writeTextExt(380,32,"CALLS: %d",render_calls_count);*/
}

void MainGamestate::action(VI_Panel *source) {
	mainGamestate->playSound(SOUND_BLIP);
	// always reset to normal input mode
	mainGamestate->setInputMode(INPUTMODE_NORMAL);
	/*if( game->mainGUILocked ) {
	return;
	}*/
	if( source == mainGamestate->endturnButton ) {
		//game->endturn();
		mainGamestate->endPlayerTurn();
	}
	else if( source == mainGamestate->newgameButton ) {
		mainGamestate->newGame();
	}
	else if( source == mainGamestate->loadgameButton ) {
		mainGamestate->loadGame();
	}
	else if( source == mainGamestate->savegameButton ) {
		mainGamestate->saveGame();
	}
	else if( source == mainGamestate->quitgameButton ) {
		mainGamestate->quitGame();
	}
	else if( source == mainGamestate->citiesAdvisorButton ) {
		mainGamestate->citiesAdvisor();
	}
	else if( source == mainGamestate->unitsAdvisorButton ) {
		mainGamestate->unitsAdvisor();
	}
	else if( source == mainGamestate->civsAdvisorButton ) {
		mainGamestate->civsAdvisor();
	}
	else if( source == mainGamestate->techsAdvisorButton ) {
		mainGamestate->techsAdvisor();
	}
	else if( source == mainGamestate->scoresAdvisorButton ) {
		mainGamestate->scoresAdvisor();
	}
	else if( source == mainGamestate->mapButton ) {
		mainGamestate->showMap();
	}
	else if( source == mainGamestate->gridButton ) {
		mainGamestate->grid = !mainGamestate->grid;
	}
	else if( source == mainGamestate->zoomOutButton ) {
		mainGamestate->zoom(false);
	}
	else if( source == mainGamestate->zoomResetButton ) {
		//Pos2D pos = mainGamestate->getMapPos(game_g->getGraphicsEnvironment()->getWidth()/2, game_g->getGraphicsEnvironment()->getHeight()/2);
		float mx = 0.0, my = 0.0;
		mainGamestate->getMapPosf(&mx, &my, game_g->getGraphicsEnvironment()->getWidth()/2, game_g->getGraphicsEnvironment()->getHeight()/2);
		if( mainGamestate->view_3d ) {
			/*VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
			Vector3D pos = vp->getPosition();
			pos.y = def_vp_height_c;
			vp->setPosition(&pos);*/
			float preferred_width = def_view3d_scale * game_g->getGraphicsEnvironment()->getWidth() / mainGamestate->def_sq_width;
			mainGamestate->reset3DZoom(preferred_width);
		}
		else {
			mainGamestate->scale_width = 1.0f;
			mainGamestate->scale_height = 1.0f;
		}
		//mainGamestate->centre(pos);
		mainGamestate->centre(mx, my);
	}
	else if( source == mainGamestate->zoomInButton ) {
		mainGamestate->zoom(true);
	}
	else if( source == mainGamestate->viewButton ) {
		mainGamestate->switchView();
	}
	else if( source == mainGamestate->waitButton ) {
		ASSERT( mainGamestate->active_unit != NULL );
		mainGamestate->active_unit = NULL;
		mainGamestate->updateUnitButtons();
	}
	else if( source == mainGamestate->skipTurnButton ) {
		ASSERT( mainGamestate->active_unit != NULL );
		mainGamestate->active_unit->useMoves();
		mainGamestate->active_unit = NULL;
		mainGamestate->updateUnitButtons();
	}
	else if( source == mainGamestate->fortifyButton ) {
		ASSERT( mainGamestate->active_unit != NULL );
		mainGamestate->active_unit->setStatus(Unit::STATUS_FORTIFIED);
		mainGamestate->active_unit = NULL;
		mainGamestate->updateUnitButtons();
	}
	else if( source == mainGamestate->buildCityButton ) {
		ASSERT( mainGamestate->active_unit != NULL );
		ASSERT( mainGamestate->active_unit->getTemplate()->canBuildCity() );
		//ASSERT( game->map->canBuildCity(game->active_unit->getX(), game->active_unit->getY()) );
		ASSERT( mainGamestate->map->getSquare(mainGamestate->active_unit->getX(), mainGamestate->active_unit->getY())->canBuildCity() );
		// build city
		string city_name = mainGamestate->active_unit->getCivilization()->makeCityName();

		// get city name
		string buttons[] = {"Build city", "Cancel"};
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), mainGamestate->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), buttons, 2, infowindow_def_x_c, infowindow_def_y_c, infowindow_def_w_c, 128);
		VI_Panel *windowPanel = window->getPanel();

		VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), infowindow_def_w_c - 32, 32);
		textfield->addText("Build new city:");
		windowPanel->add(textfield, 16, 16);

		VI_Stringgadget *stringgadget = VI_createStringgadget(game_g->getFont(), infowindow_def_w_c - 32);
		stringgadget->addText(city_name.c_str());
		stringgadget->setReadOnly(false);
		windowPanel->add(stringgadget, 16, 64);

		game_g->getGraphicsEnvironment()->flushInput(); // so we don't get an "B" keypress...
		int res = window->doModal();
		if( res == 0 ) {
			mainGamestate->active_unit->getCivilization()->advanceCityName();
			City *city = new City(mainGamestate, mainGamestate->active_unit->getCivilization(), stringgadget->getText(), mainGamestate->active_unit->getX(), mainGamestate->active_unit->getY());
			delete mainGamestate->active_unit;
			mainGamestate->active_unit = NULL;
			mainGamestate->updateUnitButtons();
			new CityWindow(mainGamestate, city);
		}

		delete window;
		delete textfield;
		delete stringgadget;
	}
	else if( source == mainGamestate->buildRoadButton ) {
		ASSERT( mainGamestate->active_unit != NULL );
		ASSERT( mainGamestate->active_unit->getTemplate()->canBuildRoads() );
		ASSERT( mainGamestate->map->getSquare(mainGamestate->active_unit->getX(), mainGamestate->active_unit->getY())->getRoad() == MapSquare::ROAD_NONE );
		ASSERT( mainGamestate->map->getSquare(mainGamestate->active_unit->getX(), mainGamestate->active_unit->getY())->canBeImproved( mainGamestate->active_unit->getCivilization() ) );
		// build road
		//square->setRoad(MapSquare::ROAD_BASIC);
		//active_unit->useMoves();
		//active_unit = NULL;
		mainGamestate->playSound(SOUND_WORKER);
		mainGamestate->active_unit->setStatus(Unit::STATUS_BUILDING_ROAD);
		mainGamestate->active_unit->update();
		mainGamestate->active_unit = NULL;
		mainGamestate->updateUnitButtons();
	}
	else if( source == mainGamestate->buildRailwaysButton ) {
		ASSERT( mainGamestate->active_unit != NULL );
		ASSERT( mainGamestate->active_unit->getTemplate()->canBuildRoads() );
		ASSERT( mainGamestate->map->getSquare(mainGamestate->active_unit->getX(), mainGamestate->active_unit->getY())->getRoad() == MapSquare::ROAD_BASIC );
		ASSERT( mainGamestate->map->getSquare(mainGamestate->active_unit->getX(), mainGamestate->active_unit->getY())->canBeImproved( mainGamestate->active_unit->getCivilization() ) );
		// build railways
		mainGamestate->playSound(SOUND_WORKER);
		mainGamestate->active_unit->setStatus(Unit::STATUS_BUILDING_RAILWAYS);
		mainGamestate->active_unit->update();
		mainGamestate->active_unit = NULL;
		mainGamestate->updateUnitButtons();
	}
	else if( source == mainGamestate->automateButton ) {
		ASSERT( mainGamestate->active_unit != NULL );
		ASSERT( mainGamestate->active_unit->getTemplate()->canBuildRoads() );
		mainGamestate->active_unit->setAutomated(true);
		mainGamestate->active_unit->automate();
		mainGamestate->cache_active_unit = mainGamestate->active_unit; // so we can cope with additional turns of this unit
		mainGamestate->active_unit = NULL;
		mainGamestate->updateUnitButtons();
	}
	else if( source == mainGamestate->gotoButton ) {
		ASSERT( mainGamestate->active_unit != NULL );
		ASSERT( !mainGamestate->active_unit->getTemplate()->isAir() );
		/*mainGamestate->active_unit->setAutomated(true);
		mainGamestate->active_unit->automate();
		mainGamestate->cache_active_unit = mainGamestate->active_unit; // so we can cope with additional turns of this unit
		mainGamestate->active_unit = NULL;
		mainGamestate->updateUnitButtons();*/
		mainGamestate->setInputMode(INPUTMODE_GOTO);
		// set up goto targets
		// get dists - n.b., don't include travel
		Pos2D pos = mainGamestate->active_unit->getPos();
		const Map::Distance *dists = mainGamestate->getMap()->calculateDistanceMap(mainGamestate->active_unit->getCivilization(), mainGamestate->active_unit, false, true, pos.x, pos.y);
		for(int y=0;y<mainGamestate->getMap()->getHeight();y++) {
			for(int x=0;x<mainGamestate->getMap()->getWidth();x++) {
				int this_dist = dists[y * mainGamestate->getMap()->getWidth() + x].cost;;
				if( ( pos.x != x || pos.y != y ) && this_dist != -1 && mainGamestate->player->isExplored(x, y) ) {
					mainGamestate->map->getSquare(x, y)->setTarget(true);
				}
			}
		}
		delete [] dists;
	}
	else if( source == mainGamestate->travelButton ) {
		mainGamestate->setInputMode(INPUTMODE_TRAVEL);
		ASSERT( mainGamestate->active_unit != NULL );
		City *city = mainGamestate->map->findCity(mainGamestate->active_unit->getPos());
		ASSERT( city != NULL );
		city->initTravel(false, mainGamestate->active_unit->getTemplate());
	}
	else if( source == mainGamestate->airRaidButton ) {
		mainGamestate->setInputMode(INPUTMODE_BOMBARD);
		// set up bombard targets
		ASSERT( mainGamestate->active_unit != NULL );
		Pos2D unit_pos = mainGamestate->active_unit->getPos();
		for(vector<Civilization *>::iterator iter = mainGamestate->civilizations.begin(); iter != mainGamestate->civilizations.end(); ++iter) {
			Civilization *civilization = *iter;
			if( civilization == mainGamestate->player )
				continue;
			for(int i=0;i<civilization->getNCities();i++) {
				City *city = civilization->getCity(i);
				Pos2D city_pos = city->getPos();
				//int dist = unit_pos.distanceChebyshev(&city_pos);
				int dist = mainGamestate->getMap()->distanceChebyshev(city_pos, unit_pos);
				if( dist <= mainGamestate->active_unit->getTemplate()->getAirRange() ) {
					mainGamestate->map->getSquare(city_pos)->setTarget(true);
				}
			}
		}
	}

	game_g->getGraphicsEnvironment()->flushInput(); // needed to avoid problem where the mouse click is registered after clicking - is there a better way to do this?
	/*if( source != game->travelButton ) {
		game->setInputMode(INPUTMODE_NORMAL);
	}*/
}

void MainGamestate::zoom(bool zoom_in) {
	//Pos2D pos = mainGamestate->getMapPos(game_g->getGraphicsEnvironment()->getWidth()/2, game_g->getGraphicsEnvironment()->getHeight()/2);
	float mx = 0.0f, my = 0.0f;
	mainGamestate->getMapPosf(&mx, &my, game_g->getGraphicsEnvironment()->getWidth()/2, game_g->getGraphicsEnvironment()->getHeight()/2);
	if( mainGamestate->view_3d ) {
		VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
		Vector3D pos = vp->getPosition();
		pos.y += zoom_in ? -0.4f : 0.4f;
		pos.y = max(pos.y, 4.0f);
		pos.y = min(pos.y, 25.0f);
		vp->setPosition(&pos);
	}
	else {
		if( zoom_in ) {
			mainGamestate->scale_width += 0.1f;
			mainGamestate->scale_height += 0.1f;
			mainGamestate->scale_width = min(mainGamestate->scale_width, 2.0f);
			mainGamestate->scale_height = min(mainGamestate->scale_height, 2.0f);
		}
		else {
			mainGamestate->scale_width -= 0.1f;
			mainGamestate->scale_height -= 0.1f;
			mainGamestate->scale_width = max(mainGamestate->scale_width, 0.3f);
			mainGamestate->scale_height = max(mainGamestate->scale_height, 0.3f);
		}
	}
	//mainGamestate->centre(pos);
	mainGamestate->centre(mx, my);
	/*VI_log("%f, %f\n", mx, my);
	mainGamestate->getMapPosf(&mx, &my, game_g->getGraphicsEnvironment()->getWidth()/2, game_g->getGraphicsEnvironment()->getHeight()/2);
	VI_log(" -> %f, %f\n", mx, my);*/
}

void MainGamestate::newGame() {
	/*string buttons[] = { "New Game", "Cancel" };
	InfoWindow *window = new InfoWindow(this->getPanel(), "Are you sure you wish to start a new game?", buttons, 2, infowindow_def_x_c, infowindow_def_y_c, infowindow_def_w_c, infowindow_def_h_c);
	int res = window->doModal();
	delete window;
	if( res == 0 ) {*/
	if( InfoWindow::confirm(game_g->getGraphicsEnvironment(), this->getPanel(), game_g->getFont(), game_g->getPanelTexture(), "Are you sure you wish to start a new game?", "New Game", "Cancel") ) {
		this->clearGameData();
		//if( !this->initGameData(true) ) {
		if( !game_g->setupNewGame(this) ) {
			quit = true;
		}
		else {
			//beginTurn();
			// beginTurn() now called by setupNewGame()
			if( this->gamestate_started ) {
				this->generateTerrain();
			}
		}
	}
}

void MainGamestate::quitGame() {
	VI_log("MainGamestate::quitGame()\n");
	if( !this->isGUILocked() ) {
		//VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), 32, 64);
		//VI_Button *button = VI_createButton("", game_g->getFont());
		//VI_createSceneGraphNode();
		//delete textfield;
		//VI_addtag();
		this->setInputMode(INPUTMODE_NORMAL);
		if( InfoWindow::confirm(game_g->getGraphicsEnvironment(), this->getPanel(), game_g->getFont(), game_g->getPanelTexture(), "Are you sure you wish to quit?", "Quit", "Cancel") ) {
			quit = true;
		}
	}
}

vector<MainGamestate::FilenameInfo> MainGamestate::getSaveGameFiles(bool is_loading) const {
/*#ifdef _WIN32
	HCURSOR cursor = LoadCursor( NULL, IDC_WAIT );
	SetCursor(cursor);
	SetClassLong(game_g->getGraphicsEnvironment()->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
	game_g->setCursor(cursor_wait);

	vector<FilenameInfo> savegamefiles;

	if( access(savegames_dirname, 0) != 0 ) {
		VI_log("save game folder doesn't seem to exist\n");
	}
	else {
#ifdef _WIN32
		WIN32_FIND_DATAA findFileData;
		char savegames_dirname_w[256] = "";
		sprintf(savegames_dirname_w, "%s/*.sav", savegames_dirname);
		HANDLE handle = FindFirstFileA(savegames_dirname_w, &findFileData);
		if( handle == INVALID_HANDLE_VALUE ) {
			VI_log("Invalid File Handle. GetLastError reports %d\n", GetLastError());
			// try to create the folder
		}
		else {
			while(true) {
				//VI_log("found file: %s\n", findFileData.cFileName);
				SYSTEMTIME stUTC;
				FileTimeToSystemTime(&findFileData.ftLastWriteTime, &stUTC);
				stringstream filedescription_str;
				filedescription_str << std::setfill('0') << "[" << stUTC.wYear << "/" << std::setw(2) << stUTC.wMonth << "/" << std::setw(2) << stUTC.wDay << " - " << std::setw(2) << stUTC.wHour << ":" << std::setw(2) << stUTC.wMinute << "] " << findFileData.cFileName;
				FilenameInfo filenameInfo;
				filenameInfo.filename = findFileData.cFileName;
				filenameInfo.filedescription = filedescription_str.str();
				savegamefiles.push_back(filenameInfo);
				if( FindNextFileA(handle, &findFileData) == 0 ) {
					FindClose(handle);
					DWORD error = GetLastError();
					if( error != ERROR_NO_MORE_FILES ) {
						VI_log("error reading directory: %d\n", error);
					}
					break;
				}
			}
		}
#elif __linux
		DIR *dir = opendir(savegames_dirname);
		if( dir == NULL ) {
			VI_log("failed to open directory: %s\n", savegames_dirname);
		}
		else {
			for(;;) {
				errno = 0;
				dirent *ent = readdir(dir);
				if( ent == NULL ) {
					closedir(dir);
					if( errno ) {
						VI_log("error reading directory: %d\n", errno);
					}
					break;
				}
				if( strstr(ent->d_name, ".sav") != NULL ) {
					//VI_log("found file: %s\n", ent->d_name);
					stringstream fullfilename;
					fullfilename << savegames_dirname << "/" << ent->d_name;
                    struct stat attrib;
                    if( stat(fullfilename.str().c_str(), &attrib) == -1 ) {
                        VI_log("stat failed for: %s\n", fullfilename.str().c_str());
                    }
                    struct tm *clock = localtime(&(attrib.st_mtime));
                    /*VI_log("%d : %d : %d\n", attrib.st_atime, attrib.st_mtime, attrib.st_ctime);
                    VI_log("%d : %d : %d\n", clock->tm_year + 1900, clock->tm_mon+1, clock->tm_mday+1);
                    VI_log("The date/time is: %s", asctime(clock));*/
                    stringstream filedescription_str;
                    filedescription_str << std::setfill('0') << "[" << clock->tm_year + 1900 << "/" << std::setw(2) << clock->tm_mon+1 << "/" << std::setw(2) << clock->tm_mday+1 << " - " << std::setw(2) << clock->tm_hour << ":" << std::setw(2) << clock->tm_min << "] " << ent->d_name;
					FilenameInfo filenameInfo;
					filenameInfo.filename = ent->d_name;
					filenameInfo.filedescription = filedescription_str.str();
					savegamefiles.push_back(filenameInfo);
				}
			}
		}
#endif
	}

/*#ifdef _WIN32
	cursor = LoadCursor( NULL, IDC_ARROW );
	SetCursor(cursor);
	SetClassLong(game_g->getGraphicsEnvironment()->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
	game_g->setCursor(cursor_arrow);
	return savegamefiles;
}

bool MainGamestate::loadGame() {
	bool ok = false;

	vector<FilenameInfo> savegamefiles = getSaveGameFiles(true);

	if( savegamefiles.size() > 0 ) {
		string buttons[] = {"Load game", "Cancel"};
		//string buttons[] = {"Load game", "Delete file", "Cancel"};
		//const int window_h_c = 512;
		int window_w_c = 504;
		int window_h_c = game_g->getGraphicsEnvironment()->getHeight() - infowindow_def_y_c - 32;
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), game_g->getGamestate()->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), buttons, 2, infowindow_def_x_c, infowindow_def_y_c, window_w_c, window_h_c);
		VI_Panel *windowPanel = window->getPanel();

		vector<string> strings;
		for(vector<FilenameInfo>::const_iterator iter = savegamefiles.begin(); iter != savegamefiles.end(); ++iter ) {
			const FilenameInfo filenameInfo = *iter;
			strings.push_back( filenameInfo.filedescription );
		}
		//VI_Listbox *listbox = VI_createListbox(window_w_c - 32, window_h_c - 64, &*savegamefiles.begin(), savegamefiles.size(), game_g->getFont());
		VI_Listbox *listbox = VI_createListbox(window_w_c - 32, window_h_c - 64, &*strings.begin(), strings.size(), game_g->getFont());
		//VI_Listbox *listbox = VI_createListbox(infowindow_def_w_c/2 - 16, window_h_c - 128, &*technology_names.begin(), technology_names.size(), game_g->getFont());
		listbox->setHasInputFocus(true);
		listbox->setForeground(1.0f, 1.0f, 1.0f);
		listbox->setActive(0);
		windowPanel->add(listbox, 16, 16);

		int res = window->doModal();
		if( res == 0 ) {
			int active = listbox->getActive();
			ASSERT( active >= 0 && active < savegamefiles.size() );
			string filename = savegames_dirname;
			//filename += "/" + savegamefiles.at(active);
			filename += "/" + savegamefiles.at(active).filename;
			VI_log("%d: loading: %s\n", active, filename.c_str());
			delete window;
			ok = this->load(filename.c_str());
		}
		else {
			delete window;
		}

		//delete textfield;
		delete listbox;
	}
	else {
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), game_g->getGamestate()->getPanel(), "No save game files available", game_g->getFont(), game_g->getPanelTexture());
		window->doModal();
		delete window;
	}

	/*DIR *dir = opendir(savegames_dirname);
	if( dir == NULL ) {
	VI_log("failed to open savegames folder\n");
	}*/


	/*#ifdef _WIN32
	char filename[1024] = "";
	OPENFILENAMEA openfilename;
	memset(&openfilename, 0, sizeof(OPENFILENAME));
	openfilename.lStructSize = sizeof(OPENFILENAME);
	openfilename.hwndOwner = game_g->getGraphicsEnvironment()->getHWND();
	openfilename.lpstrFilter = filter;
	openfilename.nFilterIndex = 1;
	openfilename.lpstrDefExt = ext;
	openfilename.lpstrFile = filename;
	openfilename.nMaxFile = sizeof(filename)/sizeof(*filename);
	openfilename.lpstrInitialDir = ".";
	openfilename.Flags = OFN_SHOWHELP | OFN_OVERWRITEPROMPT;
	//openfilename.lpstrTitle = szTitle;
	//ShowCursor(TRUE);
	if( game_g->getGraphicsEnvironment()->isTrueFullscreen() ) {
	ShowWindow(game_g->getGraphicsEnvironment()->getHWND(), SW_MINIMIZE); // needed for D3D9 fullscreen mode
	}

	char current_directory[65536];
	getcwd(current_directory, sizeof(current_directory));

	if( GetOpenFileNameA(&openfilename) ) {
	chdir(current_directory);
	game_g->getGraphicsEnvironment()->flushInput(); // needed because we miss key releases when dialog is open, meaning we may think a key is still being pressed!
	ok = this->load(filename);
	}
	else {
	chdir(current_directory);
	game_g->getGraphicsEnvironment()->flushInput(); // needed because we miss key releases when dialog is open, meaning we may think a key is still being pressed!
	VI_log("file dialog failed/cancelled\n");
	ok = false;
	}

	if( game_g->getGraphicsEnvironment()->isTrueFullscreen() ) {
	ShowWindow(game_g->getGraphicsEnvironment()->getHWND(), SW_RESTORE);
	}
	//ShowCursor(FALSE);
	//this->player->revealMap(); // uncomment to reveal map
	#else
	ok = false;
	#endif*/
	if( CHEAT ) {
		//this->player->revealMap(); // uncomment to reveal map
	}
	return ok;
}

void MainGamestate::saveGame() {
	VI_log("save game requested...\n");

	char filename[1024] = "";
	sprintf(filename, "%s Year %d", player->getName(), year);

	string buttons[] = {"Save game", "Cancel"};
	InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), buttons, 2, infowindow_def_x_c, infowindow_def_y_c, infowindow_def_w_c, 128);
	VI_Panel *windowPanel = window->getPanel();

	VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), infowindow_def_w_c - 32, 32);
	textfield->addText("Save game as:");
	windowPanel->add(textfield, 16, 16);

	VI_Stringgadget *stringgadget = VI_createStringgadget(game_g->getFont(), infowindow_def_w_c - 32);
	stringgadget->addText(filename);
	stringgadget->setReadOnly(false);
	windowPanel->add(stringgadget, 16, 64);

	game_g->getGraphicsEnvironment()->flushInput(); // so we don't get an "S" keypress...
	int res = window->doModal();
	if( res == 0 ) {
		string filename = savegames_dirname;
		filename += "/";
		filename += stringgadget->getText();
		filename += ext;
		this->save(filename.c_str(), false);
	}

	delete window;
	delete textfield;
	delete stringgadget;

/*#ifdef _WIN32
	//const int ext_len = strlen(ext);
	char filename[1024] = "";
	sprintf(filename, "%s Year %d.sav", player->getName(), year);
	OPENFILENAMEA openfilename;
	memset(&openfilename, 0, sizeof(OPENFILENAME));
	openfilename.lStructSize = sizeof(OPENFILENAME);
	openfilename.hwndOwner = game_g->getGraphicsEnvironment()->getHWND();
	openfilename.lpstrFilter = filter;
	openfilename.nFilterIndex = 1;
	openfilename.lpstrDefExt = ext;
	openfilename.lpstrFile = filename;
	//openfilename.nMaxFile = sizeof(filename)/sizeof(*filename) - ext_len; // -ext_len to allow adding extension
	openfilename.nMaxFile = sizeof(filename)/sizeof(*filename);
	openfilename.lpstrInitialDir = ".";
	openfilename.Flags = OFN_SHOWHELP | OFN_OVERWRITEPROMPT;
	//openfilename.lpstrTitle = szTitle;
	//ShowCursor(TRUE);
	if( game_g->getGraphicsEnvironment()->isTrueFullscreen() ) {
		ShowWindow(game_g->getGraphicsEnvironment()->getHWND(), SW_MINIMIZE); // needed for D3D9 fullscreen mode
	}

	char current_directory[65536];
	getcwd(current_directory, sizeof(current_directory));

	if( GetSaveFileNameA(&openfilename) ) {
		chdir(current_directory);
		if( game_g->getGraphicsEnvironment()->isTrueFullscreen() ) {
			ShowWindow(game_g->getGraphicsEnvironment()->getHWND(), SW_RESTORE);
		}
		game_g->getGraphicsEnvironment()->flushInput(); // needed because we miss key releases when dialog is open, meaning we may think a key is still being pressed!
		this->save(filename);
		stringstream text;
		text << "Successfully saved game under filename:\n\n" << filename;
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture());
		window->doModal();
		delete window;
	}
	else {
		chdir(current_directory);
		if( game_g->getGraphicsEnvironment()->isTrueFullscreen() ) {
			ShowWindow(game_g->getGraphicsEnvironment()->getHWND(), SW_RESTORE);
		}
		game_g->getGraphicsEnvironment()->flushInput(); // needed because we miss key releases when dialog is open, meaning we may think a key is still being pressed!
		VI_log("file dialog failed/cancelled\n");
	}

	//ShowCursor(FALSE);
#endif*/
}

const char file_header_string_c[] = "CONQUESTSSAVEGAME";

void MainGamestate::save(const char *filename, bool autosave) {
	VI_log("MainGamestate::save(%s, %d)\n", filename, autosave);

	if( access(savegames_dirname, 0) != 0 ) {
		VI_log("save game folder doesn't seem to exist - try creating it\n");
#if _WIN32
		int res = mkdir(savegames_dirname);
#elif __linux
        /*mode_t mask = umask(0);
        VI_log("mask %d\n", mask);
        VI_log("mask %d\n", umask(0));*/
		int res = mkdir(savegames_dirname, S_IRWXU | S_IRWXG | S_IRWXO);
		//umask(mask);
#endif
		if( res != 0 ) {
			VI_log("mkdir failed: %d\n", errno);
			if( !autosave ) {
				InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), game_g->getGamestate()->getPanel(), "Unable to create save game folder", game_g->getFont(), game_g->getPanelTexture());
				window->doModal();
				delete window;
			}
			return;
		}
		else if( access(savegames_dirname, 0) != 0 ) {
			VI_log("still can't access it!\n");
			if( !autosave ) {
				InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), game_g->getGamestate()->getPanel(), "Unable to access save game folder", game_g->getFont(), game_g->getPanelTexture());
				window->doModal();
				delete window;
			}
			return;
		}
	}

	if( !autosave && access(filename, 0) == 0 ) {
		// file already exists!
		stringstream str;
		str << "The file " << filename << " already exists. Do you wish to overwrite it?";
		if( !InfoWindow::confirm(game_g->getGraphicsEnvironment(), game_g->getGamestate()->getPanel(), game_g->getFont(), game_g->getPanelTexture(), str.str().c_str(), "Yes, overwrite", "No, cancel") ) {
			return;
		}
	}

	//FILE *file = fopen(filename, "wb");
	FILE *file = fopen(filename, "w");
	if( file == NULL ) {
		VI_log("failed to open file for writing\n");
/*#ifdef _WIN32
		MessageBoxA(NULL, "Failed to open file for writing", "Error", MB_OK|MB_ICONEXCLAMATION);
#endif*/
		if( !autosave ) {
			InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), game_g->getGamestate()->getPanel(), "Failed to open file for writing", game_g->getFont(), game_g->getPanelTexture());
			window->doModal();
			delete window;
		}
		return;
	}
	fprintf(file, "%s\n", file_header_string_c);
	fprintf(file, "major_version = %d\n", versionMajor);
	fprintf(file, "minor_version = %d\n", versionMinor);
	fprintf(file, "year=%d\n", this->getYear());
	fprintf(file, "seed=%d\n", this->seed);
	fprintf(file, "difficulty=%d\n", (int)this->difficulty);
	fprintf(file, "aiaggression=%d\n", (int)this->aiaggression);
	fprintf(file, "help_buildcity=%d\n", help_buildcity?1:0);
	fprintf(file, "help_buildroad=%d\n", help_buildroad?1:0);
	fprintf(file, "help_bomb=%d\n", help_bomb?1:0);
	fprintf(file, "player=%s\n", player->getName());
	fprintf(file, "map_width=%d\n", map->getWidth());
	fprintf(file, "map_height=%d\n", map->getHeight());
	fprintf(file, "map_topology=%d\n", map->getTopology());
	fprintf(file, "map=\n");
	for(int y=0;y<map->getHeight();y++) {
		for(int x=0;x<map->getWidth();x++) {
			const MapSquare *square = map->getSquare(x, y);
			fprintf(file, "%d", square->getType());
			if( square->getRoad() == MapSquare::ROAD_BASIC ) {
				fprintf(file, "r");
			}
			else if( square->getRoad() == MapSquare::ROAD_RAILWAYS ) {
				fprintf(file, "R");
			}
			int progress_road = square->getProgressRoad();
			if( progress_road > 0 ) {
				ASSERT( progress_road < 10 ); // can only cope with saving single digit!
				fprintf(file, "p%d", progress_road);
			}
		}
		fprintf(file, "\n");
	}
	for(vector<Civilization *>::const_iterator iter = civilizations.begin(); iter != civilizations.end(); ++iter) {
		const Civilization *civilization = *iter;
		civilization->save(file);
	}
	for(vector<Relationship *>::const_iterator iter = relationships.begin(); iter != relationships.end(); ++iter) {
		const Relationship *relationship = *iter;
		relationship->save(file);
	}
	fclose(file);
}

bool MainGamestate::load(const char *filename) {
	VI_log("MainGamestate::load(%s)\n", filename);
	//FILE *file = fopen(filename, "rb");
	FILE *file = fopen(filename, "r");
	if( file == NULL ) {
		VI_log("failed to open file for reading\n");
/*#ifdef _WIN32
		MessageBoxA(NULL, "Failed to open file", "Error", MB_OK|MB_ICONEXCLAMATION);
#endif*/
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), game_g->getGamestate()->getPanel(), "Failed to open file", game_g->getFont(), game_g->getPanelTexture());
		window->doModal();
		delete window;
		return false;
	}
/*#ifdef _WIN32
	HCURSOR cursor = LoadCursor( NULL, IDC_WAIT );
	SetCursor(cursor);
	SetClassLong(game_g->getGraphicsEnvironment()->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
	game_g->setCursor(cursor_wait);

	this->clearGameData();

	cache_active_unit = NULL;
	active_unit = NULL;
	//updateUnitButtons(); // disabled for when called from options screen? But is this needed during standard game play?
	moved_units_this_turn = false;

	//map = new Map(100, 100);
	/*map = new Map();
	const char mapfile[] = "conquests/data/world.map";
	if( !map->readFile(mapfile) ) {
	VI_log("Failed to read map file: %s\n", mapfile);
	quit = true;
	return;
	}*/

	const int max_line_c = 4095;
	char line[max_line_c+1] = "";
	char word[max_line_c+1] = "";
	bool ok = true;
	const char field_version_major_c[] = "major_version=";
	const char field_version_minor_c[] = "minor_version=";
	const char field_year_c[] = "year=";
	const char field_seed_c[] = "seed=";
	const char field_difficulty_c[] = "difficulty=";
	const char field_aiaggression_c[] = "aiaggression=";
	const char field_help_buildcity_c[] = "help_buildcity=";
	const char field_help_buildroad_c[] = "help_buildroad=";
	const char field_help_bomb_c[] = "help_bomb=";
	const char field_player_c[] = "player=";
	const char field_map_width_c[] = "map_width=";
	const char field_map_height_c[] = "map_height=";
	const char field_map_topology_c[] = "map_topology=";
	const char field_map_c[] = "map=";
	int map_width = -1, map_height = -1;
	Map::Topology map_topology = Map::TOPOLOGY_FLAT; // default to flat for backwards compatibility
	string player_civ;
	bool first_line = true;
	bool invalid_header = false;
	bool invalid_version = false;

	while( ok && fgets(line, max_line_c, file) != NULL ) {
		if( first_line ) {
			if( matchFileLine(line, file_header_string_c) ) {
				// ok
				first_line = false;
				continue;
			}
			else {
				VI_log("header id missing, failed to recognise file\n");
				invalid_header = true;
				ok = false;
				break;
			}
		}
		if( matchFileLine(line, field_version_major_c) ) {
			parseFileLine(word, line, field_version_major_c);
			int this_versionMajor = atoi(word);
			VI_log("filename has major version %d\n", this_versionMajor);
			if( this_versionMajor != versionMajor ) {
				invalid_version = true;
				ok = false;
			}
		}
		else if( matchFileLine(line, field_version_minor_c) ) {
			parseFileLine(word, line, field_version_minor_c);
			int this_versionMinor = atoi(word);
			VI_log("filename has minor version %d\n", this_versionMinor);
			if( this_versionMinor != versionMinor ) {
				invalid_version = true;
				ok = false;
			}
		}
		else if( matchFileLine(line, field_year_c) ) {
			parseFileLine(word, line, field_year_c);
			year = atoi(word);
			VI_log("year = %d\n", year);
		}
		else if( matchFileLine(line, field_seed_c) ) {
			parseFileLine(word, line, field_seed_c);
			seed = atoi(word);
			VI_log("seed = %d\n", seed);
		}
		else if( matchFileLine(line, field_difficulty_c) ) {
			parseFileLine(word, line, field_difficulty_c);
			int difficulty_int = atoi(word);
			VI_log("difficulty = %d\n", difficulty_int);
			if( difficulty_int < 0 || difficulty_int >= N_DIFFICULTY ) {
				VI_log("unrecognised difficulty!\n");
				ok = false;
			}
			else {
				difficulty = (Difficulty)difficulty_int;
			}
		}
		else if( matchFileLine(line, field_aiaggression_c) ) {
			parseFileLine(word, line, field_aiaggression_c);
			int aiaggression_int = atoi(word);
			VI_log("aiaggression = %d\n", aiaggression_int);
			if( aiaggression_int < 0 || aiaggression_int >= N_AIAGGRESSION ) {
				VI_log("unrecognised aiaggression!\n");
				ok = false;
			}
			else {
				aiaggression = (AIAggression)aiaggression_int;
			}
		}
		else if( matchFileLine(line, field_help_buildcity_c) ) {
			parseFileLine(word, line, field_help_buildcity_c);
			help_buildcity = parseBool(&ok, word);
		}
		else if( matchFileLine(line, field_help_buildroad_c) ) {
			parseFileLine(word, line, field_help_buildroad_c);
			help_buildroad = parseBool(&ok, word);
		}
		else if( matchFileLine(line, field_help_bomb_c) ) {
			parseFileLine(word, line, field_help_bomb_c);
			help_bomb = parseBool(&ok, word);
		}
		else if( matchFileLine(line, field_player_c) ) {
			parseFileLine(word, line, field_player_c);
			player_civ = word;
		}
		else if( matchFileLine(line, field_map_width_c) ) {
			parseFileLine(word, line, field_map_width_c);
			map_width = atoi(word);
		}
		else if( matchFileLine(line, field_map_height_c) ) {
			parseFileLine(word, line, field_map_height_c);
			map_height = atoi(word);
		}
		else if( matchFileLine(line, field_map_topology_c) ) {
			parseFileLine(word, line, field_map_topology_c);
			map_topology = (Map::Topology)atoi(word);
		}
		else if( matchFileLine(line, field_map_c) ) {
			if( map_width == -1 || map_height == -1 ) {
				VI_log("map width/height not set\n");
				ok = false;
				break;
			}
			map = new Map(this, map_width, map_height);
			map->setTopology(map_topology);
			for(int y=0;y<map->getHeight() && ok;y++ ) {
				if( fgets(line, max_line_c, file) == NULL ) {
					VI_log("MainGamestate::load unexpected eof when reading map\n");
					ok = false;
					break;
				}
				for(int x=-1,c=0;ok;c++) {
					if( line[c] == '\r' || line[c] == '\n' ) {
						if( x != map->getWidth()-1 ) {
							VI_log("MainGamestate::load unexpected newline when reading map\n");
							ok = false;
						}
						break;
					}
					else if( line[c] >= '0' && line[c] <= '9' ) {
						x++;
						if( x >= map->getWidth() ) {
							VI_log("MainGamestate::load unexpected line too long when reading map\n");
							ok = false;
						}
						else {
							MapSquare *square = map->getSquare(x, y);
							square->setType((MapSquare::Type)(line[c] - '0'));
						}
					}
					else if( line[c] == 'r' || line[c] == 'R' ) {
						MapSquare *square = map->getSquare(x, y);
						if( line[c] == 'r' )
							square->setRoad(MapSquare::ROAD_BASIC);
						else
							square->setRoad(MapSquare::ROAD_RAILWAYS);
					}
					else if( line[c] == 'p' ) {
						MapSquare *square = map->getSquare(x, y);
						if( line[c+1] >= '1' && line[c+1] <= '9' ) {
							// read progress_road
							int progress_road = (int)(line[c+1] - '0');
							square->setProgressRoad(progress_road);
							c++;
						}
						else {
							VI_log("Civilization::load unrecognised progress_road\n");
							ok = false;
						}
					}
					else {
						VI_log("Civilization::load unrecognised map symbol\n");
						ok = false;
					}
				}
			}
			map->createAlphamap();
		}
		//else if( strcmp(line, "[civilization]\n") == 0 ) {
		else if( matchFileLine(line, "[civilization]") ) {
			if( map == NULL ) {
				VI_log("MainGamestate::load map should be present before any civilizations\n");
				ok = false;
				break;
			}
			Civilization *civilization = Civilization::load(this, file);
			if( civilization == NULL ) {
				VI_log("MainGamestate::load failed to load civilization\n");
				ok = false;
				break;
			}
			else if( strcmp(civilization->getName(), player_civ.c_str()) == 0 ) {
				player = civilization;
				VI_log("found player civilization: %s\n", player->getName());
			}
		}
		//else if( strcmp(line, "[relationship]\n") == 0 ) {
		else if( matchFileLine(line, "[relationship]") ) {
			Relationship *relationship = Relationship::load(this, file);
			if( relationship == NULL ) {
				VI_log("MainGamestate::load failed to load relationship\n");
				ok = false;
				break;
			}
			else {
				this->relationships.push_back(relationship);
			}
		}
		else if( line[0] == '[' ) {
			VI_log("MainGamestate::load: Unexpected '[' start of new section!\n");
			ok = false;
			break;
		}
	}
	VI_log("done loading save game file, ok?: %d\n", ok);
	// TODO: check all fields have been set
	if( player == NULL ) {
		VI_log("MainGamestate::load failed to find player civilization\n");
		ok = false;
	}
	if( map == NULL ) {
		VI_log("MainGamestate::load failed to find map\n");
		ok = false;
	}
	// check all relationships present
	if( !checkRelationships() ) {
		VI_log("MainGamestate::load not all relationships present\n");
		ok = false;
	}

	fclose(file);

	if( ok ) {
		// shouldn't be needed, but a safeguard to repair corrupt save game files!
		for(vector<Civilization *>::const_iterator iter = this->civilizations.begin(); iter != this->civilizations.end(); ++iter) {
			const Civilization *civilization = *iter;
			for(int i=0;i<civilization->getNCities();i++) {
				const City *city = civilization->getCity(i);
				city->initTravel(true, NULL);
			}
		}
	}

	VI_log("MainGamestate::load ok? %d\n", ok);

/*#ifdef _WIN32
	cursor = LoadCursor( NULL, IDC_ARROW );
	SetCursor(cursor);
	SetClassLong(game_g->getGraphicsEnvironment()->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
	game_g->setCursor(cursor_arrow);

	if( !ok ) {
		if( invalid_header )
			VI_log("Failed to load - not a valid save game file\n");
		else if( invalid_version )
			VI_log("Failed to load - save game file is from a different version of Conquests\n");
		else
			VI_log("Failed to load - game data corrupt or outdated\n");
/*#ifdef _WIN32
		if( invalid_header )
			MessageBoxA(NULL, "Failed to load - not a valid save game file", "Error", MB_OK|MB_ICONEXCLAMATION);
		else if( invalid_version )
			MessageBoxA(NULL, "Failed to load - save game file is from a different version of Conquests", "Error", MB_OK|MB_ICONEXCLAMATION);
		else
			MessageBoxA(NULL, "Failed to load - game data corrupt or outdated", "Error", MB_OK|MB_ICONEXCLAMATION);
#endif*/
		string message;
		if( invalid_header )
			message = "Failed to load - not a valid save game file";
		else if( invalid_version )
			message = "Failed to load - save game file is from a different version of Conquests";
		else
			message = "Failed to load - game data corrupt or outdated";
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), game_g->getGamestate()->getPanel(), message.c_str(), game_g->getFont(), game_g->getPanelTexture());
		window->doModal();
		delete window;
		if( this->gamestate_started ) {
			// can't recover from corrupt game file when in main game, need to quit!
			quit = true;
		}
		else {
			this->clearGameData();
		}
	}

	if( ok && this->gamestate_started ) {
		this->generateTerrain();
		this->home();
	}

	return ok;
}

void MainGamestate::cityListClickAction(VI_Panel *source) {
	vector<const City *> *cities = static_cast<vector<const City *> *>(source->getUserData());
	ASSERT( cities != NULL );
	VI_Listbox *listbox = dynamic_cast<VI_Listbox *>(source);
	int active = listbox->getActive();
	ASSERT( active >= 0 && active < cities->size() );
	const City *city = cities->at(active);
	ASSERT( city != NULL );
	mainGamestate->centre(city->getPos());
}

void MainGamestate::citiesAdvisor() {
	vector<const City *> cities;
	vector<string> city_names;
	for(int i=0;i<player->getNCities();i++) {
		const City *city = player->getCity(i);
		stringstream text;
		text << city->getName() << ": Size " << city->getSize() << ", Production: " << city->getProduction() << ", Science: " << city->getScience();
		if( city->getBuildable() != NULL ) {
			text << ", Building " << city->getBuildable()->getName() << " ( " << city->getProgressBuildable() << " / " << city->getBuildable()->getCost() << " )";
		}
		cities.push_back( city );
		city_names.push_back( text.str() );
	}

	string buttons[2];
	int n_buttons = 0;
	if( city_names.size() == 0 ) {
		buttons[0] = "Okay";
		n_buttons = 1;
	}
	else {
		buttons[0] = "View";
		buttons[1] = "Close";
		n_buttons = 2;
	}
	const int window_x_c = city_names.size() == 0 ? infowindow_def_x_c : 16;
	const int window_w_c = city_names.size() == 0 ? infowindow_def_w_c : game_g->getGraphicsEnvironment()->getWidth() - 2*window_x_c;
	const int window_h_c = city_names.size() == 0 ? infowindow_def_h_c : game_g->getGraphicsEnvironment()->getHeight() - infowindow_def_y_c - 32;
	InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), buttons, n_buttons, window_x_c, infowindow_def_y_c, window_w_c, window_h_c);
	VI_Panel *windowPanel = window->getPanel();

	VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), window_w_c - 32, city_names.size() == 0 ? window_h_c - 64 : 64);
	if( city_names.size() == 0 ) {
		textfield->addText("We have yet to build any cities.");
	}
	else {
		textfield->addText("You have the following cities:");
	}
	textfield->setForeground(1.0f, 1.0f, 1.0f);
	windowPanel->add(textfield, 16, 16);

	VI_Listbox *listbox = NULL;
	if( city_names.size() > 0 ) {
		listbox = VI_createListbox(window_w_c - 32, window_h_c - 128, &*city_names.begin(), city_names.size(), game_g->getFont());
		listbox->setHasInputFocus(true);
		listbox->setForeground(1.0f, 1.0f, 1.0f);
		listbox->setActive(0);
		listbox->setAction(cityListClickAction);
		listbox->setUserData(&cities);
		windowPanel->add(listbox, 16, 80);
	}

	int res = window->doModal();
	int active = -1;
	if( listbox != NULL ) {
		active = listbox->getActive() ;
		ASSERT( active >= 0 && active < player->getNCities() );
	}

	delete window;
	delete textfield;
	if( listbox != NULL ) {
		delete listbox;
	}

	if( res == 0 && active != -1 ) {
		City *city = player->getCity(active);
		this->centre(city->getPos());
		new CityWindow(this, city);
	}
}

void MainGamestate::unitListClickAction(VI_Panel *source) {
	vector<Unit *> *units = static_cast<vector<Unit *> *>(source->getUserData());
	ASSERT( units != NULL );
	VI_Listbox *listbox = dynamic_cast<VI_Listbox *>(source);
	int active = listbox->getActive();
	ASSERT( active >= 0 && active < units->size() );
	const Unit *unit = units->at(active);
	if( unit != NULL ) {
		mainGamestate->centre(unit->getPos());
	}
}

void MainGamestate::unitsAdvisorGetUnits(vector<Unit *> *units, vector<string> *unit_names, vector< vector<Unit *> > list_of_lists) {
	units->clear();
	unit_names->clear();
	for(vector< vector<Unit *> >::iterator iter = list_of_lists.begin(); iter != list_of_lists.end(); ++iter) {
		vector<Unit *> *list = &*iter;
		bool any = false;
		for(vector<Unit *>::iterator iter2 = list->begin(); iter2 != list->end(); ++iter2) {
			Unit *unit = *iter2;
			if( !any ) {
				any = true;
				units->push_back(NULL);
				stringstream str;
				str << unit->getTemplate()->getName() << " (" << list->size() << " units):";
				unit_names->push_back(str.str().c_str());
			}
			units->push_back(unit);
			stringstream name;
			name << "    " << getUnitName(unit);
			unit_names->push_back(name.str().c_str());
		}
	}
}

void MainGamestate::unitsAdvisor() {
	//vector<string> unit_names = game->getUnitNames( player->getUnits() );
	vector< vector<Unit *> > list_of_lists;
	// full with initial dummy lists
	/*for(vector<const UnitTemplate *>::iterator iter = this->unit_templates.begin(); iter != this->unit_templates.end(); ++iter) {
		const UnitTemplate *unit_template = *iter;*/
	for(int i=0;i<this->getNUnitTemplates();i++) {
		const UnitTemplate *unit_template = this->getUnitTemplate(i);
		vector<Unit *> list;
		list_of_lists.push_back(list);
	}
	// populate lists
	for(int i=0;i<player->getNUnits();i++) {
		Unit *unit = player->getUnit(i);
		const UnitTemplate *unit_template = unit->getTemplate();
		bool found = false;
		/*for(int j=0;j<this->unit_templates.size() && !found;j++) {
			const UnitTemplate *store_unit_template = this->unit_templates.at(j);*/
		for(int j=0;j<this->getNUnitTemplates() && !found;j++) {
			const UnitTemplate *store_unit_template = this->getUnitTemplate(j);
			if( store_unit_template == unit_template ) {
				found = true;
				vector<Unit *> *list = &list_of_lists.at(j);
				list->push_back(unit);
			}
		}
		ASSERT( found );
		/*for(vector< vector<Unit *> >::iterator iter = list_of_lists.begin(); iter != list_of_lists.end() && !found; ++iter) {
		vector<Unit *> *list = &*iter;
		if( unit_template == list->at(0)->getTemplate() ) {
		list->push_back(unit);
		found = true;
		}
		}
		if( !found ) {
		vector<Unit *> list;
		list.push_back(unit);
		list_of_lists.push_back(list);
		}*/
	}
	vector<Unit *> units;
	vector<string> unit_names;
	unitsAdvisorGetUnits(&units, &unit_names, list_of_lists);

	//vector<string> unit_names = game->getUnitNames( &units );
	/*vector<string> unit_names;
	for(vector<Unit *>::const_iterator iter = units.begin();iter != units.end(); ++iter) {
	const Unit *unit = *iter;
	string name = getUnitName(unit);
	unit_names.push_back(name.c_str());
	}*/

	string buttons[3];
	int n_buttons = 0;
	if( unit_names.size() == 0 ) {
		buttons[0] = "Okay";
		n_buttons = 1;
	}
	else {
		buttons[0] = "Activate";
		buttons[1] = "Activate All Of This Type";
		buttons[2] = "Close";
		n_buttons = 3;
	}
	const int window_x_c = unit_names.size() == 0 ? infowindow_def_x_c : 16;
	const int window_w_c = unit_names.size() == 0 ? infowindow_def_w_c : game_g->getGraphicsEnvironment()->getWidth() - 2*window_x_c;
	//const int window_w_c = 560;
	const int window_h_c = unit_names.size() == 0 ? infowindow_def_h_c : 400;
	//const int window_h_c = infowindow_def_h_c;
	InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), buttons, n_buttons, window_x_c, infowindow_def_y_c, window_w_c, window_h_c);
	//window->clearActiveWindow(); // shouldn't be set as the active window, otherwise may get deleted!
	VI_Panel *windowPanel = window->getPanel();

	VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), window_w_c - 32, unit_names.size() == 0 ? window_h_c - 64 : 32);
	if( unit_names.size() == 0 ) {
		textfield->addText("We have no units.");
	}
	else {
		textfield->addText("You have the following units:");
	}
	textfield->setForeground(1.0f, 1.0f, 1.0f);
	windowPanel->add(textfield, 16, 16);

	VI_Listbox *listbox = NULL;
	if( unit_names.size() > 0 ) {
		listbox = VI_createListbox(window_w_c - 32, window_h_c - 96, &*unit_names.begin(), unit_names.size(), game_g->getFont());
		listbox->setHasInputFocus(true);
		listbox->setForeground(1.0f, 1.0f, 1.0f);
		listbox->setActive(0);
		listbox->setAction(unitListClickAction);
		listbox->setUserData(&units);
		windowPanel->add(listbox, 16, 48);
	}

	bool done = false;
	int res = -1;
	int active = -1;
	while( !done ) {
		done = true;
		res = window->doModal();
		if( listbox != NULL ) {
			active = listbox->getActive() ;
			//ASSERT( active >= 0 && active < player->getNUnits() );
			ASSERT( active >= 0 && active < units.size() );
			if( res == 0 ) {
				//Unit *unit = player->getUnit(active);
				Unit *unit = units.at(active);
				if( unit != NULL && unit->getStatus() != Unit::STATUS_NORMAL ) {
					done = false; // don't close window, only activate the unit
				}
				if( unit == NULL ) {
					// no unit
					done = false; // don't close window
				}
				else if( unit->hasMovesLeft() ) {
					this->activateUnit(unit, true);
				}
				else {
					//unit->setStatus(Unit::STATUS_NORMAL);
					this->wakeupUnit(unit);
				}
			}
			else if( res == 1 ) {
				Unit *selected_unit = units.at(active);
				if( selected_unit == NULL ) {
					// no unit
					done = false; // don't close window
				}
				else {
					done = false; // don't close window, only activate the unit(s)
					bool found = false;
					for(int i=0;i<player->getNUnits();i++) {
						Unit *unit = player->getUnit(i);
						if( unit->getTemplate() != selected_unit->getTemplate() ) {
							// not of same type
						}
						else if( !found && unit->hasMovesLeft() ) {
							// activate the first one we can
							found = true;
							this->activateUnit(unit, false);
						}
						else {
							// deal with the rest
							//unit->setStatus(Unit::STATUS_NORMAL);
							this->wakeupUnit(unit);
						}
					}
				}
			}
		}
		if( !done ) {
			// reset
			unitsAdvisorGetUnits(&units, &unit_names, list_of_lists);
			listbox->reset(&*unit_names.begin(), unit_names.size());
		}
	}

	delete window;
	delete textfield;
	if( listbox != NULL ) {
		delete listbox;
	}
}

int MainGamestate::getMinTurnsWar(const Civilization *civ1, const Civilization *civ2) const {
	// how long until civ1 considers making peace with civ2?
	const int min_turns_war_c = 8;
	int min_turns_war = min_turns_war_c;
	if( this->aiaggression == AIAGGRESSION_AGGRESSIVE )
		min_turns_war *= 2;
	int dummy = 0;
	int civ1_strength = civ1->getMilitary(&dummy);
	int civ2_strength = civ2->getMilitary(&dummy);
	int civ2_nuclear_strength = civ2->getNuclear();
	if( civ2_nuclear_strength >= 2 ) {
		min_turns_war = 1;
	}
	else if( 1.5 * civ1_strength < civ2_strength ) {
		min_turns_war /= 2;
	}
	else if( civ1_strength > 1.5 * civ2_strength ) {
		min_turns_war *= 2;
	}
	return min_turns_war;
}

// n.b., TalkCode is a global rather than local to MainGamestate, otherwise on Linux we are unable to use it in a vector
enum TalkCode {
	TALKCODE_DECLAREWAR = 0,
	TALKCODE_MAKEPEACE = 1,
	TALKCODE_TECHEXCHANGE = 2,
	TALKCODE_MAPSEXCHANGE = 3,
	TALKCODE_GOODBYE = 4
};

void MainGamestate::civsAdvisor() {
	if( !CHEAT && player->isDead() ) {
		return;
	}
	vector<Civilization *> contact_civs;
	vector<string> civilization_names;
	for(vector<Civilization *>::const_iterator iter = this->civilizations.begin();iter != this->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		if( civilization != player && !civilization->isDead() ) {
			stringstream name;
			name << civilization->getName();
			//if( !CHEAT ) {
			if( !player->isDead() ) {
				const Relationship *relationship = this->findRelationship(player, civilization);
				if( !CHEAT && !relationship->madeContact() ) {
					continue;
				}
				if( relationship->getStatus() == Relationship::STATUS_PEACE )
					name << " (PEACE)";
				else if( relationship->getStatus() == Relationship::STATUS_WAR )
					name << " (AT WAR)";
			}
			bool any = false;
			for(vector<Civilization *>::const_iterator iter2 = this->civilizations.begin();iter2 != this->civilizations.end(); ++iter2) {
				Civilization *civilization2 = *iter2;
				if( civilization2 != civilization && civilization2 != player && !civilization2->isDead() ) {
					Relationship *relationship2 = this->findRelationship(civilization2, civilization);
					if( relationship2->getStatus() == Relationship::STATUS_WAR ) {
						if( !any ) {
							any = true;
							name << " (At war with ";
						}
						else {
							name << ", ";
						}
						name << civilization2->getName();
					}
				}
			}
			if( any ) {
				name << ")";
			}
			civilization_names.push_back(name.str().c_str());
			contact_civs.push_back(civilization);
		}
	}

	string buttons[2];
	int n_buttons = 0;
	if( civilization_names.size() == 0 || player->isDead() ) {
		buttons[0] = "Okay";
		n_buttons = 1;
	}
	else {
		buttons[0] = "Contact";
		buttons[1] = "Cancel";
		n_buttons = 2;
	}

	int res = -1;
	int active = -1;
	{
		const int window_x_c = civilization_names.size() == 0 ? infowindow_def_x_c : 16;
		const int window_w_c = civilization_names.size() == 0 ? infowindow_def_w_c : game_g->getGraphicsEnvironment()->getWidth() - 2*window_x_c;
		const int window_h_c = civilization_names.size() == 0 ? infowindow_def_h_c : game_g->getGraphicsEnvironment()->getHeight() - infowindow_def_y_c - 32;
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), buttons, n_buttons, window_x_c, infowindow_def_y_c, window_w_c, window_h_c);
		VI_Panel *windowPanel = window->getPanel();

		VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), window_w_c - 32, civilization_names.size() == 0 ? window_h_c - 64 : 64);
		if( civilization_names.size() == 0 ) {
			textfield->addText("We know of no other Civilizations!");
		}
		else {
			textfield->addText("We have made contact with the following civilizations:");
		}
		//textfield->setForeground(1.0f, 1.0f, 0.0f);
		textfield->setForeground(1.0f, 1.0f, 1.0f);
		windowPanel->add(textfield, 16, 16);

		VI_Listbox *listbox = NULL;
		if( civilization_names.size() > 0 ) {
			listbox = VI_createListbox(window_w_c - 32, window_h_c - 128, &*civilization_names.begin(), civilization_names.size(), game_g->getFont());
			listbox->setHasInputFocus(true);
			//listbox->setForeground(1.0f, 1.0f, 0.0f);
			listbox->setForeground(1.0f, 1.0f, 1.0f);
			listbox->setActive(0);
			windowPanel->add(listbox, 16, 80);
		}

		res = window->doModal();
		if( listbox != NULL ) {
			active = listbox->getActive() ;
			ASSERT( active >= 0 && active < contact_civs.size() );
		}

		delete window;
		delete textfield;
		if( listbox != NULL ) {
			delete listbox;
		}
	}

	if( res == 0 && active != -1 && !player->isDead() ) {
		Civilization *civilization = contact_civs.at(active);
		Relationship *relationship = this->findRelationship(player, civilization);
		if( !relationship->madeContact() ) {
			ASSERT( CHEAT );
		}
		else {
			//ASSERT( relationship->madeContact() );
			//string buttons[] = {"Declare War", "Cancel"};
			stringstream text;
			vector<string> talkButtons;
			vector<TalkCode> talkCodes;
			vector<const Technology *> player_distinct_technologies;
			vector<const Technology *> civ_distinct_technologies;

			if( relationship->getStatus() == Relationship::STATUS_PEACE ) {
				text << "Greetings from the ";
				text << civilization->getNameAdjective();
				text << " people. How may we help you?";
				talkButtons.push_back("Declare War");
				talkCodes.push_back(TALKCODE_DECLAREWAR);
				Civilization::getDistinctTechnologies(this, &player_distinct_technologies, &civ_distinct_technologies, player, civilization);
				if( player_distinct_technologies.size() > 0 && civ_distinct_technologies.size() > 0 ) {
					talkButtons.push_back("Exchange Technology");
					talkCodes.push_back(TALKCODE_TECHEXCHANGE);
				}
				if( player->canShareMaps(civilization) && player->compareMaps(civilization) ) {
					talkButtons.push_back("Exchange Maps");
					talkCodes.push_back(TALKCODE_MAPSEXCHANGE);
				}
			}
			else if( relationship->getStatus() == Relationship::STATUS_WAR ) {
				/*text << "You are an enemy of ";
				text << civilization->getName();
				text << " and we will wipe you from the face of the Earth!";*/
				text << "What do you want?";
				talkButtons.push_back("Make Peace");
				talkCodes.push_back(TALKCODE_MAKEPEACE);
			}
			else {
				ASSERT( false );
			}

			talkButtons.push_back("Goodbye");
			talkCodes.push_back(TALKCODE_GOODBYE);

			VI_Texture *civ_texture = civilization->getTexture();
			const int talk_window_h_c = civ_texture->getHeight() + 96;
			InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), &*talkButtons.begin(), talkButtons.size(), infowindow_def_x_c, infowindow_def_y_c, infowindow_def_w_c, talk_window_h_c);
			window->getButton(0)->disableKeyShortcut(); // disable key shortcut - probably not best to do something like declaring war if the user hits Return!
			VI_Panel *windowPanel = window->getPanel();

			VI_ImageButton *imagebutton = VI_createImageButton(civ_texture);
			windowPanel->add(imagebutton, 16, 48);

			//VI_Stringgadget *title = VI_createStringgadget(this->getFont(), infowindow_def_w_c - 32);
			VI_Stringgadget *title = VI_createStringgadget(game_g->getFont(), imagebutton->getWidth());
			const unsigned char *rgb = civilization->getColor();
			title->setBackground(rgb[0]/255.0f, rgb[1]/255.0f, rgb[2]/255.0f);
			title->setForeground(0.0f, 0.0f, 0.0f);
			//title->setForeground(rgb[0]/255.0f, rgb[1]/255.0f, rgb[2]/255.0f);
			title->addText(civilization->getName());
			windowPanel->add(title, 16, 16);

			VI_Textfield *textfield = VI_createTextfield(game_g->getFont(),  infowindow_def_w_c - imagebutton->getWidth() - 48, talk_window_h_c - 64);
			//textfield->setForeground(1.0f, 1.0f, 0.0f);
			textfield->setForeground(1.0f, 1.0f, 1.0f);
			textfield->addText(text.str().c_str());
			windowPanel->add(textfield, imagebutton->getWidth() + 32, 16);

			res = window->doModal();
			delete window;
			delete imagebutton;

			TalkCode talkCode = talkCodes.at(res);
			if( talkCode == TALKCODE_DECLAREWAR ) {
				relationship->setStatus(Relationship::STATUS_WAR);
				InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), "Prepare to die!", game_g->getFont(), game_g->getPanelTexture());
				window->doModal();
				delete window;
			}
			else if( talkCode == TALKCODE_TECHEXCHANGE ) {
				ASSERT( player_distinct_technologies.size() > 0 && civ_distinct_technologies.size() > 0 );
				vector<string> technology_names;
				for(vector<const Technology *>::const_iterator iter = civ_distinct_technologies.begin(); iter != civ_distinct_technologies.end(); ++iter) {
					const Technology *technology = *iter;
					technology_names.push_back(technology->getName());
				}
				int window_h_c = game_g->getGraphicsEnvironment()->getHeight() - infowindow_def_y_c - 32;
				int active = InfoWindow::select(game_g->getGraphicsEnvironment(), this->getPanel(), game_g->getFont(), game_g->getPanelTexture(), "Which technology would you like?", "Okay", "Cancel", window_h_c, &technology_names);
				if( active != -1 ) {
					ASSERT( active >= 0 && active < civ_distinct_technologies.size() );
					player->addTechnology( civ_distinct_technologies.at( active ) );

					// now AI picks one of the player's technologies
					int choose = rand() % player_distinct_technologies.size();
					const Technology *technology = player_distinct_technologies.at( choose );
					civilization->addTechnology( technology );
					stringstream text;
					text << civilization->getName() << " take " << technology->getName() << ".";
					InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture());
					window->doModal();
					delete window;
				}
			}
			else if( talkCode == TALKCODE_MAPSEXCHANGE ) {
				player->shareMaps(civilization);
			}
			else if( talkCode == TALKCODE_MAKEPEACE ) {
				string text;
				int min_turns_war = this->getMinTurnsWar(civilization, player);
				if( year >= relationship->getYearDeclaredWar() + min_turns_war ) {
					text = "Let us put an end to this bloodshed.";
					relationship->setStatus(Relationship::STATUS_PEACE);
				}
				else {
					text = "Never! We will keep fighting until you are destroyed!";
				}
				InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text, game_g->getFont(), game_g->getPanelTexture());
				window->doModal();
				delete window;
			}
			else if( talkCode == TALKCODE_GOODBYE ) {
				// simply close window
			}
			else {
				ASSERT( false );
			}
		}
	}
}

void MainGamestate::techsAdvisor() {
	vector<string> technology_names;
	vector<const Technology *> technologies;
	/*for(vector<Technology *>::const_iterator iter = this->technologies.begin();iter != this->technologies.end(); ++iter) {
		const Technology *technology = *iter;*/
	for(int i=0;i<this->getNTechnologies();i++) {
		const Technology *technology = this->getTechnology(i);
		if( player->hasTechnology(technology) ) {
			stringstream name;
			name << technology->getName();
			name << " (" << technology->getAgeName() << ")";
			technology_names.push_back(name.str().c_str());
			technologies.push_back(technology); // need to make a separate list, as the order maintained in the Civilization class will in general be a different order!
		}
	}

	//int window_h = technology_names.size() == 0 ? 128 : 512;
	int window_h = technology_names.size() == 0 ? 128 : game_g->getGraphicsEnvironment()->getHeight() - infowindow_def_y_c - 16;
	const int cheat_info_h_c = 256;
	if( technology_names.size() == 0 && CHEAT ) {
		window_h += cheat_info_h_c + 16;
	}
	int list_h = technology_names.size() == 0 ? 0 : window_h - 128;
	if( CHEAT && technology_names.size() > 0 ) {
		list_h -= cheat_info_h_c - 16;
	}
	string okay = "Okay";
	int width_c = game_g->getGraphicsEnvironment()->getWidth() - 128;
	width_c = min(width_c, 640);
	InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), &okay, 1, 64, infowindow_def_y_c, width_c, window_h);
	VI_Panel *windowPanel = window->getPanel();

	VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), width_c - 32, 64);
	if( technology_names.size() == 0 ) {
		textfield->addText("We have yet to discover anything!");
	}
	else {
		textfield->addText("We have discovered the following technologies:");
	}
	//textfield->setForeground(1.0f, 1.0f, 0.0f);
	textfield->setForeground(1.0f, 1.0f, 1.0f);
	windowPanel->add(textfield, 16, 16);

	VI_Listbox *listbox = NULL;
	if( technology_names.size() > 0 ) {
		//listbox = VI_createListbox(width_c - 32, list_h, &*technology_names.begin(), technology_names.size(), game_g->getFont());
		listbox = VI_createListbox((width_c - 32)/2, list_h, &*technology_names.begin(), technology_names.size(), game_g->getFont());
		listbox->setHasInputFocus(true);
		//listbox->setForeground(1.0f, 1.0f, 0.0f);
		listbox->setForeground(1.0f, 1.0f, 1.0f);
		listbox->setActive(0);
		windowPanel->add(listbox, 16, 80);

		VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), (width_c - 32)/2, list_h);
		windowPanel->add(textfield, 16 + (width_c - 32)/2, 80);

		AskTechnologyInfo askTechnologyInfo;
		askTechnologyInfo.textfieldInfo = textfield;
		//askTechnologyInfo.player_technologies = player->getTechnologies();
		askTechnologyInfo.player_technologies = &technologies;
		listbox->setUserData(&askTechnologyInfo);
		listbox->setAction(askTechnologyAction);
		askTechnologyAction(listbox);
	}

	VI_Textfield *textfield_cheat = NULL;
	if( CHEAT ) {
		textfield_cheat = VI_createTextfield(game_g->getFont(), width_c - 32, cheat_info_h_c);
		stringstream str;
		for(vector<Civilization *>::const_iterator iter = civilizations.begin(); iter != civilizations.end(); ++iter) {
			const Civilization *civilization = *iter;
			if( civilization == player || civilization->isDead() ) {
				continue;
			}
			str << civilization->getName() << " (researching ";
			const Technology *tech = civilization->getTechnology();
			if( tech == NULL )
				str << "NOTHING";
			else {
				str << tech->getName();
				str << " " << civilization->getProgressTechnology() << " / " << tech->getCost(this->mainGamestate);
			}
			str << ") :\n\n";
			for(int i=0;i<this->getNTechnologies();i++) {
				const Technology *technology = this->getTechnology(i);
				if( civilization->hasTechnology(technology) ) {
					str << technology->getName() << "\n";
				}
			}
			str << "\n";
		}
		textfield_cheat->addText(str.str().c_str());
		textfield_cheat->setCPos(0);
		windowPanel->add(textfield_cheat, 16, 80 + list_h);
	}

	window->doModal();

	delete window;
	delete textfield;
	if( listbox != NULL ) {
		delete listbox;
	}
	if( textfield_cheat != NULL ) {
		delete textfield_cheat;
	}
}

void MainGamestate::scoresAdvisor() {
	if( !CHEAT && player->isDead() ) {
		return;
	}
	string okay = "Okay";
	stringstream text;
	text << "Difficulty Level: ";
	text << this->getDifficultyString();
	text << "\n\n";

	for(vector<Civilization *>::const_iterator iter = this->civilizations.begin();iter != this->civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		if( civilization->isDead() ) {
			continue;
		}
		if( !CHEAT && civilization != player ) {
			const Relationship *relationship = this->findRelationship(player, civilization);
			if( !relationship->madeContact() ) {
				continue;
			}
		}
		text << civilization->getName() << ":\n";
		text << "Population: " << civilization->getTotalPopulation() << "\n";
		text << "Cities: " << civilization->getNCities() << "\n";
		text << "Power: " << civilization->getPower() << " (" << civilization->getPowerPerTurn() << " per turn)\n";
		int n_land_units = 0;
		int land_power = civilization->getMilitary(&n_land_units);
		text << "Military Strength: " << land_power << " (" << n_land_units << " units)\n";
		int n_air_units = 0;
		int air_power = civilization->getAir(&n_air_units);
		if( air_power > 0 ) {
			text << "Air Strength: " << air_power << " (" << n_air_units << " units)\n";
		}
		int nuclear_power = civilization->getNuclear();
		if( nuclear_power > 0 ) {
			text << "Nuclear Strength: " << nuclear_power << "\n";
		}
		text << "Science: " << civilization->getTotalScience() << "\n";
		text << "Production: " << civilization->getTotalProduction() << "\n";
		text << "\n";
	}
	const int window_h_c = game_g->getGraphicsEnvironment()->getHeight() - 2*infowindow_def_y_c;
	InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), &okay, 1, infowindow_def_x_c, infowindow_def_y_c, infowindow_def_w_c, window_h_c);
	window->doModal();
	delete window;
}
struct MapInfo {
	const MainGamestate *mainGamestate;
	int image_w, image_h;
	MapInfo() : mainGamestate(NULL), image_w(-1), image_h(-1) {
	}
};

void MainGamestate::mapPaintPanel(VI_Graphics2D *g2d, void *data) {
	const MapInfo *mapInfo = static_cast<const MapInfo *>(data);
	const MainGamestate *mainGamestate = mapInfo->mainGamestate;
	const Map *map = mainGamestate->getMap();

	if( mainGamestate->view_3d ) {
		float mx[4] = {0.0f, 0.0f, 0.0f, 0.0f};
		float my[4] = {0.0f, 0.0f, 0.0f, 0.0f};
		mainGamestate->getMapPosf(&mx[0], &my[0], 0, 0);
		mainGamestate->getMapPosf(&mx[1], &my[1], game_g->getGraphicsEnvironment()->getWidth()-1, 0);
		mainGamestate->getMapPosf(&mx[2], &my[2], 0, game_g->getGraphicsEnvironment()->getHeight()-1);
		mainGamestate->getMapPosf(&mx[3], &my[3], game_g->getGraphicsEnvironment()->getWidth()-1, game_g->getGraphicsEnvironment()->getHeight()-1);
		for(int i=0;i<4;i++) {
			mx[i] *=  mapInfo->image_w / (float)map->getWidth();
			my[i] *=  mapInfo->image_h / (float)map->getHeight();
		}
		g2d->setColor3i(255, 255, 255);
		bool clip_left = true;
		if( map->getTopology() == Map::TOPOLOGY_FLAT && mx[0] < 0 ) {
			// so that we can use the same clipping code for the left hand side
			clip_left = false;
			for(int i=0;i<4;i++)
				mx[i] += mapInfo->image_w;
		}
		if( mx[2] >= mapInfo->image_w ) {
			// partial left wraparound
			float scale = ((float)(mx[2] - mapInfo->image_w)) / (float)(mx[2] - mx[0]);
			int iy = (int)(my[2] + (my[0] - my[2])*scale);
			if( map->getTopology() == Map::TOPOLOGY_CYLINDER || !clip_left ) {
				g2d->draw(0, my[0], mx[1] - mapInfo->image_w, my[1]);
				g2d->draw(mx[1] - mapInfo->image_w, my[1], mx[3] - mapInfo->image_w, my[3]);
				g2d->draw(mx[3] - mapInfo->image_w, my[3], mx[2] - mapInfo->image_w, my[2]);
				g2d->draw(mx[2] - mapInfo->image_w, my[2], 0, iy);
			}
			if( map->getTopology() == Map::TOPOLOGY_CYLINDER || clip_left ) {
				g2d->draw(mx[0], my[0], mapInfo->image_w, my[1]);
				g2d->draw(mapInfo->image_w, iy, mx[0], my[0]);
			}
		}
		else if( mx[3] >= mapInfo->image_w ) {
			// full wraparound
			if( map->getTopology() == Map::TOPOLOGY_CYLINDER || !clip_left ) {
				g2d->draw(0, my[0], mx[1] - mapInfo->image_w, my[1]);
				g2d->draw(mx[1] - mapInfo->image_w, my[1], mx[3] - mapInfo->image_w, my[3]);
				g2d->draw(mx[3] - mapInfo->image_w, my[3], 0, my[2]);
			}
			if( map->getTopology() == Map::TOPOLOGY_CYLINDER || clip_left ) {
				g2d->draw(mx[0], my[0], mapInfo->image_w, my[1]);
				g2d->draw(mapInfo->image_w, my[3], mx[2], my[2]);
				g2d->draw(mx[2], my[2], mx[0], my[0]);
			}
		}
		else if( mx[1] >= mapInfo->image_w ) {
			// partial right wraparound
			float scale = ((float)(mx[1] - mapInfo->image_w)) / (float)(mx[1] - mx[3]);
			int iy = (int)(my[1] + (my[3] - my[1])*scale);
			if( map->getTopology() == Map::TOPOLOGY_CYLINDER || !clip_left ) {
				g2d->draw(0, my[0], mx[1] - mapInfo->image_w, my[1]);
				g2d->draw(mx[1] - mapInfo->image_w, my[1], 0, iy);
			}
			if( map->getTopology() == Map::TOPOLOGY_CYLINDER || clip_left ) {
				g2d->draw(mx[0], my[0], mapInfo->image_w, my[1]);
				g2d->draw(mapInfo->image_w, iy, mx[3], my[3]);
				g2d->draw(mx[3], my[3], mx[2], my[2]);
				g2d->draw(mx[2], my[2], mx[0], my[0]);
			}
		}
		else {
			g2d->draw(mx[0], my[0], mx[1], my[1]);
			g2d->draw(mx[1], my[1], mx[3], my[3]);
			g2d->draw(mx[3], my[3], mx[2], my[2]);
			g2d->draw(mx[2], my[2], mx[0], my[0]);
		}
	}
	else {
		float n_x = ((float)game_g->getGraphicsEnvironment()->getWidth()) / (float)mainGamestate->getSqWidth();
		float n_y = ((float)game_g->getGraphicsEnvironment()->getHeight()) / (float)mainGamestate->getSqHeight();
		float frac_w = n_x / (float)map->getWidth();
		float frac_h = n_y / (float)map->getHeight();
		int box_w = frac_w * mapInfo->image_w;
		int box_h = frac_h * mapInfo->image_h;
		float frac_x = mainGamestate->map_pos_x / (float)map->getWidth();
		float frac_y = mainGamestate->map_pos_y / (float)map->getHeight();
		int box_x = frac_x * mapInfo->image_w;
		int box_y = frac_y * mapInfo->image_h;

		int wraparound_box_w = 0;
		// handle wraparound!
		if( map->getTopology() == Map::TOPOLOGY_CYLINDER ) {
			if( box_x + box_w > mapInfo->image_w-1 ) {
				int new_box_w = mapInfo->image_w-1 - box_x;
				wraparound_box_w = box_w - new_box_w;
				box_w = new_box_w;
			}
		}

		g2d->setColor3i(255, 255, 255);
		g2d->drawRect(box_x, box_y, box_w, box_h);
		if( wraparound_box_w > 0 ) {
			g2d->drawRect(0, box_y, wraparound_box_w, box_h);
		}
	}
}

void MainGamestate::showMap() {
	VI_Image *image = VI_createImage(map->getWidth(), map->getHeight());
	for(int y=0;y<map->getHeight();y++) {
		for(int x=0;x<map->getWidth();x++) {
			unsigned char r = 0, g = 0, b = 0;
			if( player->isExplored(x, y) ) {
				const MapSquare *square = map->getSquare(x, y);
				const City *city = square->getCity();
				if( city != NULL ) {
					const unsigned char *rgb = city->getCivilization()->getColor();
					r = rgb[0];
					g = rgb[1];
					b = rgb[2];
				}
				else if( square->isLand() ) {
					g = 127;
				}
				else {
					g = 64;
					b = 255;
				}
			}
			int fy = map->getHeight() - 1 - y; // texture coords are inverted
			image->putRGB(x, fy, r, g, b);
		}
	}
	if( !image->makePowerOf2() ) {
	    VI_log("Failed to create image\n");
#ifdef _WIN32
		MessageBoxA(NULL,"Failed to create image","Error",MB_OK|MB_ICONEXCLAMATION);
#endif
		delete image;
		return;
	}
	VI_Texture *texture = VI_createTexture(image);

	int window_x = 64;
	int window_y = 64;
	int border_x = 16, border_y = 16;
	int extra_x = 2 * border_x;
	int extra_y = 2 * border_y + 32;
	/*int image_w = map->getWidth();
	int image_h = map->getHeight();*/
	int image_w = 10 * map->getWidth();
	int image_h = 10 * map->getHeight();
	if( image_w > game_g->getGraphicsEnvironment()->getWidth() - 2*window_x - extra_x ) {
		image_w = game_g->getGraphicsEnvironment()->getWidth() - 2*window_x - extra_x;
	}
	if( image_h > game_g->getGraphicsEnvironment()->getHeight() - 2*window_y - extra_y ) {
		image_h = game_g->getGraphicsEnvironment()->getHeight() - 2*window_y - extra_y;
	}
	/*int image_w = 2 * max(map->getWidth(), map->getHeight());
	int image_h = image_w;*/
	string okay = "Okay";
	int window_w = image_w + extra_x;
	int window_h = image_h + extra_y;
	InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), &okay, 1, window_x, window_y, window_w, window_h);
	VI_Panel *windowPanel = window->getPanel();
	VI_ImageButton *imageButton = VI_createImageButton(texture, image_w, image_h);

	MapInfo mapInfo;
	mapInfo.mainGamestate = this;
	mapInfo.image_w = image_w;
	mapInfo.image_h = image_h;
	imageButton->setPaintFunc(mapPaintPanel, &mapInfo);

	windowPanel->add(imageButton, border_x, border_y);

	window->doModal();

	delete window;
	delete texture;
	delete imageButton;
}

void MainGamestate::checkMapView() {
	//VI_log("was: %f, %f\n", this->map_pos_x, this->map_pos_y);

	if( view_3d ) {
		VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
		/*Vector3D pos = vp->getPosition();
		Vector3D dir_l = game_g->getGraphicsEnvironment()->getDirectionOfPoint(vp, 0, 0);
		Vector3D dir_r = game_g->getGraphicsEnvironment()->getDirectionOfPoint(vp, game_g->getGraphicsEnvironment()->getWidth()-1, 0);*/
		float mx_l = 0.0f, mx_r = 0.0f, my_t = 0.0f, my_b = 0.0f, dummy = 0.0f;
		getMapPosf(&dummy, &my_t, 0, 0);
		getMapPosf(&mx_l, &my_b, 0, game_g->getGraphicsEnvironment()->getHeight()-1);
		getMapPosf(&mx_r, &dummy, game_g->getGraphicsEnvironment()->getWidth()-1, game_g->getGraphicsEnvironment()->getHeight()-1);
		if( map->getTopology() == Map::TOPOLOGY_FLAT ) {
			if( mx_l < 0.0f ) {
				vp->move(-mx_l, 0.0f, 0.0f);
			}
			else if( mx_r > (float)map->getWidth() ) {
				vp->move(((float)map->getWidth()) - mx_r, 0.0f, 0.0f);
			}
		}
		if( my_t < 0.0f ) {
				vp->move(0.0f, 0.0f, -my_t);
		}
		else if( my_b > (float)map->getHeight() ) {
			vp->move(0.0f, 0.0f, ((float)map->getHeight()) - my_b);
		}

		getMapPosf(&mx_l, &my_t, 0, 0);
		float old_mx_l = mx_l;
		float old_my_t = my_t;
		this->getMap()->reduceToBase(&mx_l, &my_t);
		vp->move(mx_l - old_mx_l, 0.0f, my_t - old_my_t);
	}
	else {
		if( map->getTopology() == Map::TOPOLOGY_FLAT ) {
			float map_limit_x = max(0.0f, map->getWidth() - ((float)game_g->getGraphicsEnvironment()->getWidth())/(float)this->getSqWidth());
			this->map_pos_x = max(this->map_pos_x, 0.0f);
			this->map_pos_x = min(this->map_pos_x, map_limit_x);
		}

		float map_limit_y = max(0.0f, map->getHeight() - ((float)game_g->getGraphicsEnvironment()->getHeight())/(float)this->getSqHeight());
		this->map_pos_y = max(this->map_pos_y, 0.0f);
		this->map_pos_y = min(this->map_pos_y, map_limit_y);
		//VI_log("    now: %f, %f\n", this->map_pos_x, this->map_pos_y);

		this->getMap()->reduceToBase(&this->map_pos_x, &this->map_pos_y);
	}
}

bool MainGamestate::canBomb(const UnitTemplate *unit_template, Pos2D source, Pos2D target) const {
	if( !unit_template->canBomb() ) {
		return false;
	}
	int dist = this->getMap()->distanceChebyshev(source, target);
	if( dist <= unit_template->getAirRange() ) {
		return true;
	}
	return false;
}

void MainGamestate::bomb(Unit *unit, Pos2D target) {
	{
		// checks
		City *src_city = map->findCity(unit->getPos());
		ASSERT( src_city != NULL );
		ASSERT( src_city->canLaunch(unit->getTemplate()) );
		ASSERT( unit->getTemplate()->canBomb() );
		ASSERT( this->canBomb(unit->getTemplate(), unit->getPos(), target) );
	}
	//const MapSquare *square = map->getSquare(target);
	//City *city = square->getCity();
	City *city = map->findCity(target);
	ASSERT( city != NULL );
	VI_log("%s %s bombards %s city of %s\n", unit->getCivilization()->getNameAdjective(), unit->getTemplate()->getName(), city->getCivilization()->getNameAdjective(), city->getName());
	ASSERT( city->getCivilization() != unit->getCivilization() );
	Relationship *relationship = this->findRelationship(city->getCivilization(), unit->getCivilization());
	ASSERT( relationship->getStatus() == Relationship::STATUS_WAR ); // should already have declared war before calling this function
	bool is_missile = unit->getTemplate()->isMissile();
	ASSERT( is_missile || unit->getTemplate()->getAirAttack() > 0 );

	if( city->getCivilization() == player || unit->getCivilization() == player ) {
		this->centre(city->getPos());
	}

	unit->incNBombed();

	if( is_missile && unit->getTemplate()->getNuclearType() != UnitTemplate::NUCLEARTYPE_NONE && city->hasImprovement(IMPROVEMENT_SDILASERS) ) {
		int r = rand() % 100;
		VI_log("SDI Lasers roll: %d\n", r);
		if( r < 80 ) {
			stringstream text;
			bool show_window = false;
			if( unit->getCivilization() == player ) {
				text << "Your nuke was shot down by enemy SDI Lasers!";
				show_window = true;
			}
			else if( city->getCivilization() == player ) {
				text << city->getName() << ": Your SDI Lasers have shot down ";
				text << unit->getCivilization()->getNameAdjective();
				text << " nukes!";
				show_window = true;
			}
			if( show_window ) {
				InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), 64, 16, infowindow_def_w_c, 180);
				window->doModal();
				delete window;
			}
			if( unit->getCivilization() == player && unit == active_unit ) {
				active_unit = NULL;
				updateUnitButtons();
			}
			delete unit;
			unit = NULL;
			return; // attack destroyed, so quit
		}
	}

	const Improvement *air_defence_improvement = city->getBestAirDefenceImprovement();
	if( air_defence_improvement != NULL && !is_missile ) {
		int best_defence = air_defence_improvement->getAirDefence();
		ASSERT( best_defence > 0 );
		VI_log("%s %s attack %s %s\n", unit->getCivilization()->getNameAdjective(), unit->getTemplate()->getName(), city->getCivilization()->getNameAdjective(), air_defence_improvement->getName());
		int attack = unit->getTemplate()->getAirAttack();
		VI_log("attack: %d\n", attack);
		VI_log("defence: %d\n", best_defence);
		int total = attack + best_defence;
		int r = rand() % total;
		if( r < attack ) {
			// attacker wins! Improvement isn't destroyed however; no need to inform player
		}
		else {
			// attacker loses
			stringstream text;
			bool show_window = false;
			if( unit->getCivilization() == player ) {
				text << "You were shot down by enemy ";
				text << air_defence_improvement->getName();
				text << ".";
				show_window = true;
			}
			else if( city->getCivilization() == player ) {
				text << city->getName() << ": Your ";
				text << air_defence_improvement->getName();
				text << " have shot down ";
				text << unit->getCivilization()->getNameAdjective();
				text << " ";
				text << unit->getTemplate()->getName();
				text << ".";
				show_window = true;
			}
			if( show_window ) {
				InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), 64, 16, infowindow_def_w_c, 180);
				window->doModal();
				delete window;
			}
			if( unit->getCivilization() == player && unit == active_unit ) {
				active_unit = NULL;
				updateUnitButtons();
			}
			delete unit;
			unit = NULL;
			return; // attack destroyed, so quit
		}
	}

	const vector<Unit *> *units = map->findUnitsAt(city->getPos());
	if( units->size() > 0 && !is_missile ) {
		// check for defending air units
		Unit *defender = NULL;
		int best_defence = 0;
		for(vector<Unit *>::const_iterator iter = units->begin(); iter != units->end(); ++iter) {
			Unit *this_unit = *iter;
			if( this_unit->getTemplate()->isAir() && !this_unit->getTemplate()->isMissile() && this_unit->getTemplate()->getAirDefence() > 0 ) {
				if( !city->canLaunch(this_unit->getTemplate()) ) {
					// fighter unit can't be launched, as airport no longer available
					continue;
				}
				int defence = this_unit->getTemplate()->getAirDefence();
				// find best air defender
				if( defender == NULL || defence > best_defence ) {
					defender = this_unit;
					best_defence = defence;
				}
			}
		}
		if( defender != NULL ) {
			VI_log("%s %s attack %s %s\n", unit->getCivilization()->getNameAdjective(), unit->getTemplate()->getName(), defender->getCivilization()->getNameAdjective(), defender->getTemplate()->getName());
			ASSERT( defender->getCivilization() == city->getCivilization() );
			int attack = unit->getTemplate()->getAirAttack();
			VI_log("attack: %d\n", attack);
			VI_log("defence: %d\n", best_defence);
			int total = attack + best_defence;
			stringstream text;
			bool show_window = false;
			bool attacker_wins = false;
			int r = rand() % total;
			if( r < attack ) {
				// attacker wins!
				if( unit->getCivilization() == player ) {
					text << "You have shot down enemy ";
					text << defender->getTemplate()->getName();
					text << ".";
					show_window = true;
				}
				else if( defender->getCivilization() == player ) {
					text << city->getName() << ": Your ";
					text << defender->getTemplate()->getName();
					text << " have been shot down by ";
					text << unit->getCivilization()->getNameAdjective();
					text << " ";
					text << unit->getTemplate()->getName();
					text << "!";
					show_window = true;
				}
				attacker_wins = true;
			}
			else {
				// attacker loses
				if( unit->getCivilization() == player ) {
					text << "You were shot down by enemy ";
					text << defender->getTemplate()->getName();
					text << ".";
					show_window = true;
				}
				else if( defender->getCivilization() == player ) {
					text << city->getName() << ": Your ";
					text << defender->getTemplate()->getName();
					text << " have been victorious against an attack by ";
					text << unit->getCivilization()->getNameAdjective();
					text << " ";
					text << unit->getTemplate()->getName();
					text << ".";
					show_window = true;
				}
			}
			if( show_window ) {
				InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), 64, 16, infowindow_def_w_c, 180);
				window->doModal();
				delete window;
			}
			// delete the units afterwards, so messages about wiping out civilizations comes after the message about winning or losing the attack
			if( attacker_wins ) {
				delete defender;
				defender = NULL;
			}
			else {
				if( unit->getCivilization() == player && unit == active_unit ) {
					active_unit = NULL;
					updateUnitButtons();
				}
				delete unit;
				unit = NULL;
				return; // attack destroyed, so quit
			}
		}
	}

	ASSERT( unit != NULL );
	if( unit->getTemplate()->getNuclearType() == UnitTemplate::NUCLEARTYPE_NONE ) {
		VI_log("conventional attack\n");
		ASSERT( unit->getTemplate()->getBombard() > 0 );
		stringstream text;
		bool success = false;
		int size = city->getSize();
		if( size > 1 && unit->getTemplate()->getBombardPower() > 100 && rand() % 2 == 0 ) {
			// population damage
			int r = rand() % 100;
			VI_log("r = %d\n", r);
			if( r < unit->getTemplate()->getBombard() ) {
				success = true;
				size--;
				ASSERT( size >= 1 );
				int old_pop = city->getPopulation();
				city->setSize(size);
				VI_log("bombing reduces city of %s to size %d\n", city->getName(), size);
				int new_pop = city->getPopulation();
				if( city->getCivilization() == player || unit->getCivilization() == player ) {
					// inform player
					string havehas = is_missile ? "has" : "have";
					if( city->getCivilization() == player ) {
						text << unit->getCivilization()->getNameAdjective() << " " << unit->getTemplate()->getName() << " " << havehas << " killed " << (old_pop - new_pop) << " citizens of " << city->getName() << "!";
					}
					else {
						text << "Your " << unit->getTemplate()->getName() << " " << havehas << " killed " << (old_pop - new_pop) << " citizens of " << city->getName() << "!";
					}
				}
			}
		}
		else {
			// improvement damage
			const Improvement *destroy_improvement = NULL;
			int r = rand() % 100;
			VI_log("r = %d\n", r);
			if( r < unit->getTemplate()->getBombard() ) {
				int bombard_power = unit->getTemplate()->getBombardPower();
				vector<const Improvement *> candidates;
				for(int i=0;i<city->getNImprovements();i++) {
					const Improvement *improvement = city->getImprovement(i);
					if( improvement->isImmuneFromBombing() )
						continue;
					if( improvement->getCost() <= bombard_power ) {
						candidates.push_back(improvement);
					}
				}
				if( candidates.size() > 0 ) {
					int r = rand() % candidates.size();
					destroy_improvement = candidates.at(r);
					if( player != NULL && player->isExplored(city->getX(), city->getY()) ) {
						/*Effect *effect = new Effect(this->fireEntity, city->getX(), city->getY(), fire_effect_duration_c);
						this->effects.push_back(effect);*/
						this->addFireEffect(city->getPos());
					}
				}
				else {
					VI_log("no candidates\n");
				}
			}
			if( destroy_improvement != NULL ) {
				success = true;
				VI_log("bombard destroys %s\n", destroy_improvement->getName());
				city->removeImprovement(destroy_improvement);
			}
			if( city->getCivilization() == player || unit->getCivilization() == player ) {
				// inform player
				if( destroy_improvement != NULL ) {
					string havehas = is_missile ? "has" : "have";
					if( city->getCivilization() == player ) {
						text << unit->getCivilization()->getNameAdjective() << " " << unit->getTemplate()->getName() << " " << havehas << " destroyed " << city->getName() << "'s " << destroy_improvement->getName() << "!";
					}
					else {
						text << "Your " << unit->getTemplate()->getName() << " " << havehas << " destroyed the city's " << destroy_improvement->getName() << "!";
					}
				}
			}
		}
		if( city->getCivilization() == player || unit->getCivilization() == player ) {
			// inform player
			mainGamestate->playSound(SOUND_WOODBRK); // TODO better sound
			if( !success ) {
				if( city->getCivilization() == player ) {
					if( is_missile ) {
						text << unit->getCivilization()->getNameAdjective() << " " << unit->getTemplate()->getName() << " lands on " << city->getName() << ", but nothing is destroyed.";
					}
					else {
						text << unit->getCivilization()->getNameAdjective() << " " << unit->getTemplate()->getName() << " bomb " << city->getName() << ", but nothing is destroyed.";
					}
				}
				else {
					if( is_missile ) {
						text << "The missile was unsuccessful.";
					}
					else {
						text << "The air raid was unsuccessful.";
					}
				}
			}
			InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), 64, 16, infowindow_def_w_c, 180);
			window->doModal();
			delete window;
		}
	}
	else {
		VI_log("nuclear attack\n");
		ASSERT( is_missile );
		if( player != NULL && player->isExplored(city->getX(), city->getY()) ) {
			/*Effect *effect = new Effect(this->fireEntity, city->getX(), city->getY(), fire_effect_duration_c);
			this->effects.push_back(effect);*/
			this->addFireEffect(city->getPos());
		}
		int percent_chance = unit->getTemplate()->getNuclearType() == UnitTemplate::NUCLEARTYPE_FISSION ? 25 : 50;
		int pop_damage = unit->getTemplate()->getNuclearType() == UnitTemplate::NUCLEARTYPE_FISSION ? 1 : 3;
		vector<const Improvement *> destroyed;
		for(int i=0;i<city->getNImprovements();i++) {
			const Improvement *improvement = city->getImprovement(i);
			if( improvement->isImmuneFromBombing() )
				continue;
			int r = rand() % 100;
			if( r < percent_chance ) {
				destroyed.push_back(improvement);
			}
		}
		for(vector<const Improvement *>::iterator iter = destroyed.begin(); iter != destroyed.end(); ++iter) {
			const Improvement *improvement = *iter;
			VI_log("nuke destroys %s\n", improvement->getName());
			city->removeImprovement(improvement);
		}
		bool bombshelter_helped = false;
		if( city->hasImprovement(IMPROVEMENT_BOMBSHELTER) ) {
			if( rand() % 2 == 0 ) {
				VI_log("bomb shelter helped against nuclear attack\n");
				bombshelter_helped = true;
				pop_damage--;
			}
		}
		int size = city->getSize();
		int old_pop = city->getPopulation();
		ASSERT( pop_damage >= 0 );
		size -= pop_damage;
		size = max(size, 1);
		city->setSize(size);
		VI_log("nuke reduces city of %s to size %d\n", city->getName(), size);
		int new_pop = city->getPopulation();
		//if( city->getCivilization() == player || unit->getCivilization() == player )
		{
			// inform player
			mainGamestate->playSound(SOUND_NUKE_EXPLOSION);
			this->centre(city->getPos());
			stringstream text;

			if( city->getCivilization() == player ) {
				text << unit->getCivilization()->getName() << " have nuked the city of " << city->getName() << "!\n";
			}
			else if( unit->getCivilization() == player ) {
				text << "You have nuked the " << city->getCivilization()->getNameAdjective() << " city of " << city->getName() << "!\n";
			}
			else {
				text << unit->getCivilization()->getName() << " have nuked the " << city->getCivilization()->getNameAdjective() << " city of " << city->getName() << "!\n";
			}
			if( new_pop != old_pop ) {
				text << "\nDeaths: " << (old_pop - new_pop) << "\n";
			}
			if( bombshelter_helped ) {
				text << "(Bomb shelter helped against attack.)\n";
			}
			if( destroyed.size() > 0 ) {
				text << "\nImprovements Destroyed:\n\n";
				for(vector<const Improvement *>::iterator iter = destroyed.begin(); iter != destroyed.end(); ++iter) {
					const Improvement *improvement = *iter;
					text << improvement->getName() << "\n";
				}
			}

			int window_h = destroyed.size() > 0 ? 440 : 180;
			InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), 64, 16, infowindow_def_w_c, window_h);
			window->doModal();
			delete window;
		}
	}

	if( is_missile ) {
		if( unit->getCivilization() == player && unit == active_unit ) {
			active_unit = NULL;
			updateUnitButtons();
		}
		delete unit;
		unit = NULL;
	}
	else {
		unit->useMove(1);
	}
}

bool MainGamestate::askDeclareWar(const Civilization *civilization) {
	stringstream text;
	text << "Are you sure you wish to declare war on " << civilization->getName() << "?";
	if( InfoWindow::confirm(game_g->getGraphicsEnvironment(), this->getPanel(), game_g->getFont(), game_g->getPanelTexture(), text.str(), "Declare War", "Cancel") ) {
		Relationship *relationship = this->findRelationship(civilization, player);
		relationship->setStatus(Relationship::STATUS_WAR);
		return true;
	}
	return false;
}

void MainGamestate::scrollView(int time) {
	if( game_g->getGraphicsEnvironment()->isActive() ) {
		int m_x = 0, m_y = 0;
		bool m_buttons[3];
		game_g->getGraphicsEnvironment()->getMouseInfo(&m_x, &m_y, m_buttons);

		float scroll_speed = 10.0f;
		float scroll_dist = scroll_speed * time / CLOCKS_PER_SEC;
		const int gap = 8;
		//float scroll_dist = 0.02f;
		bool moved = false;
		//VI_log("mouse pos: %d, %d\n", m_x, m_y);
		//VI_log("time %d, scroll dist: %f\n", time, scroll_dist);
		if( m_x <= gap ) {
			//this->map_pos_x = max(this->map_pos_x, 0);
			moved = true;
			if( view_3d ) {
				VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
				vp->move(-scroll_dist, 0.0f, 0.0f);
			}
			else
				this->map_pos_x -= scroll_dist;
		}
		else if( m_x >= game_g->getGraphicsEnvironment()->getWidth()-1-gap ) {
			//float map_limit_x = max(0, map->getWidth() - ((float)genv->getWidth())/(float)this->sq_width);
			//this->map_pos_x = min(this->map_pos_x, map_limit_x);
			moved = true;
			if( view_3d ) {
				VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
				vp->move(scroll_dist, 0.0f, 0.0f);
			}
			else
				this->map_pos_x += scroll_dist;
		}
		if( m_y <= gap ) {
			//this->map_pos_y = max(this->map_pos_y, 0);
			moved = true;
			if( view_3d ) {
				VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
				vp->move(0.0f, 0.0f, -scroll_dist);
			}
			else
				this->map_pos_y -= scroll_dist;
		}
		else if( m_y >= game_g->getGraphicsEnvironment()->getHeight()-1-gap ) {
			//float map_limit_y = max(0, map->getHeight() - ((float)genv->getHeight())/(float)this->sq_height);
			//this->map_pos_y = min(this->map_pos_y, map_limit_y);
			moved = true;
			if( view_3d ) {
				VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
				vp->move(0.0f, 0.0f, scroll_dist);
			}
			else
				this->map_pos_y += scroll_dist;
		}
		if( moved )
			checkMapView();
	}
}

void MainGamestate::update(int time) {
	time = min(time, (int)(0.05f * CLOCKS_PER_SEC));
	//time = 25;
	//VI_log("%d\n", time);
	if( view_3d ) {
		smokeEntity_3d->update(time, false);
		fireEntity_3d->update(time, false);
	}
	else {
		smokeEntity_2d->update(time, false);
		fireEntity_2d->update(time, false);
	}
	/*if( VI_keyany() ) {
	VI_log("!\n");
	}*/
	int m_x = 0, m_y = 0;
	bool m_buttons[3];
	game_g->getGraphicsEnvironment()->getMouseInfo(&m_x, &m_y, m_buttons);
	game_g->getGraphicsEnvironment()->getMouseClick(m_buttons); // needed to avoid bug where city window reopens as soon as closed, if city is at same location as CityWindow's Okay button!

	this->scrollView(time);

	if( this->isGUILocked() ) {
		return;
	}

	// check for moving units (TODO: speed up)
	//int time_s = clock();
	for(int i=0;i<player->getNUnits();i++) {
		const Unit *unit = player->getUnit(i);
		if( unit->isMoving() ) {
			this->endturnButton->setVisible(false);
			return;
		}
	}
	//VI_log("### %d\n", clock() - time_s);

	// update unit buttons (TODO: avoid doing this every frame?)
	updateUnitButtons();

	// update Game buttons (TODO: avoid doing this every frame?)
	bool buttons_visible = this->phase == PHASE_PLAYERTURN;
	this->endturnButton->setVisible(buttons_visible);
	this->newgameButton->setVisible(buttons_visible);
	this->loadgameButton->setVisible(buttons_visible);
	this->savegameButton->setVisible(buttons_visible);
	this->citiesAdvisorButton->setVisible(buttons_visible);
	this->unitsAdvisorButton->setVisible(buttons_visible);
	this->civsAdvisorButton->setVisible(buttons_visible);
	this->techsAdvisorButton->setVisible(buttons_visible);
	this->scoresAdvisorButton->setVisible(buttons_visible);
	this->mapButton->setVisible(buttons_visible);
	this->gridButton->setVisible(buttons_visible);
	this->zoomOutButton->setVisible(buttons_visible);
	this->zoomResetButton->setVisible(buttons_visible);
	this->zoomInButton->setVisible(buttons_visible);
	this->viewButton->setVisible(buttons_visible);

	if( phase == PHASE_CPUTURN ) {
		// do Unit based AI here...
		bool all_done = true;
		int count = 0;
		for(vector<Civilization *>::iterator iter = civilizations.begin(); iter != civilizations.end() && all_done; ++iter) {
			Civilization *civilization = *iter;
			if( civilization == player || civilization->doneAI() ) {
				continue;
			}
			// do missiles first
			for(int i=civilization->getNUnits()-1;i>=0 && all_done;i--) { // go backwards, to allow for deletion
				Unit *unit = civilization->getUnit(i);
				if( !unit->getTemplate()->isMissile() )
					continue;
				if( unit->hasMovesLeft() ) {
					if( unit->doAI() ) { // unit may be deleted (e.g., loses a battle)!
						all_done = false;
					}
					if( this->isGUILocked() ) {
						// need to recheck in case set by AI
						return;
					}
				}
			}
			// do all remaining units
			for(int i=civilization->getNUnits()-1;i>=0 && all_done;i--) { // go backwards, to allow for deletion
				Unit *unit = civilization->getUnit(i);
				if( unit->hasMovesLeft() ) {
					if( unit->doAI() ) { // unit may be deleted (e.g., loses a battle)!
						all_done = false;
					}
					if( this->isGUILocked() ) {
						// need to recheck in case set by AI
						return;
					}
					count++;
					if( count == 20 ) {
						all_done = false; // force an update to the screen drawing / UI occasionally, to avoid long pauses!
					}
				}
			}
			if( all_done ) {
				civilization->setDoneAI();
			}
		}
		if( all_done ) {
			this->endTurn();
			return;
		}
	}
	if( phase == PHASE_ENDTURN ) {
		// If the main GUI is locked, we must be waiting on user input, so we temporarily suspend updating (otherwise
		// we risk, e.g., multiple requesters opening).
		// We must be careful to check the flag for every call to an update function, because any of them could open
		// a new requester!
		bool any_updated = false;
		for(vector<Civilization *>::iterator iter = civilizations.begin(); iter != civilizations.end() && !this->isGUILocked(); ++iter) {
			Civilization *civilization = *iter;
			for(int i=0;i<civilization->getNCities() && !this->isGUILocked();i++) {
				City *city = civilization->getCity(i);
				if( city->needsUpdate() ) {
					city->update();
					any_updated = true;
				}
			}
		}
		if( !this->isGUILocked() && !any_updated ) {
			//the GUI isn't locked, and nothing needs to be updated - we can pass control back to the player!
			this->beginTurn();
		}
		else {
			// still not the player's turn yet
			return;
		}
	}

	// non unit key-presses [most of these now done via button key shortcuts]:

	if( VI_keypressed(V_H) ) {
		this->home();
	}

	if( this->phase == PHASE_PLAYERTURN && m_buttons[2] ) {
		// right click disables input mode
		this->setInputMode(INPUTMODE_NORMAL);
	}

	if( game_g->getGraphicsEnvironment()->readMouseWheelUp() ) {
		//this->action(this->zoomInButton);
		this->zoom(true);
	}
	else if( game_g->getGraphicsEnvironment()->readMouseWheelDown() ) {
		//this->action(this->zoomOutButton);
		this->zoom(false);
	}

	if( this->inputMode == INPUTMODE_TRAVEL || this->inputMode == INPUTMODE_BOMBARD || this->inputMode == INPUTMODE_GOTO ) {
		Pos2D m_map_pos = this->getMapPos(m_x, m_y);
		if( map->isValid(m_map_pos.x, m_map_pos.y) ) {
			const MapSquare *square = map->getSquare(m_map_pos);
/*#ifdef _WIN32
			HCURSOR cursor = NULL;
			if( square->isTarget() )
				cursor = LoadCursor( NULL, IDC_CROSS );
			else
				cursor = LoadCursor( NULL, IDC_NO );
			SetCursor(cursor);
			SetClassLong(game_g->getGraphicsEnvironment()->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
			if( square->isTarget() )
				game_g->setCursor(cursor_target);
			else
				game_g->setCursor(cursor_block);
		}
	}

	bool request_move[8] = {false, false, false, false, false, false, false, false}; // start from numpad 1, then go clockwise
	if( this->active_unit != NULL ) {
		ASSERT( this->phase == PHASE_PLAYERTURN );
		const Pos2D pos = this->active_unit->getPos();
		if( ( VI_keypressed(V_LEFT) || VI_keypressed(V_NUMPAD4) ) /*&& pos.x > 0*/ ) {
			request_move[1] = true;
		}
		else if( ( VI_keypressed(V_RIGHT) || VI_keypressed(V_NUMPAD6) ) /*&& pos.x < map->getWidth()-1*/ ) {
			request_move[5] = true;
		}
		else if( ( VI_keypressed(V_UP) || VI_keypressed(V_NUMPAD8) ) /*&& pos.y > 0*/ ) {
			request_move[3] = true;
		}
		else if( ( VI_keypressed(V_DOWN) || VI_keypressed(V_NUMPAD2) ) /*&& pos.y < map->getHeight()-1*/ ) {
			request_move[7] = true;
		}
		else if( VI_keypressed(V_NUMPAD7) /*&& pos.x > 0 && pos.y > 0*/ ) {
			request_move[2] = true;
		}
		else if( VI_keypressed(V_NUMPAD9) /*&& pos.x < map->getWidth()-1 && pos.y > 0*/ ) {
			request_move[4] = true;
		}
		else if( VI_keypressed(V_NUMPAD3) /*&& pos.x < map->getWidth()-1 && pos.y < map->getHeight()-1*/ ) {
			request_move[6] = true;
		}
		else if( VI_keypressed(V_NUMPAD1) /*&& pos.x > 0 && pos.y < map->getHeight()-1*/ ) {
			request_move[0] = true;
		}
		else if( VI_keypressed(V_C) ) {
			centreUnit();
		}
		// actual processing of moves done later
	}

	if( ( m_buttons[0] || m_buttons[2] ) && this->phase == PHASE_PLAYERTURN && game_g->getGraphicsEnvironment()->isActive() ) {
		//ASSERT( !mainGUILocked );
		ASSERT( !this->isGUILocked() );
		Pos2D m_map_pos = this->getMapPos(m_x, m_y);
		map->reduceToBase(&m_map_pos.x, &m_map_pos.y); // important!
		VI_log("clicked %d, %d (%d, %d)\n", m_map_pos.x, m_map_pos.y, m_buttons[0], m_buttons[2]);
		bool done_click = false;
		if( !map->isValid(m_map_pos.x, m_map_pos.y) )
			done_click = true;

		if( !done_click && inputMode == INPUTMODE_TRAVEL && m_buttons[0] ) {
			ASSERT( active_unit != NULL );
			const MapSquare *square = map->getSquare(m_map_pos);
			//if( square->isLand() && !square->hasEnemies(player) ) {
			if( square->isTarget() ) {
				// need to call setInputMode after calling isTarget(), as setting the inputmode resets the targets
				// but we also call it now rather than later, so the crosshairs pointer isn't showing when we display request windows
				setInputMode(INPUTMODE_NORMAL);
				done_click = true;
				ASSERT( square->isLand() );
				//ASSERT( !square->hasForeignUnits(player) );
				bool travel_ok = true;
				const Civilization *enemy_civ = this->getMap()->findCivilizationAt(m_map_pos);
				if( enemy_civ == player )
					enemy_civ = NULL;
				else if( enemy_civ != NULL ) {
					const Relationship *relationship = this->findRelationship(enemy_civ, player);
					if( relationship->getStatus() != Relationship::STATUS_PEACE ) {
						enemy_civ = NULL;
					}
				}
				if( enemy_civ != NULL ) {
					// break alliance?
					if( !askDeclareWar(enemy_civ) ) {
						travel_ok = false;
					}
				}
				if( travel_ok ) {
					this->moveTo(active_unit, m_map_pos, true);
					if( active_unit != NULL ) {
						//active_unit->useMoves(); // now done in moveTo()
						active_unit = NULL;
					}
				}
			}
			else {
				setInputMode(INPUTMODE_NORMAL);
			}
		}
		else if( !done_click && inputMode == INPUTMODE_BOMBARD && m_buttons[0] ) {
			ASSERT( active_unit != NULL );
			ASSERT( active_unit->getTemplate()->getBombard() > 0 || active_unit->getTemplate()->getNuclearType() != UnitTemplate::NUCLEARTYPE_NONE );
			const MapSquare *square = map->getSquare(m_map_pos);
			if( square->isTarget() ) {
				// need to call setInputMode after calling isTarget(), as setting the inputmode resets the targets
				// but we also call it now rather than later, so the crosshairs pointer isn't showing when we display request windows
				setInputMode(INPUTMODE_NORMAL);
				done_click = true;
				const City *city = square->getCity();
				ASSERT( city != NULL );
				const Relationship *relationship = this->findRelationship(city->getCivilization(), active_unit->getCivilization());
				bool bomb_ok = true;
				if( relationship->getStatus() == Relationship::STATUS_PEACE ) {
					if( !askDeclareWar(city->getCivilization()) ) {
						bomb_ok = false;
					}
				}
				if( bomb_ok ) {
					//this->bomb(this->active_unit, square);
					this->bomb(this->active_unit, m_map_pos);
					if( active_unit != NULL && !active_unit->hasMovesLeft() ) {
						this->active_unit = NULL;
						updateUnitButtons();
					}
					else {
						centreUnit();
					}
				}
			}
			else {
				setInputMode(INPUTMODE_NORMAL);
			}
		}
		else if( !done_click && inputMode == INPUTMODE_GOTO && m_buttons[0] ) {
			ASSERT( active_unit != NULL );
			ASSERT( !active_unit->getTemplate()->isAir() > 0 );
			ASSERT( !active_unit->isAutomated() );
			const MapSquare *square = map->getSquare(m_map_pos);
			if( square->isTarget() ) {
				// need to call setInputMode after calling isTarget(), as setting the inputmode resets the targets
				// but we also call it now rather than later, so the crosshairs pointer isn't showing when we display request windows
				setInputMode(INPUTMODE_NORMAL);
				done_click = true;
				active_unit->setGotoTarget(m_map_pos);
				this->active_unit->automate();
				Pos2D dummy;
				if( this->active_unit->hasGotoTarget(&dummy) ) { // only do this is we haven't reached destination yet
					this->cache_active_unit = this->active_unit; // so we can cope with additional turns of this unit
					this->active_unit = NULL;
					this->updateUnitButtons();
				}
				else if( !active_unit->hasMovesLeft() ) {
					// important - if we reached our destination straight away, but used up all the moves, this should no longer be the active unit!
					this->active_unit = NULL;
					this->updateUnitButtons();
				}

				if( cache_active_unit->isMoving() ) {
					// wait for smooth movement to finish!
					return;
				}
			}
			else {
				setInputMode(INPUTMODE_NORMAL);
			}
		}

		if( !done_click && m_buttons[0] ) {
			City *city = map->findCity(m_map_pos.x, m_map_pos.y);
			if( city != NULL ) {
				if( SPY || city->getCivilization() == player ) {
					VI_log("CITY!\n");
					new CityWindow(this, city);
					done_click = true;
				}
			}
		}
		if( !done_click && m_buttons[0] && player->isExplored(m_map_pos.x, m_map_pos.y) ) {
			stringstream terrain_info;
			const MapSquare *square = map->getSquare(m_map_pos);
			terrain_info << "Terrain: " << square->getName();
			MapSquare::Road road = square->getRoad();
			if( road == MapSquare::ROAD_BASIC )
				terrain_info << " (Road)";
			else if( road == MapSquare::ROAD_RAILWAYS )
				terrain_info << " (Railways)";
			terrain_info << ".\n";
			if( /*CHEAT &&*/ square->getProgressRoad() > 0 ) {
				if( road == MapSquare::ROAD_NONE )
					terrain_info << "Road progress: ";
				else if( road == MapSquare::ROAD_BASIC )
					terrain_info << "Railways progress: ";
				else {
					ASSERT( false );
				}
				terrain_info << square->getProgressRoad() << " / " << square->getRoadCost() << ".\n";
			}
			terrain_info << "(" << m_map_pos.x << "," << m_map_pos.y << ")";
			City *city = map->findCity(m_map_pos.x, m_map_pos.y);
			const vector<Unit *> *units = map->findUnitsAt(m_map_pos.x, m_map_pos.y);
			// don't show units at enemy city
			if( city == NULL && units->size() > 0 ) {
				//const int window_h_c = 512;
				int window_h_c = game_g->getGraphicsEnvironment()->getHeight() - infowindow_def_y_c - 32;
				Unit *first_unit = units->at(0);
				//if( first_unit->getCivilization() == player )
				{
					done_click = true;
					Unit *selected_unit = NULL;
					enum SelectType {
						SELECT_NONE = 0,
						SELECT_ONE = 1,
						SELECT_ALL = 2,
						FORTIFY_ALL = 3
					};
					SelectType selectType = SELECT_NONE;
					const int window_width_c = 500;
					bool show_single_info_window = false;
					if( first_unit->getCivilization() == player )
						show_single_info_window = units->size() == 1 && ( first_unit == active_unit || !first_unit->hasMovesLeft() );
					else
						show_single_info_window = true;
					if( first_unit->getCivilization() == player && units->size() > 1 ) {
						// choose which unit
						string buttons[] = {"Activate Unit", "Activate All", "Fortify All", "Cancel"};
						InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), terrain_info.str().c_str(), game_g->getFont(), game_g->getPanelTexture(), buttons, 4, infowindow_def_x_c, infowindow_def_y_c, window_width_c, window_h_c);
						/*InfoWindow *window = NULL;
						if( units->size() > 1 ) {
						string buttons[] = {"Activate Unit", "Activate All", "Cancel"};
						window = new InfoWindow(this->getPanel(), "", buttons, 3, infowindow_def_x_c, infowindow_def_y_c, infowindow_def_w_c, window_h_c);
						}
						else {
						string buttons[] = {"Okay"};
						window = new InfoWindow(this->getPanel(), "", buttons, 1, infowindow_def_x_c, infowindow_def_y_c, infowindow_def_w_c, window_h_c);
						}*/
						VI_Panel *windowPanel = window->getPanel();

						vector<string> unit_names = this->getUnitNames(units);
						ASSERT( unit_names.size() > 0 );
						//VI_Listbox *listbox = VI_createListbox(window_width_c - 32, window_h_c - 96, &*unit_names.begin(), unit_names.size(), game_g->getFont());
						VI_Listbox *listbox = VI_createListbox(window_width_c - 32, window_h_c - 112, &*unit_names.begin(), unit_names.size(), game_g->getFont());
						listbox->setHasInputFocus(true);
						//listbox->setForeground(1.0f, 1.0f, 0.0f);
						listbox->setForeground(1.0f, 1.0f, 1.0f);
						listbox->setActive(0);
						// find first unit that still has moves
						for(int i=0;i<units->size();i++) {
							Unit *unit = units->at(i);
							if( unit->hasMovesLeft() ) {
								listbox->setActive(i);
								break;
							}
						}
						windowPanel->add(listbox, 16, 64);

						int res = window->doModal();
						/*if( units->size() == 1 ) {
						// keep with the selected unit
						}
						else*/ if( res == 0 ) {
							int active = listbox->getActive();
							ASSERT( active >= 0 && active < units->size() );
							selected_unit = units->at(active);
							selectType = SELECT_ONE;
						}
						else if( res == 1 ) {
							selected_unit = NULL;
							selectType = SELECT_ALL;
						}
						else if( res == 2 ) {
							selected_unit = NULL;
							selectType = FORTIFY_ALL;
						}
						else {
							selected_unit = NULL;
							selectType = SELECT_NONE;
						}
						delete window;
						delete listbox;
					}
					else {
						bool activate = true;
						if( first_unit->getCivilization() == player && ( first_unit->getStatus() == Unit::STATUS_BUILDING_ROAD || first_unit->getStatus() == Unit::STATUS_BUILDING_RAILWAYS ) ) {
							// warn
							stringstream text;
							text << "This " << first_unit->getTemplate()->getName() << " unit is currently building " << (first_unit->getStatus() == Unit::STATUS_BUILDING_ROAD ? "a road" : "railways") << ". Are you sure you wish to wake it?";
							if( !InfoWindow::confirm(game_g->getGraphicsEnvironment(), this->getPanel(), game_g->getFont(), game_g->getPanelTexture(), text.str(), "Activate", "Cancel") ) {
								activate = false;
							}
						}
						if( activate ) {
							selected_unit = first_unit;
							selectType = SELECT_ONE;
						}
						else {
							selected_unit = NULL;
							selectType = SELECT_NONE;
							show_single_info_window = false;
						}
					}
					if( first_unit->getCivilization() != player ) {
						// nothing
					}
					else if( selectType == SELECT_ONE ) {
						ASSERT( selected_unit != NULL );
						if( selected_unit->hasMovesLeft() ) {
							this->activateUnit(selected_unit, false);
						}
						else {
							//selected_unit->setStatus(Unit::STATUS_NORMAL);
							this->wakeupUnit(selected_unit);
						}
					}
					else if( selectType == SELECT_ALL ) {
						bool found = false;
						for(vector<Unit *>::const_iterator iter = units->begin();iter != units->end();++iter) {
							Unit *unit = *iter;
							if( !found && unit->hasMovesLeft() ) {
								// activate the first one we can
								found = true;
								this->activateUnit(unit, false);
							}
							else {
								// deal with the rest
								//unit->setStatus(Unit::STATUS_NORMAL);
								this->wakeupUnit(unit);
							}
						}
					}
					else if( selectType == FORTIFY_ALL ) {
						for(vector<Unit *>::const_iterator iter = units->begin();iter != units->end();++iter) {
							Unit *unit = *iter;
							unit->setStatus(Unit::STATUS_FORTIFIED);
							if( unit == this->active_unit ) {
								this->active_unit = NULL;
								this->updateUnitButtons();
							}
						}
					}

					// now show info window if there was only one unit (we do afterwards, to allow for activation of workers), or if viewing enemy player
					if( ( first_unit->getCivilization() != player || units->size() == 1 ) && show_single_info_window ) {
						string buttons[] = {"Okay"};
						InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), terrain_info.str().c_str(), game_g->getFont(), game_g->getPanelTexture(), buttons, 1, infowindow_def_x_c, infowindow_def_y_c, window_width_c, window_h_c);
						VI_Panel *windowPanel = window->getPanel();

						vector<string> unit_names = this->getUnitNames(units);
						ASSERT( unit_names.size() > 0 );
						//VI_Listbox *listbox = VI_createListbox(window_width_c - 32, window_h_c - 96, &*unit_names.begin(), unit_names.size(), game_g->getFont());
						VI_Listbox *listbox = VI_createListbox(window_width_c - 32, window_h_c - 112, &*unit_names.begin(), unit_names.size(), game_g->getFont());
						listbox->setHasInputFocus(true);
						//listbox->setForeground(1.0f, 1.0f, 0.0f);
						listbox->setForeground(1.0f, 1.0f, 1.0f);
						listbox->setActive(0);
						windowPanel->add(listbox, 16, 64);

						window->doModal();
						delete window;
						//delete listbox;
					}

					/*bool made_active = false;
					for(vector<Unit *>::const_iterator iter = units->begin();iter != units->end(); ++iter) {
					Unit *unit = *iter;
					if( unit->hasMovesLeft() ) {
					if( made_active ) {
					unit->setStatus(Unit::STATUS_NORMAL);
					// un-fortify the remainder
					}
					else {
					// make the first available unit active
					activateUnit(unit, false);
					made_active = true;
					}
					}
					}*/
				}
			}
			else {
				done_click = true;
				InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), terrain_info.str().c_str(), game_g->getFont(), game_g->getPanelTexture());
				window->doModal();
				delete window;
			}
			//delete units;
		}

		if( m_buttons[2] && this->active_unit != NULL ) {
			// right click to move?
			ASSERT( this->phase == PHASE_PLAYERTURN );
			const Pos2D pos = this->active_unit->getPos();
			if( pos != m_map_pos ) {
				Pos2D diff = m_map_pos - pos;
				this->map->reduceDiff(&diff.x, &diff.y);
				if( abs(diff.x) <= 1 && abs(diff.y) <= 1 ) {
					VI_log("right click move requested: %d, %d\n", diff.x, diff.y);
					if( diff.x == -1 && diff.y == 1 )
						request_move[0] = true;
					else if( diff.x == -1 && diff.y == 0 )
						request_move[1] = true;
					else if( diff.x == -1 && diff.y == -1 )
						request_move[2] = true;
					else if( diff.x == 0 && diff.y == -1 )
						request_move[3] = true;
					else if( diff.x == 1 && diff.y == -1 )
						request_move[4] = true;
					else if( diff.x == 1 && diff.y == 0 )
						request_move[5] = true;
					else if( diff.x == 1 && diff.y == 1 )
						request_move[6] = true;
					else if( diff.x == 0 && diff.y == 1 )
						request_move[7] = true;
				}
			}
		}
		//if( mainGUILocked ) {
		if( this->isGUILocked() ) {
			return;
		}
	}

	if( this->active_unit != NULL ) {
		ASSERT( this->phase == PHASE_PLAYERTURN );
		Pos2D pos = this->active_unit->getPos();
		bool move = false;
		if( request_move[1] ) {
			pos.x--;
			move = true;
		}
		else if( request_move[5] ) {
			pos.x++;
			move = true;
		}
		else if( request_move[3] ) {
			pos.y--;
			move = true;
		}
		else if( request_move[7] ) {
			pos.y++;
			move = true;
		}
		else if( request_move[2] ) {
			pos.x--;
			pos.y--;
			move = true;
		}
		else if( request_move[4] ) {
			pos.x++;
			pos.y--;
			move = true;
		}
		else if( request_move[6] ) {
			pos.x++;
			pos.y++;
			move = true;
		}
		else if( request_move[0] ) {
			pos.x--;
			pos.y++;
			move = true;
		}
		if( move ) {
			if( !this->map->isValid(pos.x, pos.y) ) {
				move = false;
			}
		}
		if( move ) {
			map->reduceToBase(&pos.x, &pos.y);
			bool checkEnemies = ( this->active_unit->getTemplate()->getAttack() == 0 );
			// check for enemies blocking, if we can't attack
			if( this->active_unit->canMoveTo( pos, checkEnemies ) ) {
				bool move_ok = true;
				if( !checkEnemies ) {
					// check for attacking enemies (TODO: improve performance?
					/*Civilization *enemy_civ = NULL;
					const vector<Unit *> *units = this->getMap()->findUnitsAt(pos.x, pos.y);
					if( units->size() > 0 && units->at(0)->getCivilization() != this->active_unit->getCivilization() ) {
						const Relationship *relationship = this->findRelationship(units->at(0)->getCivilization(), this->active_unit->getCivilization());
						if( relationship->getStatus() == Relationship::STATUS_PEACE ) {
							enemy_civ = units->at(0)->getCivilization();
						}
					}
					if( enemy_civ == NULL ) {
						const City *city = this->getMap()->findCity(pos.x, pos.y);
						if( city != NULL && city->getCivilization() != this->active_unit->getCivilization() ) {
							const Relationship *relationship = this->findRelationship(city->getCivilization(), this->active_unit->getCivilization());
							if( relationship->getStatus() == Relationship::STATUS_PEACE ) {
								enemy_civ = city->getCivilization();
							}
						}
					}
					//delete units;
					*/
					const Civilization *enemy_civ = this->getMap()->findCivilizationAt(pos);
					if( enemy_civ == player )
						enemy_civ = NULL;
					else if( enemy_civ != NULL ) {
						const Relationship *relationship = this->findRelationship(enemy_civ, player);
						if( relationship->getStatus() != Relationship::STATUS_PEACE ) {
							enemy_civ = NULL;
						}
					}

					if( enemy_civ != NULL ) {
						// break alliance?
						if( !askDeclareWar(enemy_civ) ) {
							move_ok = false;
						}
					}
				}

				if( move_ok ) {
					//this->active_unit->moveTo(pos);
					this->moveTo(active_unit, pos, false); // unit may be deleted (e.g., loses a battle), in which case, active_unit will be set to NULL
					bool smooth_moving = active_unit != NULL && active_unit->isMoving();
					if( active_unit != NULL && !active_unit->hasMovesLeft() ) {
						this->active_unit = NULL;
						updateUnitButtons();
					}
					else if( active_unit != NULL ) {
						centreUnit();
					}
					//if( mainGUILocked ) {
					if( this->isGUILocked() ) {
						return;
					}
					if( smooth_moving ) {
						return;
					}
				}
			}
			else {
				mainGamestate->playSound(SOUND_BLIP);
			}
		}
	}

	if( this->active_unit == NULL && this->phase == PHASE_PLAYERTURN )  {
		bool found_unit = false;
		ASSERT( !this->isGUILocked() );
		// find new active unit
		//VI_log("find new active unit\n");
		if( cache_active_unit != NULL ) {
			// search after cache first
			bool found_cache = false;
			for(int i=0;i<player->getNUnits() && !found_unit;i++) {
				Unit *unit = player->getUnit(i);
				if( unit == cache_active_unit ) {
					found_cache = true;
					// if non-automated, we want to move onto the next unit (so that Wait works), otherwise we stick with this one, so that the automated unit can do multiple moves
					/*if( unit->isAutomated() && unit->hasMovesLeft() && unit->getStatus() == Unit::STATUS_NORMAL )
						; // keep with this unit
					else*/
					if( !unit->isAutomated() )
						continue;
				}
				if( found_cache && unit->hasMovesLeft() && unit->getStatus() == Unit::STATUS_NORMAL ) {
					if( unit->isAutomated() ) {
						/*while( unit->hasMovesLeft() && unit->automate() ) {
						}*/
						/// ?
						centre( unit->getPos() );
						this->endturnButton->setVisible(false);
						unit->automate();
						cache_active_unit = unit;
						found_unit = true;
					}
					else {
						activateUnit(unit, true);
						found_unit = true;
					}
				}
			}
		}
		// either no cached unit, or need wraparound
		for(int i=0;i<player->getNUnits() && !found_unit;i++) {
			Unit *unit = player->getUnit(i);
			if( unit->hasMovesLeft() && unit->getStatus() == Unit::STATUS_NORMAL ) {
				if( unit->isAutomated() ) {
					/*while( unit->hasMovesLeft() && unit->automate() ) {
					}*/
					centre( unit->getPos() );
					unit->automate();
					this->endturnButton->setVisible(false);
					cache_active_unit = unit;
					found_unit = true;
				}
				else {
					activateUnit(unit, true);
					found_unit = true;
				}
			}
		}
		if( active_unit == NULL ) {
			// what does this do???
			/*if( InfoWindow::getActiveWindow() != NULL ) {
				delete InfoWindow::getActiveWindow();
			}*/
			if( infoWindow != NULL ) {
				delete infoWindow;
				infoWindow = NULL;
			}
		}
		// moved_units_this_turn is a flag to stop us immediately ending the turn, when the player has no units to move!
		//if( active_unit == NULL && moved_units_this_turn ) {
		if( !found_unit && moved_units_this_turn ) {
			// no more units left to move
			this->endPlayerTurn();
		}
		//else if( active_unit == NULL ) {
		else if( !found_unit ) {
			// waiting for user to end turn
			if( VI_keypressed( V_RETURN ) ) {
				// Return also ends turn, if all units are moved
				game_g->getGraphicsEnvironment()->flushInput(); // need to fix bug where requesters would disappear, due to repeated detection of Return key!
				this->endPlayerTurn();
			}
		}
	}
}

void MainGamestate::moveTo(Unit *unit, Pos2D newPos, bool travelling) {
	map->reduceToBase(&newPos.x, &newPos.y);
	ASSERT( unit->canMoveTo( newPos ) );
	ASSERT( unit->movesLeft() > 0 );

	/*int xdist = abs(newPos.x - unit->getX());
	int ydist = abs(newPos.y - unit->getY());
	ASSERT(xdist > 0 || ydist > 0);
	bool travelling = xdist > 1 || ydist > 1;*/

	const vector<Unit *> *units = map->findUnitsAt(newPos.x, newPos.y);
	Civilization *o_civ = units->size() == 0 ? NULL : units->at(0)->getCivilization();
	if( o_civ == NULL || o_civ == unit->getCivilization() ) {
		//VI_log("unit %d (%s) moves to %d, %d\n", unit, unit->getTemplate()->getName(), newPos.x, newPos.y);
		// move
		/*int ox = unit->getX();
		int oz = unit->getY();*/
		if( unit->getCivilization() == player && map->getSquare(newPos)->getRoad() != MapSquare::ROAD_RAILWAYS ) {
			mainGamestate->playSound(SOUND_FOOT);
		}
		unit->setPos(newPos);
		if( travelling )
			unit->useMoves();
		else {
			Rational movement_cost = map->getSquare(newPos.x, newPos.y)->getMovementCost(unit->getCivilization());
			unit->useMove(movement_cost);
		}
		//board->updateArmy(army);
		//if( army->getCivilization() == player g)
		/*{
		board->setArmyVisibles(ox, oz); // old pos
		board->setArmyVisibles(army->pos_x, army->pos_z); // new pos
		}*/
		//bool smooth_move = unit->getCivilization() == player && !unit->isAutomated();
		bool smooth_move = unit->getCivilization() == player;

		City *city = map->findCity(newPos.x, newPos.y);
		if( city != NULL ) {
			if( city->getCivilization() != unit->getCivilization() ) {
				// capture city
				smooth_move = false;
				/*char text[MAX_LONG_TEXT+1];
				if( army->getCivilization() == player ) {
				char *racename = Race::getNameAdjective(city->getCivilization()->race);
				sprintf(text, "We have conquered the %s city of %s!\n", racename, city->getName());
				}
				else {
				char *racename = Race::getNameAdjective(army->getCivilization()->race);
				sprintf(text, "The %s have captured the city of %s!\n", racename, city->getName());
				}*/

				//delete city;
				/*Listselect *ls = new Listselect(listselect_x, listselect_y, listselect_w, listselect_h, text, NULL, NULL, NULL, false);
				ls->doModal();*/

				bool lose_population = city->losePopulation();
				bool destroy = false;
				int size = city->getSize();
				if( lose_population && size == 1 ) {
					destroy = true;
				}

				if( unit->getCivilization() == player ) {
					//string text = "You have captured the ";
					stringstream text;
					text << "You have ";
					text << ( destroy ? "destroyed" : "captured" );
					text << " the ";
					text << city->getCivilization()->getNameAdjective();
					text << " city of ";
					text << city->getName();
					text << ".";
					//text.append("Here is a lot of long very long long long text blah blah long text ending in a new line:\n");
					//text.append("Here is another lot of long very long long long text blah blah long text, but not ending in a new line.");
					InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture());
					window->doModal();
					delete window;
				}
				else if( city->getCivilization() == player ) {
					this->centre( unit->getPos() );
					stringstream text;
					text << unit->getCivilization()->getName();
					text << " have ";
					text << ( destroy ? "destroyed" : "captured" );
					text << " the city of ";
					text << city->getName();
					text << "!";
					InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture());
					window->doModal();
					delete window;
				}

				if( player != NULL && player->isExplored(city->getX(), city->getY()) ) {
					/*Effect *effect = new Effect(mainGamestate->fireEntity, city->getX(), city->getY(), fire_effect_duration_c);
					mainGamestate->effects.push_back(effect);*/
					this->addFireEffect(city->getPos());
				}
				if( destroy ) {
					// destroy city
					delete city;
					city = NULL;
				}
				else {
					if( lose_population ) {
						// reduce population
						city->setSize(size-1);
					}
					city->capture(unit->getCivilization());
				}
			}
		}

		if( smooth_move ) {
			unit->setMoving();
		}

		//setNextArmyButton();
		/*if( map->uncover(army->pos_x, army->pos_z) ) {
		board->updateMap();
		}*/
		//city_active = NULL; // we clear to NULL, even if the new square has a city
		//V_PANEL_set_visible(changeBuildButton, city_active!=NULL);

#if 0
		if( army->getCivilization() == player && !army->movesLeft() ) {
			Army *new_army_active = nextArmy();
			if( new_army_active != NULL ) {
				//board->updateArmy(army_active);
				setArmyActive(new_army_active);
			}
			else {
				endTurn();
			}
		}
#endif

	}
	else {
		VI_log("unit %d (%s) attacks at %d, %d\n", unit, unit->getTemplate()->getName(), newPos.x, newPos.y);
		ASSERT( unit->getTemplate()->getAttack() > 0 );
		const Relationship *relationship = this->findRelationship(unit->getCivilization(), o_civ);
		ASSERT( relationship->madeContact() );
		ASSERT( relationship->getStatus() == Relationship::STATUS_WAR );
		// attack
		/*Unit *defender = NULL;
		int best_defence = 0;
		for(vector<Unit *>::const_iterator iter = units->begin(); iter != units->end(); ++iter) {
			Unit *this_unit = *iter;
			int defence = this_unit->getTemplate()->getDefence();
			// find best defender - for equal defences, choose one with lower attack (so only the weaker unit is put at risk of being destroyed)
			if( defender == NULL || defence > best_defence || ( defence == best_defence && this_unit->getTemplate()->getAttack() < defender->getTemplate()->getAttack() ) ) {
				defender = this_unit;
				best_defence = defence;
			}
		}*/
		Unit *defender = map->getSquare(newPos)->getBestDefender();
		int best_defence = defender->getTemplate()->getDefence();
		ASSERT( defender != NULL );

		best_defence *= 100; // to allow for fractional comparisons
		City *city = map->findCity(newPos.x, newPos.y);
		int city_bonus = 0;
		if( city != NULL ) {
			city_bonus = city->getDefenceBonus(unit);
			VI_log("City defence bonus %d\n", city_bonus);
			best_defence = ( best_defence * ( 100 + city_bonus ) ) / 100;
		}

		if( defender->getCivilization() == player ) {
			this->centre( defender->getPos() );
		}
		if( unit->getCivilization() == player || defender->getCivilization() == player ) {
			this->playSound(SOUND_WOODBRK);
		}
		int attack = unit->getTemplate()->getAttack() * 100;
		const float medicine_bonus_c = 1.2f;
		const Technology *medicineTech = this->findTechnology("Medicine");
		if( unit->getCivilization()->hasTechnology(medicineTech) ) {
			VI_log("attacker has medicine\n");
			attack *= medicine_bonus_c;
		}
		if( defender->getCivilization()->hasTechnology(medicineTech) ) {
			VI_log("defender has medicine\n");
			best_defence *= medicine_bonus_c;
		}
		const float satellites_bonus_c = 1.2f;
		const Technology *satellitesTech = this->findTechnology("Satellites");
		if( unit->getCivilization()->hasTechnology(satellitesTech) ) {
			VI_log("attacker has satellites\n");
			attack *= satellites_bonus_c;
		}
		if( defender->getCivilization()->hasTechnology(satellitesTech) ) {
			VI_log("defender has satellites\n");
			best_defence *= satellites_bonus_c;
		}
		if( defender->getCivilization()->hasBonus(Civilization::BONUS_MILITARY) ) {
			VI_log("defender has military bonus\n");
			best_defence *= 1.5f;
		}
		if( travelling ) {
			VI_log("defence bonus due to travelling by sea\n");
			if( city != NULL && city->hasImprovement(IMPROVEMENT_COASTALDEFENCE) ) {
				VI_log("...and with coastal defence\n");
				best_defence *= 1.5;
			}
			else
				best_defence *= 1.25;
		}
		VI_log("attack: %d\n", attack);
		VI_log("defence: %d\n", best_defence);
		int total = attack + best_defence;
		stringstream text;
		bool show_window = false;
		bool attacker_wins = false;
		int r = rand() % total;
		const Civilization *cpu = NULL;
		if( travelling && defender->getCivilization() == player ) {
			// if travelling, this should always be by sea, as we don't allow travelling by air to attack (if this is ever changed, we need to rethink this code!)
			text << unit->getCivilization()->getName() << " invade ";
			if( city != NULL ) {
				text << "the city of " << city->getName() << " ";
			}
			text << "by sea!";
			text << "\n\n";
		}
		if( r < attack ) {
			// attacker wins!
			if( unit->getCivilization() == player ) {
				text << "You have defeated the enemy ";
				text << defender->getTemplate()->getName();
				text << ".";
				show_window = true;
				cpu = defender->getCivilization();
			}
			else if( defender->getCivilization() == player ) {
				text << "Your ";
				text << defender->getTemplate()->getName();
				text << " have been destroyed by ";
				text << unit->getCivilization()->getNameAdjective();
				text << " ";
				text << unit->getTemplate()->getName();
				text << "!";
				show_window = true;
				cpu = unit->getCivilization();
			}
			attacker_wins = true;
		}
		else {
			// attacker loses
			if( unit->getCivilization() == player ) {
				text << "You were defeated by the enemy ";
				text << defender->getTemplate()->getName();
				text << ".";
				show_window = true;
				cpu = defender->getCivilization();
			}
			else if( defender->getCivilization() == player ) {
				text << "Your ";
				text << defender->getTemplate()->getName();
				text << " have been victorious against an attack by ";
				text << unit->getCivilization()->getNameAdjective();
				text << " ";
				text << unit->getTemplate()->getName();
				text << ".";
				show_window = true;
				cpu = unit->getCivilization();
			}
		}
		if( show_window ) {
			text << "\n";
			if( player->hasTechnology(medicineTech) ) {
				text << "\nYour medicine bonus: 20%";
			}
			if( player->hasTechnology(satellitesTech) ) {
				text << "\nYour satellites bonus: 20%";
			}
			if( player == defender->getCivilization() && defender->getCivilization()->hasBonus(Civilization::BONUS_MILITARY) ) {
				text << "\nYour military defence bonus: 50%";
			}
			if( city_bonus > 0 ) {
				text << "\nCity Defence Bonus: ";
				text << city_bonus;
				text << "%";
			}
			if( travelling ) {
				if( player == defender->getCivilization() )
					text << "\nYour";
				else
					text << "\nEnemy's";
				text << " sea defence bonus: ";
				if( city != NULL && city->hasImprovement(IMPROVEMENT_COASTALDEFENCE) )
					text << "50%";
				else
					text << "25%";
			}
			if( cpu->hasTechnology(medicineTech) ) {
				text << "\nEnemy's medicine bonus: 20%";
			}
			if( cpu->hasTechnology(satellitesTech) ) {
				text << "\nEnemy's satellites bonus: 20%";
			}
			if( cpu == defender->getCivilization() && defender->getCivilization()->hasBonus(Civilization::BONUS_MILITARY) ) {
				text << "\nEnemy's military defence bonus: 50%";
			}
		}
		if( show_window ) {
			InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), 64, 16, infowindow_def_w_c, 360);
			window->doModal();
			delete window;
		}
		// delete the units/cities afterwards, so messages about wiping out civilizations comes after the message about winning or losing the attack
		if( attacker_wins ) {
			delete defender;
			defender = NULL;

			if( city != NULL && city->losePopulation() ) {
				/*Effect *effect = new Effect(mainGamestate->fireEntity, city->getX(), city->getY(), fire_effect_duration_c);
				mainGamestate->effects.push_back(effect);*/
				this->addFireEffect(city->getPos());
				int size = city->getSize();
				if( size == 1 ) {
					// destroy city
					if( show_window ) {
						text.str("");
						text << "The city of " << city->getName() << " has been destroyed!";
						InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), 64, 16, infowindow_def_w_c, 180);
						window->doModal();
						delete window;
					}
					delete city;
					city = NULL;
				}
				else {
					// reduce population
					city->setSize(size-1);
				}
			}
		}
		else {
			if( unit->getCivilization() == player && unit == active_unit ) {
				active_unit = NULL;
				updateUnitButtons();
			}
			delete unit;
			unit = NULL;
		}
		if( attacker_wins && units->size() == 0 ) {
			// no defenders left (note that "units" is automatically updated, as it is a pointer to the Map's list!)
			City *city = map->findCity(newPos.x, newPos.y);
			if( city == NULL ) {
				// okay to move
				unit->setPos(newPos);
				if( travelling )
					unit->useMoves();
				else {
					Rational movement_cost = map->getSquare(newPos.x, newPos.y)->getMovementCost(unit->getCivilization());
					if( movement_cost < 1 ) {
						// must cost at least 1 full move, due to the attack
						movement_cost = 1;
					}
					unit->useMove(movement_cost);
				}
			}
			else {
				if( travelling )
					unit->useMoves();
				else {
					unit->useMove(1);
				}
			}
		}
		else if( attacker_wins ) {
			if( travelling )
				unit->useMoves();
			else {
				unit->useMove(1);
			}
		}
#if 0
		bool next = false;
		army->moves_taken++;
		if( !army->movesLeft() ) {
			next = true;
		}

		Army *army1 = army;
		//Army *army2 = Army::combineArmies(armies);
		Army *army2 = armies->at(0);
		/*float def1 = army1->unit->defence;
		float def2 = army1->unit->defence;
		def1 *= army1->getDefenceModifier();
		def2 *= army1->getDefenceModifier();
		ASSERT( def1 > 0 );
		ASSERT( def2 > 0 );*/
		float def1 = army1->getDefence(army1->unit);
		float def2 = army2->getDefence(army2->unit);
		int number1 = army1->unit->getNumber();
		int number2 = army2->unit->getNumber();
		int max_kills1 = ( number1 * army1->unit->attack ) /  def2;
		int max_kills2 = ( number2 * army2->unit->attack ) /  def1;
		/*ASSERT( max_kills1 > 0 );
		ASSERT( max_kills2 > 0 );*/
		int kills1 = max_kills1 == 0 ? 0 : rand() % max_kills1;
		int kills2 = max_kills2 == 0 ? 0 : rand() % max_kills2;
		kills2 = army1->unit->kill(kills2);
		kills1 = army2->unit->kill(kills1);
		char *racename1 = Race::getNameAdjective(army1->getCivilization()->race);
		char *racename2 = Race::getNameAdjective(army2->getCivilization()->race);
		char text[MAX_LONG_TEXT+1];
		sprintf(text, "%d %s %s vs %d %s %s:\n\n%s casualties: %d\n%s casualties: %d\n",
			number1, racename1, army1->unit->getName(),
			number2, racename2, army2->unit->getName(),
			racename1, kills2,
			racename2, kills1
			);
		bool is_player = army->getCivilization() != player;
		if( army1->unit->getNumber() == 0 ) {
			/*V_ENTITY_kill(army1->ent);
			delete army1;*/
			army1->kill();
			if( army1 == army ) {
				next = true;
			}
		}
		if( army2->unit->getNumber() == 0 ) {
			/*V_ENTITY_kill(army2->ent);
			delete army2;*/
			army2->kill();
			if( army2 == army ) {
				next = true;
			}
		}
		if( next && is_player )
			next = false;
		Listselect *ls = new Listselect(listselect_x, listselect_y, listselect_w, listselect_h, text, NULL, NULL, next ? switchToNextArmyLS : NULL, false);
		ls->doModal();
		/*if( next ) {
		setArmyActive(NULL);
		}*/
#endif
	}
	//delete units;
}

/*void Game::generateRelationships() {
ASSERT( relationships.size() == 0 );
for(int i=0;i<this->getNCivilizations();i++) {
Civilization *civ1 = this->getCivilization(i);
for(int j=i+1;j<this->getNCivilizations();j++) {
Civilization *civ2 = this->getCivilization(j);
Relationship *relationship = new Relationship(civ1, civ2);
relationships.push_back(relationship);
}
}
}*/

bool MainGamestate::checkRelationships() {
	bool ok = true;
	for(int i=0;i<this->getNCivilizations() && ok;i++) {
		const Civilization *civ1 = this->getCivilization(i);
		if( civ1->isDead() ) {
			continue;
		}
		for(int j=i+1;j<this->getNCivilizations() && ok;j++) {
			const Civilization *civ2 = this->getCivilization(j);
			if( civ2->isDead() ) {
				continue;
			}
			const Relationship *relationship = this->findRelationship(civ1, civ2);
			// findRelationship currently ASSERTS if found, but put check here just to be sure...
			if( relationship == NULL ) {
				VI_log("MainGamestate::checkRelationships not all relationships present\n");
				ok = false;
			}
		}
	}
	return ok;
}

/*void MainGamestate::removeCivilization(Civilization *civilization) {
for(vector<Civilization *>::iterator iter = civilizations.begin(); iter != civilizations.end(); ++iter) {
Civilization *this_civilization = *iter;
if( this_civilization == civilization ) {
civilizations.erase(iter);
return;
}
}
VI_log("Failed to find Civilization to remove\n");
ASSERT(false);
}*/

void MainGamestate::playSound(SoundID soundID) const {
	if( game_g->getSound() != NULL ) {
		game_g->getSound()->PlaySound(soundID, false, 0);
	}
}

bool MainGamestate::isGUILocked() const {
	//return mainGUILocked || genv->getModalPanel() != NULL;
	return game_g->getGraphicsEnvironment()->getModalPanel() != NULL;
}

GameData *MainGamestate::getGameData() {
	return game_g->getGameData();
}

const GameData *MainGamestate::getGameData() const {
	return game_g->getGameData();
}

const Race *GameData::findRace(const char *name) const {
	for(vector<const Race *>::const_iterator iter = races.begin();iter != races.end(); ++iter) {
		const Race *race = *iter;
		if( race->equals(name) ) {
			return race;
		}
	}
	VI_log("GameData::findRace(%s) failed\n", name);
	ASSERT( false );
	return NULL;
}

const Technology *GameData::findTechnology(const char *name) const {
	for(vector<Technology *>::const_iterator iter = technologies.begin();iter != technologies.end(); ++iter) {
		const Technology *technology = *iter;
		if( technology->equals(name) ) {
			return technology;
		}
	}
	VI_log("GameData::findTechnology(%s) failed\n", name);
	ASSERT( false );
	return NULL;
}

Buildable *GameData::findBuildable(const char *name) {
	for(vector<Buildable *>::iterator iter = buildables.begin();iter != buildables.end(); ++iter) {
		Buildable *buildable = *iter;
		if( buildable->equals(name) ) {
			return buildable;
		}
	}
	VI_log("GameData::findBuildable(%s) failed\n", name);
	ASSERT( false );
	return NULL;
}

const Improvement *GameData::findImprovement(const char *name) const {
	for(vector<const Improvement *>::const_iterator iter = improvements.begin();iter != improvements.end(); ++iter) {
		const Improvement *improvement = *iter;
		if( improvement->equals(name) ) {
			return improvement;
		}
	}
	VI_log("GameData::findImprovement(%s) failed\n", name);
	ASSERT( false );
	return NULL;
}

const Improvement *GameData::findImprovement(ImprovementType improvementType) const {
	for(vector<const Improvement *>::const_iterator iter = improvements.begin();iter != improvements.end(); ++iter) {
		const Improvement *improvement = *iter;
		if( improvement->getImprovementType() == improvementType ) {
			return improvement;
		}
	}
	VI_log("GameData::findImprovement(%d) failed\n", improvementType);
	ASSERT( false );
	return NULL;
}

const UnitTemplate *GameData::findUnitTemplate(const char *name) const {
	/*for(vector<const UnitTemplate *>::const_iterator iter = unit_templates.begin();iter != unit_templates.end(); ++iter) {
		const UnitTemplate *unit_template = *iter;*/
	for(int i=0;i<this->getNUnitTemplates();i++) {
		const UnitTemplate *unit_template = this->getUnitTemplate(i);
		if( unit_template->equals(name) ) {
			return unit_template;
		}
	}
	VI_log("GameData::findUnitTemplate(%s) failed\n", name);
	ASSERT( false );
	return NULL;
}

const Relationship *MainGamestate::findRelationship(const Civilization *civ1, const Civilization *civ2) const {
	ASSERT( civ1 != civ2 );
	ASSERT( !civ1->isDead() );
	ASSERT( !civ2->isDead() );
	for(vector<Relationship *>::const_iterator iter = relationships.begin(); iter != relationships.end(); ++iter) {
		Relationship *relationship = *iter;
		if( relationship->equals(civ1, civ2) ) {
			return relationship;
		}
	}
	VI_log("GameData::findRelationship(%s, %s) failed\n", civ1->getName(), civ2->getName());
	ASSERT( false );
	return NULL;
}

Relationship *MainGamestate::findRelationship(const Civilization *civ1, const Civilization *civ2) {
	ASSERT( civ1 != civ2 );
	ASSERT( !civ1->isDead() );
	ASSERT( !civ2->isDead() );
	for(vector<Relationship *>::const_iterator iter = relationships.begin(); iter != relationships.end(); ++iter) {
		Relationship *relationship = *iter;
		if( relationship->equals(civ1, civ2) ) {
			return relationship;
		}
	}
	VI_log("GameData::findRelationship(%s, %s) failed\n", civ1->getName(), civ2->getName());
	ASSERT( false );
	return NULL;
}

/*bool MainGamestate::initWorld() {
	//VI_flush(0);

	world = VI_createWorld();

	// setup world here

	game_g->getGraphicsEnvironment()->linkWorld(world);

	VI_check();
	VI_debug_mem_usage_all();
	return true;
}*/

bool MainGamestate::init() {
	/*VisionPrefs visionPrefs;
	visionPrefs.load();
	visionPrefs.want_std_shaders = false;
	//visionPrefs.shaders = false; // no need for shaders
	//genv = VI_createGraphicsEnvironment(width, height, multisample, fullscreen, hdr);
	genv = VI_createGraphicsEnvironment(&visionPrefs);
	if( genv == NULL ) {
		bool done = false;
		if( visionPrefs.fullscreen ) {
			MessageBoxA(NULL,"Failed to run in fullscreen mode","Error",MB_OK|MB_ICONEXCLAMATION);
			visionPrefs.fullscreen = false;
			genv = VI_createGraphicsEnvironment(&visionPrefs);
			if( genv != NULL )
				done = true;
		}
		if( !done ) {
			MessageBoxA(NULL,"Failed to open graphics window","Error",MB_OK|MB_ICONEXCLAMATION);
			return false;
		}
	}

	sound = VI_Sound::create( game_g->getGraphicsEnvironment()->getHWND() );
	if( sound == NULL ) {
		MessageBoxA(NULL,"Failed to initialise sound","Error",MB_OK|MB_ICONEXCLAMATION);
	}

	//ShowCursor(FALSE); // always hide cursor
	char buffer[4096] = "";
	sprintf_s(buffer, "Conquests v%d.%d | (c) 2009-2010 Mark Harman", versionMajor, versionMinor);
	SetWindowTextA(game_g->getGraphicsEnvironment()->getHWND(), buffer);

	//VI_Texture *cursor = VI_createTextureWithMask("textures/arrow.png", 255, 255, 255);
	//genv->setMouseTexture(cursor);

	//font = genv->createFont("Verdana",12,FW_BOLD);
	font = game_g->getGraphicsEnvironment()->createFont("Verdana",12,FW_NORMAL);*/

	// create game panels

	VI_Font *font = game_g->getFont();

	gamePanel = VI_createPanel();
	gamePanel->setPaintFunc(paintPanel, NULL);
	game_g->getGraphicsEnvironment()->setPanel(gamePanel);
	game_g->getGraphicsEnvironment()->setRenderFunc(renderFunc);

	endturnButton = VI_createButton("End turn", font);
	initButton(endturnButton);
	endturnButton->setAction(action);
	//endturnButton->setKeyShortcut(V_RETURN);
	endturnButton->setPopupText("Finish your turn", font);
	//gamePanel->add(endturnButton, 16, 96);
	gamePanel->add(endturnButton, 16, 80);

	int xpos = game_g->getGraphicsEnvironment()->getWidth() - 96;
	const int xdiff = 16, ydiff = 30;
	int ypos = 16;

	newgameButton = VI_createButton("New Game", font);
	initButton(newgameButton);
	newgameButton->setAction(action);
	newgameButton->setKeyShortcut(V_N);
	newgameButton->setPopupText("Start a new game (N)", font);
	gamePanel->add(newgameButton, xpos, ypos); ypos += ydiff;

	loadgameButton = VI_createButton("Load", font);
	initButton(loadgameButton);
	loadgameButton->setAction(action);
	loadgameButton->setKeyShortcut(V_L);
	loadgameButton->setPopupText("Load a previously saved game (L)", font);
	gamePanel->add(loadgameButton, xpos, ypos); ypos += ydiff;

	savegameButton = VI_createButton("Save", font);
	initButton(savegameButton);
	savegameButton->setAction(action);
	savegameButton->setKeyShortcut(V_S);
	savegameButton->setPopupText("Save the current game (S)", font);
	gamePanel->add(savegameButton, xpos, ypos); ypos += ydiff;

	quitgameButton = VI_createButton("Quit", font);
	initButton(quitgameButton);
	quitgameButton->setAction(action);
	quitgameButton->setPopupText("Quit the game (Escape)", font);
	gamePanel->add(quitgameButton, xpos, ypos); ypos += ydiff;

	//xpos -= 64;

	citiesAdvisorButton = VI_createButton("Cities", font);
	initButton(citiesAdvisorButton);
	citiesAdvisorButton->setAction(action);
	citiesAdvisorButton->setKeyShortcut(V_F1);
	citiesAdvisorButton->setPopupText("List your cities (F1)", font);
	gamePanel->add(citiesAdvisorButton, xpos, ypos); ypos += ydiff;

	unitsAdvisorButton = VI_createButton("Units", font);
	initButton(unitsAdvisorButton);
	unitsAdvisorButton->setAction(action);
	unitsAdvisorButton->setKeyShortcut(V_F2);
	unitsAdvisorButton->setPopupText("List your units (F2)", font);
	gamePanel->add(unitsAdvisorButton, xpos, ypos); ypos += ydiff;

	civsAdvisorButton = VI_createButton("Civilizations", font);
	initButton(civsAdvisorButton);
	civsAdvisorButton->setAction(action);
	civsAdvisorButton->setKeyShortcut(V_F3);
	civsAdvisorButton->setPopupText("View and contact other Civilizations you have met (F3)", font);
	gamePanel->add(civsAdvisorButton, xpos, ypos); ypos += ydiff;

	techsAdvisorButton = VI_createButton("Technologies", font);
	initButton(techsAdvisorButton);
	techsAdvisorButton->setAction(action);
	techsAdvisorButton->setKeyShortcut(V_F4);
	techsAdvisorButton->setPopupText("List your technologies (F4)", font);
	gamePanel->add(techsAdvisorButton, xpos, ypos); ypos += ydiff;

	scoresAdvisorButton = VI_createButton("Scores", font);
	initButton(scoresAdvisorButton);
	scoresAdvisorButton->setAction(action);
	scoresAdvisorButton->setKeyShortcut(V_F5);
	scoresAdvisorButton->setPopupText("Compare civilizations (F5)", font);
	gamePanel->add(scoresAdvisorButton, xpos, ypos); ypos += ydiff;

	mapButton = VI_createButton("Map", font);
	initButton(mapButton);
	mapButton->setAction(action);
	mapButton->setKeyShortcut(V_M);
	mapButton->setPopupText("Show map (M)", font);
	gamePanel->add(mapButton, xpos, ypos); ypos += ydiff;

	gridButton = VI_createButton("Grid", font);
	initButton(gridButton);
	gridButton->setAction(action);
	gridButton->setPopupText("Toggle Grid", font);
	gamePanel->add(gridButton, xpos, ypos); ypos += ydiff;

	zoomOutButton = VI_createButton("Zoom Out", font);
	initButton(zoomOutButton);
	zoomOutButton->setAction(action);
	zoomOutButton->setKeyShortcut(V_Z);
	zoomOutButton->setPopupText("Zoom Out (Z)", font);
	gamePanel->add(zoomOutButton, xpos, ypos); ypos += ydiff;

	zoomResetButton = VI_createButton("Reset Zoom", font);
	initButton(zoomResetButton);
	zoomResetButton->setAction(action);
	zoomResetButton->setPopupText("Reset Zoom To Default", font);
	gamePanel->add(zoomResetButton, xpos, ypos); ypos += ydiff;

	zoomInButton = VI_createButton("Zoom In", font);
	initButton(zoomInButton);
	zoomInButton->setAction(action);
	zoomInButton->setKeyShortcut(V_X);
	zoomInButton->setPopupText("Zoom In (X)", font);
	gamePanel->add(zoomInButton, xpos, ypos); ypos += ydiff;

	viewButton = VI_createButton("View 3D/2D", font);
	initButton(viewButton);
	viewButton->setAction(action);
	viewButton->setPopupText("Switch between 3D and 2D view", font);
	gamePanel->add(viewButton, xpos, ypos); ypos += ydiff;

	xpos = 16;
	ypos = game_g->getGraphicsEnvironment()->getHeight() - 40; // need to also allow room for scrolling down "gap"

	waitButton = VI_createButton("Wait", font);
	initButton(waitButton);
	waitButton->setAction(action);
	waitButton->setKeyShortcut(V_W);
	waitButton->setPopupText("Wait, and move another unit first (W)", font);
	gamePanel->add(waitButton, xpos, ypos); xpos += waitButton->getWidth() + xdiff;

	skipTurnButton = VI_createButton("Skip", font);
	initButton(skipTurnButton);
	skipTurnButton->setAction(action);
	skipTurnButton->setKeyShortcut(V_SPACE);
	skipTurnButton->setPopupText("Skip this unit's turn (SPACE)", font);
	gamePanel->add(skipTurnButton, xpos, ypos); xpos += skipTurnButton->getWidth() + xdiff;

	fortifyButton = VI_createButton("Fortify", font);
	initButton(fortifyButton);
	fortifyButton->setAction(action);
	fortifyButton->setKeyShortcut(V_F);
	fortifyButton->setPopupText("Causes the unit to wait until activated again (F)", font);
	gamePanel->add(fortifyButton, xpos, ypos); xpos += fortifyButton->getWidth() + xdiff;

	buildCityButton = VI_createButton("Build City", font);
	initButton(buildCityButton);
	buildCityButton->setAction(action);
	buildCityButton->setKeyShortcut(V_B);
	buildCityButton->setPopupText("Build a city at this location (B)", font);
	gamePanel->add(buildCityButton, xpos, ypos); xpos += buildCityButton->getWidth() + xdiff;

	buildRoadButton = VI_createButton("Build Road", font);
	initButton(buildRoadButton);
	buildRoadButton->setAction(action);
	buildRoadButton->setKeyShortcut(V_R);
	buildRoadButton->setPopupText("Roads allow faster movement along them (R)", font);
	gamePanel->add(buildRoadButton, xpos, ypos); // don't update xpos

	buildRailwaysButton = VI_createButton("Build Railways", font);
	initButton(buildRailwaysButton);
	buildRailwaysButton->setAction(action);
	buildRailwaysButton->setKeyShortcut(V_R);
	buildRailwaysButton->setPopupText("Railways allow instantaneous movement along them (R)", font);
	gamePanel->add(buildRailwaysButton, xpos, ypos);

	xpos += max(buildRoadButton->getWidth(), buildRailwaysButton->getWidth() ) + xdiff;

	automateButton = VI_createButton("Automate", font);
	initButton(automateButton);
	automateButton->setAction(action);
	automateButton->setKeyShortcut(V_A);
	automateButton->setPopupText("Automate this unit (A)", font);
	gamePanel->add(automateButton, xpos, ypos); xpos += automateButton->getWidth() + xdiff;

	gotoButton = VI_createButton("Goto", font);
	initButton(gotoButton);
	gotoButton->setAction(action);
	gotoButton->setKeyShortcut(V_G);
	gotoButton->setPopupText("Automatically move this unit to another square (G)", font);
	gamePanel->add(gotoButton, xpos, ypos); xpos += gotoButton->getWidth() + xdiff;

	travelButton = VI_createButton("Travel", font);
	initButton(travelButton);
	travelButton->setAction(action);
	travelButton->setKeyShortcut(V_T);
	travelButton->setPopupText("Travel by sea (T)", font);
	gamePanel->add(travelButton, xpos, ypos); xpos += travelButton->getWidth() + xdiff;

	airRaidButton = VI_createButton("Bomb", font);
	initButton(airRaidButton);
	airRaidButton->setAction(action);
	airRaidButton->setKeyShortcut(V_B);
	airRaidButton->setPopupText("Bomb enemy city (B)", font);
	gamePanel->add(airRaidButton, xpos, ypos); xpos += airRaidButton->getWidth() + xdiff;

	/*
	// load shaders
	this->splat_pixel_shader_ambient = NULL;
	if( game_g->getGraphicsEnvironment()->hasShaders() && visionPrefs.terrainsplatting ) {
		try {
			this->splat_pixel_shader_ambient = Shader::createShader(game_g->getGraphicsEnvironment(), "conquests/shaders/shaders.cg", "main_splat_ambient_ps", false);
		}
		catch(VisionException *ve) {
			// carry on without shaders if we can't load it
			VI_log("Failed to load terrain splatting shader\n");
			delete ve;
		}
	}*/

	// load textures
	try {
		//MapSquare::loadTextures();
		//cityTexture = VI_createTexture("conquests/data/gfx/icons.png");
		cityTexture = VI_createTextureWithMask("conquests/data/gfx/city.png", 255, 0, 255);
		cityModernTexture = VI_createTextureWithMask("conquests/data/gfx/city_modern.png", 255, 0, 255);

		/*
		//panelTexture = VI_createTexture("conquests/data/gfx/land.png");
		unsigned char filter_max_panel[3] = {220, 172, 40};
		//unsigned char filter_min_panel[3] = {127, 100, 20};
		unsigned char filter_min_panel[3] = {120, 90, 20};
		VI_Image *image = VI_createImageNoise(64, 64, 16.0f, filter_max_panel, filter_min_panel, V_NOISEMODE_PERLIN, 4);
		panelTexture = VI_createTexture(image);
		*/

		cityPanelTexture = VI_createTexture("conquests/data/gfx/marble.jpg");

		// particle systems

		//VI_Texture *smokeTexture = VI_createTexture("conquests/data/gfx/smoke.png");
		//smokeTexture = VI_createTexture("conquests/data/gfx/spare/xsmoke.png");
		unsigned char smoke_min[] = {0, 0, 0};
		//unsigned char smoke_max[] = {255, 255, 255};
		//unsigned char smoke_max[] = {181, 181, 181};
		unsigned char smoke_max[] = {186, 186, 186};
		VI_Image *smokeImage = VI_createImageRadial(64, 64, smoke_max, smoke_min, true);
		//smokeImage->save("test.png", "png");
		smokeTexture = VI_createTexture(smokeImage);
		const unsigned char grey = 51;
		//const unsigned char grey = 255;
		const float scale_3d = 1.2f;
		const float scale_2d = 48.0f;
		// &ps_ent,&ps,30,ps_tx,color,color,255,0,0.2*0.5*scale,1.0*0.5*scale,10,3.0,-0.5,0.1);
		smokeParticlesystem_2d = VI_GeneralParticlesystem::create(30);
		smokeParticlesystem_2d->setColoriStart(grey, grey, grey);
		smokeParticlesystem_2d->setColoriEnd(grey, grey, grey);
		smokeParticlesystem_2d->setAlphaiEnd(255);
		smokeParticlesystem_2d->setAlphaiEnd(127);
		smokeParticlesystem_2d->setSizeStart(0.3f*scale_2d);
		smokeParticlesystem_2d->setSizeEnd(scale_2d);
		smokeParticlesystem_2d->setBirthRate(10);
		smokeParticlesystem_2d->setLife(3.0f);
		smokeParticlesystem_2d->setMass(1.0f);
		smokeParticlesystem_2d->setInitialVelocity(0.05f);
		smokeParticlesystem_2d->setTexture(smokeTexture);
		smokeEntity_2d = VI_createEntity(smokeParticlesystem_2d);

		smokeParticlesystem_3d = VI_GeneralParticlesystem::create(30);
		smokeParticlesystem_3d->setColoriStart(grey, grey, grey);
		smokeParticlesystem_3d->setColoriEnd(grey, grey, grey);
		smokeParticlesystem_3d->setAlphaiEnd(255);
		smokeParticlesystem_3d->setAlphaiEnd(127);
		smokeParticlesystem_3d->setSizeStart(0.3f*scale_3d);
		smokeParticlesystem_3d->setSizeEnd(scale_3d);
		smokeParticlesystem_3d->setBirthRate(5);
		smokeParticlesystem_3d->setLife(2.5f);
		smokeParticlesystem_3d->setMass(-0.03125f*1.5f);
		smokeParticlesystem_3d->setInitialVelocity(-0.0015625f*2.0f);
		smokeParticlesystem_3d->setTexture(smokeTexture);
		smokeEntity_3d = VI_createEntity(smokeParticlesystem_3d);

		fireParticlesystem_2d = VI_GeneralParticlesystem::create(50);
		fireParticlesystem_2d->setColoriStart(255, 0, 0);
		fireParticlesystem_2d->setColoriEnd(255, 127, 0);
		fireParticlesystem_2d->setAlphaiStart(255);
		fireParticlesystem_2d->setAlphaiEnd(127);
		fireParticlesystem_2d->setSizeStart(0.2f*scale_2d);
		fireParticlesystem_2d->setSizeEnd(2.0f*scale_2d);
		fireParticlesystem_2d->setBirthRate(10);
		fireParticlesystem_2d->setLife(4.0f);
		fireParticlesystem_2d->setMass(1.0f);
		fireParticlesystem_2d->setInitialVelocity(0.05f);
		fireParticlesystem_2d->setTexture(smokeTexture);
		fireEntity_2d = VI_createEntity(fireParticlesystem_2d);

		fireParticlesystem_3d = VI_GeneralParticlesystem::create(50);
		fireParticlesystem_3d->setColoriStart(255, 0, 0);
		fireParticlesystem_3d->setColoriEnd(255, 127, 0);
		fireParticlesystem_3d->setAlphaiStart(255);
		fireParticlesystem_3d->setAlphaiEnd(127);
		fireParticlesystem_3d->setSizeStart(0.2f*scale_3d);
		fireParticlesystem_3d->setSizeEnd(2.0f*scale_3d);
		fireParticlesystem_3d->setBirthRate(3);
		fireParticlesystem_3d->setLife(4.0f);
		fireParticlesystem_3d->setMass(-0.015625f*1.5f);
		fireParticlesystem_3d->setInitialVelocity(-0.0015625f*2.0f);
		fireParticlesystem_3d->setTexture(smokeTexture);
		fireEntity_3d = VI_createEntity(fireParticlesystem_3d);

		// load sounds
		if( game_g->getSound() != NULL ) {
			game_g->getSound()->LoadSound(SOUND_FOOT, "conquests/data/sound/foot3.wav");
			game_g->getSound()->LoadSound(SOUND_WOODBRK, "conquests/data/sound/woodbrk.wav");
			game_g->getSound()->LoadSound(SOUND_WORKER, "conquests/data/sound/worker.wav");
			game_g->getSound()->LoadSound(SOUND_BLIP, "conquests/data/sound/pulsed808bd.wav");
			game_g->getSound()->LoadSound(SOUND_NUKE_EXPLOSION, "conquests/data/sound/explosion_6.wav");
		}

		//VI_Texture *map_texture = game->map->getSquare(0, 0)->getTexture();
		VI_Texture *map_texture = MapSquare::getTexture(MapSquare::TYPE_GRASSLAND);
		def_sq_width = map_texture->getWidth();
		def_sq_height = map_texture->getHeight();
	}
	catch(VisionException *ve) {
		VI_log("Failed - something went wrong loading main game state data!\n");
#ifdef _WIN32
		MessageBoxA(NULL,"Failed to load Main Game State data","Error",MB_OK|MB_ICONEXCLAMATION);
#endif
		delete ve;
		return false;
	}

	if( map != NULL ) {
		VI_log("map is %d x %d\n", map->getWidth(), map->getHeight());
	}

	VI_World *world = VI_createWorld();
	game_g->getGraphicsEnvironment()->linkWorld(world);
	VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
	vp->rotateToEuler(-60.0f, 0.0f, 0.0f);
	if( view_3d ) {
		if( this->map != NULL )
			generateTerrain();
		float preferred_width = def_view3d_scale * game_g->getGraphicsEnvironment()->getWidth() / this->def_sq_width;
		reset3DZoom(preferred_width);
	}
	{
		unsigned char filter_max[3] = {32, 32, 64};
		unsigned char filter_min[3] = {0, 0, 0};
		VI_Image *image = VI_createImageNoise(256, 256, 4.0f, filter_max, filter_min, V_NOISEMODE_SMOKE, 4);
		VI_Texture *texture = VI_createTexture(image);
		VI_Sky *sky = VI_createSky1(texture);
		sky->setMode(V_SKYMODE_FULL);
		world->setSky(sky);
	}

	//VI_set_persistence_all(-1);
	/*if( !initWorld() ) {
		return false;
	}*/

	return true;
}

void MainGamestate::generateTerrain() {
	VI_log("MainGamestate::generateTerrain()\n");
	if( game_g->getGraphicsEnvironment()->getWorld() == NULL ) {
		return;
	}
	if( view_3d ) {
		ASSERT( game_g->getGraphicsEnvironment()->getWorld()->getTerrain() == NULL );
		VI_lock();
		int extended_width = map->getTopology() == Map::TOPOLOGY_CYLINDER ? 2*map->getWidth()+1 : map->getWidth()+1;
		VI_Terrain *terrain = VI_createTerrain(game_g->getGraphicsEnvironment(), extended_width, map->getHeight()+1, xscale, zscale, 1.0f);
		terrain->setLOD(false);
		terrain->setLighting(false);
		/*for(int i=0;i<MapSquare::N_TYPES;i++) {
			terrain->setTexture(i, MapSquare::getTexture((MapSquare::Type)i));
		}*/
		terrain->setTexture(0, MapSquare::getTexture(MapSquare::TYPE_GRASSLAND));
		terrain->setTexture(1, MapSquare::getTexture(MapSquare::TYPE_OCEAN));
		terrain->setTexture(2, MapSquare::getTexture(MapSquare::TYPE_DESERT));
		terrain->setTexture(3, MapSquare::getTexture(MapSquare::TYPE_ARTIC));
		for(int y=0;y<map->getHeight();y++) {
			for(int x=0;x<map->getWidth();x++) {
				const MapSquare *square = map->getSquare(x, y);
				/*unsigned char type = (unsigned char)square->getType();
				terrain->setTextureMap(x, y, type);*/
				unsigned char v = 0;
				if( square->getType() == MapSquare::TYPE_OCEAN ) {
					v = 1;
				}
				else if( square->getType() == MapSquare::TYPE_DESERT ) {
					v = 2;
				}
				else if( square->getType() == MapSquare::TYPE_ARTIC ) {
					v = 3;
				}
				terrain->setTextureMap(x, y, v);
				if( map->getTopology() == Map::TOPOLOGY_CYLINDER ) {
					terrain->setTextureMap(x + map->getWidth(), y, v);
				}
				if( this->player != NULL && !this->player->isExplored(x, y) ) {
					terrain->setVisible(x, y, false);
					if( map->getTopology() == Map::TOPOLOGY_CYLINDER ) {
						terrain->setVisible(x + map->getWidth(), y, false);
					}
				}
			}
		}
		VI_unlock();

		game_g->getGraphicsEnvironment()->getWorld()->setTerrain(terrain);

		/*VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
		vp->setPosition(0.5f*terrain->getWidth(), def_vp_height_c, 0.75f*terrain->getDepth());
		vp->rotateToEuler(-60.0f, 0.0f, 0.0f);*/
	}

}

void MainGamestate::reset3DZoom(float preferred_width) {
	ASSERT( view_3d );
	// TODO: fix calculation, not quite correct!
	//vp->setPosition(0.5f*terrain->getWidth(), def_vp_height_c, 0.75f*terrain->getDepth());
	VI_SceneGraphNode *vp = game_g->getGraphicsEnvironment()->getWorld()->getViewpoint();
	Vector3D vp_pos = vp->getPosition();
	// move to reference height
	vp_pos.y = 10.0f;
	vp->setPosition(&vp_pos);

	float mx0 = 0.0f, mx1 = 0.0f, my = 0.0f;
	getMapPosf(&mx0, &my, 0, 0);
	getMapPosf(&mx1, &my, game_g->getGraphicsEnvironment()->getWidth()-1, 0);
	float width = mx1 - mx0;
	vp_pos.y *= preferred_width / width;
	vp->setPosition(&vp_pos);

}

/*void MainGamestate::run() {
	if( !init() ) {
		clear();
		return;
	}
	if( !initGameData(true) ) {
		clear();
		return;
	}
	game_g->getGraphicsEnvironment()->setVisible(true);
	beginTurn();

	bool timedemo = false;
	//timedemo = true;
	int last_update_time = clock();
	int t1 = clock();
	const int n_frames = 500;

	// program main loop
	while( !quit ) {
		// check for messages
		if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {
			// handle or dispatch messages
			if( msg.message == WM_QUIT ) {
				if( !this->isGUILocked() ) {
					this->setInputMode(INPUTMODE_NORMAL);
					quitGame();
				}
			}
			else {
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		}
		//else if( genv->isActive() )
		else
		{
			if( game_g->getGraphicsEnvironment()->isActive() ) {
				//ASSERT( _heapchk() == _HEAPOK );
				game_g->getGraphicsEnvironment()->drawFrame();
				//ASSERT( _heapchk() == _HEAPOK );
				game_g->getGraphicsEnvironment()->swapBuffers();
				frames++;
				if( timedemo && frames >= n_frames )
					quit = true;
			}
			else {
				Sleep(10); // don't consume 100% CPU
			}

			int time_diff = clock() - last_update_time;
			if( time_diff > 0 ) {
				update(time_diff);
				last_update_time = clock();
				if( game_g->getGraphicsEnvironment()->isActive() ) {
					game_g->getGraphicsEnvironment()->handleInput(); // must be done last due to resetting input
				}
			}
		}
		//else {
		//WaitMessage();
		//}
	}
	int t2 = clock();

	clear();

	//ShowCursor(TRUE);
	char buffer[256] = "";
	sprintf_s(buffer,"%d frames in time %d\nFPS : %f",frames,t2-t1,(frames*1000)/((float)(t2-t1)));
	MessageBoxA(NULL,buffer,"Finished",NULL);
}*/

bool MainGamestate::start() {
	bool ok = true;

/*#ifdef _WIN32
	HCURSOR cursor = LoadCursor( NULL, IDC_WAIT );
	SetCursor(cursor);
	SetClassLong(game_g->getGraphicsEnvironment()->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
	game_g->setCursor(cursor_wait);

	/*if( !initGameData(true) ) {
		ok = false;
	}
	else*/
	if( !init() ) {
		ok = false;
	}
	/*else if( !initGameData(true) ) {
		ok = false;
	}*/
	else {
		VI_check();
		VI_debug_mem_usage_all();

		game_g->getGraphicsEnvironment()->setVisible(true);
		game_g->getGraphicsEnvironment()->setFade(fade_duration_c, 0, 255);
		home();
		//beginTurn();
		// don't beginTurn() here, as we may be starting from a loaded file!
	}

	if( !ok ) {
		clear();
	}

/*#ifdef _WIN32
	cursor = LoadCursor( NULL, IDC_ARROW );
	SetCursor(cursor);
	SetClassLong(game_g->getGraphicsEnvironment()->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
	game_g->setCursor(cursor_arrow);

	if( ok )
		gamestate_started = true;
	return ok;
}

void MainGamestate::stop() {
	clear();
	gamestate_started = false;
}

void MainGamestate::clear() {
	clearGameData();

	VI_flush(0); // delete all the gamestate objects, but leave the game level objects (which should be set at persistence level -1)
	VI_GraphicsEnvironment *genv = game_g->getGraphicsEnvironment();
	//VI_log("%d\n", genv);
	game_g->getGraphicsEnvironment()->setPanel(NULL); // as the main panel is now destroyed
	game_g->getGraphicsEnvironment()->linkWorld(NULL); // as the world is now destroyed
	game_g->getGraphicsEnvironment()->setRenderFunc(NULL);
}

void MainGamestate::clearGameData() {
	for(vector<Effect *>::iterator iter = effects.begin(); iter != effects.end(); ++iter) {
		Effect *effect = *iter;
		delete effect;
	}
	effects.clear();

	/*if( InfoWindow::getActiveWindow() != NULL ) {
		delete InfoWindow::getActiveWindow();
	}*/
	if( infoWindow != NULL ) {
		delete infoWindow;
		infoWindow = NULL;
	}

	for(vector<Relationship *>::iterator iter = relationships.begin(); iter != relationships.end(); ++iter) {
		Relationship *relationship = *iter;
		delete relationship;
	}
	relationships.clear();
	for(vector<Civilization *>::iterator iter = civilizations.begin(); iter != civilizations.end(); ++iter) {
		Civilization *civilization = *iter;
		delete civilization;
	}
	civilizations.clear();
	player = NULL;
	help_buildcity = false;
	help_buildroad = false;
	help_bomb = false;

	if( map != NULL ) {
		delete map;
		map = NULL;
	}

	difficulty = DIFFICULTY_EASY;
	aiaggression = AIAGGRESSION_NORMAL;

	scale_width = 1.0f;
	scale_height = 1.0f;

	VI_World *world = game_g->getGraphicsEnvironment()->getWorld();
	if( world != NULL && world->getTerrain() != NULL ) {
		VI_log("about to delete terrain\n");
		delete world->getTerrain();
		VI_log("done\n");
		world->setTerrain(NULL);
	}
}

bool MainGamestate::initGameData(bool std_game, int map_width, int map_height, Map::Topology topology, const Race *player_race, int n_opponent_civs) {
	if( RANDOMSEED ) {
		time_t seconds = time(NULL) % 65536;
		this->seed = clock() + seconds;
		//this->seed = 82915; // "extra" terrain bug with texture splatting on small flat world.
	}
	else {
		//this->seed = 456;
		//this->seed = 3331;
		this->seed = 1271;
		//this->seed = 2465;
	}
	VI_log("set seed %d\n", seed);
	srand(seed);
	year = 0;
	help_buildcity = false;
	help_buildroad = false;
	help_bomb = false;
	cache_active_unit = NULL;
	active_unit = NULL;
	//updateUnitButtons();
	setInputMode(INPUTMODE_NORMAL);
	moved_units_this_turn = false;

	ASSERT( map == NULL );
	if( std_game ) {
		/*map = new Map(this);
		const char mapfile[] = "conquests/data/world.map";
		//const char mapfile[] = "conquests/data/world_32x32.map";
		if( !map->readFile(mapfile) ) {
			VI_log("Failed to read map file: %s\n", mapfile);
			MessageBoxA(NULL,"Failed to load map","Error",MB_OK|MB_ICONEXCLAMATION);
			return false;
		}*/
		map = new Map(this, map_width, map_height);
		map->setTopology(topology);
		if( !map->createRandom() ) {
			VI_log("Failed to create random map\n");
#ifdef _WIN32
			MessageBoxA(NULL,"Failed to create map","Error",MB_OK|MB_ICONEXCLAMATION);
#endif
			return false;
		}
	}

	// generate Civilizations

	if( std_game ) {
		ASSERT( player_race != NULL );
		const UnitTemplate *unit_template = NULL;
		//City *city = NULL;
		Civilization *cpu = NULL;
		Pos2D start;

		player = new Civilization( this, player_race );
		if( CHEAT ) {
			player->revealMap();
		}
		//player->addTechnology(findTechnology("Sailing"));
		//start = Pos2D(25, 9);
		start = map->findRandomFreeSquare();
		map->ensureNearbyTerrain(start, MapSquare::TYPE_FOREST);
		//new City(player, 0, 0);
		//city = new City(player, 10, 20);
		//city->setName("Rome");
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		new Unit(this, player, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, player, unit_template, start);
		/*unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Modern Infantry"));
		new Unit(this, player, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Modern Tanks"));
		new Unit(this, player, unit_template, start);*/
		/*unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Biplanes"));
		new Unit(this, player, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Fighters"));
		new Unit(this, player, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Bombers"));
		new Unit(this, player, unit_template, start);*/
		/*unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Riflemen"));
		new Unit(this, player, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Infantry"));
		new Unit(this, player, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Tanks"));
		new Unit(this, player, unit_template, start);*/
		/*unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Militia"));
		new Unit(this, player, unit_template, start);*/
		/*unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Cannon"));
		new Unit(player, unit_template, start);*/
		/*unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Pikemen"));
		new Unit(player, unit_template, start);*/
		/*unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Men At Arms"));
		new Unit(player, unit_template, start);*/
		/*unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
		new Unit(player, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
		new Unit(player, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Horsemen"));
		new Unit(player, unit_template, start);*/

		/*for(int i=0;i<game_g->getGameData()->getNRaces();i++) {
			const Race *race = game_g->getGameData()->getRace(i);
			if( race == player_race )
				continue;

			cpu = new Civilization( this, race );

			start = map->findRandomFreeSquare();
			map->ensureNearbyTerrain(start, MapSquare::TYPE_FOREST);

			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
			new Unit(this, cpu, unit_template, start);
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
			new Unit(this, cpu, unit_template, start);
		}*/
		vector<const Race *> races;
		for(int i=0;i<game_g->getGameData()->getNRaces();i++) {
			const Race *race = game_g->getGameData()->getRace(i);
			if( race == player_race )
				continue;
			races.push_back(race);
		}

		ASSERT( races.size() >= n_opponent_civs );
		for(int i=0;i<n_opponent_civs;i++) {
			int r = rand () % races.size();
			const Race *race = races.at(r);

			cpu = new Civilization( this, race );

			start = map->findRandomFreeSquare();
			map->ensureNearbyTerrain(start, MapSquare::TYPE_FOREST);

			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
			new Unit(this, cpu, unit_template, start);
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
			new Unit(this, cpu, unit_template, start);

			// remove so we don't get duplicates!
			races.erase( races.begin() + r );
		}

		/*
		cpu = new Civilization( this, findRace("The English") );

		//start = Pos2D(9, 22);
		start = map->findRandomFreeSquare();
		map->ensureNearbyTerrain(start, MapSquare::TYPE_FOREST);
		//city = new City(cpu, 14, 14);
		//city->setName("London");
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		new Unit(this, cpu, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, cpu, unit_template, start);

		cpu = new Civilization( this, findRace("The Egyptians") );
		//start = Pos2D(43, 19);
		start = map->findRandomFreeSquare();
		map->ensureNearbyTerrain(start, MapSquare::TYPE_FOREST);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		new Unit(this, cpu, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, cpu, unit_template, start);

		cpu = new Civilization( this, findRace("The Americans") );
		//start = Pos2D(60, 13);
		start = map->findRandomFreeSquare();
		map->ensureNearbyTerrain(start, MapSquare::TYPE_FOREST);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		new Unit(this, cpu, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, cpu, unit_template, start);

		cpu = new Civilization( this, findRace("The Greeks") );
		start = map->findRandomFreeSquare();
		map->ensureNearbyTerrain(start, MapSquare::TYPE_FOREST);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		new Unit(this, cpu, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, cpu, unit_template, start);

		cpu = new Civilization( this, findRace("The French") );
		start = map->findRandomFreeSquare();
		map->ensureNearbyTerrain(start, MapSquare::TYPE_FOREST);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		new Unit(this, cpu, unit_template, start);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, cpu, unit_template, start);
		*/

		//generateRelationships();
		if( !checkRelationships() ) {
			return false;
		}
	}

	return true;
}

/**
 *
 * Tests
 *
 *
 *
 * TEST_SETTLERS_MOVE_0 - test that a Settler moves away from adjacent enemy unit, and doesn't move into the adjacent unprotected city.
 * TEST_SETTLERS_MOVE_1 - test that a Settler moves away from adjacent enemy unit.
 * TEST_SETTLERS_MOVE_2 - test that a Settler ignores an adjacent friendly unit, and builds a city.
 * TEST_SETTLERS_MOVE_3 - test that a Settler ignores an adjacent enemy Settler unit, and builds a city.
 * TEST_SETTLERS_MOVE_4 - test that a Settler moves away from adjacent enemy unit, when building a road.
 * TEST_MAKE_CONTACT - test that a unit moves towards a foreign unit to make contact.
 * TEST_ATTACK_UNIT - tests that a unit attacks an adjacent settler unit.
 * TEST_TRAVEL_EXPLORE_3 - tests that we can explore via harbour, across map wraparound
 * TEST_SETTLER_BUILDCITY_0 - build city at current location: grasslands, coastal.
 * TEST_SETTLER_BUILDCITY_1 - move from grasslands, non-coastal to grasslands, coastal.
 * TEST_SETTLER_BUILDCITY_2 - move from desert, coastal to grasslands, non-coastal.
 * TEST_SETTLER_BUILDCITY_3 - build city at current location: grasslands, non-coastal (compared with desert, coastal).
 * TEST_SETTLER_BUILDCITY_4 - move from desert, non-coastal to desert, coastal.
 * TEST_SETTLER_BUILDCITY_5 - build city at current location: grasslands, non-coastal (compared with grasslands, non-coastal).
 * TEST_SETTLER_BUILDCITY_6 - build city at current location: desert, non-coastal (compared with desert, non-coastal).
 * TEST_SETTLER_BUILDCITY_7 - as TEST_SETTLER_BUILDCITY_1 but with hills, coastal as target square.
 * TEST_SETTLER_BUILDCITY_8 - as TEST_SETTLER_BUILDCITY_1, but with forest such that we build city at original location.
 * TEST_SETTLER_BUILDCITY_9 - test that we can't build city at current location, due to city that is across the wraparound seam (cylinder topology)
 * TEST_BUILDROAD_0 - test that a Peasants unit near a city starts building roads
 * TEST_BUILDROAD_1 - test that a Peasants unit near a city starts building roads (automation)
 * TEST_BUILDROAD_2 - test that a Peasants unit between cities moves in order to start building roads
 * TEST_BUILDROAD_3 - test that a Peasants unit between cities moves in order to start building roads (automation)
 * TEST_BUILDROAD_4 - test that a Peasants unit between cities starts building roads
 * TEST_BUILDROAD_5 - test that a Peasants unit between cities starts building roads (automation)
 * TEST_BUILDROAD_6 - test that a Peasants unit near a small city doesn't start building roads
 * TEST_BUILDROAD_7 - test that a Peasants unit at a small city does start building roads
 * TEST_BUILDRAILWAYS_0 - test that a Peasants unit near a city starts building railways
 * TEST_BUILDRAILWAYS_1 - test that a Peasants unit near a city starts building railways (automation)
 * TEST_BUILDRAILWAYS_2 - test that a Peasants unit between cities moves in order to start building railways
 * TEST_BUILDRAILWAYS_3 - test that a Peasants unit between cities moves in order to start building railways (automation)
 * TEST_BUILDRAILWAYS_4 - test that a Peasants unit between cities starts building railways
 * TEST_BUILDRAILWAYS_5 - test that a Peasants unit between cities starts building railways (automation)
 * TEST_UNIT_MOVE_0 - test that a unit finds the shortest path along road
 * TEST_UNIT_MOVE_1 - test that a unit finds the shortest path along railways
 * TEST_UNIT_MOVE_2 - test that a unit finds the shortest path along road
 * TEST_UNIT_MOVE_3 - test that a unit finds the shortest path along railways
 * TEST_UNIT_MOVE_4 - test that a unit finds the shortest path along railways (infinite loop check)
 * TEST_UNIT_MOVE_5 - test that a unit finds the shortest path along railways (infinite loop check - case where the railway loop is much longer in terms of number of squares)
 * TEST_UNIT_MOVE_6 - test that a unit finds the shortest path along railways (infinite loop check - case where an enemy blocks the route for a Settlers unit)
 * TEST_UNIT_MOVE_7 - as TEST_UNIT_MOVE_1, with friendly civ territory
 * TEST_UNIT_MOVE_8 - as TEST_UNIT_MOVE_1, with enemy civ territory
 * TEST_UNIT_MOVE_9 - tests that a unit moves over the periodic seam of map with cylinder topology.
 * TEST_UNIT_MOVE_10 - movement from a city with port, with road along coast, unit with 1 move, move 2 squares away: check we move along road, with 1/3 moves left. Also check distance to dest square.
 * TEST_UNIT_MOVE_11 - movement from a city with port, with road along coast, unit with 1 move, move 4 squares away: check we travel, with 0 moves left. Also check distance to dest square.
 * TEST_UNIT_MOVE_12 - movement from a city with port, with road along coast, unit with 2 move, move 4 squares away: check we move along road, with 2/3 moves left. Also check distance to dest square.
 * TEST_UNIT_MOVE_13 - movement from a city with port, with railways along coast, unit with 1 move, move 6 squares away: check we move along railways, with 1 move left. Also check distance to dest square.
 * TEST_UNIT_RETREAT_0 - tests that a tanks unit retreats to a city, prefer city that's further, but with port.
 * TEST_UNIT_RETREAT_1 - as TEST_UNIT_RETREAT_0, but prefer the city that has fewer tank units.
 * TEST_UNIT_RETREAT_2 - as TEST_UNIT_RETREAT_1, but the unit is located at the best city.
 * TEST_AIR_BOMB_0 - Test for bombing enemy city with bomber (with airport).
 * TEST_AIR_BOMB_1 - Test for city with bomber and no airport, check that we can't bomb.
 * TEST_AIR_BOMB_2 - Test for bombing enemy city with fission bomb (with airport).
 * TEST_AIR_BOMB_3 - Test for city with fission bomb and no airport, check that we can't bomb.
 * TEST_AIR_BOMB_4 - Test for launching cruise missile against enemy city (with missile launcher and no airport).
 * TEST_AIR_BOMB_5 - Test for city with cruise missile, airport and no missile launcher, check that we can't bomb.
 * TEST_AIR_BOMB_6 - Test for launching nuclear missile against enemy city (with missile launcher and no airport).
 * TEST_AIR_BOMB_7 - Test for city with nuclear missile, airport and no missile launcher, check that we can't bomb.
 * TEST_AIR_REBASE_0 - Test for two cities and one bomber, each with airports, check that we can rebase.
 * TEST_AIR_REBASE_1 - Test for two cities and one jet bomber, only source has airport, check that we can't rebase.
 * TEST_AIR_REBASE_2 - Test for two cities and one stealth bomber, only destination has airport, check that we can't rebase.
 * TEST_AIR_REBASE_3 - Test for two cities and one cruise missile, each with airports, check that we can't rebase.
 * TEST_AIR_REBASE_4 - Test for land unit travelling between two cities, each with airports.
 * TEST_AIR_REBASE_5 - Test for land unit not travelling between two cities, only source has airport.
 * TEST_AIR_REBASE_6 - Test for land unit not travelling between two cities, only destination has airport.
 * TEST_ATTACK_0 - City defended with two spearmen. Test for swordsmen at city to attack adjacent catapults.
 * TEST_ATTACK_1 - City defended with three spearmen; also add one swordsmen. Test for one of the spearmen at city to attack adjacent catapults. (As it's better odds than being attacked.)
 * TEST_ATTACK_2 - City defended with three infantry. Test that we do attack an adjacent infantry unit. (As it's the best attacker unit.)
 * TEST_ATTACK_3 - City defended with three infantry; also add one tank. Test that we don't attack an adjacent infantry unit. (As it's better odds to be attacked.)
 * TEST_UNIT_EXPLORE_0 - Tests that a cavalry unit moves towards an unexplored square.
 * TEST_UNIT_EXPLORE_1 - Tests that a settlers unit adjacent to a city explores rather than building a road.
 * TEST_UNIT_EXPLORE_2 - Settlers and catapults at same square. Test that the catapults wait, then settlers build a city.
 * TEST_CITYBUILD_0 - Tests that a city starts building peasants to defend with. City is on coast, cpu has Sailing - shouldn't build a harbour.
 * TEST_CITYBUILD_1 - Tests that city starts building horsemen to explore.
 * TEST_CITYBUILD_2 - Tests that city starts building farmland.
 * TEST_CITYBUILD_3 - Tests that city starts building peasants as a "cheap unit".
 * TEST_CITYBUILD_4 - Tests that city starts building harbour.
 * TEST_CITYBUILD_5 - Tests that city starts building peasants as worker unit for roads.
 * TEST_CITYBUILD_6 - Tests that city starts building peasants as worker unit for railways.
 * TEST_CITYBUILD_7 - As TEST_CITYBUILD_4, but with railways covering the map.
 * TEST_CITYBUILD_8 - Tests that city rushbuilds a nuke.
 * TEST_CITYBUILD_9 - Tests that a city starts building spearmen to defend with. City already has two peasants units. City is on coast, cpu has Sailing - shouldn't build a harbour.
 * TEST_CITYBUILD_10 - Tests that a city starts building tanks (when nothing else to build).
 * TEST_CITYBUILD_11 - As TEST_CITYBUILD_0, but size 2 city, already has one peasants defender. However for peasants, one defender is enough, so we should build settlers.
 * TEST_CITYBUILD_12 - As TEST_CITYBUILD_9, but size 2 city, already has one spearmen defender. Also have settlers, so we build the second spearman.
 * TEST_CITYBUILD_13 - Tests that a school. There's an enemy unit and city nearby, but civilizations are at peace, so improvement should take priority over attacker units.
 * TEST_CITYBUILD_14 - As TEST_CITYBUILD_13, but civilizations at war, so build tanks.
 * TEST_CITYBUILD_15 - As TEST_CITYBUILD_0, but with Rocketry and Agriculture - should build Farmland, as better to grow due to Modern Infantry taking too long to build.
 * TEST_CITYBUILD_16 - As TEST_CITYBUILD_12, but with no settlers unit, and so 2 peasants are enough. So start building settlers instead of spearmen.
 * TEST_CITYBUILD_17 - As TEST_CITYBUILD_5, where 3 nearby squares already have roads. Size 3 city, so we build peasants as worker. Make sure all improvements already built, due to random chance of building 1st phase improvements earlier for size >= 3 cities.
 * TEST_CITYBUILD_18 - As TEST_CITYBUILD_17, but with size 2 city, so we should build horsemen as explorers.
 * TEST_AIPERF_0 - Repeatedly call the AI routine, to move a unit across to attack an enemy unit. Also checks dist to enemy unit.
 * TEST_AIPERF_1 - Repeatedly call the AI routine, to move a unit across to attack an enemy unit, followed by a second enemy unit. Also checks dist, namely that we can't reach the second unit to begin with.
 * TEST_AIPERF_2 - As TEST_AIPERF_1, but the first unit is a friendly civ (which we also first make contact with), so we have to go round that unit to get to the second unit.
 * TEST_AIPERF_3 - As TEST_AIPERF_2, but first unit is now enemy civ again, but we force moving to the second unit, to ensure we navigate around it.
 * TEST_AIPERF_4 - As TEST_AIPERF_0, but this time resetting the ai_has_dest to force recalculation.
 * TEST_AIPERF_5 - Test with range of cities and units; tests performance of AI code for cities.
 * TEST_AIPERF_6 - As TEST_AIPERF_5, but testing performance of AI code for units.
 * TEST_AIPERF_7 - As TEST_AIPERF_6, but without revealing map to the AI (tests exploration).
 * TEST_DISTANCES_0 - performance test for calculating distance map
 * TEST_DISTANCES_1 - as TEST_DISTANCES_0, but with added roads and railways
 */

void MainGamestate::runTest(const char *filename, TestID index) {
	if( !initGameData(false, 0, 0, Map::TOPOLOGY_UNDEFINED, NULL, 0) ) {
		clear();
		return;
	}
	this->grid = true;
	VI_log(">>> run test %d\n", index);

	game_g->getGraphicsEnvironment()->setVisible(true);

	// setup a game world for this test
	Civilization *cpu1 = NULL, *cpu2 = NULL, *cpu3 = NULL;
	const UnitTemplate *unit_template = NULL;
	City *city = NULL;
	if( index == TEST_SETTLERS_MOVE_0 || index == TEST_MAKE_CONTACT || index == TEST_ATTACK_CITY ) {
		map = new Map(this, 16, 16);
		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		cpu1->revealMap();
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		if( index == TEST_MAKE_CONTACT )
			new Unit(this, cpu1, unit_template, Pos2D(6, 8));
		else
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));

		cpu2 = new Civilization( this, findRace("The English") );
		cpu2->revealMap();
		if( index == TEST_MAKE_CONTACT )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
		else
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		new Unit(this, cpu2, unit_template, Pos2D(6, 6));
		// create a nearby City, so we can't build a new City
		city = new City(this, cpu2, "", 6, 5);
	}
	else if( index == TEST_SETTLERS_MOVE_1 || index == TEST_SETTLERS_MOVE_2 || index == TEST_SETTLERS_MOVE_3 || index == TEST_SETTLERS_MOVE_4 || index == TEST_ATTACK_UNIT ) {
		map = new Map(this, 16, 16);
		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		cpu1->revealMap();
		if( index == TEST_SETTLERS_MOVE_3 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		else
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, cpu1, unit_template, Pos2D(5, 5));

		cpu2 = new Civilization( this, findRace("The English") );
		cpu2->revealMap();
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		new Unit(this, cpu2, unit_template, Pos2D(6, 6));
	}
	else if( index == TEST_TRAVEL_EXPLORE_0 || index == TEST_TRAVEL_EXPLORE_1 || index == TEST_TRAVEL_EXPLORE_2 || index == TEST_TRAVEL_EXPLORE_3 ) {
		// N.B. - also tests for city close to boundary of map
		if( index == TEST_TRAVEL_EXPLORE_3 ) {
			map = new Map(this, 1000, 16);
			map->setTopology(Map::TOPOLOGY_CYLINDER);
		}
		else
			map = new Map(this, 16, 16);
		map->setAllTo(MapSquare::TYPE_OCEAN);

		map->getSquare(5, 0)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(5, 1)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(5, 2)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(5, 3)->setType(MapSquare::TYPE_GRASSLAND);

		map->getSquare(4, 4)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(5, 4)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(6, 4)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(4, 5)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(5, 5)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(6, 5)->setType(MapSquare::TYPE_GRASSLAND);

		int offx = 0; // n.b., also defined below!
		if( index == TEST_TRAVEL_EXPLORE_3 )
			offx = 993;
		else
			offx = 4;
		map->getSquare(offx, 8)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(offx+1, 8)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(offx+2, 8)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(offx, 9)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(offx+1, 9)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(offx+2, 9)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(offx, 10)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(offx+1, 10)->setType(MapSquare::TYPE_GRASSLAND);
		map->getSquare(offx+2, 10)->setType(MapSquare::TYPE_GRASSLAND);

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Horsemen"));
		if( index == TEST_TRAVEL_EXPLORE_0 || index == TEST_TRAVEL_EXPLORE_3 ) {
			// horsemen at city
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else if( index == TEST_TRAVEL_EXPLORE_1 ) {
			// horsemen near city
			new Unit(this, cpu1, unit_template, Pos2D(5, 4));
		}
		else if( index == TEST_TRAVEL_EXPLORE_2 ) {
			// horsemen near north city
			new Unit(this, cpu1, unit_template, Pos2D(5, 2));
		}
		// add Spearmen so the city is defended
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		city = new City(this, cpu1, "", 5, 5);
		if( index == TEST_TRAVEL_EXPLORE_3 )
			city->addImprovementTest(findImprovement("Port")); // need a Port, as we're testing a larger distance
		else
			city->addImprovementTest(findImprovement("Harbour"));
		city = new City(this, cpu1, "", 5, 0);
		new Unit(this, cpu1, unit_template, Pos2D(5, 0));
		new Unit(this, cpu1, unit_template, Pos2D(5, 0));
	}
	else if( index == TEST_SETTLER_BUILDCITY_0 || index == TEST_SETTLER_BUILDCITY_1 || index == TEST_SETTLER_BUILDCITY_2 || index == TEST_SETTLER_BUILDCITY_3 || index == TEST_SETTLER_BUILDCITY_4 || index == TEST_SETTLER_BUILDCITY_5 || index == TEST_SETTLER_BUILDCITY_6 || index == TEST_SETTLER_BUILDCITY_7 || index == TEST_SETTLER_BUILDCITY_8 || index == TEST_SETTLER_BUILDCITY_9 ) {
		map = new Map(this, 16, 16);
		if( index == TEST_SETTLER_BUILDCITY_9 ) {
			map->setTopology(Map::TOPOLOGY_CYLINDER);
		}

		if( index == TEST_SETTLER_BUILDCITY_5 || index == TEST_SETTLER_BUILDCITY_9 ) {
			//map->setAllTo(MapSquare::TYPE_GRASSLAND);
			// grassland by default
		}
		else if( index == TEST_SETTLER_BUILDCITY_6 )
			map->setAllTo(MapSquare::TYPE_DESERT);
		else {
			map->setAllTo(MapSquare::TYPE_OCEAN);
			map->getSquare(4, 4)->setType(MapSquare::TYPE_DESERT);
			map->getSquare(5, 4)->setType(MapSquare::TYPE_DESERT);
			if( index == TEST_SETTLER_BUILDCITY_8 )
				map->getSquare(6, 4)->setType(MapSquare::TYPE_FOREST);
			else
				map->getSquare(6, 4)->setType(MapSquare::TYPE_DESERT);
			map->getSquare(4, 5)->setType(MapSquare::TYPE_DESERT);
			map->getSquare(5, 5)->setType(index == TEST_SETTLER_BUILDCITY_4 ? MapSquare::TYPE_DESERT : MapSquare::TYPE_GRASSLAND);
			map->getSquare(6, 5)->setType(MapSquare::TYPE_DESERT);
			if( index == TEST_SETTLER_BUILDCITY_3 || index == TEST_SETTLER_BUILDCITY_4 )
				map->getSquare(4, 6)->setType(MapSquare::TYPE_DESERT);
			else if( index == TEST_SETTLER_BUILDCITY_7 )
				map->getSquare(4, 6)->setType(MapSquare::TYPE_HILLS);
			else
				map->getSquare(4, 6)->setType(MapSquare::TYPE_GRASSLAND);
			map->getSquare(5, 6)->setType(MapSquare::TYPE_DESERT);
			map->getSquare(6, 6)->setType(MapSquare::TYPE_DESERT);
		}

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		cpu1->revealMap(); // needed so the AI can see the terrain, when making the decisions for best city location
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
		if( index == TEST_SETTLER_BUILDCITY_0 ) {
			new Unit(this, cpu1, unit_template, Pos2D(4, 6));
		}
		else if( index == TEST_SETTLER_BUILDCITY_1 || index == TEST_SETTLER_BUILDCITY_3 || index == TEST_SETTLER_BUILDCITY_4 || index == TEST_SETTLER_BUILDCITY_5 || index == TEST_SETTLER_BUILDCITY_6 || index == TEST_SETTLER_BUILDCITY_7 || index == TEST_SETTLER_BUILDCITY_8 ) {
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else if( index == TEST_SETTLER_BUILDCITY_2 ) {
			new Unit(this, cpu1, unit_template, Pos2D(6, 5));
		}
		else if( index == TEST_SETTLER_BUILDCITY_9 ) {
			new Unit(this, cpu1, unit_template, Pos2D(15, 5));
			new City(this, cpu1, "", 1, 5);
			cpu1->revealMap(); // so the Settler unit can move to a sensible square to build city, rather than exploring
		}
		else {
			ASSERT( false );
		}
	}
	else if( index == TEST_DEFEND_0 || index == TEST_DEFEND_1 ) {
		map = new Map(this, 16, 16);

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		city = new City(this, cpu1, "", 5, 5);
		city->setSize(3);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Knights")); // not swordsmen, as that can work squares
		new Unit(this, cpu1, unit_template, Pos2D(5, 5));
	}
	else if( index == TEST_BUILDROAD_0 || index == TEST_BUILDROAD_1 || index == TEST_BUILDROAD_6 || index == TEST_BUILDROAD_7 || index == TEST_BUILDRAILWAYS_0 || index == TEST_BUILDRAILWAYS_1 ) {
		map = new Map(this, 16, 16);

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		const Technology *tech = this->findTechnology("Steam Power");
		cpu1->addTechnology(tech);
		city = new City(this, cpu1, "", 5, 5);
		if( index == TEST_BUILDROAD_6 || index == TEST_BUILDROAD_7 ) {
			city->setSize(2);
		}
		else {
			city->setSize(3);
		}
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		if( index == TEST_BUILDROAD_7 ) {
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else {
			new Unit(this, cpu1, unit_template, Pos2D(5, 6));
		}
		if( index == TEST_BUILDRAILWAYS_0 || index == TEST_BUILDRAILWAYS_1 ) {
			map->getSquare(5, 6)->setRoad(MapSquare::ROAD_BASIC);
		}
		map->getSquare(4, 4)->setRoad(MapSquare::ROAD_RAILWAYS);
		map->getSquare(5, 4)->setRoad(MapSquare::ROAD_RAILWAYS);
		map->getSquare(6, 4)->setRoad(MapSquare::ROAD_RAILWAYS);
	}
	else if( index == TEST_BUILDROAD_2 || index == TEST_BUILDROAD_3 || index == TEST_BUILDROAD_4 || index == TEST_BUILDROAD_5 || index == TEST_BUILDRAILWAYS_2 || index == TEST_BUILDRAILWAYS_3 || index == TEST_BUILDRAILWAYS_4 || index == TEST_BUILDRAILWAYS_5 ) {
		map = new Map(this, 16, 16);

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		const Technology *tech = this->findTechnology("Steam Power");
		cpu1->addTechnology(tech);
		city = new City(this, cpu1, "", 5, 5);
		city->setSize(3);
		city = new City(this, cpu1, "", 11, 5);
		city->setSize(3);
		if( index == TEST_BUILDROAD_4 || index == TEST_BUILDROAD_5 || index == TEST_BUILDRAILWAYS_4 || index == TEST_BUILDRAILWAYS_5 ) {
			for(int i=0;i<cpu1->getNCities();i++) {
				city = cpu1->getCity(i);
				for(int j=0;j<city->getNCitySquares();j++) {
					Pos2D pos = city->getCitySquare(j);
					if( map->isValid(pos.x, pos.y) ) {
						map->getSquare(pos)->setRoad(MapSquare::ROAD_BASIC);
					}
				}
			}
		}
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, cpu1, unit_template, Pos2D(8, 5));
		if( index == TEST_BUILDRAILWAYS_2 || index == TEST_BUILDRAILWAYS_3 ) {
			map->getSquare(8, 5)->setRoad(MapSquare::ROAD_RAILWAYS);
		}
		else if( index == TEST_BUILDRAILWAYS_4 || index == TEST_BUILDRAILWAYS_5 ) {
			map->getSquare(8, 5)->setRoad(MapSquare::ROAD_BASIC);
		}
		if( index == TEST_BUILDROAD_2 || index == TEST_BUILDROAD_3 || index == TEST_BUILDRAILWAYS_2 || index == TEST_BUILDRAILWAYS_3 ) {
			cpu1->revealMap(); // needed so that the unit moves to build roads/railways - otherwise exploring will take priority
		}
	}
	else if( index == TEST_UNIT_MOVE_0 || index == TEST_UNIT_MOVE_1 || index == TEST_UNIT_MOVE_2 || index == TEST_UNIT_MOVE_3 || index == TEST_UNIT_MOVE_4 || index == TEST_UNIT_MOVE_5 || index == TEST_UNIT_MOVE_6 || index == TEST_UNIT_MOVE_7 || index == TEST_UNIT_MOVE_8 || index == TEST_UNIT_MOVE_9 ) {
		map = new Map(this, 16, 16);

		if( index == TEST_UNIT_MOVE_9 ) {
			map->setTopology(Map::TOPOLOGY_CYLINDER);
		}

		cpu1 = new Civilization( this, findRace("The Romans") );
		cpu1->revealMap();

		if( index == TEST_UNIT_MOVE_0 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Horsemen"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			map->getSquare(5, 5)->setRoad(MapSquare::ROAD_BASIC);
			map->getSquare(6, 6)->setRoad(MapSquare::ROAD_BASIC);
			map->getSquare(7, 5)->setRoad(MapSquare::ROAD_BASIC);
		}
		else if( index == TEST_UNIT_MOVE_1 || index == TEST_UNIT_MOVE_7 || index == TEST_UNIT_MOVE_8 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Horsemen"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			for(int y=0;y<map->getHeight();y++) {
				for(int x=0;x<map->getWidth();x++) {
					map->getSquare(x, y)->setRoad(MapSquare::ROAD_BASIC);
				}
			}
			map->getSquare(5, 5)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(6, 6)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(7, 5)->setRoad(MapSquare::ROAD_RAILWAYS);

			if( index == TEST_UNIT_MOVE_7 || index == TEST_UNIT_MOVE_8 ) {
				cpu2 = new Civilization(this, findRace("The Greeks") );
				new City(this, cpu2, "", 5, 4);
			}
			if( index == TEST_UNIT_MOVE_8 ) {
				Relationship *relationship = this->findRelationship(cpu1, cpu2);
				relationship->makeContact();
				relationship->setStatus(Relationship::STATUS_WAR);
			}
		}
		else if( index == TEST_UNIT_MOVE_2 || index == TEST_UNIT_MOVE_3 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Horsemen")); // need unit with moves 2, for test TEST_UNIT_MOVE_2
			new Unit(this, cpu1, unit_template, Pos2D(4, 6));
			MapSquare::Road road = index == TEST_UNIT_MOVE_2 ? MapSquare::ROAD_BASIC : MapSquare::ROAD_RAILWAYS;
			map->getSquare(5, 5)->setRoad(road);
			map->getSquare(4, 6)->setRoad(road);
			map->getSquare(5, 7)->setRoad(road);
			map->getSquare(6, 7)->setRoad(road);
		}
		else if( index == TEST_UNIT_MOVE_4 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants")); // need a unit with moves 1, to test an infinite loop bug
			new Unit(this, cpu1, unit_template, Pos2D(9, 10));
			map->getSquare(7, 9)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(7, 10)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(8, 10)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(9, 10)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(8, 9)->setType(MapSquare::TYPE_FOREST);
			map->getSquare(9, 9)->setType(MapSquare::TYPE_FOREST);
			map->getSquare(10, 9)->setType(MapSquare::TYPE_FOREST);
			map->getSquare(10, 8)->setType(MapSquare::TYPE_HILLS);
		}
		else if( index == TEST_UNIT_MOVE_5 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			map->getSquare(5, 5)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(5, 6)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(5, 7)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(5, 8)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(6, 8)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(7, 8)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(7, 7)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(7, 6)->setRoad(MapSquare::ROAD_RAILWAYS);
			map->getSquare(7, 5)->setRoad(MapSquare::ROAD_RAILWAYS);
		}
		else if( index == TEST_UNIT_MOVE_6 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers")); // must be a unit with 0 defence
			new Unit(this, cpu1, unit_template, Pos2D(0, 3));
			cpu2 = new Civilization( this, findRace("The English") );
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants")); // must be a unit capable of attack
			new Unit(this, cpu2, unit_template, Pos2D(1, 1));
			for(int y=0;y<=3;y++) {
				for(int x=0;x<=3;x++) {
					map->getSquare(x, y)->setRoad(MapSquare::ROAD_RAILWAYS);
				}
			}
			Relationship *relationship = this->findRelationship(cpu1, cpu2);
			// check we've made contact (as adjacent units)
			relationship->makeContact();
			relationship->setStatus(Relationship::STATUS_WAR);
		}
		else if( index == TEST_UNIT_MOVE_9 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Horsemen"));
			new Unit(this, cpu1, unit_template, Pos2D(0, 5));
		}
		else {
			ASSERT( false );
		}

		map->createAlphamap();

	}
	else if( index == TEST_UNIT_MOVE_10 || index == TEST_UNIT_MOVE_11 || index == TEST_UNIT_MOVE_12 || index == TEST_UNIT_MOVE_13 ) {
		map = new Map(this, 16, 16);
		map->setAllTo(MapSquare::TYPE_OCEAN);

		const int ypos = 5;
		for(int x=1;x<=14;x++) {
			map->getSquare(x, ypos)->setType(MapSquare::TYPE_GRASSLAND);
			if( x > 1 ) {
				if( index == TEST_UNIT_MOVE_13 )
					map->getSquare(x, ypos)->setRoad(MapSquare::ROAD_RAILWAYS);
				else
					map->getSquare(x, ypos)->setRoad(MapSquare::ROAD_BASIC);
			}
		}
		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Vikings") );
		city = new City(this, cpu1, "City", 1, ypos);
		city->addImprovementTest(findImprovement("Port"));

		if( index == TEST_UNIT_MOVE_12 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Horsemen"));
		else
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, cpu1, unit_template, Pos2D(1, ypos));
	}
	else if( index == TEST_UNIT_RETREAT_0 || index == TEST_UNIT_RETREAT_1 || index == TEST_UNIT_RETREAT_2 ) {
		map = new Map(this, 16, 16);
		for(int y=0;y<map->getHeight();y++) {
			for(int x=0;x<map->getWidth();x++) {
				if( x == 0 || x == map->getWidth()-1 || y == 0 || y == map->getHeight()-1 ) {
					map->getSquare(x, y)->setType(MapSquare::TYPE_OCEAN);
				}
			}
		}

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		cpu1->revealMap();

		city = new City(this, cpu1, "Rome", 1, 1);
		if( index == TEST_UNIT_RETREAT_0 ) {
			city->addImprovementTest(findImprovement("Port"));
		}

		city = new City(this, cpu1, "Pompeii", 14, 1);

		if( index == TEST_UNIT_RETREAT_0 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Modern Tanks"));
			new Unit(this, cpu1, unit_template, Pos2D(12, 2));
		}
		else if( index == TEST_UNIT_RETREAT_1 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Modern Tanks"));
			new Unit(this, cpu1, unit_template, Pos2D(3, 2));
		}
		else if( index == TEST_UNIT_RETREAT_2 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Modern Tanks"));
			new Unit(this, cpu1, unit_template, Pos2D(14, 1));
		}
		else {
			ASSERT(false);
		}

		if( index == TEST_UNIT_RETREAT_1 || index == TEST_UNIT_RETREAT_2 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Modern Tanks"));
			new Unit(this, cpu1, unit_template, Pos2D(1, 1));
			new Unit(this, cpu1, unit_template, Pos2D(1, 1));
			new Unit(this, cpu1, unit_template, Pos2D(14, 1));
		}
	}
	else if(
		index == TEST_AIR_BOMB_0 ||
		index == TEST_AIR_BOMB_1 ||
		index == TEST_AIR_BOMB_2 ||
		index == TEST_AIR_BOMB_3 ||
		index == TEST_AIR_BOMB_4 ||
		index == TEST_AIR_BOMB_5 ||
		index == TEST_AIR_BOMB_6 ||
		index == TEST_AIR_BOMB_7 )
	{
		map = new Map(this, 16, 16);

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		city = new City(this, cpu1, "Rome", 5, 5);

		if( index == TEST_AIR_BOMB_0 || index == TEST_AIR_BOMB_2 )
			city->addImprovementTest(findImprovement("Airport"));
		else {
			cpu1->revealMap(); // only need to reveal map if no airport
		}

		if( index == TEST_AIR_BOMB_4 || index == TEST_AIR_BOMB_6 )
			city->addImprovementTest(findImprovement("Missile Launcher"));

		if( index == TEST_AIR_BOMB_0 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Bombers"));
		else if( index == TEST_AIR_BOMB_1 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Jet Bombers"));
		else if( index == TEST_AIR_BOMB_2 || index == TEST_AIR_BOMB_3 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Fission Bomb"));
		else if( index == TEST_AIR_BOMB_4 || index == TEST_AIR_BOMB_5 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Cruise Missile"));
		else if( index == TEST_AIR_BOMB_6 || index == TEST_AIR_BOMB_7 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Nuclear Missile"));
		else {
			ASSERT(false);
		}

		new Unit(this, cpu1, unit_template, Pos2D(5, 5));

		cpu2 = new Civilization( this, findRace("The French") );
		city = new City(this, cpu2, "Paris", 10, 10);
		city->setSize(20); // make worth bombing
		// add some improvements to potentially destroy
		city->addImprovementTest(findImprovement("Mine"));
		city->addImprovementTest(findImprovement("Factory"));

		Relationship *relationship = this->findRelationship(cpu1, cpu2);
		relationship->makeContact();
		relationship->setStatus(Relationship::STATUS_WAR);
	}
	else if(
		index == TEST_AIR_REBASE_0 ||
		index == TEST_AIR_REBASE_1 ||
		index == TEST_AIR_REBASE_2 ||
		index == TEST_AIR_REBASE_3 ||
		index == TEST_AIR_REBASE_4 ||
		index == TEST_AIR_REBASE_5 ||
		index == TEST_AIR_REBASE_6 )
	{
		map = new Map(this, 16, 16);

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );

		city = new City(this, cpu1, "", 5, 5);
		city->addImprovementTest(findImprovement("Port"));
		if( index == TEST_AIR_REBASE_0 || index == TEST_AIR_REBASE_1 || index == TEST_AIR_REBASE_3 || index == TEST_AIR_REBASE_4 || index == TEST_AIR_REBASE_5 )
			city->addImprovementTest(findImprovement("Airport"));

		city = new City(this, cpu1, "", 10, 10);
		if( index == TEST_AIR_REBASE_0 || index == TEST_AIR_REBASE_2 || index == TEST_AIR_REBASE_3 || index == TEST_AIR_REBASE_4 || index == TEST_AIR_REBASE_6 )
			city->addImprovementTest(findImprovement("Airport"));

		if( index == TEST_AIR_REBASE_0 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Bombers"));
		else if( index == TEST_AIR_REBASE_1 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Jet Bombers"));
		else if( index == TEST_AIR_REBASE_2 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Stealth Bombers"));
		else if( index == TEST_AIR_REBASE_3 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Cruise Missile"));
		else if( index == TEST_AIR_REBASE_4 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		else if( index == TEST_AIR_REBASE_5 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Tanks"));
		else if( index == TEST_AIR_REBASE_6 )
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Cavalry"));
		else {
			ASSERT(false);
		}

		new Unit(this, cpu1, unit_template, Pos2D(5, 5));
	}
	else if( index == TEST_ATTACK_0 || index == TEST_ATTACK_1 || index == TEST_ATTACK_2 || index == TEST_ATTACK_3 ) {
		map = new Map(this, 16, 16);

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Romans") );
		cpu1->revealMap(); // so the AI doesn't decide to explore instead

		city = new City(this, cpu1, "", 5, 5);
		city->setSize(3); // so min defenders is 2

		if( index == TEST_ATTACK_0 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Swordsmen"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else if( index == TEST_ATTACK_1 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Swordsmen"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else if( index == TEST_ATTACK_2 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Infantry"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else if( index == TEST_ATTACK_3 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Infantry"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Tanks"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else {
			ASSERT( false );
		}
		map->getSquare(5, 5)->setRoad(MapSquare::ROAD_BASIC); // so the units don't try to work the city square

		cpu2 = new Civilization( this, findRace("The Vikings") );

		if( index == TEST_ATTACK_0 || index == TEST_ATTACK_1 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
		}
		else if( index == TEST_ATTACK_2 || index == TEST_ATTACK_3 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Infantry"));
		}
		else {
			ASSERT( false );
		}
		new Unit(this, cpu2, unit_template, Pos2D(6, 5));

		Relationship *relationship = this->findRelationship(cpu1, cpu2);
		relationship->makeContact();
		relationship->setStatus(Relationship::STATUS_WAR);
	}
	else if( index == TEST_UNIT_EXPLORE_0 || index == TEST_UNIT_EXPLORE_1 || index == TEST_UNIT_EXPLORE_2 ) {
		map = new Map(this, 16, 16);

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Americans") );

		if( index == TEST_UNIT_EXPLORE_0 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Cavalry"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			for(int y=0;y<map->getHeight();y++) {
				for(int x=0;x<map->getWidth();x++) {
					if( x == 8 && y == 2 )
						continue;
					cpu1->uncover(x, y, 0);
				}
			}
		}
		else if( index == TEST_UNIT_EXPLORE_1 ) {
			city = new City(this, cpu1, "Washington", 6, 5);
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else if( index == TEST_UNIT_EXPLORE_2 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else {
			ASSERT(false);
		}

	}
	else if( index == TEST_CITYBUILD_0 || index == TEST_CITYBUILD_1 || index == TEST_CITYBUILD_2 || index == TEST_CITYBUILD_3 || index == TEST_CITYBUILD_4 || index == TEST_CITYBUILD_5 || index == TEST_CITYBUILD_6 || index == TEST_CITYBUILD_7 || index == TEST_CITYBUILD_8 || index == TEST_CITYBUILD_9 || index == TEST_CITYBUILD_10 || index == TEST_CITYBUILD_11 || index == TEST_CITYBUILD_12 || index == TEST_CITYBUILD_13 || index == TEST_CITYBUILD_14 || index == TEST_CITYBUILD_15 || index == TEST_CITYBUILD_16 || index == TEST_CITYBUILD_17 || index == TEST_CITYBUILD_18 ) {
		map = new Map(this, 16, 16);
		map->setAllTo(MapSquare::TYPE_OCEAN);
		for(int y=3;y<=5;y++) {
			for(int x=3;x<=5;x++) {
				map->getSquare(x, y)->setType(MapSquare::TYPE_GRASSLAND);
				if( index == TEST_CITYBUILD_1 || index == TEST_CITYBUILD_2 || index == TEST_CITYBUILD_3 || index == TEST_CITYBUILD_4 || index == TEST_CITYBUILD_6 || index == TEST_CITYBUILD_10 || index == TEST_CITYBUILD_13 || index == TEST_CITYBUILD_14 ) {
					map->getSquare(x, y)->setRoad(MapSquare::ROAD_BASIC);
				}
				else if( index == TEST_CITYBUILD_7 ) {
					map->getSquare(x, y)->setRoad(MapSquare::ROAD_RAILWAYS);
				}
			}
		}
		if( index == TEST_CITYBUILD_13 || index == TEST_CITYBUILD_14 ) {
			for(int y=0;y<=2;y++) {
				for(int x=3;x<=5;x++) {
					map->getSquare(x, y)->setType(MapSquare::TYPE_GRASSLAND);
				}
			}
		}
		if( index == TEST_CITYBUILD_17 || index == TEST_CITYBUILD_18 ) {
			map->getSquare(4, 4)->setRoad(MapSquare::ROAD_BASIC);
			map->getSquare(5, 4)->setRoad(MapSquare::ROAD_BASIC);
			map->getSquare(6, 4)->setRoad(MapSquare::ROAD_BASIC);
		}

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Americans") );
		cpu1->removeAllTechnologies(); // to get rid of starting technology, for testing purposes
		if( index == TEST_CITYBUILD_2 || index == TEST_CITYBUILD_3 || index == TEST_CITYBUILD_10 || index == TEST_CITYBUILD_13 || index == TEST_CITYBUILD_14 ) {
			cpu1->revealMap();
		}
		else if( index == TEST_CITYBUILD_4 || index == TEST_CITYBUILD_7 ) {
			for(int y=2;y<=6;y++) {
				for(int x=2;x<=6;x++) {
					cpu1->uncover(x, y, 1);
				}
			}
		}

		cpu1->addTechnology( game_g->findTechnology("Sailing") );
		cpu1->addTechnology( game_g->findTechnology("Horse Riding") );
		cpu1->addTechnology( game_g->findTechnology("Mathematics") );
		cpu1->addTechnology( game_g->findTechnology("Agriculture") );
		if( index == TEST_CITYBUILD_6 || index == TEST_CITYBUILD_7 ) {
			cpu1->addTechnology( game_g->findTechnology("Steam Power") );
		}
		if( index == TEST_CITYBUILD_0 || index == TEST_CITYBUILD_8 || index == TEST_CITYBUILD_9 || index == TEST_CITYBUILD_11 || index == TEST_CITYBUILD_12 || index == TEST_CITYBUILD_15 || index == TEST_CITYBUILD_16 ) {
			cpu1->addTechnology( game_g->findTechnology("Rocketry") );
			cpu1->addTechnology( game_g->findTechnology("Nuclear Missiles") );
		}
		if( index == TEST_CITYBUILD_9 || index == TEST_CITYBUILD_12 || index == TEST_CITYBUILD_16 ) {
			cpu1->addTechnology( game_g->findTechnology("Bronze Working") );
		}
		if( index == TEST_CITYBUILD_10 || index == TEST_CITYBUILD_13 || index == TEST_CITYBUILD_14 ) {
			cpu1->addTechnology( game_g->findTechnology("Industrial Revolution") );
			cpu1->addTechnology( game_g->findTechnology("Motorised Warfare") );
		}
		if( index == TEST_CITYBUILD_15 ) {
			cpu1->addTechnology( game_g->findTechnology("Industrial Revolution") );
		}

		if( index == TEST_CITYBUILD_0 || index == TEST_CITYBUILD_15 ) {
		}
		else if( index == TEST_CITYBUILD_9 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else if( index == TEST_CITYBUILD_11 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}
		else if( index == TEST_CITYBUILD_12 || index == TEST_CITYBUILD_16 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			if( index == TEST_CITYBUILD_12 ) {
				unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
				new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			}
		}
		else {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
			new Unit(this, cpu1, unit_template, Pos2D(5, 5));
		}

		if( index == TEST_CITYBUILD_2 || index == TEST_CITYBUILD_9 || index == TEST_CITYBUILD_11 || index == TEST_CITYBUILD_12 ) {
			// if no settler units at all, with only 1 city, we'd give preferences to settlers
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers"));
			new Unit(this, cpu1, unit_template, Pos2D(5, 4));
		}

		if( index == TEST_CITYBUILD_13 || index == TEST_CITYBUILD_14 ) {
			cpu2 = new Civilization( this, findRace("The Vikings") );
			city = new City(this, cpu2, "Enemy city", 5, 0);
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
			new Unit(this, cpu2, unit_template, Pos2D(5, 0));
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
			new Unit(this, cpu2, unit_template, Pos2D(5, 2));
			Relationship *relationship = this->findRelationship(cpu1, cpu2);
			relationship->makeContact();
			if( index == TEST_CITYBUILD_14 ) {
				relationship->setStatus(Relationship::STATUS_WAR);
			}
		}

		// city must be created after units, so we know if it's being defended!
		int size = 1;
		if( index == TEST_CITYBUILD_2 ) {
			size = size_limit_farmland_c;
			this->year = 100;
		}
		else if( index == TEST_CITYBUILD_5 || index == TEST_CITYBUILD_6 || index == TEST_CITYBUILD_10 || index == TEST_CITYBUILD_13 || index == TEST_CITYBUILD_14 || index == TEST_CITYBUILD_17 || index == TEST_CITYBUILD_18 ) {
			size = 2; // remove possibility of building peasants as "cheap unit" while growing
			this->year = 150; // prefer worker/attacker units to settlers
			if( index == TEST_CITYBUILD_17 ) {
				size = 3; // so we want to work all the city squares
			}
		}
		else if( index == TEST_CITYBUILD_8 ) {
			size = 25; // so we have enough production to consider rush-building nuclear missile
		}
		else if( index == TEST_CITYBUILD_11 || index == TEST_CITYBUILD_12 || index == TEST_CITYBUILD_16 ) {
			size = 3; // so we require 2 defenders instead of 1 (though depends on situation)
		}
		else if( index == TEST_CITYBUILD_15 ) {
			size = size_limit_farmland_c;
			this->year = 100;
		}
		city = new City(mainGamestate, cpu1, "Washington", 5, 5, size);
		if( index == TEST_CITYBUILD_0 || index == TEST_CITYBUILD_8 || index == TEST_CITYBUILD_9 || index == TEST_CITYBUILD_11 || index == TEST_CITYBUILD_12 || index == TEST_CITYBUILD_15 || index == TEST_CITYBUILD_16 ) {
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_MISSILELAUNCHER)); // needed to build nuclear missile!
			if( index == TEST_CITYBUILD_15 ) {
				city->addImprovementTest(this->findImprovement(IMPROVEMENT_FACTORY)); // needed to allow possibility of Modern Infantry
			}
		}
		else if( index == TEST_CITYBUILD_10 ) {
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_FORT));
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_HARBOUR));
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_FARMLAND));
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_SCHOOL));
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_FACTORY)); // needed to build tanks!
		}
		else if( index == TEST_CITYBUILD_13 || index == TEST_CITYBUILD_14 ) {
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_FORT));
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_HARBOUR));
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_FARMLAND));
			// leave school available to build
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_FACTORY)); // needed to build tanks!
		}
		else if( index == TEST_CITYBUILD_17 ) {
			// so we don't build any improvements in "1st phase"
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_FORT));
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_HARBOUR));
			city->addImprovementTest(this->findImprovement(IMPROVEMENT_FARMLAND));
		}
		city->setBuildable(NULL);
		city->doAI(); // should force choice again, to take into account improvements
	}
	else if( index == TEST_AIPERF_0 || index == TEST_AIPERF_1 || index == TEST_AIPERF_2 || index == TEST_AIPERF_3 || index == TEST_AIPERF_4 ) {
		map = new Map(this, 100, 40);

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Americans") );
		cpu1->revealMap();
		cpu2 = new Civilization( this, findRace("The Romans") );
		if( index == TEST_AIPERF_2 ) {
			cpu3 = new Civilization( this, findRace("The French") );
		}

		if( index == TEST_AIPERF_0 || index == TEST_AIPERF_4 ) {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
		}
		else {
			unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Horsemen"));
		}
		new Unit(this, cpu1, unit_template, Pos2D(4, 4));
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Settlers")); // must be a unit we are guaranteed to defeat
		new Unit(this, cpu2, unit_template, Pos2D(96, 4));
		if( index == TEST_AIPERF_1 || index == TEST_AIPERF_3 ) {
			new Unit(this, cpu2, unit_template, Pos2D(80, 4));
		}
		if( index == TEST_AIPERF_2 ) {
			new Unit(this, cpu3, unit_template, Pos2D(80, 4));
		}
		for(int x=4;x<=96;x++) {
			map->getSquare(x, 4)->setRoad(MapSquare::ROAD_RAILWAYS);
		}

		Relationship *relationship = this->findRelationship(cpu1, cpu2);
		relationship->makeContact();
		relationship->setStatus(Relationship::STATUS_WAR);
	}
	else if( index == TEST_AIPERF_5 || index == TEST_AIPERF_6 || index == TEST_AIPERF_7 ) {
		map = new Map(this, 50, 20);
		for(int y=0;y<map->getHeight();y++) {
			for(int x=0;x<map->getWidth();x++) {
				if( x == 0 || x == map->getWidth()-1 || y == 0 || y == map->getHeight()-1 ) {
					map->getSquare(x, y)->setType(MapSquare::TYPE_OCEAN);
				}
			}
		}

		map->createAlphamap();

		cpu1 = new Civilization( this, findRace("The Americans") );
		if( index == TEST_AIPERF_5 || index == TEST_AIPERF_6 ) {
			cpu1->revealMap();
		}
		cpu1->addTechnology( game_g->findTechnology("Sailing") );
		cpu1->addTechnology( game_g->findTechnology("Horse Riding") );
		cpu1->addTechnology( game_g->findTechnology("Mathematics") );
		cpu1->addTechnology( game_g->findTechnology("Agriculture") );
		cpu1->addTechnology( game_g->findTechnology("Bronze Working") );

		city = new City(this, cpu1, "Washington", 1, 1, 3);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu1, unit_template, city->getPos());
		new Unit(this, cpu1, unit_template, city->getPos());

		city = new City(this, cpu1, "New York", 14, 1, 1);

		city = new City(this, cpu1, "Boston", 7, 2, 2);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu1, unit_template, city->getPos());
		new Unit(this, cpu1, unit_template, city->getPos());

		city = new City(this, cpu1, "Los Angeles", 1, 6, 3);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu1, unit_template, city->getPos());
		new Unit(this, cpu1, unit_template, city->getPos());

		city = new City(this, cpu1, "Seattle", 6, 7, 5);
		city->addImprovementTest(this->findImprovement("Farmland"));
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu1, unit_template, city->getPos());
		new Unit(this, cpu1, unit_template, city->getPos());
		new Unit(this, cpu1, unit_template, city->getPos());

		// add units

		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Peasants"));
		new Unit(this, cpu1, unit_template, Pos2D(3, 1));
		new Unit(this, cpu1, unit_template, Pos2D(5, 4));

		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Cavalry"));
		new Unit(this, cpu1, unit_template, Pos2D(4, 2));

		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
		new Unit(this, cpu1, unit_template, Pos2D(6, 2));
		new Unit(this, cpu1, unit_template, Pos2D(3, 6));

		cpu2 = new Civilization( this, findRace("The Romans") );

		city = new City(this, cpu2, "Rome", 11, 6, 3);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu2, unit_template, city->getPos());
		new Unit(this, cpu2, unit_template, city->getPos());

		city = new City(this, cpu2, "Pompeii", 14, 11, 2);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu2, unit_template, city->getPos());
		new Unit(this, cpu2, unit_template, city->getPos());

		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
		new Unit(this, cpu2, unit_template, Pos2D(3, 3));

		Relationship *relationship = this->findRelationship(cpu1, cpu2);
		relationship->makeContact();
		relationship->setStatus(Relationship::STATUS_WAR);

		cpu3 = new Civilization( this, findRace("The English") );

		city = new City(this, cpu3, "London", 1, 14, 3);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu3, unit_template, city->getPos());
		new Unit(this, cpu3, unit_template, city->getPos());

		city = new City(this, cpu3, "Cambridge", 6, 12, 3);
		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Spearmen"));
		new Unit(this, cpu3, unit_template, city->getPos());
		new Unit(this, cpu3, unit_template, city->getPos());

		unit_template = static_cast<const UnitTemplate *>(this->findBuildable("Catapults"));
		new Unit(this, cpu3, unit_template, Pos2D(5, 11));

		this->year = 200; // better to test performance for later stages, when turns are slower due to more cities/units
	}
	else if( index == TEST_DISTANCES_0 || index == TEST_DISTANCES_1 ) {
		map = new Map(this);
		if( !map->readFile("conquests/data/test_world0.map") ) {
			VI_log("failed to read test map file\n");
			ASSERT(false);
		}
		else {
			cpu1 = new Civilization( this, findRace("The Romans") );
			cpu1->revealMap(); // important! so we calculate the entire distance map
			if( index == TEST_DISTANCES_1 ) {
				for(int y=0;y<map->getHeight();y++) {
					for(int x=0;x<map->getWidth();x++) {
						MapSquare *square = map->getSquare(x, y);
						if( square->isLand() ) {
							MapSquare::Road roads[9] = {MapSquare::ROAD_NONE, MapSquare::ROAD_RAILWAYS, MapSquare::ROAD_NONE, MapSquare::ROAD_BASIC, MapSquare::ROAD_BASIC, MapSquare::ROAD_RAILWAYS, MapSquare::ROAD_NONE, MapSquare::ROAD_RAILWAYS, MapSquare::ROAD_BASIC};
							int indx = ( y * y + 3 * x - 4 ) % 9;
							MapSquare::Road road = roads[indx];
							square->setRoad(road);
						}
					}
				}
			}
			// to initialise threads before we do the timing
			/*if( index == TEST_DISTANCES_1 ) {
				int x = 0, y = 0;
				const MapSquare *square = map->getSquare(x, y);
				Map::Distance *dists = map->calculateDistanceMap(cpu1, NULL, false, true, x, y, -1, -1, !square->isLand());
				delete [] dists;
			}*/
		}
	}
	else {
		ASSERT( false );
	}

	// now need to generate terrain, after creating the map!
	if( view_3d && map != NULL ) {
		this->generateTerrain();
	}

	//generateRelationships();
	if( !checkRelationships() ) {
		clear();
		return;
	}
	beginTurn();

	//const bool show = false;
	const bool show = true;
	if( show ) {
		stringstream text;
		text << "Starting test: " << (int)index;
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), infowindow_def_x_c, infowindow_def_y_c, 256, 128);
		window->doModal();
		delete window;
	}

	FILE *file = fopen(filename, "a+");
	ASSERT(file != NULL);

	// do the test
	bool test_result = true;
	int time_s = clock();
	bool perf_test = false;
	try {
		if( index == TEST_SETTLERS_MOVE_0 || index == TEST_SETTLERS_MOVE_1 || index == TEST_SETTLERS_MOVE_2 || index == TEST_SETTLERS_MOVE_3 || index == TEST_SETTLERS_MOVE_4 || index == TEST_MAKE_CONTACT ) {
			// Tests making contact between adjacent Civs at start, and AI for Settlers moving away from enemies.
			// Tests 0 and 1 are similar, but test 1 has no nearby City, so checks that the AI Settler doesn't build a City right next to the enemy.
			Relationship *relationship = this->findRelationship(cpu1, cpu2);
			ASSERT( relationship != NULL );
			// check we've made contact (as adjacent units)
			if( index == TEST_MAKE_CONTACT ) {
				if( relationship->madeContact() ) {
					throw "Shouldn't have yet set contact flag between civs with adjacent units";
				}
				cpu2->revealMap();
			}
			else {
				if( !relationship->madeContact() ) {
					throw "Failed to set contact flag between civs with adjacent units";
				}
			}
			if( index != TEST_SETTLERS_MOVE_2 && index != TEST_MAKE_CONTACT )
				relationship->setStatus(Relationship::STATUS_WAR);
			Unit *unit = cpu2->getUnit(0);
			if( index == TEST_SETTLERS_MOVE_4 )
				unit->setStatus(Unit::STATUS_BUILDING_ROAD);
			unit->doAI();
			if( index == TEST_SETTLERS_MOVE_0 || index == TEST_SETTLERS_MOVE_1 || index == TEST_SETTLERS_MOVE_4 ) {
				// expect unit to move away
				// does the unit still exist?
				if( cpu1->getNUnits() != 1 || cpu2->getNUnits() != 1 ) {
					throw "Expected 1 unit on each civ";
				}
				Pos2D pos = unit->getPos();
				VI_log("unit moved to %d, %d\n", pos.x, pos.y);
				Unit *unit1 = cpu1->getUnit(0);
				Pos2D pos1 = unit1->getPos();
				if( abs(pos.x - pos1.x) <= 1 && abs(pos.y - pos1.y) <= 1 ) {
					throw "AI Settlers should move away from enemy";
				}
			}
			else if( index == TEST_SETTLERS_MOVE_2 || index == TEST_SETTLERS_MOVE_3 ) {
				// expect unit to build new city
				if( cpu1->getNUnits() != 1 || cpu2->getNUnits() != 0 ) {
					throw "Unexpected number of units for each civ";
				}
				else if( map->getSquare(6, 6)->getCity() == NULL )
					throw "No city here";
			}
			else if( index == TEST_MAKE_CONTACT ) {
				// expected to move towards other civ, to make contact
				// does the unit still exist?
				if( cpu1->getNUnits() != 1 || cpu2->getNUnits() != 1 ) {
					throw "Expected 1 unit on each civ";
				}
				Pos2D pos = unit->getPos();
				VI_log("unit moved to %d, %d\n", pos.x, pos.y);
				if( pos.y != 7 ) {
					throw "AI Settlers should move towards other civ to make contact";
				}
				else if( !relationship->madeContact() ) {
					throw "Failed to set contact flag between civs with adjacent units";
				}
			}
			else {
				ASSERT( false );
			}
		}
		else if( index == TEST_ATTACK_CITY || index == TEST_ATTACK_UNIT ) {
			// Tests attacking an enemy City/Unit.
			// Same map as test 0/1, but we do AI on the other Civ.
			Relationship *relationship = this->findRelationship(cpu1, cpu2);
			// check we've made contact (as adjacent units)
			ASSERT( relationship != NULL );
			if( !relationship->madeContact() ) {
				throw "Failed to set contact flag between civs with adjacent units";
			}
			relationship->setStatus(Relationship::STATUS_WAR); // so the Peasants should attack
			Unit *unit = cpu1->getUnit(0);
			unit->doAI();
			if( index == TEST_ATTACK_CITY ) {
				// expected the city to be destroyed
				if( cpu1->getNUnits() != 1 || cpu2->getNUnits() != 1 ) {
					throw "Expected both units to remain";
				}
				else if( cpu1->getNCities() != 0 || cpu2->getNCities() != 0 ) {
					throw "Expected city to be destroyed";
				}
			}
			else if( index == TEST_ATTACK_UNIT ) {
				// expected a unit to be defeated
				if( cpu1->getNUnits() + cpu2->getNUnits() != 1 ) {
					throw "Expected only 1 unit remaining in total";
				}
			}
			else {
				ASSERT( false );
			}
		}
		else if( index == TEST_TRAVEL_EXPLORE_0 || index == TEST_TRAVEL_EXPLORE_1 || index == TEST_TRAVEL_EXPLORE_2 || index == TEST_TRAVEL_EXPLORE_3 ) {
			// Tests using the harbour to explore
			Unit *unit = cpu1->getUnit(0);
			unit->doAI();
			if( unit->hasMovesLeft() ) {
				unit->doAI();
			}
			Pos2D pos = unit->getPos();
			VI_log("unit moved to %d, %d\n", pos.x, pos.y);
			if( ( index == TEST_TRAVEL_EXPLORE_0 || index == TEST_TRAVEL_EXPLORE_1 || index == TEST_TRAVEL_EXPLORE_3 ) && pos.y < 8 ) {
				throw "AI should have moved to other island";
			}
			else if( index == TEST_TRAVEL_EXPLORE_2 && pos.y != 4 ) {
				throw "AI should have moved south towards city with harbour";
			}
			int offx = 0; // n.b., also defined above!
			if( index == TEST_TRAVEL_EXPLORE_3 )
				offx = 990;
			else
				offx = 4;
			if( pos.x == offx+1 && pos.y == 9 ) {
				throw "How did AI move to here?";
			}
			/*if( abs(pos.x - pos1.x) <= 1 && abs(pos.y - pos1.y) <= 1 ) {
			throw "AI Settlers should move away from enemy";
			}*/
		}
		else if( index == TEST_SETTLER_BUILDCITY_0 || index == TEST_SETTLER_BUILDCITY_1 || index == TEST_SETTLER_BUILDCITY_2 || index == TEST_SETTLER_BUILDCITY_3 || index == TEST_SETTLER_BUILDCITY_4 || index == TEST_SETTLER_BUILDCITY_5 || index == TEST_SETTLER_BUILDCITY_6 || index == TEST_SETTLER_BUILDCITY_7 || index == TEST_SETTLER_BUILDCITY_8 || index == TEST_SETTLER_BUILDCITY_9 ) {
			// Tests moving Settlers to a suitable City location
			Unit *unit = cpu1->getUnit(0);
			if( index == TEST_SETTLER_BUILDCITY_9 ) {
				// extra check
				if( this->map->getSquare( unit->getPos() )->canBuildCity() ) {
					throw "Didn't expect to be able to build a city here";
				}
				// also check some further squares
				for(int x=6;x<=12;x++) {
					for(int y=0;y<map->getHeight();y++) {
						if( !this->map->getSquare(x, y)->canBuildCity() ) {
							VI_log("failed check at %d, %d\n", x, y);
							throw "Expected to be able to build a city here";
						}
					}
				}
				for(int x=13;x<=21;x++) {
					for(int y=3;y<=7;y++) {
						if( this->map->getSquare(x, y)->canBuildCity() ) {
							VI_log("failed check at %d, %d\n", x, y);
							throw "Didn't expect to be able to build a city here";
						}
					}
				}
			}
			else {
				// extra check
				if( !this->map->getSquare( unit->getPos() )->canBuildCity() ) {
					throw "Expected to be able to build a city here";
				}
			}
			unit->doAI();
			if( index == TEST_SETTLER_BUILDCITY_0 ) {
				// Coastal grasslands - ideal spot
				if( cpu1->getNUnits() > 0 )
					throw "Should have built city here";
				else if( map->getSquare(4, 6)->getCity() == NULL )
					throw "No city here";
			}
			else if( index == TEST_SETTLER_BUILDCITY_1 ) {
				// Prefer coastal to non-coastal
				if( cpu1->getNUnits() != 1 )
					throw "Expected 1 unit";
				else if( unit->getPos() != Pos2D(4, 6) )
					throw "Move to unexpected location";
			}
			else if( index == TEST_SETTLER_BUILDCITY_2 ) {
				// Prefer grasslands to non-grasslands, even if the non-grasslands have coastal advantage
				if( cpu1->getNUnits() != 1 )
					throw "Expected 1 unit";
				else if( unit->getPos() != Pos2D(5, 5) )
					throw "Move to unexpected location";
			}
			else if( index == TEST_SETTLER_BUILDCITY_3 ) {
				// Non-coastal grasslands - better than coastal non-grasslands
				if( cpu1->getNUnits() > 0 )
					throw "Should have built city here";
				else if( map->getSquare(5, 5)->getCity() == NULL )
					throw "No city here";
			}
			else if( index == TEST_SETTLER_BUILDCITY_4 ) {
				// Prefer coastal to non-coastal, even when all are deserts
				if( cpu1->getNUnits() != 1 )
					throw "Expected 1 unit";
				else if( unit->getPos() == Pos2D(5, 5) )
					throw "Move to unexpected location";
			}
			else if( index == TEST_SETTLER_BUILDCITY_5 ) {
				// All grasslands, no coast, build city
				if( cpu1->getNUnits() > 0 )
					throw "Should have built city here";
				else if( map->getSquare(5, 5)->getCity() == NULL )
					throw "No city here";
			}
			else if( index == TEST_SETTLER_BUILDCITY_6 ) {
				// All desert, no coast, build city
				if( cpu1->getNUnits() > 0 )
					throw "Should have built city here";
				else if( map->getSquare(5, 5)->getCity() == NULL )
					throw "No city here";
			}
			else if( index == TEST_SETTLER_BUILDCITY_7 ) {
				// Prefer coastal to non-coastal
				if( cpu1->getNUnits() != 1 )
					throw "Expected 1 unit";
				else if( unit->getPos() != Pos2D(4, 6) )
					throw "Move to unexpected location";
			}
			else if( index == TEST_SETTLER_BUILDCITY_8 ) {
				// prefer forest, build city
				if( cpu1->getNUnits() > 0 )
					throw "Should have built city here";
				else if( map->getSquare(5, 5)->getCity() == NULL )
					throw "No city here";
			}
			else if( index == TEST_SETTLER_BUILDCITY_9 ) {
				// can't build city here
				if( cpu1->getNUnits() != 1 )
					throw "Expected 1 unit";
				else if( unit->getPos().x < 14 || unit->getPos().x > 15 )
					throw "Move to unexpected location";
			}
			else {
				ASSERT( false );
			}
		}
		else if( index == TEST_DEFEND_0 || index == TEST_DEFEND_1 ) {
			Unit *unit = NULL;
			if( index == TEST_DEFEND_0 )
				unit = cpu1->getUnit(0); // spearmen
			else if( index == TEST_DEFEND_1 )
				unit = cpu1->getUnit(2); // knights
			else {
				ASSERT( false );
			}
			unit->doAI();
			if( index == TEST_DEFEND_0 ) {
				// expect to defend city
				if( unit->getPos() != Pos2D(5, 5) ) {
					throw "Unit should have defended city";
				}
			}
			else if( index == TEST_DEFEND_1 ) {
				// expect to leave city to explore
				if( unit->getPos() == Pos2D(5, 5) ) {
					throw "Unit should have left city";
				}
			}
			else {
				ASSERT( false );
			}
		}
		else if( index == TEST_BUILDROAD_0 || index == TEST_BUILDROAD_1 || index == TEST_BUILDROAD_2 || index == TEST_BUILDROAD_3 || index == TEST_BUILDROAD_4 || index == TEST_BUILDROAD_5 || index == TEST_BUILDROAD_6 || index == TEST_BUILDROAD_7 || index == TEST_BUILDRAILWAYS_0 || index == TEST_BUILDRAILWAYS_1 || index == TEST_BUILDRAILWAYS_2 || index == TEST_BUILDRAILWAYS_3 || index == TEST_BUILDRAILWAYS_4 || index == TEST_BUILDRAILWAYS_5 ) {
			Unit *unit = cpu1->getUnit(0);
			Pos2D pos = unit->getPos();
			if( index == TEST_BUILDROAD_0 || index == TEST_BUILDROAD_2 || index == TEST_BUILDROAD_4 || index == TEST_BUILDROAD_6 || index == TEST_BUILDROAD_7 || index == TEST_BUILDRAILWAYS_0 || index == TEST_BUILDRAILWAYS_2 || index == TEST_BUILDRAILWAYS_4 )
				unit->doAI();
			else
				unit->automate();
			if( index == TEST_BUILDROAD_2 || index == TEST_BUILDROAD_3 || index == TEST_BUILDRAILWAYS_2 || index == TEST_BUILDRAILWAYS_3 ) {
				// expect to move
				Pos2D new_pos = unit->getPos();
				if( new_pos.x != pos.x - 1 && new_pos.x != pos.x + 1 ) {
					throw "Expected unit to move towards a city";
				}
			}
			else if( index == TEST_BUILDROAD_6 ) {
				// expect to move
				Pos2D new_pos = unit->getPos();
				if( pos == new_pos || unit->getStatus() != Unit::STATUS_NORMAL ) {
					throw "Expected unit to move";
				}
			}
			else {
				// expect to stay and build road here
				Pos2D new_pos = unit->getPos();
				if( new_pos != pos ) {
					throw "Unit shouldn't have moved";
				}
				else if( index == TEST_BUILDROAD_0 || index == TEST_BUILDROAD_1 || index == TEST_BUILDROAD_4 || index == TEST_BUILDROAD_5 || index == TEST_BUILDROAD_7 ) {
					if( unit->getStatus() != Unit::STATUS_BUILDING_ROAD ) {
						throw "Unit should have started to build road";
					}
				}
				else {
					if( unit->getStatus() != Unit::STATUS_BUILDING_RAILWAYS ) {
						throw "Unit should have started to build railways";
					}
				}
			}
		}
		else if( index == TEST_UNIT_MOVE_0 || index == TEST_UNIT_MOVE_1 || index == TEST_UNIT_MOVE_7 || index == TEST_UNIT_MOVE_8 ) {
			Unit *unit = cpu1->getUnit(0);
			int target_x = 7, target_y = 5;
			unit->moveTowardsAI(target_x, target_y);
			Pos2D new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( index == TEST_UNIT_MOVE_8 ) {
				if( new_pos != Pos2D(6, 5) ) {
					throw "Unit in unexpected position after 1st move";
				}
				else if( unit->movesUsed() != Rational(1) ) {
					throw "Unexpected number of used moves after 1st move";
				}
				else {
					unit->moveTowardsAI(target_x, target_y);
					new_pos = unit->getPos();
					VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
					if( new_pos != Pos2D(target_x, target_y) ) {
						throw "Unit in unexpected position after 2nd move";
					}
					else if( unit->movesUsed() != Rational(2) ) {
						throw "Unexpected number of used moves after 2nd move";
					}
				}
			}
			else {
				if( new_pos != Pos2D(6, 6) ) {
					throw "Unit in unexpected position after 1st move";
				}
				else if( index == TEST_UNIT_MOVE_0 && unit->movesUsed() != Rational(1, 3) ) {
					throw "Unexpected number of used moves after 1st move";
				}
				else if( ( index == TEST_UNIT_MOVE_1 || index == TEST_UNIT_MOVE_7 ) && unit->movesUsed() != Rational(0) ) {
					throw "Unexpected number of used moves after 1st move";
				}
				else {
					unit->moveTowardsAI(target_x, target_y);
					new_pos = unit->getPos();
					VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
					if( new_pos != Pos2D(target_x, target_y) ) {
						throw "Unit in unexpected position after 2nd move";
					}
					else if( index == TEST_UNIT_MOVE_0 && unit->movesUsed() != Rational(2, 3) ) {
						throw "Unexpected number of used moves after 2nd move";
					}
					else if( ( index == TEST_UNIT_MOVE_1 || index == TEST_UNIT_MOVE_7 ) && unit->movesUsed() != Rational(0) ) {
						throw "Unexpected number of used moves after 2nd move";
					}
				}
			}
		}
		else if( index == TEST_UNIT_MOVE_2 ) {
			Unit *unit = cpu1->getUnit(0);
			int target_x = 7, target_y = 7;
			unit->moveTowardsAI(target_x, target_y);
			Pos2D new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( new_pos != Pos2D(5, 7) ) {
				throw "Unit in unexpected position after 1st move";
			}
			else if( unit->movesUsed() != Rational(1, 3) ) {
				throw "Unexpected number of used moves after 1st move";
			}
			else {
				unit->moveTowardsAI(target_x, target_y);
				new_pos = unit->getPos();
				VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
				if( new_pos != Pos2D(6, 7) ) {
					throw "Unit in unexpected position after 2nd move";
				}
				else if( unit->movesUsed() != Rational(2, 3) ) {
					throw "Unexpected number of used moves after 2nd move";
				}
				else {
					unit->moveTowardsAI(target_x, target_y);
					new_pos = unit->getPos();
					VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
					if( new_pos != Pos2D(target_x, target_y) ) {
						throw "Unit in unexpected position after 3rd move";
					}
					else if( unit->movesUsed() != Rational(5, 3) ) {
						throw "Unexpected number of used moves after 3rd move";
					}
				}
			}
		}
		else if( index == TEST_UNIT_MOVE_3 ) {
			Unit *unit = cpu1->getUnit(0);
			int target_x = 7, target_y = 7;
			unit->moveTowardsAI(target_x, target_y);
			Pos2D new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( new_pos != Pos2D(5, 7) ) {
				throw "Unit in unexpected position after 1st move";
			}
			else if( unit->movesUsed() != Rational(0) ) {
				throw "Unexpected number of used moves after 1st move";
			}
			else {
				unit->moveTowardsAI(target_x, target_y);
				new_pos = unit->getPos();
				VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
				if( new_pos != Pos2D(6, 7) ) {
					throw "Unit in unexpected position after 2nd move";
				}
				else if( unit->movesUsed() != Rational(0) ) {
					throw "Unexpected number of used moves after 2nd move";
				}
				else {
					unit->moveTowardsAI(target_x, target_y);
					new_pos = unit->getPos();
					VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
					if( new_pos != Pos2D(target_x, target_y) ) {
						throw "Unit in unexpected position after 3rd move";
					}
					else if( unit->movesUsed() != Rational(1) ) {
						throw "Unexpected number of used moves after 3rd move";
					}
				}
			}
		}
		else if( index == TEST_UNIT_MOVE_4 ) {
			Unit *unit = cpu1->getUnit(0);
			int target_x = 11, target_y = 7;
			unit->moveTowardsAI(target_x, target_y);
			Pos2D new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( new_pos != Pos2D(9, 9) ) {
				throw "Unit in unexpected position after 1st move";
			}
			else if( unit->movesUsed() != Rational(2) ) {
				throw "Unexpected number of used moves after 1st move";
			}
		}
		else if( index == TEST_UNIT_MOVE_5 ) {
			Unit *unit = cpu1->getUnit(0);
			int target_x = 7, target_y = 5;
			unit->moveTowardsAI(target_x, target_y);
			Pos2D new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( new_pos != Pos2D(5, 6) ) {
				throw "Unit in unexpected position after 1st move";
			}
			else if( unit->movesUsed() != Rational(0) ) {
				throw "Unexpected number of used moves after 1st move";
			}
			else {
				unit->moveTowardsAI(target_x, target_y);
				new_pos = unit->getPos();
				VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
				if( new_pos != Pos2D(5, 7) ) {
					throw "Unit in unexpected position after 2nd move";
				}
				else if( unit->movesUsed() != Rational(0) ) {
					throw "Unexpected number of used moves after 2nd move";
				}
				else {
					unit->moveTowardsAI(target_x, target_y);
					new_pos = unit->getPos();
					VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
					if( new_pos != Pos2D(6, 8) ) {
						throw "Unit in unexpected position after 3rd move";
					}
					else if( unit->movesUsed() != Rational(0) ) {
						throw "Unexpected number of used moves after 3rd move";
					}
					else {
						unit->moveTowardsAI(target_x, target_y);
						new_pos = unit->getPos();
						VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
						if( new_pos != Pos2D(7, 7) ) {
							throw "Unit in unexpected position after 4th move";
						}
						else if( unit->movesUsed() != Rational(0) ) {
							throw "Unexpected number of used moves after 4th move";
						}
						else {
							unit->moveTowardsAI(target_x, target_y);
							new_pos = unit->getPos();
							VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
							if( new_pos != Pos2D(7, 6) ) {
								throw "Unit in unexpected position after 5th move";
							}
							else if( unit->movesUsed() != Rational(0) ) {
								throw "Unexpected number of used moves after 5th move";
							}
							else {
								unit->moveTowardsAI(target_x, target_y);
								new_pos = unit->getPos();
								VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
								if( new_pos != Pos2D(target_x, target_y) ) {
									throw "Unit in unexpected position after 6th move";
								}
								else if( unit->movesUsed() != Rational(0) ) {
									throw "Unexpected number of used moves after 6th move";
								}
							}
						}
					}
				}
			}
		}
		else if( index == TEST_UNIT_MOVE_6 ) {
			Unit *unit = cpu1->getUnit(0);
			int target_x = 0, target_y = 0;
			for(int x=0;x<=2;x++) {
				unit->setPos(Pos2D(x, 3));
				unit->moveTowardsAI(target_x, target_y);
				Pos2D new_pos = unit->getPos();
				VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
				if( new_pos != Pos2D(x, 3) ) {
					VI_log("moved from %d, %d to %d, %d\n", x, 3, new_pos.x, new_pos.y);
					throw "Unit in unexpected position";
				}
				else if( unit->movesUsed() != Rational(0) ) {
					throw "Unexpected number of used moves";
				}
			}
			for(int y=0;y<=2;y++) {
				unit->setPos(Pos2D(3, y));
				unit->moveTowardsAI(target_x, target_y);
				Pos2D new_pos = unit->getPos();
				VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
				if( new_pos != Pos2D(3, y) ) {
					VI_log("moved from %d, %d to %d, %d\n", 3, y, new_pos.x, new_pos.y);
					throw "Unit in unexpected position";
				}
				else if( unit->movesUsed() != Rational(0) ) {
					throw "Unexpected number of used moves";
				}
			}
			// here we do move, as we can reduce the "squares" count
			unit->setPos(Pos2D(3, 3));
			unit->moveTowardsAI(target_x, target_y);
			Pos2D new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( new_pos != Pos2D(3, 2) ) {
				VI_log("moved from %d, %d to %d, %d\n", 3, 3, new_pos.x, new_pos.y);
				throw "Unit in unexpected position";
			}
			else if( unit->movesUsed() != Rational(0) ) {
				throw "Unexpected number of used moves";
			}
		}
		else if( index == TEST_UNIT_MOVE_9 ) {
			Unit *unit = cpu1->getUnit(0);
			int target_x = 10, target_y = 5;
			unit->moveTowardsAI(target_x, target_y);
			Pos2D new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( new_pos != Pos2D(15, 5) ) {
				throw "Unit in unexpected position after 1st move";
			}
			else if( unit->movesUsed() != Rational(1) ) {
				throw "Unexpected number of used moves after 1st move";
			}

			unit->moveTowardsAI(target_x, target_y);
			new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( new_pos != Pos2D(14, 5) ) {
				throw "Unit in unexpected position after 2nd move";
			}
			else if( unit->movesUsed() != Rational(2) ) {
				throw "Unexpected number of used moves after 2nd move";
			}

			unit->reset(); // reset moves
			if( unit->movesUsed() != Rational(0) ) {
				throw "Failed to reset moves after 2nd move";
			}

			// now test moving back again
			target_x = 0;

			unit->moveTowardsAI(target_x, target_y);
			new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( new_pos != Pos2D(15, 5) ) {
				throw "Unit in unexpected position after 3rd move";
			}
			else if( unit->movesUsed() != Rational(1) ) {
				throw "Unexpected number of used moves after 3rd move";
			}

			unit->moveTowardsAI(target_x, target_y);
			new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( new_pos != Pos2D(0, 5) ) {
				throw "Unit in unexpected position after 4th move";
			}
			else if( unit->movesUsed() != Rational(2) ) {
				throw "Unexpected number of used moves after 4th move";
			}
		}
		else if( index == TEST_UNIT_MOVE_10 || index == TEST_UNIT_MOVE_11 || index == TEST_UNIT_MOVE_12 || index == TEST_UNIT_MOVE_13 ) {
			Unit *unit = cpu1->getUnit(0);
			Pos2D target = unit->getPos();
			if( index == TEST_UNIT_MOVE_10 )
				target.x += 2;
			else if( index == TEST_UNIT_MOVE_11 )
				target.x += 4;
			else if( index == TEST_UNIT_MOVE_12 )
				target.x += 4;
			else if( index == TEST_UNIT_MOVE_13 )
				target.x += 6;

			const Map::Distance *dists = map->calculateDistanceMap(cpu1, unit, true, true, unit->getX(), unit->getY());
			Map::Distance dist = dists[target.y*map->getWidth() + target.x];
			Map::Distance expected_dist;
			Rational expected_moves_left;
			if( index == TEST_UNIT_MOVE_10 ) {
				expected_dist.cost = 2;
				expected_dist.squares = 2;
				expected_moves_left = Rational(1, 3);
			}
			else if( index == TEST_UNIT_MOVE_11 ) {
				expected_dist.cost = 3;
				expected_dist.squares = 1;
				expected_moves_left = Rational(0);
			}
			else if( index == TEST_UNIT_MOVE_12 ) {
				expected_dist.cost = 4;
				expected_dist.squares = 4;
				expected_moves_left = Rational(2, 3);
			}
			else if( index == TEST_UNIT_MOVE_13 ) {
				expected_dist.cost = 0;
				expected_dist.squares = 6;
				expected_moves_left = Rational(1);
			}
			VI_log("dist cost %d, squares %d\n", dist.cost, dist.squares);

			if( dist.cost != expected_dist.cost ) {
				throw "Unexpected distance cost";
			}
			else if( dist.squares != expected_dist.squares ) {
				throw "Unexpected distance squares";
			}

			while( unit->hasMovesLeft() && unit->getPos() != target ) {
				unit->moveTowardsAI(target.x, target.y);
				Pos2D new_pos = unit->getPos();
				VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			}
			if( unit->getPos() != target ) {
				throw "Failed to reach target";
			}

			Rational moves_left = unit->movesLeft();
			if( moves_left != expected_moves_left ) {
				throw "Unexpected moves left";
			}

			delete [] dists;
		}
		else if( index == TEST_UNIT_RETREAT_0 || index == TEST_UNIT_RETREAT_1 || index == TEST_UNIT_RETREAT_2 ) {
			Unit *unit = cpu1->getUnit(0);
			Pos2D old_pos = unit->getPos();
			bool moved = unit->doAI();
			Pos2D new_pos = unit->getPos();
			VI_log("unit moves to %d, %d\n", new_pos.x, new_pos.y);
			if( index == TEST_UNIT_RETREAT_0 ) {
				if( new_pos.x - old_pos.x != -1 ) {
					throw "Expected unit to move west towards city";
				}
			}
			else if( index == TEST_UNIT_RETREAT_1 ) {
				if( new_pos.x - old_pos.x != 1 ) {
					throw "Expected unit to move east towards city";
				}
			}
			else if( index == TEST_UNIT_RETREAT_2 ) {
				if( moved ) {
					throw "Didn't expect unit to move";
				}
				else if( new_pos != old_pos ) {
					throw "Didn't expect unit to change position";
				}
			}
			else {
				ASSERT( false );
			}
		}
		else if(
			index == TEST_AIR_BOMB_0 ||
			index == TEST_AIR_BOMB_1 ||
			index == TEST_AIR_BOMB_2 ||
			index == TEST_AIR_BOMB_3 ||
			index == TEST_AIR_BOMB_4 ||
			index == TEST_AIR_BOMB_5 ||
			index == TEST_AIR_BOMB_6 ||
			index == TEST_AIR_BOMB_7 )
		{
			bool expected_can_bomb = index == TEST_AIR_BOMB_0 || index == TEST_AIR_BOMB_2 || index == TEST_AIR_BOMB_4 || index == TEST_AIR_BOMB_6;
			City *city = cpu1->getCity(0);
			Unit *unit = cpu1->getUnit(0);
			if( !unit->getTemplate()->canBomb() ) {
				throw "Unit can't bomb";
			}
			else if( city->canLaunch(unit->getTemplate()) != expected_can_bomb ) {
				throw "Unexpected launch ability";
			}
			Pos2D old_pos = unit->getPos();
			bool is_missile = unit->getTemplate()->isMissile();
			int old_n_bombed = unit->getNBombed();
			unit->doAI();
			if( expected_can_bomb ) {
				if( is_missile ) {
					unit = NULL; // unit should now be deleted!
					if( cpu1->getNUnits() != 0 ) {
						throw "Didn't use missile";
					}
				}
				else {
					int new_n_bombed = unit->getNBombed();
					if( unit->movesLeft() != 0 ) {
						VI_log("moves left: %d\n", unit->movesLeft());
						throw "Expected 0 moves left";
					}
					else if( new_n_bombed - old_n_bombed != 1 ) {
						throw "Unexpected n_bombed count";
					}
				}
			}
			else {
				int new_n_bombed = unit->getNBombed();
				VI_log("n_bombed was %d, now %d\n", old_n_bombed, new_n_bombed);
				/*if( unit->movesLeft() != 1 ) {
					VI_log("moves left: %d\n", unit->movesLeft());
					throw "Expected 1 move left";
				}*/
				// we now test we didn't bomb via the n_bombed counts; as we now use up all the moves, even if we can't bomb (i.e., action NOTHING rather than WAIT)
				if( new_n_bombed != old_n_bombed ) {
					throw "Didn't expect to bomb";
				}
			}
			if( unit != NULL && unit->getPos() != old_pos ) {
				throw "Didn't expect air unit to move";
			}
		}
		else if(
			index == TEST_AIR_REBASE_0 ||
			index == TEST_AIR_REBASE_1 ||
			index == TEST_AIR_REBASE_2 ||
			index == TEST_AIR_REBASE_3 ||
			index == TEST_AIR_REBASE_4 ||
			index == TEST_AIR_REBASE_5 ||
			index == TEST_AIR_REBASE_6 )
		{
			bool expected_can_rebase0 = index == TEST_AIR_REBASE_0 || index == TEST_AIR_REBASE_1 || index == TEST_AIR_REBASE_4 || index == TEST_AIR_REBASE_5;
			bool expected_can_rebase1 = index == TEST_AIR_REBASE_0 || index == TEST_AIR_REBASE_2 || index == TEST_AIR_REBASE_4 || index == TEST_AIR_REBASE_6;
			City *city0 = cpu1->getCity(0);
			City *city1 = cpu1->getCity(1);
			Unit *unit = cpu1->getUnit(0);
			int sea_range0 = 0;
			int air_range0 = 0;
			int sea_range1 = 0;
			int air_range1 = 0;
			city0->getTravelRanges(&sea_range0, &air_range0, unit->getTemplate());
			city1->getTravelRanges(&sea_range1, &air_range1, unit->getTemplate());
			if( index == TEST_AIR_REBASE_0 || index == TEST_AIR_REBASE_1 || index == TEST_AIR_REBASE_2 || index == TEST_AIR_REBASE_3 ) {
				if( sea_range0 != 0 || sea_range1 != 0 ) {
					throw "Air units shouldn't be able to travel by sea";
				}
			}
			bool can_rebase0 = air_range0 != 0;
			bool can_rebase1 = air_range1 != 0;
			if( can_rebase0 != expected_can_rebase0 || can_rebase1 != expected_can_rebase1 ) {
				throw "Unexpected rebase ability";
			}
			if( air_range0 != 0 ) {
				city0->initTravel(false, unit->getTemplate());
				bool can_travel_to = this->getMap()->getSquare(city1->getPos())->isTarget();
				if( can_travel_to != expected_can_rebase1 ) {
					throw "Unexpected travel to ability";
				}
			}
			if( can_rebase0 && can_rebase1 ) {
				this->moveTo(unit, city1->getPos(), true);
			}
		}
		else if( index == TEST_ATTACK_0 || index == TEST_ATTACK_1 || index == TEST_ATTACK_2 || index == TEST_ATTACK_3 ) {
			int n_units1 = cpu1->getNUnits();
			int n_units2 = cpu2->getNUnits();
			Unit *unit = cpu1->getUnit(0);
			Pos2D pos = unit->getPos();
			unit->doAI();
			bool expected_attack = false;
			if( index == TEST_ATTACK_0 || index == TEST_ATTACK_1 || index == TEST_ATTACK_2 ) {
				expected_attack = true;
			}
			bool did_attack = false;
			if( n_units1 != cpu1->getNUnits() ) {
				// unit must have attacked and lost
				did_attack = true;
			}
			else if( n_units2 != cpu2->getNUnits() ) {
				// unit must have attacked and won
				did_attack = true;
			}
			if( expected_attack != did_attack ) {
				throw "Unexpected attack behaviour";
			}
		}
		else if( index == TEST_UNIT_EXPLORE_0 || index == TEST_UNIT_EXPLORE_1 || index == TEST_UNIT_EXPLORE_2 ) {
			Unit *unit = cpu1->getUnit(0);
			Pos2D old_pos = unit->getPos();
			unit->doAI();
			Pos2D pos = unit->getPos();
			if( index == TEST_UNIT_EXPLORE_0 ) {
				//if( pos != Pos2D(6, 5) ) {
				if( pos != Pos2D(6, 4) ) {
					throw "Unit didn't move towards unexplored square";
				}
			}
			else if( index == TEST_UNIT_EXPLORE_1 ) {
				if( pos == old_pos ) {
					throw "Unit didn't move towards unexplored square";
				}
			}
			else if( index == TEST_UNIT_EXPLORE_2 ) {
				if( pos != old_pos ) {
					throw "Unit should have waited";
				}
				else if( !unit->hasMovesLeft() ) {
					throw "Unit should have moves left";
				}
				else {
					unit = cpu1->getUnit(1);
					unit->doAI();
					if( cpu1->getNCities() != 1 ) {
						throw "Expected 1 city";
					}
					else if( cpu1->getNUnits() != 1 ) {
						throw "Expected 1 unit to remain";
					}
				}
			}
			else {
				ASSERT( false );
			}
		}
		else if( index == TEST_CITYBUILD_0 || index == TEST_CITYBUILD_1 || index == TEST_CITYBUILD_2 || index == TEST_CITYBUILD_3 || index == TEST_CITYBUILD_4 || index == TEST_CITYBUILD_5 || index == TEST_CITYBUILD_6 || index == TEST_CITYBUILD_7 || index == TEST_CITYBUILD_8 || index == TEST_CITYBUILD_9 || index == TEST_CITYBUILD_10 || index == TEST_CITYBUILD_11 || index == TEST_CITYBUILD_12 || index == TEST_CITYBUILD_13 || index == TEST_CITYBUILD_14 || index == TEST_CITYBUILD_15 || index == TEST_CITYBUILD_16 || index == TEST_CITYBUILD_17 || index == TEST_CITYBUILD_18 ) {
			City *city = cpu1->getCity(0);
			const Buildable *buildable = city->getBuildable();
			const Buildable *exp_buildable = NULL;
			if( index == TEST_CITYBUILD_0 )
				exp_buildable = game_g->findBuildable("Peasants");
			else if( index == TEST_CITYBUILD_1 )
				exp_buildable = game_g->findBuildable("Horsemen");
			else if( index == TEST_CITYBUILD_2 )
				exp_buildable = game_g->findBuildable("Farmland");
			else if( index == TEST_CITYBUILD_3 )
				exp_buildable = game_g->findBuildable("Peasants");
			else if( index == TEST_CITYBUILD_4 )
				exp_buildable = game_g->findBuildable("Harbour");
			else if( index == TEST_CITYBUILD_5 )
				exp_buildable = game_g->findBuildable("Peasants");
			else if( index == TEST_CITYBUILD_6 )
				exp_buildable = game_g->findBuildable("Peasants");
			else if( index == TEST_CITYBUILD_7 )
				exp_buildable = game_g->findBuildable("Harbour");
			else if( index == TEST_CITYBUILD_8 )
				exp_buildable = game_g->findBuildable("Nuclear Missile");
			else if( index == TEST_CITYBUILD_9 )
				exp_buildable = game_g->findBuildable("Spearmen");
			else if( index == TEST_CITYBUILD_10 )
				exp_buildable = game_g->findBuildable("Tanks");
			else if( index == TEST_CITYBUILD_11 )
				exp_buildable = game_g->findBuildable("Settlers");
			else if( index == TEST_CITYBUILD_12 )
				exp_buildable = game_g->findBuildable("Spearmen");
			else if( index == TEST_CITYBUILD_13 )
				exp_buildable = game_g->findBuildable("School");
			else if( index == TEST_CITYBUILD_14 )
				exp_buildable = game_g->findBuildable("Tanks");
			else if( index == TEST_CITYBUILD_15 )
				exp_buildable = game_g->findBuildable("Farmland");
			else if( index == TEST_CITYBUILD_16 )
				exp_buildable = game_g->findBuildable("Settlers");
			else if( index == TEST_CITYBUILD_17 )
				exp_buildable = game_g->findBuildable("Peasants");
			else if( index == TEST_CITYBUILD_18 )
				exp_buildable = game_g->findBuildable("Horsemen");
			else {
				ASSERT( false );
			}
			if( buildable != exp_buildable ) {
				VI_log("expected to build a %s\n", exp_buildable->getName());
				throw "Unexpected choice for city build";
			}
		}
		else if( index == TEST_AIPERF_0 || index == TEST_AIPERF_1 || index == TEST_AIPERF_2 || index == TEST_AIPERF_4 ) {
			// this tests the performance of the AI, in particular, when we repeatedly call it for the same unit in the same turn
			// on 20110212:
			// TEST_AIPERF_0 on AMD 630, old algorithm: 2105
			// TEST_AIPERF_0 on AMD 630, new algorithm (ai_dest): 1023
			// TEST_AIPERF_0 on Intel Atom N450, old algorithm: 8705
			// TEST_AIPERF_0 on Intel Atom N450, new algorithm (ai_dest): 4799
			// on 20110313 (all testing with the "new" algorithm
			// TEST_AIPERF_0 on Intel Atom N450, C++: 5692
			// TEST_AIPERF_1 on Intel Atom N450, C++: 5135
			// TEST_AIPERF_2 on Intel Atom N450, C++: 5133
			// TEST_AIPERF_3 on Intel Atom N450, C++: 4963
			// TEST_AIPERF_4 on Intel Atom N450, C++: 12569 [Time spent in AI routine: 6077; Time spent in distance map: 10267]
			// TEST_AIPERF_0 on Intel Atom N450, Lua: 6382
			// TEST_AIPERF_1 on Intel Atom N450, Lua: 5472
			// TEST_AIPERF_2 on Intel Atom N450, Lua: 5507
			// TEST_AIPERF_3 on Intel Atom N450, Lua: 5228
			// TEST_AIPERF_4 on Intel Atom N450, Lua: 22226 [Time spent in AI routine: 12002]
			perf_test = true;
			Unit *unit = cpu1->getUnit(0);

			// firstly check the distance map
			Map::Distance *dists = map->calculateDistanceMap(cpu1, unit, false, true, unit->getX(), unit->getY());
			Map::Distance dist = dists[4 * map->getWidth() + 96];
			VI_log("dist cost %d, squares %d\n", dist.cost, dist.squares);
			if( ( index == TEST_AIPERF_0 || index == TEST_AIPERF_4 ) && ( dist.cost != 0 || dist.squares != 92 ) ) {
				throw "Unexpected cost or squares";
			}
			else if( ( index == TEST_AIPERF_1 || index == TEST_AIPERF_2 ) && ( dist.cost != road_move_scale_c || dist.squares != 92 ) ) {
				throw "Unexpected cost or squares";
			}
			delete [] dists;

			for(;;) {
				Pos2D old_pos = unit->getPos();
				if( !unit->doAI() ) {
					VI_log("old pos: %d, %d\n", old_pos);
					VI_log("new pos: %d, %d\n", unit->getPos());
					throw "Failed to move";
				}
				if( index == TEST_AIPERF_4 ) {
					unit->setAIDest();
				}
				if( index == TEST_AIPERF_2 && cpu3->getNUnits() == 0 ) {
					throw "Didn't expect to attack friendly civilization!";
				}
				if( cpu2->getNUnits() == 0 ) {
					// must have defeated the enemy(s)
					if( unit->hasMovesLeft() ) {
						throw "Shouldn't have moves left after defeating enemy";
					}
					break;
				}
				else {
					if( !unit->hasMovesLeft() ) {
						throw "Didn't expect to use up all the moves";
					}
					if( ( index == TEST_AIPERF_0 || index == TEST_AIPERF_2 || index == TEST_AIPERF_4 ) && unit->getPos() == old_pos ) {
						// in test TEST_AIPERF_1, we won't be moving when attacking the 1st unit
						throw "Unit failed to move position";
					}
				}
			}
		}
		else if( index == TEST_AIPERF_5 || index == TEST_AIPERF_6 || index == TEST_AIPERF_7 ) {
			// this tests the performance of the AI, for a range of cities/units
			perf_test = true;
			const int n_iter = index == TEST_AIPERF_5 ? 10 : 100;
			for(int i=0;i<n_iter;i++) {
				if( index == TEST_AIPERF_5 ) {
					for(int j=0;j<cpu1->getNCities();j++) {
						City *city = cpu1->getCity(j);
						city->setBuildable(NULL);
						city->doAI();
					}
				}
				else if( index == TEST_AIPERF_6 || index == TEST_AIPERF_7 ) {
					for(int j=0;j<cpu1->getNUnits();j++) {
						Unit *unit = cpu1->getUnit(j);
						unit->reset();
						unit->doAI();
					}
				}
				else {
					ASSERT( false );
				}
			}
		}
		else if( index == TEST_AIPERF_3 ) {
			// this tests the performance of the AI movement routines
			perf_test = true;
			Unit *unit = cpu1->getUnit(0);

			Pos2D target(96, 4);

			for(;;) {
				Pos2D old_pos = unit->getPos();
				if( !unit->moveTowardsAI(target.x, target.y) ) {
					VI_log("old pos: %d, %d\n", old_pos);
					VI_log("new pos: %d, %d\n", unit->getPos());
					throw "Failed to move";
				}
				if( cpu2->getNUnits() == 1 ) {
					// must have defeated the enemy
					if( unit->hasMovesLeft() ) {
						throw "Shouldn't have moves left after defeating enemy";
					}
					break;
				}
				else {
					if( !unit->hasMovesLeft() ) {
						throw "Didn't expect to use up all the moves";
					}
					if( unit->getPos() == old_pos ) {
						throw "Unit failed to move position";
					}
				}
			}
		}
		else if( index == TEST_DISTANCES_0 || index == TEST_DISTANCES_1 ) {
			perf_test = true;
			//game_g->getGraphicsEnvironment()->setSMP( index == TEST_DISTANCES_1 );
			/*
			// TEST_DISTANCES_0 on AMD 630: 13019, 12987, 13294
			// TEST_DISTANCES_0 on Intel Core Duo: 24766, 24953
			*/
			// on 20100405:
			// TEST_DISTANCES_0 on AMD 630, old algorithm: 12839, 12652
			// TEST_DISTANCES_0 on AMD 630, Dijkstra's algorithm: 10359, 10327
			// TEST_DISTANCES_1 on AMD 630, old algorithm: 16318, 16302
			// TEST_DISTANCES_1 on AMD 630, Dijkstra's algorithm: 10390, 10405
			// TEST_DISTANCES_0 on Intel Core Duo, old algorithm: 26860, 26484
			// TEST_DISTANCES_0 on Intel Core Duo, Dijkstra's algorithm: 17062, 17031, 17234
			// TEST_DISTANCES_1 on Intel Core Duo, old algorithm: 33312, 33031
			// TEST_DISTANCES_1 on Intel Core Duo, Dijkstra's algorithm: 17203, 17235
			// on 20110105
			// TEST_DISTANCES_0 on Intel Atom N450, Dijkstra's algorithm: 37414
			// TEST_DISTANCES_1 on Intel Atom N450, Dijkstra's algorithm: 36956
			{
				for(int y=0;y<map->getHeight();y++)
					//int y = 0;
				{
					//VI_log("%d\n", y);
					for(int x=0;x<map->getWidth();x++)
						//int x = 0;
					{
						//VI_log(">>> %d, %d\n", x, y);
						const MapSquare *square = map->getSquare(x, y);
						Map::Distance *dists = map->calculateDistanceMap(cpu1, NULL, false, true, x, y, -1, -1, !square->isLand());
						/*for(int cy=0;cy<map->getHeight();cy++) {
							for(int cx=0;cx<map->getWidth();cx++) {
								int dist = dists[cy*map->getWidth()+cx].cost;
								if( dist != -1 ) {
									VI_log("        %d, %d: %d\n", cx, cy, dist);
								}
							}
						}*/
						delete [] dists;
					}
				}
			}
			/*for(int i=0;i<200;i++) {
			int x = 0, y = 0;
			const MapSquare *square = map->getSquare(x, y);
			Map::Distance *dists = map->calculateDistanceMap(cpu1, NULL, false, true, x, y, -1, -1, !square->isLand());
			delete [] dists;
			}*/
		}
		else {
			ASSERT( false );
		}
	}
	catch(const char *error) {
		VI_log("TEST ERROR: %s\n", error);
		fprintf(file, "%d,FAILED,%s\n", (int)index, error);
		test_result = false;
	}
	int time_e = clock();
	int time_taken = time_e - time_s;
	VI_log("Test Time: %d\n", time_taken);
	if( test_result ) {
		fprintf(file, "%d,PASSED", (int)index);
		if( perf_test ) {
			fprintf(file, ",%d", time_taken);
		}
		fprintf(file, "\n");
	}
	fclose(file);

	if( show ) {
		stringstream text;
		text << "Finishing test: " << (int)index;
		text << "\nResult: " << ( test_result ? "PASSED" : "FAILED" );
		if( test_result && perf_test ) {
			text << "\nTime: " << time_taken;
		}
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getPanel(), text.str(), game_g->getFont(), game_g->getPanelTexture(), infowindow_def_x_c, infowindow_def_y_c, 256, 128);
		window->doModal();
		delete window;
	}

	clearGameData();
}

void MainGamestate::runTests(TestID id) {
	test_mode = true;

	if( !init() ) {
		clear();
		return;
	}

	// write file header
	char filename[] = "conquests_test_results.csv";
	{
		FILE *file = fopen(filename, "w+");
		ASSERT(file != NULL);
		time_t time_val;
		time(&time_val);
		fprintf(file, "%s\n", ctime(&time_val));
		fprintf(file, "TEST,RESULT\n");
		fclose(file);
	}

	if( id == TEST_ALL ) {
		for(int i=0;i<N_TESTS;i++) {
			runTest(filename, (TestID)i);
		}
	}
	else {
		ASSERT( id >= 0 && id < N_TESTS );
		runTest(filename, id);
	}

	clear();
}

void OptionsGamestate::action(VI_Panel *source) {
	if( source == optionsGamestate->startgameButton ) {
		MainGamestate *mainGamestate = new MainGamestate();
		//if( !mainGamestate->initGameData(true) ) {
		if( !game_g->setupNewGame(mainGamestate) ) {
			VI_log("failed to setup new game\n");
			delete mainGamestate;
			quit = true;
		}
		else {
			game_g->pushNewGamestate(mainGamestate);
		}
	}
	else if( source == optionsGamestate->loadgameButton ) {
		MainGamestate *mainGamestate = new MainGamestate();
		if( !mainGamestate->loadGame() ) {
			VI_log("failed to load game (or player cancelled)\n");
			delete mainGamestate;
		}
		else {
			game_g->pushNewGamestate(mainGamestate);
		}
	}
	else if( source == optionsGamestate->quitgameButton ) {
		optionsGamestate->quitGame();
	}
}

bool OptionsGamestate::start() {
	VI_Font *font = game_g->getFont();

	gamePanel = VI_createPanel();
	//gamePanel->setPaintFunc(paintPanel, NULL);
	game_g->getGraphicsEnvironment()->setPanel(gamePanel);

	imageTexture = VI_createTexture("conquests/data/gfx/intro.jpg");
	/*VI_Image *image = VI_createImage("conquests/data/gfx/intro.jpg");
	if( !image->scale(game_g->getGraphicsEnvironment()->getWidth(), game_g->getGraphicsEnvironment()->getHeight()) ) {
		VI_log("failed to scale background image\n");
		delete image;
		return false;
	}
	VI_Texture *image_texture = VI_createTexture(image);*/
	imageButton = VI_createImageButton(imageTexture, game_g->getGraphicsEnvironment()->getWidth(), game_g->getGraphicsEnvironment()->getHeight());
	gamePanel->add(imageButton, 0, 0);

	VI_Font *largefont = game_g->getLargefont();
	stringstream titletext;
	titletext << "Conquests   (v" << versionMajor << "." << versionMinor << ")";
	titleButton = VI_createButton(titletext.str().c_str(), largefont);
	titleButton->setForeground(1.0f, 1.0f, 0.0f);
	titleButton->setBackground(0.0f, 0.0f, 0.0f);
	titleButton->setBorder(false);
	gamePanel->add(titleButton, 48, game_g->getGraphicsEnvironment()->getHeight() - 64);

	int xpos = 48;
	int ypos = 80;
	const int ydiff = 32;

	startgameButton = VI_createButton("New Game", font);
	initButton(startgameButton);
	startgameButton->setAction(action);
	startgameButton->setKeyShortcut(V_N);
	startgameButton->setPopupText("Start playing a new game (N)", font);
	gamePanel->add(startgameButton, xpos, ypos);
	ypos += ydiff;

	loadgameButton = VI_createButton("Load Game", font);
	initButton(loadgameButton);
	loadgameButton->setAction(action);
	loadgameButton->setKeyShortcut(V_L);
	loadgameButton->setPopupText("Load a previously saved game (L)", font);
	gamePanel->add(loadgameButton, xpos, ypos);
	ypos += ydiff;

	quitgameButton = VI_createButton("Exit", font);
	initButton(quitgameButton);
	quitgameButton->setAction(action);
	quitgameButton->setPopupText("Exit the game", font);
	gamePanel->add(quitgameButton, xpos, ypos);
	ypos += ydiff;

	/*VI_World *world = VI_createWorld();

	// setup world here

	game_g->getGraphicsEnvironment()->linkWorld(world);*/
	//game_g->getGraphicsEnvironment()->linkWorld(NULL);

	VI_check();
	VI_debug_mem_usage_all();

	game_g->getGraphicsEnvironment()->setVisible(true);
	game_g->getGraphicsEnvironment()->setFade(fade_duration_c, 0, 255);

	return true;
}

void OptionsGamestate::stop() {
	VI_flush(0); // delete all the gamestate objects, but leave the game level objects (which should be set at persistence level -1)
	VI_GraphicsEnvironment *genv = game_g->getGraphicsEnvironment();
	game_g->getGraphicsEnvironment()->setPanel(NULL); // as the main panel is now destroyed
}

void OptionsGamestate::quitGame() {
	VI_log("OptionsGamestate::quitGame()\n");
	quit = true;
}

void OptionsGamestate::update(int time) {
}

bool Game::setupNewGame(MainGamestate *mainGamestate) {
	// receives a new MainGamestate, or one that has clearGameData called on it
	// asks player for new game preferences, and then calls the initGameData method
	// returns false on failure
	{
		const int n_buttons_c = MainGamestate::N_DIFFICULTY;
		string buttons[n_buttons_c];
		for(int i=0;i<MainGamestate::N_DIFFICULTY;i++) {
			buttons[i] = MainGamestate::getDifficultyString((MainGamestate::Difficulty)i);
		}
		string text = "Please select a difficulty level.";
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getGamestate()->getPanel(), text, game_g->getFont(), game_g->getPanelTexture(), buttons, n_buttons_c, infowindow_def_x_c, infowindow_def_y_c, 300, 300);
		for(int i=0;i<n_buttons_c;i++) {
			//window->getButton(i)->disableKeyShortcut();
			window->getButton(i)->setKeyShortcut( MainGamestate::getDifficultyShortcut((MainGamestate::Difficulty)i) );
		}
		int difficulty_level = window->doModal();
		VI_log("difficulty level %d\n", difficulty_level);
		delete window;

		/*MainGamestate::Difficulty difficulty = MainGamestate::DIFFICULTY_UNDEFINED;
		if( difficulty_level == 0 )
			difficulty = MainGamestate::DIFFICULTY_EASY;
		else if( difficulty_level == 1 )
			difficulty = MainGamestate::DIFFICULTY_MEDIUM;
		else if( difficulty_level == 2 )
			difficulty = MainGamestate::DIFFICULTY_HARD;
		else if( difficulty_level == 3 )
			difficulty = MainGamestate::DIFFICULTY_INSANE;
		else {
			VI_log("unrecognised difficulty level\n");
			ASSERT( false );
		}
		mainGamestate->setDifficulty(difficulty);*/

		ASSERT( difficulty_level >= 0 && difficulty_level < MainGamestate::N_DIFFICULTY );
		MainGamestate::Difficulty difficulty = (MainGamestate::Difficulty)difficulty_level;
		mainGamestate->setDifficulty(difficulty);
	}

	int map_width = 0, map_height = 0;
	{
		const int n_buttons_c = 3;
		string buttons[3] = {"Small", "Medium", "Large"};
		string text = "Please select a map size.";
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getGamestate()->getPanel(), text, game_g->getFont(), game_g->getPanelTexture(), buttons, n_buttons_c, infowindow_def_x_c, infowindow_def_y_c, 300, 300);
		window->getButton(0)->setKeyShortcut(V_S);
		window->getButton(1)->setKeyShortcut(V_M);
		window->getButton(2)->setKeyShortcut(V_L);
		int map_size = window->doModal();
		VI_log("map_size %d\n", map_size);
		delete window;

		if( map_size == 0 ) {
			map_width = 75;
			map_height = 30;
		}
		else if( map_size == 1 ) {
			map_width = 100;
			map_height = 40;
		}
		else if( map_size == 2 ) {
			map_width = 150;
			map_height = 60;
		}
		else {
			VI_log("unrecognised map size\n");
			ASSERT( false );
		}
	}

	Map::Topology topology = Map::TOPOLOGY_UNDEFINED;
	{
		const int n_buttons_c = 2;
		string buttons[2] = {"Flat", "Round"};
		string text = "Please select a world type.";
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getGamestate()->getPanel(), text, game_g->getFont(), game_g->getPanelTexture(), buttons, n_buttons_c, infowindow_def_x_c, infowindow_def_y_c, 300, 300);
		window->getButton(0)->setKeyShortcut(V_F);
		window->getButton(1)->setKeyShortcut(V_R);
		int res = window->doModal();
		VI_log("topology res %d\n", res);
		if( res == 0 ) {
			topology = Map::TOPOLOGY_FLAT;
		}
		else if( res == 1 ) {
			topology = Map::TOPOLOGY_CYLINDER;
		}
		else {
			ASSERT( false );
		}
		VI_log("topology %d\n", topology);
		delete window;
	}

	int n_opponent_civs = 0;
	{
		const int min_civs_c = 4, max_civs_c = 6;
		ASSERT( max_civs_c+1 <= game_g->getGameData()->getNRaces() );
		const int n_buttons_c = max_civs_c - min_civs_c + 1;
		string buttons[n_buttons_c];
		for(int i=0;i<n_buttons_c;i++) {
			stringstream str;
			str << (min_civs_c + i) << " civs";
			buttons[i] = str.str();
		}
		string text = "Please select number of opponent civilizations.";
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getGamestate()->getPanel(), text, game_g->getFont(), game_g->getPanelTexture(), buttons, n_buttons_c, infowindow_def_x_c, infowindow_def_y_c, 300, 300);
		for(int i=0;i<n_buttons_c;i++) {
			window->getButton(i)->disableKeyShortcut();
		}
		int res = window->doModal();
		n_opponent_civs = res + min_civs_c;
		VI_log("n_opponent_civs = %d\n", n_opponent_civs);
		delete window;
	}

	{
		const int n_buttons_c = MainGamestate::N_AIAGGRESSION;
		string buttons[n_buttons_c];
		for(int i=0;i<MainGamestate::N_AIAGGRESSION;i++) {
			buttons[i] = MainGamestate::getAIAggressionString((MainGamestate::AIAggression)i);
		}
		string text = "Please select the level of AI Aggression.";
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getGamestate()->getPanel(), text, game_g->getFont(), game_g->getPanelTexture(), buttons, n_buttons_c, infowindow_def_x_c, infowindow_def_y_c, 300, 300);
		for(int i=0;i<n_buttons_c;i++) {
			window->getButton(i)->setKeyShortcut( MainGamestate::getAIAggressionShortcut((MainGamestate::AIAggression)i) );
		}
		int aiaggression_level = window->doModal();
		VI_log("aiaggression level %d\n", aiaggression_level);
		delete window;

		ASSERT( aiaggression_level >= 0 && aiaggression_level < MainGamestate::N_AIAGGRESSION );
		MainGamestate::AIAggression aiaggression = (MainGamestate::AIAggression)aiaggression_level;
		mainGamestate->setAIAggression(aiaggression);
	}

	const Race *player_race = NULL;
	{
		string okay = "Okay";
		//const int window_h_c = 512;
		int window_h_c = game_g->getGraphicsEnvironment()->getHeight() - infowindow_def_y_c - 32;
		InfoWindow *window = new InfoWindow(game_g->getGraphicsEnvironment(), this->getGamestate()->getPanel(), "", game_g->getFont(), game_g->getPanelTexture(), &okay, 1, infowindow_def_x_c, infowindow_def_y_c, infowindow_def_w_c, window_h_c);
		VI_Panel *windowPanel = window->getPanel();

		VI_Textfield *textfield = VI_createTextfield(game_g->getFont(), infowindow_def_w_c - 32, 64);
		textfield->addText("Please select a civilization that you wish to play.");
		textfield->setForeground(1.0f, 1.0f, 1.0f);
		windowPanel->add(textfield, 16, 16);

		vector<string> entries;
		for(int i=0;i<game_g->getGameData()->getNRaces();i++) {
			const Race *race = game_g->getGameData()->getRace(i);
			entries.push_back( race->getName() );
		}

		VI_Listbox *listbox = VI_createListbox(infowindow_def_w_c - 32, window_h_c - 128, &*entries.begin(), entries.size(), this->getFont());
		listbox->setHasInputFocus(true);
		listbox->setForeground(1.0f, 1.0f, 1.0f);
		listbox->setActive(0);
		windowPanel->add(listbox, 16, 80);

		window->doModal();
		int active = listbox->getActive();
		VI_log("race index %d\n", active);
		delete window;

		ASSERT( active >= 0 );
		ASSERT( active < game_g->getGameData()->getNRaces() );

		player_race = game_g->getGameData()->getRace(active);

	}

	bool ok = mainGamestate->initGameData(true, map_width, map_height, topology, player_race, n_opponent_civs);
	if( !ok ) {
		VI_log("failed to initGameData\n");
	}
	else {
		mainGamestate->beginTurn(); // initialisation for first turn
	}
	return ok;
}

class StdMessageInterface : public MessageInterface {
public:
    virtual void show(const char *message) {
        VI_log("Vision::load() message: %s\n", message);
#ifdef _WIN32
        MessageBoxA(NULL, message, "Warning", MB_OK|MB_ICONEXCLAMATION);
#endif
    }
};

bool Game::init() {
	VI_initialise("Conquests");
	const char savegames_subdirname[] = "savegames";
	savegames_dirname = VI_getApplicationFilename(savegames_subdirname);

	VisionPrefs visionPrefs;
	StdMessageInterface messageInterface;
	visionPrefs.load(&messageInterface);
	//visionPrefs.want_std_shaders = false; // no need for shaders

	genv = VI_createGraphicsEnvironment(&visionPrefs);
	if( genv == NULL ) {
		bool done = false;
		if( visionPrefs.fullscreen ) {
			VI_log("Failed to run in fullscreen mode\n");
#ifdef _WIN32
			MessageBoxA(NULL,"Failed to run in fullscreen mode","Error",MB_OK|MB_ICONEXCLAMATION);
#endif
			visionPrefs.fullscreen = false;
			genv = VI_createGraphicsEnvironment(&visionPrefs);
			if( genv != NULL )
				done = true;
		}
		if( !done ) {
			VI_log("Failed to open graphics window\n");
#ifdef _WIN32
			MessageBoxA(NULL,"Failed to open graphics window","Error",MB_OK|MB_ICONEXCLAMATION);
#endif
			return false;
		}
	}

/*#ifdef _WIN32
	HCURSOR cursor = LoadCursor( NULL, IDC_WAIT );
	SetCursor(cursor);
	SetClassLong(genv->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
	this->setCursor(cursor_wait);

	VI_log("initialising lua\n");
	ls = lua_open();
	if( ls == NULL ) {
		this->setCursor(cursor_arrow);
		VI_log("failed to open lua\n");
#ifdef _WIN32
		MessageBoxA(NULL,"Failed to open lua","Error",MB_OK|MB_ICONEXCLAMATION);
#endif
		return false;
	}
	luaL_openlibs(ls);
	registerScripts();

	//sound = VI_Sound::create( game_g->getGraphicsEnvironment()->getHWND() );
	sound = VI_Sound::create();
	if( sound == NULL ) {
	    VI_log("Failed to initialise sound\n");
#ifdef _WIN32
		MessageBoxA(NULL,"Failed to initialise sound","Error",MB_OK|MB_ICONEXCLAMATION);
#endif
	}
/*#ifdef _WIN32
	if( sound != NULL && !sound->initHWND( game_g->getGraphicsEnvironment()->getHWND() ) ) {
	    VI_log("Failed to initialise sound HWND\n");
		MessageBoxA(NULL,"Failed to initialise sound HWND","Error",MB_OK|MB_ICONEXCLAMATION);
	}
#endif*/

	//ShowCursor(FALSE); // always hide cursor
//#ifdef _WIN32
	/*char buffer[4096] = "";
	sprintf(buffer, "Conquests v%d.%d | (c) 2009-2011 Mark Harman", versionMajor, versionMinor);
	game_g->getGraphicsEnvironment()->setWindowTitle(buffer);*/
	{
		char buffer[4096] = "";
		sprintf(buffer, "Conquests v%d.%d - Loading ...", versionMajor, versionMinor);
		game_g->getGraphicsEnvironment()->setWindowTitle(buffer);
	}
//#endif

	/*VI_Texture *cursor = VI_createTextureWithMask("textures/arrow.png", 255, 255, 255);
	genv->setMouseTexture(cursor);*/

	//font = genv->createFont("Verdana",12,FW_BOLD);
#ifdef _WIN32
	/*smallfont = game_g->getGraphicsEnvironment()->createFont("Verdana",10,FW_NORMAL);
	font = game_g->getGraphicsEnvironment()->createFont("Verdana",12,FW_NORMAL);
	largefont = game_g->getGraphicsEnvironment()->createFont("Arial",24,FW_BOLD);*/
	smallfont = game_g->getGraphicsEnvironment()->createFont("c:/windows/fonts/Verdana.ttf",10,FW_NORMAL);
	font = game_g->getGraphicsEnvironment()->createFont("c:/windows/fonts/Verdana.ttf",12,FW_NORMAL);
	largefont = game_g->getGraphicsEnvironment()->createFont("c:/windows/fonts/Arial.ttf",24,FW_BOLD);
#else
	smallfont = game_g->getGraphicsEnvironment()->createFont("/usr/share/fonts/truetype/freefont/FreeSans.ttf",10,0);
	font = game_g->getGraphicsEnvironment()->createFont("/usr/share/fonts/truetype/freefont/FreeSans.ttf",12,0);
	largefont = game_g->getGraphicsEnvironment()->createFont("/usr/share/fonts/truetype/freefont/FreeSans.ttf",24,0);
#endif

	// load shaders
	this->splat_pixel_shader_ambient = NULL;
	this->splat_vertex_shader_ambient = NULL;
	if( game_g->getGraphicsEnvironment()->hasShaders() && visionPrefs.terrainsplatting ) {
		try {
			this->splat_vertex_shader_ambient = Shader::createShader(genv, "conquests/shaders/shaders.cg", "main_splat_ambient_vs", true);
			this->splat_pixel_shader_ambient = Shader::createShader(genv, "conquests/shaders/shaders.cg", "main_splat_ambient_ps", false);
		}
		catch(VisionException *ve) {
			// carry on without shaders if we can't load it
			VI_log("Failed to load terrain splatting shader\n");
			delete ve;
			if( this->splat_vertex_shader_ambient != NULL ) {
				delete this->splat_vertex_shader_ambient;
				this->splat_vertex_shader_ambient = NULL;
			}
			if( this->splat_pixel_shader_ambient != NULL ) {
				delete this->splat_pixel_shader_ambient;
				this->splat_pixel_shader_ambient = NULL;
			}
		}
	}

	try {
		MapSquare::loadTextures();

		//panelTexture = VI_createTexture("conquests/data/gfx/land.png");
		unsigned char filter_max_panel[3] = {220, 172, 40};
		//unsigned char filter_min_panel[3] = {127, 100, 20};
		unsigned char filter_min_panel[3] = {120, 90, 20};
		//VI_Image *image = VI_createImageNoise(64, 64, 16.0f, filter_max_panel, filter_min_panel, V_NOISEMODE_PERLIN, 4);
		VI_Image *image = VI_createImageNoise(256, 256, 16.0f, filter_max_panel, filter_min_panel, V_NOISEMODE_PERLIN, 4);
		panelTexture = VI_createTexture(image);

		// generate Technologies
		if( !loadScript("conquests/data/scripts/generate_technologies.lua") ) {
			throw "lua script failed";
		}

		//Technology *technology = NULL;

		// ancient
		/*
		this->addTechnology( technology = new Technology("Agriculture", 100, NULL, NULL, Technology::AGE_ANCIENT, 20) );
		this->addTechnology( technology = new Technology("Alphabet", 100, NULL, NULL, Technology::AGE_ANCIENT, 10) );
		this->addTechnology( technology = new Technology("Archery", 100, NULL, NULL, Technology::AGE_ANCIENT, 10) );
		this->addTechnology( technology = new Technology("Bronze Working", 100, NULL, NULL, Technology::AGE_ANCIENT, 20) );
		this->addTechnology( technology = new Technology("Masonry", 100, NULL, NULL, Technology::AGE_ANCIENT, 10) );
		this->addTechnology( technology = new Technology("Sailing", 100, NULL, NULL, Technology::AGE_ANCIENT, 10) );
		this->addTechnology( technology = new Technology("The Wheel", 150, NULL, NULL, Technology::AGE_ANCIENT, 10) );

		this->addTechnology( technology = new Technology("Construction", 300, this->findTechnology("Masonry"), NULL, Technology::AGE_ANCIENT, 10) );
		this->addTechnology( technology = new Technology("Iron Working", 300, this->findTechnology("Bronze Working"), NULL, Technology::AGE_ANCIENT, 30) );
		this->addTechnology( technology = new Technology("Horse Riding", 300, this->findTechnology("The Wheel"), NULL, Technology::AGE_ANCIENT, 10) );
		this->addTechnology( technology = new Technology("Mathematics", 300, this->findTechnology("Alphabet"), NULL, Technology::AGE_ANCIENT, 30) );
		this->addTechnology( technology = new Technology("Maps", 250, this->findTechnology("Sailing"), NULL, Technology::AGE_ANCIENT, 30) );
		technology->setInfo("Allows exchanging maps with other Civilizations (that also have this technology).");
		this->addTechnology( technology = new Technology("Military Tactics", 500, this->findTechnology("Archery"), this->findTechnology("Iron Working"), Technology::AGE_ANCIENT, 20) );
		this->addTechnology( technology = new Technology("Writing", 250, this->findTechnology("Alphabet"), NULL, Technology::AGE_ANCIENT, 20) );
		*/

		// medieval

		/*this->addTechnology( technology = new Technology("Castles", 800, this->findTechnology("Construction"), this->findTechnology("Mathematics"), Technology::AGE_MEDIEVAL, 30) );
		this->addTechnology( technology = new Technology("Education", 600, this->findTechnology("Mathematics"), this->findTechnology("Writing"), Technology::AGE_MEDIEVAL, 20) );
		this->addTechnology( technology = new Technology("Feudalism", 800, this->findTechnology("Military Tactics"), this->findTechnology("Horse Riding"), Technology::AGE_MEDIEVAL, 30) );
		this->addTechnology( technology = new Technology("Medieval Arms", 1000, this->findTechnology("Feudalism"), NULL, Technology::AGE_MEDIEVAL, 50) );
		this->addTechnology( technology = new Technology("Navigation", 600, this->findTechnology("Mathematics"), this->findTechnology("Maps"), Technology::AGE_MEDIEVAL, 20) );

		this->addTechnology( technology = new Technology("Gunpowder", 1200, this->findTechnology("Medieval Arms"), this->findTechnology("Castles"), Technology::AGE_MEDIEVAL, 40) );
		technology->setInfo("Enemy walls, forts and castles no longer have any effect.");
		this->addTechnology( technology = new Technology("Firearms", 1400, this->findTechnology("Gunpowder"), NULL, Technology::AGE_MEDIEVAL, 50) );
		this->addTechnology( technology = new Technology("Magnetism", 800, this->findTechnology("Navigation"), NULL, Technology::AGE_MEDIEVAL, 40) );
		technology->setInfo("Increases range of ports.");
		this->addTechnology( technology = new Technology("Physics", 600, this->findTechnology("Education"), this->findTechnology("Navigation"), Technology::AGE_MEDIEVAL, 30) );*/

		// industrial

		/*this->addTechnology( technology = new Technology("Medicine", 1400, this->findTechnology("Physics"), NULL, Technology::AGE_INDUSTRIAL, 30) );
		technology->setInfo("Increases attack and defence of units by 20%.");
		this->addTechnology( technology = new Technology("Scientific Method", 1600, this->findTechnology("Magnetism"), this->findTechnology("Medicine"), Technology::AGE_INDUSTRIAL, 30) );
		technology->setInfo("Improves effects of Universities by 10%.");
		this->addTechnology( technology = new Technology("Electricity", 1800, this->findTechnology("Scientific Method"), NULL, Technology::AGE_INDUSTRIAL, 30) );
		this->addTechnology( technology = new Technology("Steam Power", 2000, this->findTechnology("Gunpowder"), this->findTechnology("Physics"), Technology::AGE_INDUSTRIAL, 30) );
		technology->setInfo("Allows railways, and increases the range of ports to worldwide.");
		this->addTechnology( technology = new Technology("Agricultural Revolution", 2400, this->findTechnology("Steam Power"), NULL, Technology::AGE_INDUSTRIAL, 30) );
		technology->setInfo("Increases maximum city sizes, and adds 2 to city production.");
		this->addTechnology( technology = new Technology("Industrial Revolution", 3000, this->findTechnology("Agricultural Revolution"), NULL, Technology::AGE_INDUSTRIAL, 30) );
		technology->setInfo("Adds 4 to city production.");
		this->addTechnology( technology = new Technology("Nationalism", 2400, this->findTechnology("Agricultural Revolution"), this->findTechnology("Firearms"), Technology::AGE_INDUSTRIAL, 30) );
		this->addTechnology( technology = new Technology("Combustion", 3200, this->findTechnology("Industrial Revolution"), NULL, Technology::AGE_INDUSTRIAL, 30) );
		technology->setInfo("Increases maximum city sizes for cities that also have a factory.");*/

		// modern

		/*this->addTechnology( technology = new Technology("Atomic Theory", 3200, this->findTechnology("Electricity"), NULL, Technology::AGE_MODERN, 30) );
		this->addTechnology( technology = new Technology("Computer", 6400, this->findTechnology("Atomic Theory"), NULL, Technology::AGE_MODERN, 30) );
		this->addTechnology( technology = new Technology("Explosives", 3200, this->findTechnology("Combustion"), this->findTechnology("Nationalism"), Technology::AGE_MODERN, 40) );
		this->addTechnology( technology = new Technology("Motorised Warfare", 3200, this->findTechnology("Explosives"), NULL, Technology::AGE_MODERN, 40) );
		this->addTechnology( technology = new Technology("Flight", 2400, this->findTechnology("Combustion"), NULL, Technology::AGE_MODERN, 30) );
		this->addTechnology( technology = new Technology("Fission", 10000, this->findTechnology("Atomic Theory"), this->findTechnology("Explosives"), Technology::AGE_MODERN, 50) );
		technology->setAIHint(Technology::AIHINT_REQUIRES_AIRPORT);
		this->addTechnology( technology = new Technology("Advanced Flight", 6400, this->findTechnology("Flight"), this->findTechnology("Motorised Warfare"), Technology::AGE_MODERN, 30) );
		technology->setInfo("Increases the range of airports.");
		this->addTechnology( technology = new Technology("Rocketry", 12000, this->findTechnology("Advanced Flight"), NULL, Technology::AGE_MODERN, 30) );
		technology->setInfo("Increases the range of airports.");
		this->addTechnology( technology = new Technology("Nuclear Missiles", 14000, this->findTechnology("Fission"), this->findTechnology("Rocketry"), Technology::AGE_MODERN, 40) );
		this->addTechnology( technology = new Technology("Spaceflight", 12000, this->findTechnology("Computer"), this->findTechnology("Rocketry"), Technology::AGE_MODERN, 20) );
		technology->setInfo("Reveals entire world map.");
		this->addTechnology( technology = new Technology("Satellites", 20000, this->findTechnology("Spaceflight"), NULL, Technology::AGE_MODERN, 30) );
		technology->setInfo("Increases attack and defence of units by 20%.");
		this->addTechnology( technology = new Technology("Microprocessor", 16000, this->findTechnology("Spaceflight"), NULL, Technology::AGE_MODERN, 30) );
		this->addTechnology( technology = new Technology("Stealth", 20000, this->findTechnology("Microprocessor"), NULL, Technology::AGE_MODERN, 30) );
		this->addTechnology( technology = new Technology("Genetics", 25000, this->findTechnology("Microprocessor"), NULL, Technology::AGE_MODERN, 20) );
		technology->setInfo("Increases rate of city growth.");
		this->addTechnology( technology = new Technology("Space Militarisation", 25000, this->findTechnology("Satellites"), this->findTechnology("Microprocessor"), Technology::AGE_MODERN, 30) );*/

		//this->addTechnology( technology = new Technology("", 600, this->findTechnology(""), this->findTechnology(""), Technology::AGE_MODERN, 30) );
		//this->addTechnology( technology = new Technology("", 600, this->findTechnology(""), NULL, Technology::AGE_MODERN, 30) );

		// generate Improvements
		//this->addBuildable( new Improvement("Temple", 20) );
		Improvement *improvement = NULL;

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_BARRACKS, "Barracks", 40, 30) );
		{
			stringstream str;
			str << "The Barracks improvement will automatically produce one military unit of footsoldiers every " << barracks_turns_per_unit_c << " turns.";
			improvement->setInfo(str.str().c_str());
		}
		//improvement->setInfo("The Barracks improvement will automatically produce one military unit of footsoldiers every 25 turns.");
		improvement->setRequires( this->findTechnology("Military Tactics") );

		// AI weight 0, handled with specific code
		this->addBuildable( improvement = new Improvement(IMPROVEMENT_FARMLAND, "Farmland", 20, 0) );
		improvement->setInfo("Farmland is required for a city to grow beyond size 3.\n");
		improvement->setRequires( this->findTechnology("Agriculture") );

		// AI weight 0, handled with specific code
		this->addBuildable( improvement = new Improvement(IMPROVEMENT_AQUEDUCT, "Aqueduct", 60, 0) );
		improvement->setInfo("An Aqueduct is required for a city to grow beyond size 6.\n");
		improvement->setRequires( this->findTechnology("Construction") );
		improvement->setRequires( this->findImprovement("Farmland") );

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_FORT, "Fort", 40, 10) );
		improvement->setInfo("A Fort increases the defence of your units in the city by 25%, unless the attacker has discovered Gunpowder.\n");
		improvement->setObsoletedBy( this->findTechnology("Gunpowder") );

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_HARBOUR, "Harbour", 40, 15) );
		improvement->setInfo("A Harbour allows exploration and travel by sea.\n");
		improvement->setRequires( this->findTechnology("Sailing") );
		improvement->setRequiresCoastal(true);
		improvement->setTravelRange(5, false);
		improvement->setPowerPerTurn(1);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_LIBRARY, "Library", 40, 5) );
		improvement->setInfo("A Library increases the rate of scientific research in the city by 50%.");
		improvement->setRequires( this->findTechnology("Writing") );
		improvement->setPowerPerTurn(1);
		improvement->setAIHint(Improvement::AIHINT_RESEARCH);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_MINE, "Mine", 20, 40) );
		improvement->setRequires( this->findTechnology("Iron Working") );

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_TEMPLE, "Temple", 20, 30) );
		improvement->setRequires( this->findTechnology("Masonry") );
		improvement->setObsoletedBy( this->findTechnology("Scientific Method") );
		improvement->setPowerPerTurn(1);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_WALLS, "Walls", 80, 15) );
		improvement->setInfo("City Walls increase the defence of your units in the city by 50%, unless the attacker has discovered Gunpowder. They also protect against population losses when your city is attacked, for cities of size less than or equal to size 6.\n\nReplaces Fort.\n");
		improvement->setRequires( this->findTechnology("Masonry") );
		improvement->setObsoletedBy( this->findTechnology("Gunpowder") );
		this->findBuildable("Fort")->setReplacedBy(improvement);

		// medieval

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_ARMOURER, "Armourer", 40, 40) );
		improvement->setRequires( this->findTechnology("Medieval Arms") );
		improvement->setRequires( this->findImprovement("Mine") );
		improvement->setObsoletedBy( this->findTechnology("Firearms") );

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_CASTLE, "Castle", 60, 20) );
		improvement->setInfo("A Castle increases the defence of your units in the city by 50%, unless the attacker has discovered Gunpowder.");
		improvement->setRequires( this->findTechnology("Castles") );
		improvement->setObsoletedBy( this->findTechnology("Gunpowder") );
		improvement->setPowerPerTurn(1);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_CHURCH, "Church", 40, 30) );
		improvement->setRequires( this->findTechnology("Feudalism") );
		improvement->setObsoletedBy( this->findTechnology("Scientific Method") );
		improvement->setPowerPerTurn(2);
		this->findBuildable("Temple")->setReplacedBy(improvement);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_PORT, "Port", 60, 15) );
		improvement->setInfo("A Port allows exploration and travel by sea.\n");
		improvement->setRequires( this->findTechnology("Navigation") );
		improvement->setRequiresCoastal(true);
		improvement->setTravelRange(10, false);
		improvement->setPowerPerTurn(2);
		this->findBuildable("Harbour")->setReplacedBy(improvement);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_UNIVERSITY, "University", 100, 5) );
		improvement->setInfo("A University increases the rate of scientific research in the city by 50%.");
		improvement->setRequires( this->findTechnology("Education") );
		improvement->setRequires( this->findImprovement("Library") );
		improvement->setPowerPerTurn(1);
		improvement->setAIHint(Improvement::AIHINT_RESEARCH);

		// industrial

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_COASTALDEFENCE, "Coastal Defence", 100, 30) );
		improvement->setInfo("A Coastal Defence increases the defence bonus against sea attacks to 50%.");
		improvement->setRequires( this->findTechnology("Steam Power") );
		improvement->setRequiresCoastal(true);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_FACTORY, "Factory", 400, 30) );
		improvement->setInfo("A Factory increases the maximum city size, once you have discovered Combustion.");
		improvement->setRequires( this->findTechnology("Industrial Revolution") );
		improvement->setRequires( this->findImprovement("Mine") );
		this->findBuildable("Armourer")->setReplacedBy(improvement);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_HOSPITAL, "Hospital", 300, 10) );
		improvement->setInfo("A Hospital increases the rate of city growth.");
		improvement->setRequires( this->findTechnology("Medicine") );
		improvement->setPowerPerTurn(3);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_SCHOOL, "School", 300, 5) );
		improvement->setInfo("A School increases the basic rate of scientific research in the city by 2.");
		improvement->setRequires( this->findTechnology("Industrial Revolution") );
		improvement->setPowerPerTurn(3);
		improvement->setAIHint(Improvement::AIHINT_RESEARCH);

		// modern

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_AIRPORT, "Airport", 600, 30) );
		improvement->setInfo("An Airport allows air units to bomb from here ,or rebase from or to here, and allows exploration and travel by air.");
		improvement->setRequires( this->findTechnology("Flight") );
		improvement->setRequires( this->findImprovement("Factory") );
		improvement->setTravelRange(biplane_travel_range_c, true);
		improvement->setPowerPerTurn(5);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_ANTIAIRGUNS, "Anti-Air Guns", 800, 30) );
		improvement->setInfo("Can defend against air attacks, defence 4.");
		improvement->setRequires( this->findTechnology("Advanced Flight") );
		improvement->setRequires( this->findImprovement("Factory") );
		improvement->setAirDefence(4);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_LAB, "Lab", 500, 5) );
		improvement->setInfo("A Lab increases the rate of scientific research in the city by 50%.");
		improvement->setRequires( this->findTechnology("Computer") );
		improvement->setRequires( this->findImprovement("University") );
		improvement->setAIHint(Improvement::AIHINT_RESEARCH);
		improvement->setPowerPerTurn(3);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_MISSILELAUNCHER, "Missile Launcher", 800, 30) );
		improvement->setInfo("An Missile Launcher allows missile units to bomb from here.");
		improvement->setRequires( this->findTechnology("Rocketry") );
		improvement->setRequires( this->findImprovement("Factory") );
		improvement->setImmuneFromBombing(true); // so if a civilization is nuked, they can retaliate

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_SAMLAUNCHER, "SAM Launcher", 2000, 30) );
		improvement->setInfo("Can defend against air attacks, defence 12.");
		improvement->setRequires( this->findTechnology("Rocketry") );
		improvement->setRequires( this->findImprovement("Factory") );
		improvement->setAirDefence(12);
		this->findBuildable("Anti-Air Guns")->setReplacedBy(improvement);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_BOMBSHELTER, "Bomb Shelter", 1000, 30) );
		improvement->setInfo("May reduce population deaths from nuclear attacks.");
		improvement->setRequires( this->findTechnology("Nuclear Missiles") );
		improvement->setRequires( this->findImprovement("Factory") );
		improvement->setImmuneFromBombing(true);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_SPACEPORT, "Spaceport", 2000, 30) );
		improvement->setRequires( this->findTechnology("Satellites") );
		improvement->setRequires( this->findImprovement("Airport") );
		improvement->setPowerPerTurn(10);

		this->addBuildable( improvement = new Improvement(IMPROVEMENT_SDILASERS, "SDI Lasers", 2000, 40) );
		improvement->setInfo("80% chance of preventing nuclear attacks to this city.");
		improvement->setRequires( this->findTechnology("Space Militarisation") );
		improvement->setRequires( this->findImprovement("Spaceport") );

		// generate UnitTemplates

		UnitTemplate *unit_template = NULL;

		// special
		this->addBuildable( unit_template = new UnitTemplate("Settlers", 20, 0, 0, 1) );
		unit_template->setInfo("Settlers may only be built in cities of size 2 or greater.");
		unit_template->setFoot(true);
		unit_template->setCanBuildCity(true);
		unit_template->setCanBuildRoads(true);

		// ancient
		this->addBuildable( unit_template = new UnitTemplate("Archers", 15, 2, 1, 1) );
		unit_template->setFoot(true);
		unit_template->setUpgrade(true);
		unit_template->setRequires( this->findTechnology("Archery") );

		this->addBuildable( unit_template = new UnitTemplate("Catapults", 40, 6, 1, 1) );
		unit_template->setRequires( this->findTechnology("Mathematics") );

		this->addBuildable( unit_template = new UnitTemplate("Chariots", 40, 4, 1, 2) );
		unit_template->setUpgrade(true);
		unit_template->setRequires( this->findTechnology("The Wheel") );

		this->addBuildable( unit_template = new UnitTemplate("Horsemen", 20, 2, 1, 2) );
		unit_template->setRequires( this->findTechnology("Horse Riding") );

		this->addBuildable( unit_template = new UnitTemplate("Swordsmen", 30, 3, 2, 1) );
		unit_template->setFoot(true);
		unit_template->setCanBuildRoads(true);
		unit_template->setRequires( this->findTechnology("Iron Working") );
		unit_template->setRequires( this->findImprovement("Mine") );

		this->addBuildable( unit_template = new UnitTemplate("Peasants", 10, 1, 1, 1) );
		unit_template->setFoot(true);
		unit_template->setCanBuildRoads(true);
		unit_template->setUpgrade(true);

		this->addBuildable( unit_template = new UnitTemplate("Spearmen", 20, 1, 2, 1) );
		unit_template->setFoot(true);
		unit_template->setRequires( this->findTechnology("Bronze Working") );
		unit_template->setUpgrade(true);

		// medieval
		/*this->addBuildable( new UnitTemplate("Cannon", 40, 8, 1, 1) );
		this->addBuildable( new UnitTemplate("Foot soldiers", 30, 4, 2, 1) );
		this->addBuildable( new UnitTemplate("Knights", 40, 4, 2, 2) );
		this->addBuildable( new UnitTemplate("Longbowmen", 20, 4, 1, 1) );
		this->addBuildable( new UnitTemplate("Pikemen", 30, 1, 3, 1) );*/

		this->addBuildable( unit_template = new UnitTemplate("Knights", 40, 5, 2, 2) );
		unit_template->setRequires( this->findTechnology("Medieval Arms") );
		unit_template->setRequires( this->findImprovement("Armourer") );
		this->findBuildable("Chariots")->setReplacedBy(unit_template);
		this->findBuildable("Horsemen")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Men-At-Arms", 30, 3, 3, 1) );
		unit_template->setFoot(true);
		unit_template->setCanBuildRoads(true);
		unit_template->setRequires( this->findTechnology("Medieval Arms") );
		unit_template->setRequires( this->findImprovement("Armourer") );
		this->findBuildable("Swordsmen")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Pikemen", 25, 1, 3, 1) );
		unit_template->setFoot(true);
		unit_template->setUpgrade(true);
		unit_template->setRequires( this->findTechnology("Medieval Arms") );
		this->findBuildable("Spearmen")->setReplacedBy(unit_template);

		// industrial

		this->addBuildable( unit_template = new UnitTemplate("Cannon", 60, 8, 1, 1) );
		unit_template->setRequires( this->findTechnology("Gunpowder") );
		this->findBuildable("Catapults")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Cavalry", 80, 6, 3, 3) );
		unit_template->setRequires( this->findTechnology("Firearms") );
		this->findBuildable("Knights")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Militia", 20, 2, 2, 1) );
		unit_template->setFoot(true);
		unit_template->setCanBuildRoads(true);
		unit_template->setRequires( this->findTechnology("Nationalism") );
		this->findBuildable("Peasants")->setReplacedBy(unit_template);
		this->findBuildable("Archers")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Musketeers", 50, 3, 6, 1) );
		unit_template->setFoot(true);
		unit_template->setCanBuildRoads(true);
		unit_template->setUpgrade(true);
		unit_template->setRequires( this->findTechnology("Firearms") );
		this->findBuildable("Men-At-Arms")->setReplacedBy(unit_template);
		this->findBuildable("Pikemen")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Riflemen", 80, 3, 8, 1) );
		unit_template->setFoot(true);
		unit_template->setCanBuildRoads(true);
		unit_template->setUpgrade(true);
		unit_template->setRequires( this->findTechnology("Nationalism") );
		this->findBuildable("Musketeers")->setReplacedBy(unit_template);

		// modern

		this->addBuildable( unit_template = new UnitTemplate("Infantry", 200, 8, 14, 1) );
		unit_template->setFoot(true);
		unit_template->setCanBuildRoads(true);
		unit_template->setUpgrade(true);
		unit_template->setRequires( this->findTechnology("Explosives") );
		unit_template->setRequires( this->findImprovement("Factory") );
		this->findBuildable("Riflemen")->setReplacedBy(unit_template);
		this->findBuildable("Cannon")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Tanks", 800, 24, 12, 3) );
		unit_template->setRequires( this->findTechnology("Motorised Warfare") );
		unit_template->setRequires( this->findImprovement("Factory") );
		this->findBuildable("Cavalry")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Modern Infantry", 800, 16, 28, 1) );
		unit_template->setFoot(true);
		unit_template->setCanBuildRoads(true);
		unit_template->setRequires( this->findTechnology("Rocketry") );
		unit_template->setRequires( this->findImprovement("Factory") );
		this->findBuildable("Infantry")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Modern Tanks", 3200, 36, 24, 4) );
		unit_template->setRequires( this->findTechnology("Microprocessor") );
		unit_template->setRequires( this->findImprovement("Factory") );
		this->findBuildable("Tanks")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Biplanes", 100, 0, 0, 1) );
		//this->addBuildable( unit_template = new UnitTemplate("Biplanes", 5, 0, 0, 1) );
		//unit_template->setAir(true, 1, 1, 30);
		//unit_template->setVisibilityRange(30);
		unit_template->setAir(true, 1, 1, biplane_travel_range_c, false);
		unit_template->setVisibilityRange(biplane_travel_range_c);
		unit_template->setBombard(10, 40);
		unit_template->setRequires( this->findTechnology("Flight") );

		this->addBuildable( unit_template = new UnitTemplate("Fighters", 200, 0, 0, 1) );
		//unit_template->setAir(true, 0, 8, 0, false);
		unit_template->setAir(true, 0, 8, bombers_travel_range_c, false);
		unit_template->setVisibilityRange(bombers_travel_range_c);
		unit_template->setBombard(0, 0);
		unit_template->setRequires( this->findTechnology("Advanced Flight") );
		unit_template->setRequires( this->findImprovement("Airport") );
		this->findBuildable("Biplanes")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Bombers", 400, 0, 0, 1) );
		unit_template->setAir(true, 8, 0, bombers_travel_range_c, false);
		unit_template->setVisibilityRange(bombers_travel_range_c);
		unit_template->setBombard(25, 500);
		unit_template->setRequires( this->findTechnology("Advanced Flight") );
		unit_template->setRequires( this->findImprovement("Airport") );

		this->addBuildable( unit_template = new UnitTemplate("Jet Fighters", 800, 0, 0, 1) );
		//unit_template->setAir(true, 0, 24, 0, false);
		unit_template->setAir(true, 0, 24, jet_travel_range_c, false);
		unit_template->setVisibilityRange(jet_travel_range_c);
		unit_template->setBombard(0, 0);
		unit_template->setRequires( this->findTechnology("Rocketry") );
		unit_template->setRequires( this->findImprovement("Airport") );
		this->findBuildable("Fighters")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Jet Bombers", 1600, 0, 0, 1) );
		unit_template->setAir(true, 24, 0, jet_travel_range_c, false);
		unit_template->setVisibilityRange(jet_travel_range_c);
		unit_template->setBombard(33, 2000);
		unit_template->setRequires( this->findTechnology("Rocketry") );
		unit_template->setRequires( this->findImprovement("Airport") );
		this->findBuildable("Bombers")->setReplacedBy(unit_template);

		this->addBuildable( unit_template = new UnitTemplate("Stealth Bombers", 2400, 0, 0, 1) );
		unit_template->setAir(true, 36, 0, jet_travel_range_c, false);
		unit_template->setVisibilityRange(jet_travel_range_c);
		unit_template->setBombard(33, 2000);
		unit_template->setRequires( this->findTechnology("Stealth") );
		unit_template->setRequires( this->findImprovement("Airport") );

		this->addBuildable( unit_template = new UnitTemplate("Cruise Missile", 200, 0, 0, 1) );
		//this->addBuildable( unit_template = new UnitTemplate("Cruise Missile", 1, 0, 0, 1) );
		unit_template->setAir(true, 0, 0, 30, true);
		unit_template->setVisibilityRange(0);
		unit_template->setBombard(50, 2000);
		unit_template->setRequires( this->findTechnology("Rocketry") );
		unit_template->setRequires( this->findImprovement("Missile Launcher") );

		this->addBuildable( unit_template = new UnitTemplate("Fission Bomb", 800, 0, 0, 1) );
		//this->addBuildable( unit_template = new UnitTemplate("Fission Bomb", 1, 0, 0, 1) );
		unit_template->setAir(true, 0, 0, bombers_travel_range_c, true);
		unit_template->setVisibilityRange(0);
		unit_template->setNuclearType(UnitTemplate::NUCLEARTYPE_FISSION);
		unit_template->setRequires( this->findTechnology("Fission") );
		unit_template->setRequires( this->findImprovement("Airport") );

		this->addBuildable( unit_template = new UnitTemplate("Nuclear Missile", 1200, 0, 0, 1) );
		//this->addBuildable( unit_template = new UnitTemplate("Nuclear Missile", 1, 0, 0, 1) );
		unit_template->setAir(true, 0, 0, 100, true);
		unit_template->setVisibilityRange(0);
		unit_template->setNuclearType(UnitTemplate::NUCLEARTYPE_FUSION);
		unit_template->setRequires( this->findTechnology("Nuclear Missiles") );
		unit_template->setRequires( this->findImprovement("Missile Launcher") );
		this->findBuildable("Fission Bomb")->setReplacedBy(unit_template);

		// generate Races

		if( !loadScript("conquests/data/scripts/generate_races.lua") ) {
			throw "lua script failed";
		}

		// Romans (white), English (orange), Egyptians (yellow), Americans (light blue), French (green), Greeks (pink), Vikings (brown)
		// TODO: Germans?, Russians (dark red), Chinese?, Japanese, Persians?
		/*unsigned char rgb_white[] = {255, 255, 255};
		Race *race_romans = new Race("The Romans", "Roman", rgb_white, "conquests/data/gfx/civs/romans.jpg");
		race_romans->setStartingTechnology(findTechnology("Bronze Working"));
		race_romans->addCityName("Rome");
		race_romans->addCityName("Veii");
		race_romans->addCityName("Antium");
		race_romans->addCityName("Cumae");
		race_romans->addCityName("Neapolis");
		race_romans->addCityName("Pompeii");
		race_romans->addCityName("Pisae");
		race_romans->addCityName("Ravenna");
		race_romans->addCityName("Hispalis");
		race_romans->addCityName("Viroconium");
		race_romans->addCityName("Lugdunum");
		race_romans->addCityName("Lutetia");
		race_romans->addCityName("Byzantium");
		race_romans->addCityName("Brundisium");
		race_romans->addCityName("Syracuse");
		race_romans->addCityName("Caesaraugusta");
		race_romans->addCityName("Palmyra");
		race_romans->addCityName("Jerusalem");
		race_romans->addCityName("Caeserea");
		race_romans->addCityName("Tarentum");
		race_romans->addCityName("Nicomedia");
		race_romans->addCityName("Seleucia");
		race_romans->addCityName("Artaxata");
		race_romans->addCityName("Aurelianorum");
		race_romans->addCityName("Hippo Regius");
		race_romans->addCityName("Nicopolis");*/

		/*unsigned char rgb_orange[] = {255, 64, 0};
		Race *race_english = new Race("The English", "English", rgb_orange, "conquests/data/gfx/civs/english.jpg");
		race_english->setStartingTechnology(findTechnology("Sailing"));
		race_english->addCityName("London");
		race_english->addCityName("York");
		race_english->addCityName("Nottingham");
		race_english->addCityName("Hastings");
		race_english->addCityName("Canterbury");
		race_english->addCityName("Coventry");
		race_english->addCityName("Warwick");
		race_english->addCityName("Newcastle");
		race_english->addCityName("Oxford");
		race_english->addCityName("Liverpool");
		race_english->addCityName("Dover");
		race_english->addCityName("Brighton");
		race_english->addCityName("Norwich");
		race_english->addCityName("Leeds");
		race_english->addCityName("Reading");
		race_english->addCityName("Birmingham");
		race_english->addCityName("Richmond");
		race_english->addCityName("Exeter");
		race_english->addCityName("Cambridge");
		race_english->addCityName("Gloucester");
		race_english->addCityName("Manchester");
		race_english->addCityName("Bristol");
		race_english->addCityName("Leicester");
		race_english->addCityName("Carlisle");
		race_english->addCityName("Ipswich");
		race_english->addCityName("Portsmouth");*/

		/*unsigned char rgb_yellow[] = {255, 255, 0};
		Race *race_egyptians = new Race("The Egyptians", "Egyptian", rgb_yellow, "conquests/data/gfx/civs/egyptians.jpg");
		race_egyptians->setStartingTechnology(findTechnology("Agriculture"));
		race_egyptians->addCityName("Thebes");
		race_egyptians->addCityName("Memphis");
		race_egyptians->addCityName("Heliopolis");
		race_egyptians->addCityName("Elephantine");
		race_egyptians->addCityName("Alexandria");
		race_egyptians->addCityName("Pi-Ramesses");
		race_egyptians->addCityName("Giza");
		race_egyptians->addCityName("Byblos");
		race_egyptians->addCityName("El-Amarna");
		race_egyptians->addCityName("Hieraconpolis");
		race_egyptians->addCityName("Abydos");
		race_egyptians->addCityName("Asyut");
		race_egyptians->addCityName("Avaris");
		race_egyptians->addCityName("Lisht");
		race_egyptians->addCityName("Buto");
		race_egyptians->addCityName("Edfu");
		race_egyptians->addCityName("Pithom");
		race_egyptians->addCityName("Busiris");
		race_egyptians->addCityName("Kahun");
		race_egyptians->addCityName("Athribis");
		race_egyptians->addCityName("Mendes");
		race_egyptians->addCityName("El-Ashmunein");
		race_egyptians->addCityName("Tanis");
		race_egyptians->addCityName("Buhen");
		race_egyptians->addCityName("Bubastis");
		race_egyptians->addCityName("This");*/

		/*unsigned char rgb_lightblue[] = {147, 204, 234};
		Race *race_americans = new Race("The Americans", "American", rgb_lightblue, "conquests/data/gfx/civs/americans.jpg");
		race_americans->setStartingTechnology(findTechnology("Alphabet"));
		race_americans->addCityName("Washington");
		race_americans->addCityName("New York");
		race_americans->addCityName("Boston");
		race_americans->addCityName("Philadelphia");
		race_americans->addCityName("Atlanta");
		race_americans->addCityName("Chicago");
		race_americans->addCityName("San Francisco");
		race_americans->addCityName("Buffalo");
		race_americans->addCityName("St. Louis");
		race_americans->addCityName("Detroit");
		race_americans->addCityName("New Orleans");
		race_americans->addCityName("Baltimore");
		race_americans->addCityName("Denver");
		race_americans->addCityName("Cincinnati");
		race_americans->addCityName("Dallas");
		race_americans->addCityName("Los Angeles");
		race_americans->addCityName("Kansas City");
		race_americans->addCityName("San Diego");
		//race_americans->addCityName("Richmond"); // avoid duplicate with English
		race_americans->addCityName("Las Vegas");
		race_americans->addCityName("Phoenix");
		race_americans->addCityName("Seattle");
		race_americans->addCityName("Albuquerque");
		race_americans->addCityName("Portland");
		race_americans->addCityName("Minneapolis");*/

		/*unsigned char rgb_green[] = {64, 255, 64};
		Race *race_french = new Race("The French", "French", rgb_green, "conquests/data/gfx/civs/french.jpg");
		race_french->setStartingTechnology(findTechnology("Archery"));
		race_french->addCityName("Paris");
		race_french->addCityName("Orleans");
		race_french->addCityName("Lyons");
		race_french->addCityName("Rheims");
		race_french->addCityName("Tours");
		race_french->addCityName("Marseilles");
		race_french->addCityName("Chartres");
		race_french->addCityName("Avignon");
		race_french->addCityName("Besancon");
		race_french->addCityName("Rouen");
		race_french->addCityName("Grenoble");
		race_french->addCityName("Dijon");
		race_french->addCityName("Amiens");
		race_french->addCityName("Cherbourg");
		race_french->addCityName("Poitiers");
		race_french->addCityName("Toulouse");
		race_french->addCityName("Bayonne");
		race_french->addCityName("Strasbourg");
		race_french->addCityName("Brest");
		race_french->addCityName("Bordeaux");*/

		/*unsigned char rgb_red[] = {255, 64, 64};
		Race *race_greeks = new Race("The Greeks", "Greek", rgb_red, "conquests/data/gfx/civs/greeks.jpg");
		race_greeks->setStartingTechnology(findTechnology("Alphabet"));
		race_greeks->addCityName("Athens");
		race_greeks->addCityName("Sparta");
		race_greeks->addCityName("Thermopylae");
		race_greeks->addCityName("Corinth");
		race_greeks->addCityName("Delphi");
		race_greeks->addCityName("Pharsalos");
		race_greeks->addCityName("Knossos");
		race_greeks->addCityName("Argos");
		race_greeks->addCityName("Mycenae");
		race_greeks->addCityName("Herakleia");
		race_greeks->addCityName("Ephesos");
		race_greeks->addCityName("Thessalonica");
		race_greeks->addCityName("Rhodes");
		race_greeks->addCityName("Eretria");
		race_greeks->addCityName("Troy");
		race_greeks->addCityName("Marathon");
		race_greeks->addCityName("Halicarnassus");
		race_greeks->addCityName("Pergamon");
		race_greeks->addCityName("Miletos");
		race_greeks->addCityName("Artemisium");
		race_greeks->addCityName("Megara");
		race_greeks->addCityName("Phocaea");
		race_greeks->addCityName("Sicyon");
		race_greeks->addCityName("Gortyn");
		race_greeks->addCityName("Mytilene");
		race_greeks->addCityName("Tegea");*/

		/*unsigned char rgb_brown[] = {150, 75, 0};
		Race *race_vikings = new Race("The Vikings", "Viking", rgb_brown, "conquests/data/gfx/civs/vikings.jpg");
		race_vikings->setStartingTechnology(findTechnology("Sailing"));
		race_vikings->addCityName("Nidaros");
		race_vikings->addCityName("Uppsala");
		race_vikings->addCityName("Haithabu");
		race_vikings->addCityName("Birka");
		race_vikings->addCityName("Bergen");
		race_vikings->addCityName("Jelling");
		race_vikings->addCityName("Roskilde");
		race_vikings->addCityName("Tnsberg");
		race_vikings->addCityName("Sigtuna");
		race_vikings->addCityName("Ldse");
		race_vikings->addCityName("Odense");
		race_vikings->addCityName("Oslo");
		race_vikings->addCityName("Ribe");
		race_vikings->addCityName("Lindholm Hje");
		race_vikings->addCityName("Vsters");
		race_vikings->addCityName("Aarhus");
		race_vikings->addCityName("Lund");
		race_vikings->addCityName("Kaupang");
		race_vikings->addCityName("Reykjavik");
		race_vikings->addCityName("ingvellir");
		race_vikings->addCityName("Stamford");
		race_vikings->addCityName("Holmgrd");
		race_vikings->addCityName("Aldeigjuborg");
		race_vikings->addCityName("Tynwald");
		race_vikings->addCityName("Leirvik");
		race_vikings->addCityName("Vinland");*/

		// pre-load other scripts
		if( !loadScript("conquests/data/scripts/ai.lua") ) {
			throw "lua script failed";
		}
	}
	catch(VisionException *ve) {
		VI_log("Failed - something went wrong loading game data!\n");
		this->setCursor(cursor_arrow);
#ifdef _WIN32
		MessageBoxA(NULL,"Failed to load Game data","Error",MB_OK|MB_ICONEXCLAMATION);
#endif
		delete ve;
		return false;
	}
	catch(const char *error) {
		VI_log("Failed - something went wrong loading game data!\n");
		this->setCursor(cursor_arrow);
#ifdef _WIN32
		MessageBoxA(NULL,"Failed to load Game data","Error",MB_OK|MB_ICONEXCLAMATION);
#endif
		return false;
	}

	VI_set_persistence_all(-1);

/*#ifdef _WIN32
	cursor = LoadCursor( NULL, IDC_ARROW );
	SetCursor(cursor);
	SetClassLong(genv->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
	{
		char buffer[4096] = "";
		sprintf(buffer, "Conquests v%d.%d | (c) 2009-2011 Mark Harman", versionMajor, versionMinor);
		game_g->getGraphicsEnvironment()->setWindowTitle(buffer);
	}
	this->setCursor(cursor_arrow);
	return true;
}

void GameData::clear() {
	for(vector<const Race *>::const_iterator iter = races.begin(); iter != races.end(); ++iter) {
		const Race *race = *iter;
		delete race;
	}
	races.clear();

	for(vector<Buildable *>::iterator iter = buildables.begin();iter != buildables.end(); ++iter) {
		Buildable *buildable = *iter;
		delete buildable;
	}
	buildables.clear();

	for(vector<Technology *>::const_iterator iter = technologies.begin();iter != technologies.end(); ++iter) {
		const Technology *technology = *iter;
		delete technology;
	}
	technologies.clear();
}

void Game::clear() {
	gameData.clear();
	lua_close(ls);
	ls = NULL;
	VI_cleanup();
	delete [] savegames_dirname;
	savegames_dirname = NULL;
}

void Game::run() {
	if( !init() ) {
		VI_log("failed to initialise!\n");
		clear();
		return;
	}

	/*gamestate = ::game = new MainGamestate();
	::game->run();
	delete ::game;
	::game = NULL;*/
	/*MainGamestate *mainGamestate = new MainGamestate();
	gamestate = mainGamestate;
	mainGamestate->run();
	delete mainGamestate;
	gamestate = NULL;*/

	/*MainGamestate *mainGamestate = new MainGamestate();
	gamestate = mainGamestate;
	//this->setupNewGame(mainGamestate);
	mainGamestate->initGameData(true, 100, 100, Map::TOPOLOGY_CYLINDER, this->findRace("The Romans"), 4);*/

	gamestate = new OptionsGamestate();
	//gamestate->run();

	if( !gamestate->start() ) {
		VI_log("failed to start initial gamestate!\n");
		delete gamestate;
		gamestate = NULL;
		clear();
		return;
	}

	bool timedemo = false;
	//timedemo = true;
	int last_update_time = clock();
	int t1 = clock();
	const int n_frames = 500;
	//bool hack = false;

	// program main loop
	while( !quit ) {
		if( game_g->getGraphicsEnvironment()->shouldQuit() ) {
			VI_log("should quit\n");
			game_g->getGraphicsEnvironment()->clearQuitRequest();
			gamestate->quitGame();
		}
		else if( game_g->getGraphicsEnvironment()->isActive() ) {
			//ASSERT( _heapchk() == _HEAPOK );
			game_g->getGraphicsEnvironment()->drawFrame();
			//ASSERT( _heapchk() == _HEAPOK );
			game_g->getGraphicsEnvironment()->swapBuffers();
			frames++;
			if( timedemo && frames >= n_frames )
				quit = true;
		}
		else {
			// we don' call VI_GraphicsEnvironment::waitForEvents(), as we still want to process even when not active (see below)
			// however, we want to avoid consuming too much CPU when not active, so sleep for a bit
			//Sleep(10); // don't consume 100% CPU
			game_g->getGraphicsEnvironment()->sleep(10);
		}

		// we update even if not active, so that the game still continues processing (useful for between player turns)
		int time_diff = clock() - last_update_time;
		if( time_diff > 0 ) {
			gamestate->update(time_diff);
			last_update_time = clock();
			if( game_g->getGraphicsEnvironment()->isActive() )
			{
				game_g->getGraphicsEnvironment()->handleInput(); // must be done last due to resetting input
			}
		}
		game_g->getGraphicsEnvironment()->processEvents();

		if( this->newGamestate != NULL ) {
			VI_log("switch to new gamestate\n");
			// stop the current one
			gamestate->stop();
			delete gamestate;
			// assign the new one
			gamestate = newGamestate;
			newGamestate = NULL;
			// start the new one
			if( !gamestate->start() ) {
				VI_log("failed to start the new gamestate!\n");
				delete gamestate;
				gamestate = NULL;
				VI_cleanup();
				return;
			}
		}
	}

/*	while( !quit ) {
#ifdef _WIN32
		MSG msg;
		// check for messages
		if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {
			// handle or dispatch messages
			if( msg.message == WM_QUIT ) {
				gamestate->quitGame();
			}
			else {
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		}
		//else if( genv->isActive() )
		else
#endif
		{
			if( game_g->getGraphicsEnvironment()->isActive() ) {
				//ASSERT( _heapchk() == _HEAPOK );
				game_g->getGraphicsEnvironment()->drawFrame();
				//ASSERT( _heapchk() == _HEAPOK );
				game_g->getGraphicsEnvironment()->swapBuffers();
				frames++;
				if( timedemo && frames >= n_frames )
					quit = true;
			}
			else {
#ifdef _WIN32
				Sleep(10); // don't consume 100% CPU
#endif
			}

			int time_diff = clock() - last_update_time;
			if( time_diff > 0 ) {
				gamestate->update(time_diff);
				last_update_time = clock();
				//if( game_g->getGraphicsEnvironment()->isActive() )
				{
					game_g->getGraphicsEnvironment()->handleInput(); // must be done last due to resetting input; must always call even if not active, in order to read SDL events
				}
			}

			if( this->newGamestate != NULL ) {
				VI_log("switch to new gamestate\n");
				// stop the current one
				gamestate->stop();
				delete gamestate;
				// assign the new one
				gamestate = newGamestate;
				newGamestate = NULL;
				// start the new one
				if( !gamestate->start() ) {
					VI_log("failed to start the new gamestate!\n");
					delete gamestate;
					gamestate = NULL;
					VI_cleanup();
					return;
				}
			}
		}
		//else {
		//WaitMessage();
		//}
	}*/

	// display busy cursor whilst quitting
/*#ifdef _WIN32
	HCURSOR cursor = LoadCursor( NULL, IDC_WAIT );
	SetCursor(cursor);
	SetClassLong(game_g->getGraphicsEnvironment()->getHWND(), GCL_HCURSOR, (LONG)cursor);
#endif*/
	game_g->setCursor(cursor_wait);

	int t2 = clock();

	gamestate->stop();

	delete gamestate;
	gamestate = NULL;

	clear();

	//ShowCursor(TRUE);
#ifdef _DEBUG
#ifdef _WIN32
	char buffer[256] = "";
	sprintf_s(buffer,"%d frames in time %d\nFPS : %f",frames,t2-t1,(frames*1000)/((float)(t2-t1)));
	MessageBoxA(NULL,buffer,"Finished",NULL);
#endif
#endif
}

void Game::runTests(MainGamestate::TestID id) {
	if( !init() ) {
		clear();
		return;
	}

	/*gamestate = ::game = new MainGamestate();
	::game->runTests(id);
	delete ::game;
	::game = NULL;*/
	MainGamestate *mainGamestate = new MainGamestate();
	gamestate = mainGamestate;
	mainGamestate->runTests(id);
	delete mainGamestate;
	gamestate = NULL;

	clear();
}

/*void MainGamestate::doGame() {
	::game = new MainGamestate();
	::game->run();
	delete ::game;
}

void MainGamestate::doTests(TestID id) {
	::game = new MainGamestate();
	::game->runTests(id);
	delete ::game;
}*/

void Game::doGame() {
	::game_g = new Game();
	::game_g->run();
	delete ::game_g;
	::game_g = NULL;
}

void Game::doTests(MainGamestate::TestID id, bool testmode_silent) {
	::game_g = new Game();
	game_g->testmode_silent = testmode_silent;
	::game_g->runTests(id);
	delete ::game_g;
	::game_g = NULL;
}
